#include <stdio.h>

#include "Lgi.h"
#include "LgiIde.h"
#include "GDragAndDrop.h"
#include "GToken.h"
#include "resdefs.h"
#include "GProcess.h"
#include "GCombo.h"

extern char *Untitled;

#define IDM_INSERT					100
#define IDM_NEW_FOLDER				101
#define IDM_RENAME					102
#define IDM_DELETE					103
#define IDM_SETTINGS				104
#define IDM_INSERT_DEP				105
#define IDM_SORT_CHILDREN			106
#define IDM_PROPERTIES				107
#define IDM_BROWSE_FOLDER			108
#define IDM_OPEN_TERM				109
#define IDM_IMPORT_FOLDER			110

#ifdef WIN32
#define LGI_STATIC_LIBRARY_EXT		"lib"
#else
#define LGI_STATIC_LIBRARY_EXT		"a"
#endif

char *PlatformName[] =
{
	"Windows",
	"Linux",
	"Mac",
	0
};

int PlatformCtrlId[] =
{
	IDC_WIN32,
	IDC_LINUX,
	IDC_MAC,
	0
};

#define ForAllProjectNodes(Var) \
	for (ProjectNode *Var=dynamic_cast<ProjectNode*>(GetChild()); \
		Var; \
		Var=dynamic_cast<ProjectNode*>(Var->GetNext()))

enum NodeType
{
	NodeNone		= 0,
	NodeDir			= 1,
	NodeSource		= 2,
	NodeHeader		= 3,
	NodeDependancy	= 4,
	NodeResources	= 5,
	NodeGraphic		= 6
};

char SourcePatterns[] = "*.c;*.h;*.cpp;*.java;*.d";

char *ToUnixPath(char *s)
{
	if (s)
	{
		char *c;
		while (c = strchr(s, '\\'))
		{
			*c = '/';
		}
	}
	return s;
}

char *ToNativePath(char *s)
{
	if (s)
	{
		for (char *c = s; *c; c++)
		{
			#ifdef WIN32
			if (*c == '/')
				*c = '\\';
			#else
			if (*c == '\\')
				*c = '/';
			#endif
		}
	}
	return s;
}

//////////////////////////////////////////////////////////////////////////////////
class FileProps : public GDialog
{
public:
	NodeType Type;
	int Platforms;

	FileProps(GView *p, char *m, NodeType t, int plat)
	{
		Platforms = plat;
		Type = t;
		SetParent(p);
		if (LoadFromResource(IDD_FILE_PROPS))
		{
			MoveToCenter();
			SetCtrlName(IDC_MSG, m);
			
			GCombo *c = dynamic_cast<GCombo*>(FindControl(IDC_TYPE));
			if (c)
			{
				c->Insert("None");
				c->Insert("Folder");
				c->Insert("Source");
				c->Insert("Header");
				c->Insert("Dependency");
				c->Insert("Resource");
				c->Insert("Graphic");
				c->Value(Type);
			}

			for (int i=0; PlatformName[i]; i++)
			{
				SetCtrlValue(PlatformCtrlId[i], ((1 << i) & Platforms) != 0);
			}
		}
	}
	
	int OnNotify(GViewI *c, int f)
	{
		switch (c->GetId())
		{
			case IDOK:
			{
				int t = GetCtrlValue(IDC_TYPE);
				if (t >= NodeSource)
				{
					Type = (NodeType) t;
				}

				Platforms = 0;
				for (int i=0; PlatformName[i]; i++)
				{
					if (GetCtrlValue(PlatformCtrlId[i]))
					{
						Platforms |= 1 << i;
					}
				}
				
				// fall thru
			}
			case IDC_COPY_PATH:
			{
				EndModal(c->GetId());
				break;
			}
		}
		
		return 0;
	}
};

//////////////////////////////////////////////////////////////////////////////////
class ProjectNode;

class ProjectSettings
{
public:
	char *MakeFile;
	char *Exe;
	char *ExeArgs;
	char *DefineSym;
	int TargetType;
	char *TargetName;
	char *IncludePaths;
	char *LibPaths;
	char *Libs;
	char *CommentFile;
	char *CommentFunc;
	char *MakefileRules;
	int TabSize;
	int IndentSize;
	int UseTabs;
	int ShowWhitespace;
	int Compiler;

	ProjectSettings()
	{
		MakeFile = 0;
		Exe = 0;
		ExeArgs = 0;
		TargetType = 0;
		TargetName = 0;
		IncludePaths = 0;
		LibPaths = 0;
		Libs = 0;
		CommentFile = 0;
		CommentFunc = 0;
		TabSize = 8;
		IndentSize = 4;
		UseTabs = 0;
		ShowWhitespace = 0;
		DefineSym = 0;
		MakefileRules = 0;
		Compiler = 0;
	}
	
	~ProjectSettings()
	{
		DeleteArray(MakeFile);
		DeleteArray(Exe);
		DeleteArray(ExeArgs);
		DeleteArray(TargetName);
		DeleteArray(IncludePaths);
		DeleteArray(LibPaths);
		DeleteArray(Libs);
		DeleteArray(CommentFile);
		DeleteArray(CommentFunc);
		DeleteArray(DefineSym);
		DeleteArray(MakefileRules);
	}
	
	ProjectSettings &operator =(ProjectSettings &p)
	{
		#define AssignStr(n) { char *t = NewStr(p.n); DeleteArray(n); n = t; }
		
		AssignStr(MakeFile);
		AssignStr(Exe);
		AssignStr(ExeArgs);
		AssignStr(DefineSym);
		TargetType = p.TargetType;
		AssignStr(TargetName);
		AssignStr(IncludePaths);
		AssignStr(LibPaths);
		AssignStr(Libs);
		AssignStr(CommentFile);
		AssignStr(CommentFunc);
		AssignStr(MakefileRules);
		TabSize = p.TabSize;
		IndentSize = p.IndentSize;
		UseTabs = p.UseTabs;
		ShowWhitespace = p.ShowWhitespace;
		Compiler = p.Compiler;

		return *this;
	}
};

class IdeProjectPrivate
{
public:
	AppWnd *App;
	bool Dirty;
	char *FileName;
	class BuildThread *Build;
	IdeProject *ParentProject;
	ProjectSettings Settings;

	IdeProjectPrivate(AppWnd *a)
	{
		App = a;
		Dirty = false;
		FileName = 0;
		ParentProject = 0;
		Build = 0;
	}

	~IdeProjectPrivate()
	{
		DeleteArray(FileName);
	}

	void CollectAllFiles(GTreeNode *Base, List<ProjectNode> &Files, bool SubProjects, int Platform);
};

/////////////////////////////////////////////////////////////////////////////////////
class ProjectNode : public IdeCommon, public GDragDropSource
{
	IdeProject *Project;
	NodeType Type;
	int Platforms;
	char *File;
	char *Name;
	IdeProject *Dep;
	bool IgnoreExpand;

public:
	ProjectNode(IdeProject *p)
	{
		Platforms = PLATFORM_ALL;
		Project = p;
		Type = NodeNone;
		File = 0;
		Name = 0;
		IgnoreExpand = false;
		Dep = 0;
		Tag = NewStr("Node");
	}

	~ProjectNode()
	{
		if (Dep)
		{
			DeleteObj(Dep);
		}

		DeleteArray(File);
		DeleteArray(Name);
	}
	
	int GetPlatforms()
	{
		return Platforms;
	}
	
	ProjectNode *ChildNode()
	{
		return dynamic_cast<ProjectNode*>(GetChild());
	}
	
	ProjectNode *NextNode()
	{
		return dynamic_cast<ProjectNode*>(GetNext());
	}
	
	void SetClean()
	{
		if (Dep)
		{
			Dep->SetClean();
		}
		
		ForAllProjectNodes(p)
		{
			p->SetClean();
		}
	}
	
	IdeProject *GetDep()
	{
		return Dep;
	}
	
	IdeProject *GetProject()
	{
		return Project;
	}
	
	bool GetFormats(List<char> &Formats)
	{
		Formats.Insert(NewStr(NODE_DROP_FORMAT));
		return true;
	}

	bool GetData(GVariant *Data, char *Format)
	{
		if (Format AND stricmp(Format, NODE_DROP_FORMAT) == 0)
		{
			void *t = this;
			Data->SetBinary(sizeof(this), (void*) &t);
			return true;
		}
		
		return false;
	}
	
	bool OnBeginDrag(GMouse &m)
	{
		return Drag(Tree, DROPEFFECT_MOVE);
	}
	
	char *GetFileName()
	{
		return File;
	}

	void SetFileName(char *f)
	{
		DeleteArray(File);
		File = NewStr(f);

		if (File)
		{
			char MimeType[256];
			
			char *FullPath = GetFullPath();
			if (FullPath)
			{
				if (LgiGetFileMimeType(FullPath, MimeType, sizeof(MimeType)) AND
					strnicmp(MimeType, "image/", 6) == 0)
				{
					Type = NodeGraphic;
				}
				DeleteArray(FullPath);
			}
			
			if (NOT Type)
			{
				char *Ext = LgiGetExtension(File);
				if (Ext)
				{
					if (stricmp(Ext, "lr8") == 0)
					{
						Type = NodeResources;
					}
					else if (stricmp(Ext, "h") == 0)
					{
						Type = NodeHeader;
					}
				}
			}
			
			if (NOT Type)
			{
				Type = NodeSource;
			}
		}
		
		Project->SetDirty();
	}

	char *GetName()
	{
		return Name;
	}

	void SetName(char *f)
	{
		DeleteArray(Name);
		Name = NewStr(f);
		Type = NodeDir;
	}

	NodeType GetType()
	{
		return Type;
	}

	void SetType(NodeType t)
	{
		Type = t;
	}

	int GetImage(int f)
	{
		switch (Type)
		{
			case NodeSource:
				return ICON_SOURCE;
			case NodeHeader:
				return ICON_HEADER;
			case NodeDependancy:
				return ICON_DEPENDANCY;
			case NodeGraphic:
				return ICON_GRAPHIC;
			case NodeResources:
				return ICON_RESOURCE;
		}

		return ICON_FOLDER;
	}

	char *GetText(int c)
	{
		if (File)
		{
			#ifdef WIN32
			char Other = '/';
			#else
			char Other = '\\';
			#endif
			char *s;
			while (s = strchr(File, Other))
			{
				*s = DIR_CHAR;
			}
			
			char *d = strrchr(File, DIR_CHAR);
			if (d) return d + 1;
			else return File;
		}

		return Name ? Name : Untitled;
	}

	void OnExpand(bool b)
	{
		if (NOT IgnoreExpand)
		{
			Project->SetDirty();
		}
	}
	
	bool Serialize()
	{
		SerializeAttr("File", File);
		SerializeAttr("Name", Name);
		SerializeAttr("Type", (int&)Type);
		SerializeAttr("Platforms", (int&)Platforms);

		char *Ext = LgiGetExtension(File);

		if (Ext)
		{
			if (stricmp(Ext, "cpp") == 0 OR
				stricmp(Ext, "c") == 0)
			{
				Type = NodeSource;
			}
			else if (stricmp(Ext, "h") == 0)
			{
				Type = NodeHeader;
			}
			else if (stricmp(Ext, "lr8") == 0)
			{
				Type = NodeResources;
			}			
		}

		if (Type == NodeDir)
		{
			int Open = Expanded();
			SerializeAttr("Open", Open);
			if (NOT Write)
			{
				IgnoreExpand = true;
				Expanded(Open);
				IgnoreExpand = false;
			}
		}
		else if (Type == NodeDependancy)
		{
			if (NOT Write)
			{
				Dep = Project->GetApp()->OpenProject(File, false, true);
				if (Dep)
				{
					Dep->SetParentProject(Project);
				}
			}
		}
		else
		{
			if (NOT Write)
			{
				// Check that file exists.
				char *p = GetFullPath();
				DeleteArray(p);

				if (p)
				{
					if (NOT FileExists(p))
					{
						char Path[256];
						if (Project->GetBasePath(Path))
						{
							LgiTrimDir(Path);
							
							// Find the file.
							char *d = strrchr(p, DIR_CHAR);
							List<char> Files;
							List<char> Ext;
							Ext.Insert(d ? d + 1 : p);
							if (LgiRecursiveFileSearch(Path, &Ext, &Files))
							{
								if (Files.First())
								{
									GStringPipe Buf;
									Buf.Printf(	"The file:\n"
												"\n"
												"\t%s\n"
												"\n"
												"doesn't exist. Is this the right file:\n"
												"\n"
												"\t%s",
												p,
												Files.First());
									char *Msg = Buf.NewStr();
									if (Msg)
									{
										GAlert a(Project->GetApp(), "Missing File", Msg, "Yes", "No", "Browse...");
										switch (a.DoModal())
										{
											case 1: // Yes
											{
												SetFileName(Files.First());
												break;
											}
											case 2: // No
											{
												break;
											}
											case 3: // Browse
											{
												GFileSelect s;
												s.Parent(Project->GetApp());
												s.Type("Code", SourcePatterns);
												if (s.Open())
												{
													SetFileName(s.Name());
												}
												break;
											}
										}
										
										DeleteArray(Msg);
									}
								}
								else
								{
									GStringPipe Buf;
									Buf.Printf(	"The file:\n"
												"\n"
												"\t%s\n"
												"\n"
												"doesn't exist.",
												p);
									char *Msg = Buf.NewStr();
									if (Msg)
									{
										GAlert a(Project->GetApp(), "Missing File", Msg, "Skip", "Delete", "Browse...");
										switch (a.DoModal())
										{
											case 1: // Skip
											{
												break;
											}
											case 2: // Delete
											{
												Project->SetDirty();
												delete this;
												return true;
												break;
											}
											case 3: // Browse
											{
												GFileSelect s;
												s.Parent(Project->GetApp());
												s.Type("Code", SourcePatterns);
												if (s.Open())
												{
													SetFileName(s.Name());
												}
												break;
											}
										}
										
										DeleteArray(Msg);
									}
								}
							}
						}
						else
						{
							printf("%s:%i - Project::GetBasePath failed.\n", __FILE__, __LINE__);
						}
					}
					
					DeleteArray(p);
				}
			}
		}

		return true;
	}
	
	char *GetFullPath()
	{
		char *FullPath = 0;
		
		if (File AND *File == '.')
		{
			// Relitive path
			char Path[256];
			if (Project->GetBasePath(Path))
			{
				LgiMakePath(Path, sizeof(Path), Path, File);
				FullPath = NewStr(Path);
			}
		}
		else
		{
			// Absolute path
			FullPath = NewStr(File);
		}
		
		return FullPath;
	}

	IdeDoc *Open()
	{
		IdeDoc *Doc = 0;
		
		switch (Type)
		{
			case NodeDir:
			{
				Expanded(true);
				break;
			}
			case NodeResources:
			{
				char *FullPath = GetFullPath();
				if (FullPath)
				{
					char Exe[256];
					LgiGetExePath(Exe, sizeof(Exe));
					LgiTrimDir(Exe);
					#if defined WIN32
					LgiMakePath(Exe, sizeof(Exe), Exe, "Debug\\LgiRes.exe");
					#elif defined LINUX
					LgiMakePath(Exe, sizeof(Exe), Exe, "LgiRes/lgires");
					#endif
					
					if (FileExists(Exe))
					{
						LgiExecute(Exe, FullPath);
					}
				}
				else
				{
					LgiMsg(Tree, "No Path", AppName);
				}
				break;
			}
			case NodeGraphic:
			{
				char *FullPath = GetFullPath();
				if (FullPath)
				{
					LgiExecute(FullPath);
				}
				else
				{
					LgiMsg(Tree, "No Path", AppName);
				}
				break;
			}
			default:
			{
				char *FullPath = GetFullPath();
				if (FullPath)
				{
					Doc = Project->GetApp()->OpenFile(FullPath);
					if (Doc)
					{
						if (Doc AND Doc->GetEdit())
						{
							Doc->SetProject(Project);	
							Doc->GetEdit()->SetIndentSize(Project->d->Settings.IndentSize);
							Doc->GetEdit()->SetTabSize(Project->d->Settings.TabSize);
							Doc->GetEdit()->SetHardTabs(Project->d->Settings.UseTabs);
							Doc->GetEdit()->SetShowWhiteSpace(Project->d->Settings.ShowWhitespace);
							Doc->GetEdit()->Invalidate();
						}
					}
					else
					{
						LgiMsg(Tree, "Couldn't open file '%s'", AppName, MB_OK, FullPath);
					}
				}
				else
				{
					LgiMsg(Tree, "No Path", AppName);
				}
				break;
			}
		}
		
		return Doc;
	}
	
	void Delete()
	{
		Project->SetDirty();
		GXmlTag::RemoveTag();							
		delete this;
	}
	
	bool OnKey(GKey &k)
	{
		if (k.Down())
		{
			if (k.c16 == VK_RETURN)
			{
				Open();
				return true;
			}
			else if (k.c16 == VK_DELETE)
			{
				GTreeItem *s = GetNext();
				if (s OR (s = GetParent()))
				{
					s->Select(true);
				}
				
				Delete();
				return true;
			}
		}
		
		return false;
	}

	void OnMouseClick(GMouse &m)
	{
		GTreeItem::OnMouseClick(m);
		
		if (m.Left())
		{
			if
			(
				m.Double()
				AND
				(
					Type == NodeSource
					OR
					Type == NodeHeader
					OR
					Type == NodeResources
					OR
					Type == NodeGraphic
				)
				AND
				ValidStr(File)
			)
			{
				Open();
			}
		}
		else if (m.Right())
		{
			Select(true);

			GSubMenu *Sub = NEW(GSubMenu);
			if (Sub)
			{
				if (Type == NodeDir)
				{
					Sub->AppendItem("Insert File", IDM_INSERT, true);
					Sub->AppendItem("New Folder", IDM_NEW_FOLDER, true);
					Sub->AppendItem("Import Folder", IDM_IMPORT_FOLDER, true);
					Sub->AppendSeparator();
					Sub->AppendItem("Rename", IDM_RENAME, true);
				}
				Sub->AppendItem("Remove", IDM_DELETE, true);
				Sub->AppendItem("Sort", IDM_SORT_CHILDREN, true);
				Sub->AppendSeparator();
				Sub->AppendItem("Browse Folder", IDM_BROWSE_FOLDER, ValidStr(File));
				Sub->AppendItem("Open Terminal", IDM_OPEN_TERM, ValidStr(File));
				Sub->AppendItem("Properties", IDM_PROPERTIES, ValidStr(File));

				if (Tree->GetMouse(m, true))
				{
					switch (Sub->Float(Tree, m.x, m.y))
					{
						case IDM_INSERT:
						{
							GFileSelect s;
							s.Parent(Tree);
							s.Type("Source", SourcePatterns);
							s.Type("Makefiles", "*makefile");
							s.Type("All Files", LGI_ALL_FILES);
							s.MultiSelect(true);

							char Dir[256];
							if (Project->GetBasePath(Dir))
							{
								s.InitialDir(Dir);
							}

							if (s.Open())
							{
								for (int i=0; i<s.Length(); i++)
								{
									if (NOT Project->InProject(s[i], false))
									{
										ProjectNode *New = NEW(ProjectNode(Project));
										if (New)
										{
											char Rel[256];
											if (Project->RelitivePath(Rel, s[i]))
											{
												New->SetFileName(Rel);
											}
											else
											{
												New->SetFileName(s[i]);
											}
											InsertTag(New);
											SortChildren();
											Project->SetDirty();
										}
									}
									else
									{
										LgiMsg(Tree, "'%s' is already in the project.\n", AppName, MB_OK, s[i]);
									}
								}
							}
							break;
						}
						case IDM_IMPORT_FOLDER:
						{
							GFileSelect s;
							s.Parent(Tree);

							char Dir[256];
							if (Project->GetBasePath(Dir))
							{
								s.InitialDir(Dir);
							}

							if (s.OpenFolder())
							{
								List<char> Ext, Files;
								GToken e(SourcePatterns, ";");
								for (int i=0; i<e.Length(); i++)
								{
									Ext.Insert(e[i]);
								}
								
								if (LgiRecursiveFileSearch(s.Name(), &Ext, &Files))
								{
									int Start = strlen(s.Name()) + 1;
									for (char *f=Files.First(); f; f=Files.Next())
									{
										GToken p(f + Start, DIR_STR);
										ProjectNode *Insert = this;
										
										// Find sub nodes, and drill into directory heirarchy,
										// creating the nodes if needed.
										for (int i=0; Insert AND i<p.Length()-1; i++)
										{
											// Find existing node...
											bool Found = false;
											for (ProjectNode *c=Insert->ChildNode(); c; c=c->NextNode())
											{
												if (c->GetType() == NodeDir AND
													c->GetName() AND
													stricmp(c->GetName(), p[i]) == 0)
												{
													Insert = c;
													Found = true;
													break;
												}
											}
											
											if (NOT Found)
											{
												// Create the node
												IdeCommon *Com = Insert->GetSubFolder(Project, p[i], true);												
												Insert = dynamic_cast<ProjectNode*>(Com);
												LgiAssert(Insert);
											}
										}
										
										// Insert the file into the tree...
										if (Insert)
										{
											ProjectNode *New = NEW(ProjectNode(Project));
											if (New)
											{
												char Rel[256];
												if (Project->RelitivePath(Rel, f))
												{
													New->SetFileName(Rel);
												}
												else
												{
													New->SetFileName(f);
												}
												Insert->InsertTag(New);
												Insert->SortChildren();
												Project->SetDirty();
											}
										}
									}
									
									Files.DeleteArrays();
								}														
							}
							break;
						}
						case IDM_SORT_CHILDREN:
						{
							SortChildren();
							Project->SetDirty();
							break;
						}
						case IDM_NEW_FOLDER:
						{
							GInput Name(Tree, "", "Name:", AppName);
							if (Name.DoModal())
							{
								GetSubFolder(Project, Name.Str, true);
							}
							break;
						}
						case IDM_RENAME:
						{
							GInput Name(Tree, "", "Name:", AppName);
							if (Name.DoModal())
							{
								SetName(Name.Str);
								Project->SetDirty();
								Update();
							}
							break;
						}
						case IDM_DELETE:
						{
							Delete();
							return;
							break;
						}
						case IDM_BROWSE_FOLDER:
						{
							char *Path = GetFullPath();
							if (Path)
							{
								LgiTrimDir(Path);
								LgiExecute(Path);
								DeleteArray(Path);
							}
							break;
						}
						case IDM_OPEN_TERM:
						{
							if (Type == NodeDir)
							{
							}
							else
							{
								char *Path = GetFullPath();
								if (Path)
								{
									LgiTrimDir(Path);
									
									#if defined LINUX
									char s[256] = "--workdir ";
									char *o = s + strlen(s), *i = Path;
									while (*i)
									{
										if (*i == ' ')
										{
											*o++ = '\\';
										}
										*o++ = *i++;
									}
									*o++ = 0;									
									LgiExecute("konsole", s);
									#elif defined WIN32
									LgiExecute("cmd", 0, Path);
									#endif
									
									DeleteArray(Path);
								}
							}
							break;
						}
						case IDM_PROPERTIES:
						{
							if (Type == NodeDir)
							{
							}
							else
							{
								char *Path = GetFullPath();
								if (Path)
								{
									char Size[32];
									int64 FSize = LgiFileSize(Path);
									LgiFormatSize(Size, FSize);
									char Msg[512];
									sprintf(Msg, "Source Code:\n\n\t%s\n\nSize: %s (%i bytes)", Path, Size, (int32)FSize);
								
									FileProps Dlg(Tree, Msg, Type, Platforms);
									switch (Dlg.DoModal())
									{
										case IDC_COPY_PATH:
										{
											GClipBoard Clip(Tree);
											Clip.Text(Path);
											break;
										}
									}
									if (Type != Dlg.Type)
									{
										Type = Dlg.Type;
										Project->SetDirty();
									}
									if (Platforms != Dlg.Platforms)
									{
										Platforms = Dlg.Platforms;
										Project->SetDirty();
									}									
									DeleteArray(Path);
									
									Update();
								}
							}
							break;
						}
					}
				}
				DeleteObj(Sub);
			}
		}
	}
	
	ProjectNode *FindFile(char *In, char **Full)
	{
		if (File)
		{
			bool Match = false;
			if (strchr(In, DIR_CHAR))
			{
				// Match full path
				if (File[0] == '.')
				{
					char d[256];
					if (Project->GetBasePath(d))
					{
						char f[256];
						LgiMakePath(f, sizeof(f), d, File);
						
						Match = stricmp(f, In) == 0;
					}
					else
					{
						printf("%s,%i - Couldn't get full IDoc path.\n", __FILE__, __LINE__);
					}
				}
				else
				{
					Match = stricmp(File, In) == 0;
				}
			}
			else
			{
				// Match file name only
				char *Dir = strrchr(File, DIR_CHAR);
				if (Dir)
				{
					Match = stricmp(Dir + 1, In) == 0;
				}
			}
			
			if (Match)
			{
				if (Full)
				{
					if (File[0] == '.')
					{
						char d[256];
						if (Project->GetBasePath(d))
						{
							char f[256];
							LgiMakePath(f, sizeof(f), d, File);
							*Full = NewStr(f);
						}
					}
					else
					{
						*Full = NewStr(File);
					}
				}
				
				return this;
			}
		}
		
		ForAllProjectNodes(c)
		{
			ProjectNode *n = c->FindFile(In, Full);
			if (n)
			{
				return n;
			}
		}
		
		return 0;
	}
};

//////////////////////////////////////////////////////////////////////////////////
int NodeSort(GTreeItem *a, GTreeItem *b, int d)
{
	ProjectNode *A = dynamic_cast<ProjectNode*>(a);
	ProjectNode *B = dynamic_cast<ProjectNode*>(b);
	if (A AND B)
	{
		if
		(
			(A->GetType() == NodeDir)
			^
			(B->GetType() == NodeDir)
		)
		{
			return A->GetType() == NodeDir ? -1 : 1;
		}
		else
		{
			char *Sa = a->GetText(0);
			char *Sb = b->GetText(0);
			if (Sa AND Sb)
			{
				return stricmp(Sa, Sb);
			}
		}
	}

	return 0;
}

int XmlSort(GXmlTag *a, GXmlTag *b, int d)
{
	GTreeItem *A = dynamic_cast<GTreeItem*>(a);
	GTreeItem *B = dynamic_cast<GTreeItem*>(b);
	return NodeSort(A, B, d);
}

IdeCommon::~IdeCommon()
{
	Remove();
}

void IdeCommon::SortChildren()
{
	Items.Sort(NodeSort, 0);
	Children.Sort(XmlSort, 0);
	
	if (Tree)
	{
		_RePour();
		Tree->Invalidate();
	}
}

void IdeCommon::InsertTag(GXmlTag *t)
{
	GXmlTag::InsertTag(t);

	GTreeItem *i = dynamic_cast<GTreeItem*>(t);
	if (i)
	{
		Insert(i);
	}
}

void IdeCommon::RemoveTag()
{
	GXmlTag::RemoveTag();
	Detach();
}

IdeCommon *IdeCommon::GetSubFolder(IdeProject *Project, char *Name, bool Create)
{
	if (Name)
	{
		ForAllProjectNodes(c)
		{
			if (c->GetType() == NodeDir)
			{
				if (c->GetName() AND stricmp(c->GetName(), Name) == 0)
				{
					return c;
				}
			}
		}

		ProjectNode *New = NEW(ProjectNode(Project));
		if (New)
		{
			New->SetName(Name);
			InsertTag(New);
			Project->SetDirty();
			return New;
		}
	}

	return 0;
}

//////////////////////////////////////////////////////////////////////////////////
class BuildThread : public GThread, public GStream
{
	IdeProject *Proj;
	BuildThread **Me;
	char *Makefile;
	char *Args;

public:
	BuildThread(IdeProject *proj, BuildThread **ptr, char *mf, char *args = 0)
	{
		Proj = proj;
		*(Me = ptr) = this;
		Makefile = NewStr(mf);
		DeleteOnExit = true;
		Args = NewStr(args);
		
		if (Proj->GetApp())
		{
			Proj->GetApp()->PostEvent(M_APPEND_TEXT, 0, 0);
		}
		Run();
	}
	
	~BuildThread()
	{
		DeleteArray(Makefile);
		DeleteArray(Args);
	}
	
	int Write(void *Buffer, int Size, int Flags = 0)
	{
		if (Proj->GetApp())
		{
			Proj->GetApp()->PostEvent(M_APPEND_TEXT, (int)NewStr((char*)Buffer, Size), 0);
		}
		return Size;
	}

	char *FindExe()
	{
		char Exe[256] = "";
		GToken p(getenv("PATH"), LGI_PATH_SEPARATOR);
		
		switch (Proj->GetSettings()->Compiler)
		{
			case 0:
			{
				for (int i=0; i<p.Length(); i++)
				{
					LgiMakePath
					(
						Exe,
						sizeof(Exe),
						p[i],
						"make"
						#ifdef WIN32
						".exe"
						#endif
					);
					if (FileExists(Exe))
					{
						return NewStr(Exe);
					}
				}
				break;
			}
			case 1:
			{
				for (int i=0; i<p.Length(); i++)
				{
					char Path[256];
					LgiMakePath(Path, sizeof(Path), p[i], "msdev.exe");
					if (FileExists(Path))
					{
						return NewStr(Path);
					}
				}
				break;
			}
		}
		
		return 0;
	}
	
	int Main()
	{
		GProcess Make;
		char *Err = 0;
		
		char *Exe = FindExe();
		if (Exe)
		{
			bool Status = false;

			switch (Proj->GetSettings()->Compiler)
			{
				case 1:
				{
					char a[256];
					sprintf(a, "\"%s\" /make \"All - Win32 Debug\"", Makefile);
					Status = Make.Run(Exe, a, 0, true, 0, this);
					break;
				}
				case 0:
				{
					char *MakePath = NewStr(Makefile);
					if (MakePath)
					{
						LgiTrimDir(MakePath);

						GStringPipe a;
						a.Print("-f %s", Makefile);
						if (Args)
							a.Print(" %s", Args);
						char *Temp = a.NewStr();

						Status = Make.Run(Exe, Temp, MakePath, true, 0, this);

						DeleteArray(Temp);
						DeleteArray(MakePath);
					}
					break;
				}
			}

			if (NOT Status)
			{
				Err = "Running make failed";
				LgiTrace("%s,%i - %s.\n", __FILE__, __LINE__, Err);
			}
		}
		else
		{
			Err = "Couldn't find 'make'";
			LgiTrace("%s,%i - %s.\n", __FILE__, __LINE__, Err);
		}
	
		if (Proj->GetApp())
		{
			Proj->GetApp()->UpdateState(-1, false);
			if (Err)
			{
				Proj->GetApp()->PostEvent(M_BUILD_ERR, 0, (int)NewStr(Err));
			}
		}
		
		*Me = 0;
		return 0;
	}
};

//////////////////////////////////////////////////////////////////////////////////
IdeProject::IdeProject(AppWnd *App)
{
	d = NEW(IdeProjectPrivate(App));
	Tag = NewStr("Project");
}

IdeProject::~IdeProject()
{
	d->App->OnProjectDestroy(this);
	DeleteObj(d);
}

char *IdeProject::GetFileComment()
{
	return d->Settings.CommentFile;
}

char *IdeProject::GetFunctionComment()
{
	return d->Settings.CommentFunc;
}

IdeProject *IdeProject::GetParentProject()
{
	return d->ParentProject;
}

void IdeProject::SetParentProject(IdeProject *p)
{
	d->ParentProject = p;
}

void IdeCommon::CollectAllSubProjects(List<IdeProject> &c)
{
	ForAllProjectNodes(p)
	{
		if (p->GetType() == NodeDependancy AND
			p->GetDep())
		{
			c.Insert(p->GetDep());
		}
		
		p->CollectAllSubProjects(c);
	}
}

bool IdeProject::GetChildProjects(List<IdeProject> &c)
{
	CollectAllSubProjects(c);
	return c.First() != 0;
}

bool IdeProject::RelitivePath(char *Out, char *In)
{
	if (Out AND In)
	{
		char Base[256];
		if (GetBasePath(Base))
		{
			// printf(	"XmlBase='%s'\n     In='%s'\n", Base, In);
			
			GToken b(Base, DIR_STR);
			GToken i(In, DIR_STR);
			
			// printf("Len %i-%i\n", b.Length(), i.Length());
			
			int ILen = i.Length() + (DirExists(In) ? 0 : 1);
			int Max = min(b.Length(), ILen);
			int Common = 0;
			for (; Common < Max; Common++)
			{
				#ifdef WIN32
				#define StrCompare stricmp
				#else
				#define StrCompare strcmp
				#endif
				
				// printf("Cmd '%s'-'%s'\n", b[Common], i[Common]);
				if (StrCompare(b[Common], i[Common]) != 0)
				{
					break;
				}
			}
			
			// printf("Common=%i\n", Common);
			if (Common > 0)
			{
				if (Common < b.Length())
				{
					Out[0] = 0;
					int Back = b.Length() - Common;
					// printf("Back=%i\n", Back);
					for (int n=0; n<Back; n++)
					{
						strcat(Out, "..");
						if (n < Back - 1) strcat(Out, DIR_STR);
					}
				}
				else
				{
					strcpy(Out, ".");
				}
				for (int n=Common; n<i.Length(); n++)
				{
					sprintf(Out+strlen(Out), "%s%s", DIR_STR, i[n]);
				}
				
				return true;
			}	
			else
			{
				strcpy(Out, In);
				return true;
			}		
		}
	}
	
	return false;
}

bool IdeProject::GetExePath(char *Path, int Len)
{
	if (d->Settings.Exe)
	{
		if (d->Settings.Exe[0] == '.')
		{
			char b[256];
			if (GetBasePath(b))
			{
				LgiMakePath(Path, Len, b, d->Settings.Exe);
			}
			else return false;
		}
		else
		{
			strsafecpy(Path, d->Settings.Exe, Len);
		}
		
		return true;
	}
	else return false;
}

bool IdeProject::GetMakefile(char *Path, int Len)
{
	if (d->Settings.MakeFile)
	{
		if (d->Settings.MakeFile[0] == '.')
		{
			char b[256];
			if (GetBasePath(b))
			{
				LgiMakePath(Path, Len, b, d->Settings.MakeFile);
			}
			else return false;
		}
		else
		{
			strsafecpy(Path, d->Settings.MakeFile, Len);
		}
		
		return true;
	}
	else return false;
}

void IdeProject::Clean()
{
	if (NOT d->Build AND
		d->Settings.MakeFile)
	{
		char m[256];
		if (GetMakefile(m, sizeof(m)))
		{		
			NEW(BuildThread(this, &d->Build, m, "clean"));
		}
	}
}

char *QuoteStr(char *s)
{
	GStringPipe p(256);
	while (s AND *s)
	{
		if (*s == ' ')
		{
			p.Push("\\ ", 2);
		}
		else p.Push(s, 1);
		s++;
	}
	return p.NewStr();
}

class ExecuteThread : public GThread, public GStream
{
	IdeProject *Proj;
	char *Exe, *Args, *Path;
	int Len;
	ExeAction Act;

public:
	ExecuteThread(IdeProject *proj, char *exe, char *args, char *path, ExeAction act)
	{
		Len = 32 << 10;
		Proj = proj;
		Act = act;
		Exe = NewStr(exe);
		Args = NewStr(args);
		Path = NewStr(path);
		DeleteOnExit = true;
		Run();
	}
	
	~ExecuteThread()
	{
		DeleteArray(Exe);
		DeleteArray(Args);
		DeleteArray(Path);
	}
	
	int Main()
	{
		if (Proj->GetApp())
		{
			Proj->GetApp()->PostEvent(M_APPEND_TEXT, 0, 1);
		}
		
		if (Exe)
		{
			GProcess p;
			if (Act == ExeDebug)
			{
				char *a = QuoteStr(Exe);
				char *b = QuoteStr(Path);				
				
				p.Run("kdbg", a, b, true, 0, this);
				
				DeleteArray(a);
				DeleteArray(b);
			}
			else if (Act == ExeValgrind)
			{
				if (Proj->GetExecutable())
				{
					char Path[256];
					strsafecpy(Path, Exe, sizeof(Path));
					LgiTrimDir(Path);
					
					char *e = QuoteStr(Proj->GetExecutable());
					char *p = QuoteStr(Path);
					char *a = Proj->GetExeArgs() ? Proj->GetExeArgs() : (char*)"";
					char Args[512];
					sprintf(Args,
							"--workdir %s "
							"--noclose "
							"-e valgrind --tool=memcheck --num-callers=20 %s %s", p, e, a);
					
					LgiExecute("konsole", Args);
				}
			}
			else
			{
				p.Run(Exe, Args, Path, true, 0, this);
			}
		}
		
		return 0;
	}

	int Write(void *Buffer, int Size, int Flags = 0)
	{
		if (Len > 0)
		{
			if (Proj->GetApp())
			{
				Size = min(Size, Len);
				Proj->GetApp()->PostEvent(M_APPEND_TEXT, (int)NewStr((char*)Buffer, Size), 1);
				Len -= Size;
			}
			return Size;
		}
		
		return 0;
	}
};

void IdeProject::Execute(ExeAction Act)
{
	char b[256];
	if (d->Settings.Exe AND
		GetBasePath(b))
	{
		char e[256];
		if (GetExePath(e, sizeof(e)))
		{
			if (FileExists(e))
			{
				NEW(ExecuteThread(this, e, d->Settings.ExeArgs, b, Act));
			}
			else
			{
				LgiMsg(Tree, "Executable file doesn't exist.\n", AppName);
			}
		}
	}
}

void IdeProject::Build(bool All)
{
	if (NOT d->Build)
	{
		char m[256];
		if (GetMakefile(m, sizeof(m)))
		{		
			NEW(BuildThread(this, &d->Build, m));
		}
	}
}

bool IdeProject::Serialize()
{
	SerializeAttr("Makefile", d->Settings.MakeFile);
	SerializeAttr("Exe", d->Settings.Exe);
	SerializeAttr("Args", d->Settings.ExeArgs);
	SerializeAttr("Defs", d->Settings.DefineSym);
	SerializeAttr("TargetType", d->Settings.TargetType);
	SerializeAttr("TargetName", d->Settings.TargetName);
	SerializeAttr("IncludePaths", d->Settings.IncludePaths);
	SerializeAttr("LibPaths", d->Settings.LibPaths);
	SerializeAttr("Libs", d->Settings.Libs);
	SerializeAttr("FileComment", d->Settings.CommentFile);
	SerializeAttr("FuncComment", d->Settings.CommentFunc);
	SerializeAttr("MakefileRules", d->Settings.MakefileRules);
	SerializeAttr("TabSize", d->Settings.TabSize);
	SerializeAttr("IndentSize", d->Settings.IndentSize);
	SerializeAttr("UseTabs", d->Settings.UseTabs);
	SerializeAttr("ShowWhitespace", d->Settings.ShowWhitespace);
	SerializeAttr("Compiler", d->Settings.Compiler);
	
	return true;
}

bool IdeProject::GetBasePath(char *Path)
{
	if (Path AND d->FileName)
	{
		strcpy(Path, d->FileName);
		char *e = strrchr(Path, DIR_CHAR);
		if (e) *e = 0;
		return true;
	}
	
	return false;
}

AppWnd *IdeProject::GetApp()
{
	return d->App;
}

char *IdeProject::GetIncludePaths()
{
	return d->Settings.IncludePaths;
}

char *IdeProject::GetExeArgs()
{
	return d->Settings.ExeArgs;
}

char *IdeProject::GetExecutable()
{
	return d->Settings.Exe;
}

char *IdeProject::GetFileName()
{
	return d->FileName;
}

void IdeProject::CreateProject()
{
	Empty();
	
	DeleteArray(d->FileName);
	d->App->GetTree()->Insert(this);
	
	ProjectNode *f = NEW(ProjectNode(this));
	if (f)
	{
		f->SetName("Source");
		f->SetType(NodeDir);
		InsertTag(f);
	}
	f = NEW(ProjectNode(this));
	if (f)
	{
		f->SetName("Headers");
		f->SetType(NodeDir);
		InsertTag(f);
	}
	d->Dirty = true;
	Expanded(true);	
}

bool IdeProject::OpenFile(char *FileName)
{
	bool Status = false;
	
	Empty();
	DeleteArray(d->FileName);
	d->FileName = NewStr(FileName);
	if (d->FileName)
	{
		GFile f;
		if (f.Open(d->FileName, O_READWRITE))
		{
			GXmlTree x;
			if (Status = x.Read(this, &f, this))
			{
				d->App->GetTree()->Insert(this);
				Expanded(true);
			}
			else
			{
				LgiMsg(Tree, x.GetErrorMsg(), AppName);
			}
		}
	}

	return Status;
}

bool IdeProject::SaveFile(char *FileName)
{
	char *File = FileName ? FileName : d->FileName;
	if (ValidStr(File))
	{
		GFile f;
		
		if (f.Open(File, O_WRITE))
		{
			GXmlTree x;
			if (x.Write(this, &f))
			{
				d->Dirty = false;
				return true;
			}
		}
	}

	return false;
}

void IdeProject::SetDirty()
{
	d->Dirty = true;
}

void IdeProject::SetClean()
{
	if (d->Dirty)
	{
		if (NOT ValidStr(d->FileName))
		{
			GFileSelect s;
			s.Parent(Tree);
			s.Name("Project.xml");
			if (s.Save())
			{
				d->FileName = NewStr(s.Name());
				d->App->OnFile(d->FileName, true);
				Update();
			}
		}

		SaveFile(0);
	}

	ForAllProjectNodes(p)
	{
		p->SetClean();
	}
}

char *IdeProject::GetText(int Col)
{
	if (d->FileName)
	{
		char *s = strrchr(d->FileName, DIR_CHAR);
		return s ? s + 1 : d->FileName;
	}

	return Untitled;
}

int IdeProject::GetImage(int Flags)
{
	return 0;
}

void IdeProject::Empty()
{
	GXmlTag *t;
	while (t = Children.First())
	{
		ProjectNode *n = dynamic_cast<ProjectNode*>(t);
		if (n)
		{
			n->Remove();
		}
		DeleteObj(t);
	}
}

GXmlTag *IdeProject::Create(char *Tag)
{
	return NEW(ProjectNode(this));
}

class SettingsDlg : public GDialog
{
	IdeProject *Project;
	
public:
	ProjectSettings Settings;

	SettingsDlg(GView *p,
				IdeProject *project,
				ProjectSettings &s)
	{
		Settings = s;
		Project = project;
		SetParent(p);
		if (LoadFromResource(IDD_PROJECT_SETTINGS))
		{
			GCombo *c = dynamic_cast<GCombo*>(FindControl(IDC_TARGET));
			if (c)
			{
				c->Insert("Executable");
				c->Insert("Shared Library");
				c->Insert("Static Library");
			}
			c = dynamic_cast<GCombo*>(FindControl(IDC_COMPILER));
			if (c)
			{
				c->Insert("GCC");
				c->Insert("MSVC");
			}

			SetCtrlName(IDC_MAKEFILE, Settings.MakeFile);
			SetCtrlName(IDC_EXE, Settings.Exe);
			SetCtrlName(IDC_ARGS, Settings.ExeArgs);
			SetCtrlName(IDC_DEFS, Settings.DefineSym);
			SetCtrlValue(IDC_COMPILER, Settings.Compiler);
			SetCtrlValue(IDC_TARGET, Settings.TargetType);
			SetCtrlName(IDC_TARGET_NAME, Settings.TargetName);
			SetCtrlName(IDC_INCLUDE, Settings.IncludePaths);
			SetCtrlName(IDC_LIB_PATHS, Settings.LibPaths);
			SetCtrlName(IDC_LIBS, Settings.Libs);
			SetCtrlName(IDC_FILE_COMMENT, Settings.CommentFile);
			SetCtrlName(IDC_RULES, Settings.MakefileRules);
			SetCtrlName(IDC_FUNCTION_COMMENT, Settings.CommentFunc);
			SetCtrlValue(IDC_TAB_SIZE, Settings.TabSize);
			SetCtrlValue(IDC_INDENT_SIZE, Settings.IndentSize);
			SetCtrlValue(IDC_USE_TABS, Settings.UseTabs);
			SetCtrlValue(IDC_SHOW_WHITESPACE, Settings.ShowWhitespace);
			
			MoveToCenter();
		}
	}
	
	int OnNotify(GViewI *c, int f)
	{
		switch (c->GetId())
		{
			case IDC_SET_MAKEFILE:
			{
				GFileSelect s;
				s.Parent(this);
				#ifdef WIN32
				s.Type("Developer Studio Projects", "*.dsp");
				#else
				s.Type("Makefiles", "*makefile");
				#endif
				s.Type("All Files", LGI_ALL_FILES);
				if (s.Open())
				{
					char Rel[256];
					if (Project->RelitivePath(Rel, s.Name()))
					{
						SetCtrlName(IDC_MAKEFILE, Rel);
					}
					else
					{
						SetCtrlName(IDC_MAKEFILE, s.Name());
					}
				}
				break;
			}
			case IDC_SET_EXE:
			{
				GFileSelect s;
				s.Parent(this);
				s.Type("All Files", LGI_ALL_FILES);
				if (s.Open())
				{
					char Rel[256];
					if (Project->RelitivePath(Rel, s.Name()))
					{
						SetCtrlName(IDC_EXE, Rel);
					}
					else
					{
						SetCtrlName(IDC_EXE, s.Name());
					}
				}
				break;
			}
			case IDC_SET_ARGS:
			{
				GFileSelect s;
				s.Parent(this);
				s.Type("All Files", LGI_ALL_FILES);
				if (s.Open())
				{
					SetCtrlName(IDC_ARGS, s.Name());
				}
				break;
			}
			case IDC_ADD_INC:
			{
				GFileSelect s;
				s.Parent(this);
				if (s.OpenFolder())
				{
					GToken Inc(GetCtrlName(IDC_INCLUDE), ",;\r\n");
					bool Has = false;
					for (int i=0; i<Inc.Length(); i++)
					{
						if (stricmp(Inc[i], s.Name()) == 0)
						{
							Has = true;
							break;
						}
					}
					if (NOT Has)
					{
						char *Insert = 0;
						char Out[256];
						if (Project->RelitivePath(Out, s.Name()))
						{
							Insert = Out;
						}
						else
						{
							Insert = s.Name();
						}
						
						GStringPipe Buf;
						for (int i=0; i<Inc.Length(); i++)
						{
							Buf.Printf("%s\n", Inc[i]);
						}
						Buf.Printf("%s\n", Insert);
						
						char *Paths = Buf.NewStr();
						if (Paths)
						{
							SetCtrlName(IDC_INCLUDE, Paths);
							DeleteArray(Paths);
						}
					}									
				}
				break;
			}
			case IDOK:
			{
				#define SetStr(s, c) \
				{ \
					DeleteArray(s); \
					char *i = GetCtrlName(c); \
					if (ValidStr(i)) \
						s = NewStr(i); \
				}

				SetStr(Settings.MakeFile, IDC_MAKEFILE);
				SetStr(Settings.Exe,IDC_EXE);
				SetStr(Settings.ExeArgs, IDC_ARGS);
				SetStr(Settings.DefineSym, IDC_DEFS);
				Settings.TargetType = GetCtrlValue(IDC_TARGET);
				SetStr(Settings.TargetName, IDC_TARGET_NAME);
				SetStr(Settings.IncludePaths, IDC_INCLUDE);
				SetStr(Settings.LibPaths, IDC_LIB_PATHS);
				SetStr(Settings.Libs, IDC_LIBS);
				SetStr(Settings.MakefileRules, IDC_RULES);
				SetStr(Settings.CommentFile, IDC_FILE_COMMENT);
				SetStr(Settings.CommentFunc, IDC_FUNCTION_COMMENT);
				Settings.TabSize = GetCtrlValue(IDC_TAB_SIZE);
				Settings.IndentSize = GetCtrlValue(IDC_INDENT_SIZE);
				Settings.UseTabs = GetCtrlValue(IDC_USE_TABS);
				Settings.ShowWhitespace = GetCtrlValue(IDC_SHOW_WHITESPACE);
				Settings.Compiler = GetCtrlValue(IDC_COMPILER);
				
				// fall thru
			}
			case IDCANCEL:
			{
				EndModal(c->GetId() == IDOK);
				break;
			}
		}
		
		return 0;
	}
};

void IdeProject::OnMouseClick(GMouse &m)
{
	if (m.Right() AND m.Down())
	{
		GSubMenu *Sub = NEW(GSubMenu);
		if (Sub)
		{
			Sub->AppendItem("New Folder", IDM_NEW_FOLDER, true);
			Sub->AppendSeparator();
			Sub->AppendItem("Settings", IDM_SETTINGS, true);
			Sub->AppendItem("Insert Dependancy", IDM_INSERT_DEP, true);

			if (Tree->GetMouse(m, true))
			{
				switch (Sub->Float(Tree, m.x, m.y))
				{
					case IDM_NEW_FOLDER:
					{
						GInput Name(Tree, "", "Name:", AppName);
						if (Name.DoModal())
						{
							GetSubFolder(this, Name.Str, true);
						}
						break;
					}
					case IDM_SETTINGS:
					{
						SettingsDlg Dlg(Tree,
										this,
										d->Settings);
						if (Dlg.DoModal())
						{
							d->Settings = Dlg.Settings;
							SetDirty();
						}
						break;
					}
					case IDM_INSERT_DEP:
					{
						GFileSelect s;
						s.Parent(Tree);
						s.Type("Project", "*.xml");
						if (s.Open())
						{
							ProjectNode *New = NEW(ProjectNode(this));
							if (New)
							{
								New->SetFileName(s.Name());
								New->SetType(NodeDependancy);
								InsertTag(New);
								SetDirty();
							}
						}
						break;
					}
				}
			}

			DeleteObj(Sub);
		}
	}
}

char *IdeProject::FindFullPath(char *File)
{
	char *Full = 0;
	
	ForAllProjectNodes(c)
	{
		if (c->FindFile(File, &Full))
		{
			break;
		}
	}
	
	return Full;		
}

bool IdeProject::InProject(char *FullPath, bool Open, IdeDoc **Doc)
{
	ProjectNode *n = 0;

	ForAllProjectNodes(c)
	{
		if (n = c->FindFile(FullPath, 0))
		{
			break;
		}
	}
	
	if (n AND Open)
	{
		IdeDoc *d = n->Open();
		if (d)
		{
			d->SetProject(this);
			if (Doc) *Doc = d;
		}
	}
	
	return n != 0;
}

char *GetQuotedStr(char *Str)
{
	char *s = strchr(Str, '\"');
	if (s)
	{
		s++;
		char *e = strchr(s, '\"');
		if (e)
		{
			return NewStr(s, (int)e-(int)s);
		}
	}

	return 0;
}

void IdeProject::ImportDsp(char *File)
{
	if (File AND FileExists(File))
	{
		char Base[256];
		strcpy(Base, File);
		LgiTrimDir(Base);
		
		char *Dsp = ReadTextFile(File);
		if (Dsp)
		{
			GToken Lines(Dsp, "\r\n");
			IdeCommon *Current = this;
			bool IsSource = false;
			for (int i=0; i<Lines.Length(); i++)
			{
				char *L = Lines[i];
				if (strnicmp(L, "# Begin Group", 13) == 0)
				{
					char *Folder = GetQuotedStr(L);
					if (Folder)
					{
						IdeCommon *Sub = Current->GetSubFolder(this, Folder, true);
						if (Sub)
						{
							Current = Sub;
						}
						DeleteArray(Folder);
					}
				}
				else if (strnicmp(L, "# End Group", 11) == 0)
				{
					IdeCommon *Parent = dynamic_cast<IdeCommon*>(Current->GetParent());
					if (Parent)
					{
						Current = Parent;
					}
				}
				else if (strnicmp(L, "# Begin Source", 14) == 0)
				{
					IsSource = true;
				}
				else if (strnicmp(L, "# End Source", 12) == 0)
				{
					IsSource = false;
				}
				else if (Current AND IsSource AND strncmp(L, "SOURCE=", 7) == 0)
				{
					ProjectNode *New = NEW(ProjectNode(this));
					if (New)
					{
						char *Src = 0;
						if (strchr(L, '\"'))
						{
							Src = GetQuotedStr(L);
						}
						else
						{
							Src = NewStr(L + 7);
						}						
						if (Src)
						{
							// Make absolute path
							char Abs[256];
							LgiMakePath(Abs, sizeof(Abs), Base, ToUnixPath(Src));
							
							// Make relitive path
							char Rel[256];
							if (RelitivePath(Rel, Abs))
							{
								New->SetFileName(Rel);
							}
							else
							{
								New->SetFileName(Src);
							}
							DeleteArray(Src);
						}

						Current->InsertTag(New);
						SetDirty();
					}
				}
			}
			
			DeleteArray(Dsp);
		}
	}
}

ProjectSettings *IdeProject::GetSettings()
{
	return &d->Settings;
}

bool IdeProject::BuildIncludePaths(List<char> &Paths, bool Recurse)
{
	List<IdeProject> Projects;
	if (Recurse)
	{
		GetChildProjects(Projects);
	}
	Projects.Insert(this, 0);
	
	for (IdeProject *p=Projects.First(); p; p=Projects.Next())
	{
		char *IncludePaths = p->GetIncludePaths();
		if (IncludePaths)
		{
			GToken Inc(ToNativePath(IncludePaths), (char*)",;\r\n");
			for (int i=0; i<=Inc.Length(); i++)
			{
				char *Full = 0, Buf[300];
				if (NOT Inc[i] OR Inc[i][0] == '.')
				{
					char Base[256];
					p->GetBasePath(Base);
					if (Inc[i])
						LgiMakePath(Buf, sizeof(Buf), Base, Inc[i]);
					else
						strsafecpy(Buf, Base, sizeof(Buf));
					Full = Buf;
				}
				else
				{
					Full = Inc[i];
				}
				
				bool Has = false;
				for (char *s = Paths.First(); s; s = Paths.Next())
				{
					if (stricmp(s, Full) == 0)
					{
						Has = true;
						break;
					}
				}
				
				if (NOT Has)
				{
					Paths.Insert(NewStr(Full));
				}
			}
		}
	}

	return true;
}

void IdeProjectPrivate::CollectAllFiles(GTreeNode *Base, List<ProjectNode> &Files, bool SubProjects, int Platform)
{
	for (GTreeItem *i = Base->GetChild(); i; i = i->GetNext())
	{
		IdeProject *Proj = dynamic_cast<IdeProject*>(i);
		if (Proj)
		{
			if (Proj->GetParentProject() AND NOT SubProjects)
			{
				continue;
			}
		}
		else
		{
			ProjectNode *p = dynamic_cast<ProjectNode*>(i);
			if (p)
			{
				if (p->GetType() == NodeSource OR
					p->GetType() == NodeHeader)
				{
					if (p->GetPlatforms() & Platform)
					{
						Files.Insert(p);
					}
				}
			}
		}

		CollectAllFiles(i, Files, SubProjects, Platform);
	}
}

bool IdeProject::GetTargetName(char *Target, int BufSize)
{
	if (ValidStr(d->Settings.TargetName))
	{
		// Take target name from the settings
		strsafecpy(Target, d->Settings.TargetName, BufSize);
	}
	else
	{
		char *s = strrchr(d->FileName, DIR_CHAR);
		if (s)
		{
			// Generate the target executable name
			strsafecpy(Target, s + 1, BufSize);
			s = strrchr(Target, '.');
			if (s) *s = 0;
			strlwr(Target);
			s = Target;
			for (char *i = Target; *i; i++)
			{
				if (*i != '.' AND *i != ' ')
				{
					*s++ = *i;
				}
			}
			*s = 0;
		}
		else return false;
	}
	
	return true;
}

bool IdeProject::GetTargetFile(char *Buf, int BufSize)
{
	bool Status = false;
	char t[256];
	if (GetTargetName(t, sizeof(t)))
	{
		switch (d->Settings.TargetType)
		{
			// Executable
			case TARGET_TYPE_EXE:
			{
				strsafecpy(Buf, t, BufSize);
				Status = true;
				break;
			}
			// Shared library
			case TARGET_TYPE_SHARED_LIB:
			{
				snprintf(Buf, BufSize, "lib%s.%s", t, LGI_LIBRARY_EXT);
				Status = true;
				break;
			}
			// Static library
			case TARGET_TYPE_STATIC_LIB:
			{
				snprintf(Buf, BufSize, "lib%s.%s", t, LGI_STATIC_LIBRARY_EXT);
				Status = true;
				break;
			}
		}
	}
	
	return Status;
}

int StrCmp(char *a, char *b, int d)
{
	return stricmp(a, b);
}

bool IdeProject::CreateMakefile()
{
	#if defined WIN32
	char *LinkerFlags = ",--disable-auto-import";
	#else
	char *LinkerFlags = ",-soname,$(TargetFile),-export-dynamic,-R.";
	#endif

	char MakeFile[256];
	strsafecpy(MakeFile, d->FileName, sizeof(MakeFile));
	LgiMakePath(MakeFile, sizeof(MakeFile), MakeFile, "../Makefile");
	
	GFile m;
	if (m.Open(MakeFile, O_WRITE))
	{
		m.SetSize(0);
		
		m.Print("#!/usr/bin/make\n"
				"#\n"
				"# This makefile generated by LgiIde\n"
				"# http://www.memecode.com/lgi.php\n"
				"#\n"
				"\n"
				".SILENT :\n"
				"\n"
				"CC = g++\n");
		
		char Target[256];
		if (GetTargetName(Target, sizeof(Target)))
		{
			m.Print("Target = %s\n", Target);

			// Output the build mode, gui, flags and some paths
			int BuildMode = d->App->GetBuildMode();
			char *BuildModeName = BuildMode ? (char*)"Release" : (char*)"Debug";
			m.Print("ifndef Build\n"
					"	Build = %s\n"
					"endif\n",
					BuildModeName);
			
			#ifdef WIN32
			char *Platform = "WIN32";
			#else
			char *Platform = "LINUX";
			#endif

			m.Print("Gui = X\n"
					"BuildDir = $(Build)$(Gui)\n"
					"\n"
					"Flags = -fPIC -w -fno-inline\n" // -fexceptions
					"Defs= -D%s -D_REENTRANT", Platform);

			#ifdef LINUX
			m.Print(" -D$(Gui)WIN");
			#endif
			
			if (d->Settings.DefineSym)
			{
				GToken Defs(d->Settings.DefineSym, " ;,");
				for (int i=0; i<Defs.Length(); i++)
				{
					m.Print(" -D%s", Defs[i]);
				}
			}
			
			m.Print("\n"
					"\n"
					"ifeq ($(Build),Debug)\n"
					"	Defs += -D_DEBUG\n"
					"	Flags += -g\n"
					"	Tag = d\n"
					"else\n"
					"	Flags += -s -Os\n"
					"endif\n"
					"\n");
			
			// Collect all dependencies, output their lib names and paths
			m.Print("# Libraries\n"
					"Libs = \\\n"
					#ifndef WIN32
					"			-static-libgcc \\\n"
					#else
					"			-lgcc \\\n"
					#endif
					"			-lpthread \\\n"
					"			-lstdc++");

			if (ValidStr(d->Settings.LibPaths))
			{
				GToken LibPaths(d->Settings.LibPaths, " \r\n");
				for (int i=0; i<LibPaths.Length(); i++)
				{
					m.Print(" \\\n			-L%s", LibPaths[i]);
				}
			}

			if (ValidStr(d->Settings.Libs))
			{
				GToken Libs(d->Settings.Libs, " \r\n");
				for (int i=0; i<Libs.Length(); i++)
				{
					m.Print(" \\\n			-l%s", Libs[i]);
				}
			}

			List<IdeProject> Deps;
			if (GetChildProjects(Deps))
			{
				for (IdeProject *d=Deps.First(); d; d=Deps.Next())
				{
					char t[256];
					if (d->GetTargetName(t, sizeof(t)))
					{
						m.Print(" \\\n			-l%s$(Tag)", t);

						if (d->GetBasePath(t))
						{
							m.Print(" \\\n			-L%s%s$(BuildDir)", t, DIR_STR);
						}
					}
				}
			}
			m.Print("\n\n");
			
			// Collect all files that require building
			List<ProjectNode> Files;
			d->CollectAllFiles
			(
				d->App->GetTree(),
				Files,
				false,
				#if defined WIN32
				PLATFORM_WIN32
				#elif defined LINUX
				PLATFORM_LINUX
				#elif defined MAC
				PLATFORM_MAC
				#else
				#error "Not impl"
				#endif
			);
			if (Files.First())
			{
				ProjectNode *n;

				// Do include paths
				GHashTable Inc;
				if (ValidStr(d->Settings.IncludePaths))
				{
					// Add settings include paths.
					GToken Paths(d->Settings.IncludePaths, "\r\n");
					for (int i=0; i<Paths.Length(); i++)
					{
						if (NOT Inc.Find(Paths[i]))
						{
							Inc.Add(Paths[i]);
						}
					}
				}
				
				// Add paths of headers
				for (n=Files.First(); n; n=Files.Next())
				{
					if (n->GetFileName())
					{
						char *e = LgiGetExtension(n->GetFileName());
						if (e AND stricmp(e, "h") == 0)
						{
							for (char *Dir=n->GetFileName(); *Dir; Dir++)
							{
								if (*Dir == '/' OR *Dir == '\\')
								{
									*Dir = DIR_CHAR;
								}
							}						

							char Path[256];
							strsafecpy(Path, n->GetFileName(), sizeof(Path));

							LgiTrimDir(Path);
						
							char Rel[256];
							if (NOT RelitivePath(Rel, Path))
							{
								strcpy(Rel, Path);
							}
							
							if (stricmp(Rel, ".") != 0)
							{
								if (NOT Inc.Find(Rel))
								{
									Inc.Add(Rel);
								}
							}
						}
					}
				}

				// Output include paths
				#if defined LINUX
				m.Print("# Includes\n"
						"Inc =		-I/usr/include/X11\\\n"
						"			-I/usr/include/X11/Xft\\\n"
						"			-I/usr/include/X11/extensions\\\n"
						"			-I/usr/include/freetype2");
				#else
				m.Print("# Includes\n"
						"Inc =		\\\n");
				#endif
						
				List<char> Incs;
				char *i;
				for (void *b=Inc.First(&i); b; b=Inc.Next(&i))
				{
					Incs.Insert(NewStr(i));
				}
				Incs.Sort(StrCmp, 0);
				for (i = Incs.First(); i; i = Incs.Next())
				{
					m.Print(" \\\n\t\t\t-I%s", i);
				}
				
				m.Print("\n\n");
				
				List<char> IncPaths;
				if (BuildIncludePaths(IncPaths, true))
				{
					// Do dependencies
					m.Print("# Dependencies\n"
							"Depends =\t");
					int c = 0;
					for (n=Files.First(); n; n=Files.Next(), c++)
					{
						if (n->GetType() == NodeSource)
						{
							char *f = ToNativePath(n->GetFileName());
							char *d = f ? strrchr(f, DIR_CHAR) : 0;
							f = d ? d + 1 : f;
							d = f ? strrchr(f, '.') : 0;
							if (d)
							{
								if (c) m.Print(" \\\n\t\t\t");
								m.Print("%.*s.o", (int)d-(int)f, f);
							}
						}
					}
					m.Print("\n\n");

					// Write out the target stuff
					m.Print("# Target\n");
					switch (d->Settings.TargetType)
					{
						// Executable
						case TARGET_TYPE_EXE:
						{							
							m.Print("# Executable target\n"
									"$(Target) :");
									
							GStringPipe Rules;
							IdeProject *d;
							for (d=Deps.First(); d; d=Deps.Next())
							{
								char t[256], p[256];
								if (d->GetTargetFile(t, sizeof(t)) AND
									d->GetBasePath(p))
								{
									char Rel[256];
									if (NOT RelitivePath(Rel, p))
									{
										strsafecpy(Rel, p, sizeof(Rel));
									}

									char *e = LgiGetExtension(t);
									if (e)
									{
										e--;
										int Len = strlen(e);
										memmove(e+6, e, Len+1);
										memcpy(e, "$(Tag)", 6);
									}
									
									char FullTarget[256];
									sprintf(FullTarget, "%s/$(BuildDir)/%s", Rel, t);
									m.Print(" %s", FullTarget);
									
									Rules.Printf("%s : FORCE\n"
												"	export Build=$(Build); \\\n"
												"		export Gui=$(Gui); \\\n"
												"		$(MAKE) -C %s\n"
												"\n",
												FullTarget,
												Rel);
								}
							}

							m.Print(" outputfolder $(Depends)\n"
									"	@echo Linking $(Target) [$(Gui)/$(Build)]...\n"
									"	$(CC) -Wl%s -f -o \\\n"
									"		$(Target) $(addprefix $(BuildDir)/,$(Depends)) $(Libs)\n"
									"	@echo Done.\n"
									"\n",
									LinkerFlags);

							char *r = Rules.NewStr();
							if (r)
							{
								m.Write(r, strlen(r));
								DeleteArray(r);
								
								m.Print("FORCE :\n"
										"\n");
							}

							// Various fairly global rules
							m.Print("# Create the output folder\n"
									"outputfolder :\n"
									"	-mkdir -p $(BuildDir) 2> /dev/null\n"
									"\n"
									"# Clean out targets\n"
									"clean :\n"
									"	rm -f $(BuildDir)/*.o $(Target).%s\n"
									"	@echo Cleaned $(BuildDir)\n"
									"\n",
									#ifdef WIN32
									".exe"
									#else
									""
									#endif
									);
							break;
						}
						// Shared library
						case TARGET_TYPE_SHARED_LIB:
						{
							m.Print("TargetFile = lib$(Target)$(Tag).%s\n"
									"$(TargetFile) : outputfolder $(Depends)\n"
									"	@echo Linking $(TargetFile) [$(Gui)/$(Build)]...\n"
									"	$(CC) -shared \\\n"
									"		-Wl%s \\\n"
									"		-o $(BuildDir)/$(TargetFile) \\\n"
									"		$(addprefix $(BuildDir)/,$(Depends)) \\\n"
									"		$(Libs)\n"
									"	@echo Done.\n"
									"\n",
									LGI_LIBRARY_EXT,
									LinkerFlags);

							// Cleaning target
							m.Print("# Create the output folder\n"
									"outputfolder :\n"
									"	-mkdir -p $(BuildDir) 2> /dev/null\n"
									"\n"
									"# Clean out targets\n"
									"clean :\n"
									"	rm -f $(BuildDir)/*.o $(BuildDir)/$(TargetFile)\n"
									"	@echo Cleaned $(BuildDir)\n"
									"\n",
									LGI_LIBRARY_EXT);
							break;
						}
						// Static library
						case TARGET_TYPE_STATIC_LIB:
						{
							m.Print("TargetFile = lib$(Target)$(Tag).%s\n"
									"$(TargetFile) : outputfolder $(Depends)\n"
									"	@echo Linking $(TargetFile) [$(Gui)/$(Build)]...\n"
									"	ar rcs $(BuildDir)/$(TargetFile) $(addprefix $(BuildDir)/,$(Depends))\n"
									"	@echo Done.\n"
									"\n",
									LGI_STATIC_LIBRARY_EXT);

							// Cleaning target
							m.Print("# Create the output folder\n"
									"outputfolder :\n"
									"	-mkdir -p $(BuildDir) 2> /dev/null\n"
									"\n"
									"# Clean out targets\n"
									"clean :\n"
									"	rm -f $(BuildDir)/*.o $(BuildDir)/$(TargetFile)\n"
									"	@echo Cleaned $(BuildDir)\n"
									"\n",
									LGI_STATIC_LIBRARY_EXT);
							break;
						}
					}

					// Create dependency tree, starting with all the source files.
					GHashTable Deps;
					for (n=Files.First(); n; n=Files.Next())
					{
						if (n->GetType() == NodeSource)
						{
							char *Src = n->GetFullPath();
							if (Src)
							{
								char Part[256];
								
								char *d = strrchr(Src, DIR_CHAR);
								d = d ? d + 1 : Src;
								strcpy(Part, d);
								char *Dot = strrchr(Part, '.');
								if (Dot) *Dot = 0;

								char Rel[256];
								if (NOT RelitivePath(Rel, Src))
								{
						 			strcpy(Rel, Src);
								}
								
								m.Print("%s.o : %s ", Part, ToUnixPath(Rel));

								char *c8 = ReadTextFile(Src);
								if (c8)
								{
									List<char> Headers;
									if (BuildHeaderList(c8, Headers, IncPaths, false))
									{
										int n=0;
										for (char *i=Headers.First(); i; i=Headers.Next(), n++)
										{
											if (n) m.Print(" \\\n\t");
											
											if (NOT RelitivePath(Rel, i))
											{
						 						strcpy(Rel, i);
											}
											m.Print("%s", ToUnixPath(Rel));
											
											if (NOT Deps.Find(i))
											{
												Deps.Add(i);
											}
										}
										Headers.DeleteObjects();
									}
									
									DeleteArray(c8);
								}

								m.Print("\n"
									#ifdef LINUX
										"	@echo $(<F) [$(Gui)]\n"
									#else
										"	@echo $(<F)\n"
									#endif
										"	$(CC) $(Inc) $(Flags) $(Defs) -c $< -o $(BuildDir)/$(@F)\n"
										"\n");

								DeleteArray(Src);
							}
						}
					}
					
					// Do remaining include file dependencies
					bool Done = false;
					GHashTable Processed;
					while (NOT Done)
					{
						Done = true;
						char *Src;
						for (void *b=Deps.First(&Src); b; b=Deps.Next(&Src))
						{
							if (NOT Processed.Find(Src))
							{
								Done = false;
								Processed.Add(Src);
								
								char *c8 = ReadTextFile(Src);
								if (c8)
								{
									List<char> Headers;
									if (BuildHeaderList(c8, Headers, IncPaths, false))
									{
										char *d = strrchr(Src, DIR_CHAR);
										d = d ? d + 1 : Src;
										m.Print("%s : ", d);

										int n=0;
										for (char *i=Headers.First(); i; i=Headers.Next(), n++)
										{
											if (n) m.Print(" \\\n\t");
											
											char Rel[256];
											if (NOT RelitivePath(Rel, i))
											{
						 						strcpy(Rel, i);
											}
											m.Print("%s", ToUnixPath(Rel));
											
											if (NOT Deps.Find(i))
											{
												Deps.Add(i);
											}
										}
										Headers.DeleteArrays();

										m.Print("\n\n");
									}									
									DeleteArray(c8);
								}
								break;
							}
						}
					}


					// Output VPATH
					m.Print("VPATH=%%.cpp \\\n"
							"\t$(BuildDir)");
					m.Print("\n\n");
					
					
					if (ValidStr(d->Settings.MakefileRules))
					{
						m.Print("\n%s\n", d->Settings.MakefileRules);
					}					
				}
			}
			else
			{
				m.Print("# No files require building.\n");
			}			
		}
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////////////////
IdeTree::IdeTree() : GTree(100, 0, 0, 100, 100)
{
	Hit = 0;
}

void IdeTree::OnAttach()
{
	SetWindow(this);
}

void IdeTree::OnDragExit()
{
	SelectDropTarget(0);
}

int IdeTree::WillAccept(List<char> &Formats, GdcPt2 p, int KeyState)
{
	for (char *f=Formats.First(); f; )
	{
		if (stricmp(f, NODE_DROP_FORMAT) == 0)
		{
			f = Formats.Next();
		}
		else
		{
			Formats.Delete(f);
			DeleteArray(f);
			f = Formats.Current();
		}
	}
		
	if (Formats.Length() > 0)
	{
		Hit = ItemAtPoint(p.x, p.y);
		if (Hit)
		{
			IdeCommon *Src = dynamic_cast<IdeCommon*>(Selection());
			IdeCommon *Dst = dynamic_cast<IdeCommon*>(Hit);
			if (Src AND Dst)
			{
				// Check this folder is not a child of the src
				for (IdeCommon *n=Dst; n; n=dynamic_cast<IdeCommon*>(n->GetParent()))
				{
					if (n == Src)
					{
						return DROPEFFECT_NONE;
					}
				}
			}

			// Valid target
			SelectDropTarget(Hit);
			return DROPEFFECT_MOVE;
		}
	}

	return DROPEFFECT_NONE;
}

int IdeTree::OnDrop(char *Format, GVariant *Data, GdcPt2 Pt, int KeyState)
{
	SelectDropTarget(0);

	if (Hit AND
		Data AND
		Format AND
		stricmp(Format, NODE_DROP_FORMAT) == 0 AND
		Data->Type == GV_BINARY AND
		Data->Value.Binary.Length == sizeof(ProjectNode*))
	{
		ProjectNode *Src = ((ProjectNode**)Data->Value.Binary.Data)[0];
		if (Src)
		{
			ProjectNode *Folder = dynamic_cast<ProjectNode*>(Hit);
			while (Folder AND Folder->GetType() > NodeDir)
			{
				Folder = dynamic_cast<ProjectNode*>(Folder->GetParent());
			}

			IdeCommon *Dst = dynamic_cast<IdeCommon*>(Folder?Folder:Hit);
			if (Dst)
			{
				// Check this folder is not a child of the src
				for (IdeCommon *n=Dst; n; n=dynamic_cast<IdeCommon*>(n->GetParent()))
				{
					if (n == Src)
					{
						return DROPEFFECT_NONE;
					}
				}

				// Detach
				GTreeItem *i = dynamic_cast<GTreeItem*>(Src);
				i->Detach();
				if (Src->GXmlTag::Parent)
				{
					LgiAssert(Src->GXmlTag::Parent->Children.HasItem(Src));
					Src->GXmlTag::Parent->Children.Delete(Src);
				}
				
				// Attach
				Src->GXmlTag::Parent = Dst;
				Dst->Children.Insert(Src);
				Dst->Insert(Src);
				
				// Dirty
				Src->GetProject()->SetDirty();
			}
		
			return DROPEFFECT_MOVE;
		}
	}

	return DROPEFFECT_NONE;
}

