#include "Lgi.h"
#include "LgiIde.h"
#include "GList.h"

class CharItem : public GListItem
{
	GSurface *pDC;

public:
	CharItem(GSurface *pdc)
	{
		pDC = pdc;
	}

	~CharItem()
	{
		DeleteObj(pDC);
	}

	void OnPaintColumn(GSurface *pScr, GRect &r, int i, GListColumn *c)
	{
		GListItem::OnPaintColumn(pScr, r, i, c);

		if (i == 1)
		{
			pScr->Blt(r.x1, r.y1, pDC);
		}
	}

	void OnMeasure(GMeasureInfo *Info)
	{
		GListItem::OnMeasure(Info);
		if (pDC)
			Info->y = max(Info->y, pDC->Y());
	}
};

class SysCharSupportPriv : public GLgiRes
{
public:
	GList *Lst;

	SysCharSupportPriv()
	{
		Lst = 0;
	}

	void Search(char *ch)
	{
		if (ch)
		{
			int c = atoi(ch);
			GFontSystem *s = GFontSystem::Inst();
			if (s)
			{
				List<char> Fonts;
				if (s->EnumerateFonts(Fonts))
				{
					for (char *f=Fonts.First(); f; f=Fonts.Next())
					{
						GFont *Fnt = NEW(GFont);
						if (Fnt)
						{
							if (Fnt->Create(f, 20))
							{
								if (_HasUnicodeGlyph(Fnt->GetGlyphMap(), c))
								{
									GMemDC *pDC = NEW(GMemDC);
									if (pDC)
									{
										char16 Str[] = { c, 0 };
										GDisplayString d(Fnt, Str);
										if (pDC->Create(d.X(), d.Y(), 32))
										{
											pDC->Colour(LC_WORKSPACE, 24);
											pDC->Colour(LC_TEXT, 24);
											d.Draw(pDC, 0, 0);
										}
									}

									GListItem *i;
									Lst->Insert(i = NEW(CharItem(pDC)));
									if (i)
									{
										i->SetText(f);
									}
								}
							}

							DeleteObj(Fnt);
						}
					}
				}				
			}
		}
	}
};

SysCharSupport::SysCharSupport(AppWnd *app)
{
	d = NEW(SysCharSupportPriv);
#ifdef WIN32
	CreateClass("SysCharSupportPriv");
#endif
	Name("System Character Support");

	GRect r(0, 0, 600, 600);
	SetPos(r);
	MoveToCenter();

	if (Attach(0))
	{
		d->LoadFromResource(IDD_CHAR, Children);
		AttachChildren();
		d->Lst = dynamic_cast<GList*>(FindControl(IDC_RESULTS));
		OnPosChange();

		Visible(true);
	}
}

SysCharSupport::~SysCharSupport()
{
	DeleteObj(d);
}

void SysCharSupport::OnPosChange()
{
	if (d->Lst)
	{
		GRect c = GetClient();
		GRect r = d->Lst->GetPos();
		r.x2 = c.x2 - r.x1;
		r.y2 = c.y2 - r.x1;
		d->Lst->SetPos(r);
	}
}

int SysCharSupport::OnNotify(GViewI *v, int f)
{
	switch (v->GetId())
	{
		case IDOK:
		{
			d->Search(GetCtrlName(IDC_CHAR));
			break;
		}
	}

	return 0;
}

