/*hdr
**      FILE:           LgiRes.h
**      AUTHOR:         Matthew Allen
**      DATE:           22/10/00
**      DESCRIPTION:    Resource Editor App Header
**
*/

#include "Lgi.h"
#include "INet.h"
#include "GXml.h"
#include "GTabLayout.h"
#include "GDocApp.h"

#include "resource.h"
#include "GTree.h"
#include "GDataDlg.h"

////////////////////////////////////////////////////////////////////////////////////////////
// Defines

// version
#define APP_VER						1.88

// window messages
#define	IDM_NEW						110

#define	IDM_HELP					190
#define	IDM_ABOUT					199

#define	IDM_UNDO					201
#define	IDM_REDO					202
#define	IDM_CUT						203
#define	IDM_COPY					204
#define	IDM_PASTE					205
#define	IDM_FIND					206
#define IDM_NEXT					207

#define	IDM_DELETE					300
#define	IDM_RENAME					301
#define	IDM_SETTINGS				302
#define	IDM_IMPORT					303
#define	IDM_IMPORT_WIN32			304
#define	IDM_EXPORT					305
#define	IDM_EXPORT_WIN32			306
#define	IDM_NEW_LANG				307
#define	IDM_DELETE_LANG				308
#define	IDM_IMPORT_LANG				309
#define	IDM_PROPERTIES				310
#define IDM_NEW_SUB					311
#define IDM_NEW_ITEM				312
#define IDM_DELETE_ITEM				313
#define IDM_MOVE_LEFT				314
#define IDM_MOVE_RIGHT				315
#define IDM_SET_LANG				316
#define IDM_TAB_ORDER				317
#define	IDM_DUMP					318
#define IDM_COPY_TEXT				319
#define IDM_PASTE_TEXT				320
#define IDM_NEW_ID					321
#define IDM_COMPARE					322
#define IDM_SHOW_LANG				323

#define	IDM_NEW_DLG					400
#define	IDM_NEW_STRGRP				401
#define	IDM_NEW_MENU				402

#define IDM_LANG_BASE				2000

#define	IDC_CHAT_MSG				1000

#define	STATUS_NORMAL				0
#define	STATUS_INFO					1
#define	STATUS_MAX					2

// Misc
#define	MainWnd						((AppWnd*)GApp::ObjInstance()->AppWnd)

// App
#define	TYPE_BITMAP					1
#define	TYPE_ICON					2
#define	TYPE_CURSOR					3
#define	TYPE_DIALOG					4
#define	TYPE_STRING					5
#define	TYPE_MENU					6
#define	TYPE_DIR_DISABLED			7

#define OPT_ShowLanguages			"ShowLang"

#define StrDialogSymbols			"_Dialog Symbols_"
extern char TranslationStrMagic[];

////////////////////////////////////////////////////////////////////////////////////////////
// Functions
extern char *EncodeXml(char *s, int l = -1);
extern char *DecodeXml(char *s, int l = -1);

////////////////////////////////////////////////////////////////////////////////////////////
// Classes
class AppWnd;
class ObjTreeItem;
class ResString;
typedef List<ResString> StringList;
class ResMenu;
class ResDialog;
class ResStringGroup;
class ResMenuItem;

struct ErrorInfo
{
	ResString *Str;
	GLanguageId Lang;
};

class ErrorCollection
{
public:
	GArray<ErrorInfo> StrErr;
};

class Resource
{
	friend class AppWnd;

protected:
	int ResType;
	AppWnd *AppWindow;
	ObjTreeItem *Item;
	bool SysObject;

public:
	Resource(AppWnd *w, int t = 0, bool enabled = true);
	virtual ~Resource();

	AppWnd *App() { return AppWindow; }
	bool SystemObject() { return SysObject; }
	void SystemObject(bool i) { SysObject = i; }
	bool IsSelected();

	virtual GView *Wnd() { return NULL; }
	virtual bool Attach(GViewI *Parent);
	virtual int Type() { return ResType; }
	virtual void Type(int i) { ResType = i; }
	virtual void Create() {} // called when users creates
	
	// Sub classes
	virtual ResStringGroup *IsStringGroup() { return 0; }
	virtual ResDialog *IsDialog() { return 0; }
	virtual ResMenu *IsMenu() { return 0; }
	
	// Serialization
	virtual bool Test(ErrorCollection *e) = 0;
	virtual bool Read(GXmlTag *t, ResFileFormat Format) = 0;
	virtual bool Write(GXmlTag *t, ResFileFormat Format) = 0;
	virtual StringList *GetStrs() { return NULL; }

	// Clipboard
	virtual void Delete() {}
	virtual void Copy(bool Delete = false) {}
	virtual void Paste() {}

	// UI
	virtual GView *CreateUI() { return 0; }
	virtual void OnRightClick(GSubMenu *RClick) {}
	virtual void OnCommand(int Cmd) {}
	virtual void OnShowLanguages() {}
};

class ResFolder : public Resource, public GView
{
public:
	ResFolder(AppWnd *w, int t, bool enabled = true);
	GView *Wnd() { return dynamic_cast<GView*>(this); }

	bool Test(ErrorCollection *e) { return false; }
	bool Read(GXmlTag *t, ResFileFormat Format) { return false; }
	bool Write(GXmlTag *t, ResFileFormat Format) { return false; }
};

class ResFrame : public GLayout
{
	Resource *Child;

public:
	ResFrame(Resource *child);
	~ResFrame();

	bool Pour(GRegion &r);
	bool OnKey(GKey &k);
	void OnPaint(GSurface *pDC);
	bool Attach(GViewI *p);
	void OnFocus(bool b);
};

class ObjTreeItem : public GTreeItem
{
	Resource *Obj;

public:
	ObjTreeItem(Resource *Object);
	~ObjTreeItem();

	Resource *GetObj() { return Obj; }
	char *GetText(int i=0);
	int GetImage(int Flags);
	void OnSelect();
	void OnMouseClick(GMouse &m);
};

class ObjContainer : public GTree
{
	friend class AppWnd;

	ObjTreeItem *Bitmaps;
	ObjTreeItem *Icons;
	ObjTreeItem *Cursors;
	ObjTreeItem *Dialogs;
	ObjTreeItem *Strings;
	ObjTreeItem *Menus;

	GImageList *Images;
	AppWnd *Window;

	bool AppendChildren(ObjTreeItem *Item, List<Resource> &Lst);

public:
	ObjContainer(AppWnd *w);
	~ObjContainer();

	bool ListObjects(List<Resource> &Lst);
};

class FieldSource
{
public:
	ObjProperties Values;

	virtual bool GetFields(DataDlgFieldList &Fields, ObjProperties *Props = 0) { return false; }
	virtual void OnFieldChange() {}
};

class FieldView : public GLayout, public DataDlgTools {

protected:
	FieldSource *Source;
	DataDlgFieldList Fields;
	bool Ignore;
	AppWnd *App;

public:
	FieldView(AppWnd *app);
	~FieldView();

	void Serialize(bool Write);

	void OnSelect(FieldSource *s);
	int OnEvent(GMessage *m);
	void OnPaint(GSurface *pDC);
	int OnNotify(GViewI *Ctrl, int Flags);
	void OnPosChange();
};

#include "LgiRes_String.h"

class AppWnd : public GDocApp
{
protected:
	// UI
	GSplitter		*MainSplit;
	GSplitter		*SubSplit;

	GSubMenu		*Edit;
	GSubMenu		*Help;
	GSubMenu		*ViewMenu;

	GStatusBar		*Status;
	GStatusPane		*StatusInfo[STATUS_MAX];

	// App
	ObjContainer	*Objs;
	Resource		*LastRes;
	FieldView		*Fields;
	
	// Languages
	int				CurLang;
	GArray<GLanguage*> Languages;
	GHashTable		ShowLanguages;

	void SetupUi();
	void SortDialogs();
	void GetFileTypes(GFileSelect *Dlg, bool Write);

public:
	AppWnd();
	~AppWnd();

	// Languages
	bool ShowLang(GLanguageId Lang);
	void ShowLang(GLanguageId Lang, bool Show);
	GLanguage *GetCurLang();
	void SetCurLang(GLanguage *L);
	GArray<GLanguage*> *GetLanguages();
	void OnLanguagesChange(GLanguageId Lang, bool Add, bool Update = false);

	// ---------------------------------------------------------------------
	// Application
	Resource *NewObject(int Type, bool Select = true);
	bool InsertObject(int Type, Resource *r, bool Select = true);
	void DelObject(Resource *r);
	bool ListObjects(List<Resource> &Lst);
	int GetUniqueStrRef(int Start = 1);
	int GetUniqueCtrlId();
	void FindStrings(List<ResString> &Strs, char *Define = 0, int *CtrlId = 0);
	ResString *GetStrFromRef(int Ref);
	ResStringGroup *GetDialogSymbols();

	void Empty();
	void OnObjChange(FieldSource *r);
	void OnObjSelect(FieldSource *r);
	void OnResourceSelect(Resource *r);
	void GotoObject(class ResString *s, ResStringGroup *g, ResDialog *d, ResMenuItem *m, class ResDialogCtrl *c);

	// ---------------------------------------------------------------------
	// Methods
	void SetStatusText(char *Text, int Pane = 0);
	bool TestLgi(bool Quite = true);
	bool LoadLgi(char *FileName = 0);
	bool SaveLgi(char *FileName = 0);
	bool LoadWin32(char *FileName = 0);
	bool SaveWin32();
	void ImportLang();
	void Compare();
	bool WriteDefines(GFile &Defs);

	bool OpenFile(char *FileName);
	bool SaveFile(char *FileName);

	// ---------------------------------------------------------------------
	// Window
	int OnNotify(GViewI *Ctrl, int Flags);
	int OnEvent(GMessage *m);
	int OnCommand(int Cmd, int Event, OsView Handle);
	void OnReceiveFiles(GArray<char*> &Files);
	void OnCreate();
};

class Search : public GDialog
{
	AppWnd *App;

	void OnCheck();

public:
	char *Text;
	char *Define;
	GLanguageId InLang;
	GLanguageId NotInLang;
	int Ref;
	int CtrlId;

	Search(AppWnd *app);
	~Search();

	int OnNotify(GViewI *c, int f);
};

class Results : public GWindow
{
	class ResultsPrivate *d;

public:
	Results(AppWnd *app, Search *search);
	~Results();

	void OnPosChange();
	int OnNotify(GViewI *v, int f);

};

class ShowLanguagesDlg : public GDialog
{
	class ShowLanguagesDlgPriv *d;

public:
	ShowLanguagesDlg(AppWnd *app);
	~ShowLanguagesDlg();

	int OnNotify(GViewI *n, int f);
};