/// \file
/// \author Matthew Allen (fret@memecode.com)
/// \brief A combo box control

#ifndef __GCOMBO2_H
#define __GCOMBO2_H

/// Combo Box widget.
class LgiClass GCombo :
	public GView,
	public ResObject	
{
private:
	class GComboPrivate *d;

public:
	/// Constructor
	GCombo
	(
		/// The control ID
		int id,
		/// The left edge x coordinate
		int x,
		/// The top edge y coordinate
		int y,
		/// The width
		int cx,
		/// The height
		int cy,
		/// The initial text
		char *name
	);
	~GCombo();

	// Properties
	char *GetClassName() { return "GCombo"; }

	/// Returns whether to sort the list
	bool Sort();
	/// Sets whether to sort the list
	void Sort(bool s);
	/// Returns whether to classify the list into submenus
	int Sub();
	/// Makes the list of entries sort into submenus. Good for large lists.
	void Sub
	(
		/// The type of classification to use. One of #GV_INT32, #GV_DOUBLE or #GV_STRING.
		int Type
	);
	/// Sets the selected item
	void Value(int64 i);
	/// Returns the select item
	int64 Value();
	/// Sets the selected item by name
	bool Name(char *n);
	/// Returns the selected item's name
	char *Name();
	/// Gets the menu used
	GSubMenu *GetMenu();
	/// Sets the menu used
	void SetMenu(GSubMenu *m);

	// List
	
	/// Deletes the currently select item in the list
	bool Delete();
	/// Deletes the item at index 'i'
	bool Delete(int i);
	/// Deletes the item matching 'p'
	bool Delete(char *p);
	/// Inserts a new item
	bool Insert
	(
		/// The new item
		char *p,
		/// The location to insert or -1 for the end of the list
		int Index = -1
	);
	/// Gets the items in the list
	int GetItems();
	/// Returns the item at index 'i'
	char *operator [](int i);

	// Events/Window/Implementation
	int OnEvent(GMessage *Msg);
	void OnMouseClick(GMouse &m);
	bool OnKey(GKey &k);
	void OnFocus(bool f);
	void OnPaint(GSurface *pDC);
	void SetFont(GFont *Fnt, bool OwnIt = false);
	void OnAttach();
	bool SetPos(GRect &p, bool Repaint = false);
};

#endif

