/// \file
/// \author Matthew Allen (fret@memecode.com)
/// \brief Tools for getting data into and out of the UI

#ifndef __DATA_DLG_TOOLS
#define __DATA_DLG_TOOLS

#include "GProperties.h"

enum DataDlgType
{
	DATA_STR = 1,
	DATA_BOOL,
	DATA_INT,
	DATA_FLOAT,
	DATA_PASSWORD,
	DATA_STR_SYSTEM		// Operating system specific string. Has the OS
						// tag automatically appended to the option name
						// e.g. SomeProperty-Win32
};

class DataDlgTools;

/// UI <-> properties field
class LgiClass DataDlgField
{
	friend class DataDlgTools;

protected:
	int Type;
	int CtrlId;
	char *Option;
	char *Desc;

public:
	DataDlgField(int type, int ctrlid, char *opt, char *desc = 0);
	~DataDlgField();

	int GetType() { return Type; }
	int GetCtrl() { return CtrlId; }
	char *GetOption() { return Option; }
	char *GetDesc() { return Desc; }

	void SetOption(char *o) { char *Opt = NewStr(o); DeleteArray(Option); Option = Opt; }
};

typedef List<DataDlgField> DataDlgFieldList;

/// UI <-> properties mapping class
class LgiClass DataDlgTools
{
protected:
	bool DeleteEmptyStrings;
	GView *Dlg;
	ObjProperties *Options;

public:
	DataDlgTools();

	void Set(GView *Dlg, ObjProperties *Options);
	ObjProperties *GetOptions();

	bool ProcessField(DataDlgField *f, bool Write, char *OptionOverride = 0);
	bool ProcessFields(DataDlgFieldList &Field, bool Write);
	bool LoadFields(DataDlgFieldList &Field);
	bool SaveFields(DataDlgFieldList &Field);
};

class LgiClass DRecordSetObj
{
public:
	virtual bool Serialize(ObjProperties &f, bool Write)
	{
		return false;
	}
};

class LgiClass DRecordSetCtrls : public DataDlgTools
{
	// Controls
	GText *Description;
	GScrollBar *Scroll;
	int NewRecordId;
	int DeleteRecordId;

	DataDlgFieldList *Fields;
	
protected:
	DRecordSetObj *Current;
	List<DRecordSetObj> *Records;

public:
	char *RecordStringTemplate;

	DRecordSetCtrls(	GView *window,
						DataDlgFieldList *fields,
						List<DRecordSetObj> *records,
						int DescId,
						int ScrollId,
						int NewId,
						int DelId,
						char *Template);
	~DRecordSetCtrls();

	virtual DRecordSetObj *NewObj() { return 0; }
	virtual void DelObj(DRecordSetObj *Obj) {}

	int GetCurrentIndex();
	void Serialize(bool Write);
	void OnMoveRecord(DRecordSetObj *r);
	int OnNotify(GViewI *Col, int Flags);
};

template <class Record>
class LgiClass DRecordSet : public DRecordSetCtrls
{
	typedef Record *(*Allocator)(GView *);
	Allocator Alloc;

public:
	DRecordSet
	(
		GView *window,
		DataDlgFieldList *fields,
		List<Record> *records,
		int DescId,
		int ScrollId,
		int NewId,
		int DelId,
		char *Template,
		Allocator a
	)
		:
		DRecordSetCtrls(window, fields, (List<DRecordSetObj>*) records, DescId, ScrollId, NewId, DelId, Template)
	{
		Alloc = a;
	}

	DRecordSetObj *NewObj()
	{
		return Alloc ? Alloc(Dlg) : 0;
	}

	void DelObj(DRecordSetObj *Obj)
	{
		Record *r = (Record*) Obj;
		if (r)
		{
			DeleteObj(r);
		}
	}

	Record *ItemAt(int i=-1)
	{
		return dynamic_cast<Record*>((i<0 OR NOT Records) ? Current : Records->ItemAt(i));
	}

	void OnMoveRecord(Record *r)
	{
		DRecordSetCtrls::OnMoveRecord((DRecordSetObj*)r);
	}
};

#endif

