#ifndef _LGI_SCRIPTING_H_
#define _LGI_SCRIPTING_H_

#include "GVariant.h"

class GScriptContext;
class GScriptEngine;
class GScriptEnginePrivate;

typedef GArray<GVariant*> ArgumentArray;
typedef bool (GScriptContext::*ScriptCmd)(GVariant *Ret, ArgumentArray &Args);

/// Various type of methods
enum GFuncType
{
	/// No method type.
	NullFunc,
	/// This method is provided by the hosting application.
	HostFunc,
	/// This method is defined in the script itself.
	ScriptFunc,
	/// This method is defined in an external library.
	ExternFunc
};

struct GFunc
{
	GFuncType Type;
	char *Method;

	GFunc(char *m = 0, GFuncType t = NullFunc)
	{
		Type = t;
		Method = NewStr(m);
	}

	virtual ~GFunc()
	{
		DeleteArray(Method);
	}

	virtual bool Call(GScriptEnginePrivate *Priv, GVariant *Ret, ArgumentArray &Args) = 0;
	bool Call(GScriptEnginePrivate *Priv, GVariant *Ret, GArray<int> *Args);
};

struct GHostFunc : public GFunc
{
	char *Args;
	ScriptCmd Func;
	
	GHostFunc(char *method, char *args, ScriptCmd proc) : GFunc(method, HostFunc)
	{
		Args = NewStr(args);
		Func = proc;
	}

	~GHostFunc()
	{
		DeleteArray(Args);
	}

	bool Call(GScriptEnginePrivate *Priv, GVariant *Ret, ArgumentArray &Args);
};

class GScriptUtils
{
public:
	virtual ~GScriptUtils() {}

	int atoi(char16 *s);
	int64 atoi64(char16 *s);
	double atof(char16 *s);
	int htoi(char16 *s);
};

class GScriptContext : public GScriptUtils
{
public:
	virtual GHostFunc *GetCommands() = 0;
	virtual void SetEngine(GScriptEngine *Eng) = 0;

	// AddPrimitive: Add your primitive's functions in a derived class using the format:
	//
	//		bool MyPrim(GVariant *Ret, ArgumentArray &Args) { ... }
	//
	// Where you return true if successful or false on error.
};

class GScriptEngine
{
	friend class GScriptEval;
	GScriptEnginePrivate *d;

public:
	GScriptEngine(GViewI *parent, GScriptContext *context);
	~GScriptEngine();

	void Empty();
	bool Compile(char *Script, bool Add = false);
	bool Run();
	bool RunTemporary(char *Script);
	bool EvaluateExpression(GVariant *Result, GDom *VariableSource, char *Expression);

	GVariant *Var(char16 *name, bool create = true);
	GStringPipe *GetTerm();
	bool CallMethod(char *Method, GVariant *Ret, ArgumentArray &Args);

	void DumpVariables();
};

#endif

