/// \file
/// \author Matthew Allen
/// \brief A tab view

#ifndef _GTABVIEW_H_
#define _GTABVIEW_H_

// Tab control
class GTabPage;
typedef List<GTabPage> TabPageList;

/// A tab control that displays multiple pages of information in a small area.
class LgiClass GTabView :
	public GView,
	public ResObject
{
	friend class GTabPage;
	class GTabViewPrivate *d;
	int TabY();

protected:
	List<GTabPage> Tabs;

public:
	/// Creates the tab view object
	GTabView(int id, int x, int y, int cx, int cy, char *name, int Init = 0);
	~GTabView();

	/// Gets the selected tab
	int64 Value();
	/// Sets the selected tab
	void Value(int64 i);

	/// Append an existing tab
	bool Append(GTabPage *Page, int Where = -1);
	/// Append a new tab with the title 'name'
	GTabPage *Append(char *name, int Where = -1);
	/// Delete a tab
	bool Delete(GTabPage *Page);

	///.Returns the tab at position 'i'
	GTabPage *TabAt(int i) { return Tabs.ItemAt(i); }
	/// Gets a pointer to the current tab
	GTabPage *GetCurrent();
	/// Gets the number of tabs
	int GetTabs() { return Tabs.Length(); }
	
	// Impl
	bool Attach(GViewI *parent);
	int OnEvent(GMessage *Msg);
	GViewI *FindControl(int Id);
	int OnNotify(GViewI *Ctrl, int Flags);
	void OnChildrenChanged(GViewI *Wnd, bool Attaching);
	GRect &GetTabClient();

	#if defined WIN32
	GViewI *FindControl(HWND hCtrl);
	#endif

	void OnPosChange();
	void OnPaint(GSurface *pDC);
	void OnMouseClick(GMouse &m);
	bool OnKey(GKey &k);
	void OnFocus(bool f);
};

class LgiClass GTabPage :
	public GView,
	public GLgiRes,
	public ResObject
{
	friend class GTabView;

	// Vars
	GTabView *TabCtrl;
	GRect TabPos;

	// Methods
	void PaintTab(GSurface *pDC, bool Selected);
	bool Attach(GViewI *parent);

public:
	GTabPage(char *name);
	~GTabPage();

	char *Name();
	bool Name(char *n);
	GTabView *GetTabControl() { return TabCtrl; }

	int OnEvent(GMessage *Msg);
	void OnPaint(GSurface *pDC);
	bool OnKey(GKey &k);
	void OnFocus(bool b);

	void Append(GViewI *Wnd);
	bool Remove(GViewI *Wnd);
	bool AppendDialogFromResource(int Resource);
};

#endif
