#ifndef _VCARD_VCAL_H
#define _VCARD_VCAL_H

#include "GToken.h"
#include "GDateTime.h"
#include "ScribeDefs.h"
#include "GMap.h"

#define FIELD_PERMISSIONS				2000

class VObject
{
public:
	// Set this to false to stop SetField call's making the object dirty
	bool CanDirty;

	VObject()
	{
		CanDirty = true;
	}

	// Field set/get API...
	virtual bool SetField(int Field, int n			) = 0;
	virtual bool SetField(int Field, double n		) = 0;
	virtual bool SetField(int Field, char *n		) = 0;
	virtual bool SetField(int Field, GDateTime &n	) = 0;
	virtual bool GetField(int Field, int &n			) = 0;
	virtual bool GetField(int Field, double &n		) = 0;
	virtual bool GetField(int Field, char *&n		) = 0;
	virtual bool GetField(int Field, GDateTime &n	) = 0;

	// Non-mandatory methods

	// Object set/get API...
	virtual bool GetAttendees(List<VObject> &a) { return false; }
	virtual VObject *CreateAttendee() { return 0; }
};

class VIo
{
protected:
	List<char> Object;
	GToken Var;
	GToken Type;
	GToken Value;
	GMap<char*,char*> Variables;

	bool IsTyped();
	bool IsEncoded();
	bool IsBeginEnd();
	bool IsVar(char *a, char *b = 0);
	bool IsType(char *type);
	bool IsValue(char *value);
	bool IsObject(char *Obj);

	bool ParseDate(GDateTime &out, char *in);
	bool ParseDuration(GDateTime &Out, int &Sign, char *In);

	void Fold(GStream &o, char *i, int pre_chars = 0);
	char *Unfold(char *In);
	char *UnMultiLine(char *In);

	bool ReadField(char *&s, char **Name, char **Type, char **Data);
	void WriteField(GStream &s, char *Name, char *Type, char *Data);

public:
	~VIo();
};

class VCard : public VIo
{
	VObject *c;

public:
	VCard(VObject *t)
	{
		 c = t;
	}

	bool Import(char *Text);
	char *Export();
};

class VCal : public VIo
{
	VObject *c;

public:
	VCal(VObject *t)
	{
		 c = t;
	}

	bool Import(char *Text);
	char *Export();
};

#endif
