#include <stdio.h>
#include <math.h>

#include "Lgi.h"
#include "GScripting.h"
#include "GUtf8.h"

class SystemFunctions : public GScriptContext
{
public:
	/// int Strchr(string, char[, string_length]);
	/// \returns the position of 'char' in 'string' or -1 if 'string' doesn't contain 'char'. You may 
	///			 optionally supply a max number of chars to search.
	bool Strchr(GVariant *Ret, ArgumentArray &Args)
	{
		if (Args.Length() >= 2)
		{
			char *str = Args[0]->CastString();
			char *ch = Args[1]->CastString();
			int str_len = Args.Length() > 2 ? Args[2]->CastInt32() : -1;
			if (str AND ch)
			{
				GUtf8Ptr Str(str);
				GUtf8Ptr Ch(ch);
				uint32 c = Ch;

				int n = 0;
				for (uint32 s = Str; s; s = Str++, n++)
				{
					if (c == s)
					{
						*Ret = n;
						return true;
					}
				}

				*Ret = -1;
				return true;
			}
		}
		return false;
	}

	/// int Strstr(string1, string2, case_insensitive, string_length);
	/// \returns the position of 'string2' in 'string1' or -1 if 'string1' doesn't contain 'string2'. You may 
	///			 optionally supply a max number of chars to search.
	bool Strstr(GVariant *Ret, ArgumentArray &Args)
	{
		if (Args.Length() >= 2)
		{
			char *str1 = (char*) Args[0]->CastString();
			char *str2 = (char*) Args[1]->CastString();
			int case_insensitive =	Args.Length() > 2 ? Args[2]->CastInt32() : true;
			int str_len =			Args.Length() > 3 ? Args[3]->CastInt32() : -1;
			if (str1 AND str2)
			{
				char *r = 0;

				if (str_len >= 0)
				{
					if (case_insensitive)
						r = strnistr(str1, str2, str_len);
					else
						r = strnstr(str1, str2, str_len);

				}
				else
				{
					if (case_insensitive)
						r = stristr(str1, str2);
					else
						r = strstr(str1, str2);
				}

				if (r)
					*Ret = r - str1;
				else
					*Ret = -1;

				return true;
			}
		}
		return false;
	}

	/// int Strcmp(string1, string2, case_insensitive, string_length);
	/// \returns a comparision of 'string1' and 'string2'. You can optionally make it case insensitive by 
	///			 seting the 'case_insensitive' flag.
	bool Strcmp(GVariant *Ret, ArgumentArray &Args)
	{
		if (Args.Length() >= 2)
		{
			char *a = Args[0]->CastString();
			char *b = Args[1]->CastString();
			int case_insensitive =	Args.Length() > 2 ? Args[2]->CastInt32() : true;
			int str_len =			Args.Length() > 3 ? Args[3]->CastInt32() : -1;

			if (a AND b)
			{
				if (str_len > 0)
				{
					if (case_insensitive)
						*Ret = strnicmp(a, b, str_len);
					else
						*Ret = strncmp(a, b, str_len);
				}
				else
				{
					if (case_insensitive)
						*Ret = stricmp(a, b);
					else
						*Ret = strcmp(a, b);
				}

				return true;
			}
		}

		return false;
	}

	/// String Substr(string, start[, length]);
	/// \returns a sub-segment of 'string'.
	bool Substr(GVariant *Ret, ArgumentArray &Args)
	{
		if (Args.Length() >= 2)
		{
			uint8 *str = (uint8*) Args[0]->CastString();
			int start = Args[1]->CastInt32();
			int len = Args.Length() > 2 ? Args[2]->CastInt32() : -1;
			if (str)
			{
				uint32 ch;
				int slen = strlen((char*) str);
				while (start > 0 AND (ch = LgiUtf8To32(str, slen)) != 0)
				{
					start--;
				}

				if (len < 0)
				{
					*Ret = (char*)str;
				}
				else
				{
					uint8 *end = str;
					while (len > 0 AND (ch = LgiUtf8To32(end, slen)) != 0)
					{
						len--;
					}
					Ret->Empty();
					Ret->Type = GV_STRING;
					Ret->Value.String = NewStr((char*)str, end - str);
				}
				return true;
			}
		}

		return false;
	}

	bool NewHashTable(GVariant *Ret, ArgumentArray &Args)
	{
		bool CaseSen = Args.Length() == 1 ? Args[0]->CastInt32() != 0 : true;
		Ret->SetHashTable(NEW(GHashTable(0, CaseSen)));
		return true;
	}

	bool NewList(GVariant *Ret, ArgumentArray &Args)
	{
		Ret->SetList(NEW(List<GVariant>));
		return true;
	}

	bool DeleteElement(GVariant *Ret, ArgumentArray &Args)
	{
		if (Args.Length() == 2)
		{
			switch (Args[0]->Type)
			{
				case GV_HASHTABLE:
				{
					char *Key = Args[1]->CastString();
					if (Key)
					{
						return Args[0]->Value.Hash->Delete(Key);
					}
					break;
				}
				case GV_LIST:
				{
					int32 Idx = Args[1]->CastInt32();
					if (Idx >= 0 AND Idx < Args[0]->Value.Lst->Length())
					{
						return Args[0]->Value.Lst->Delete(Idx);
					}
					break;
				}
			}
		}

		return false;
	}

	bool ReadTextFile(GVariant *Ret, ArgumentArray &Args)
	{
		if (Args.Length() == 1 AND
			FileExists(Args[0]->CastString()))
		{
			if (Ret->Value.String = ::ReadTextFile(Args[0]->CastString()))
			{
				Ret->Type = GV_STRING;
				return true;
			}
		}

		return false;
	}

	bool WriteTextFile(GVariant *Ret, ArgumentArray &Args)
	{
		if (Args.Length() == 2)
		{
			GFile f;
			if (f.Open(Args[0]->CastString(), O_WRITE))
			{
				f.SetSize(0);
				
				GVariant *v = Args[1];
				if (v)
				{
					switch (v->Type)
					{
						case GV_STRING:
						{
							int Len = strlen(v->Value.String);
							*Ret = f.Write(v->Value.String, Len) == Len;
							return true;
							break;
						}
						case GV_BINARY:
						{
							*Ret = f.Write(v->Value.Binary.Data, v->Value.Binary.Length) == v->Value.Binary.Length;
							return true;
							break;
						}
					}
				}
			}
		}

		return false;
	}
};

GHostFunc SystemLibrary[] =
{
	// String handling
	GHostFunc("Strchr", 0, (ScriptCmd)&SystemFunctions::Strchr),
	GHostFunc("Strstr", 0, (ScriptCmd)&SystemFunctions::Strstr),
	GHostFunc("Strcmp", 0, (ScriptCmd)&SystemFunctions::Strcmp),
	GHostFunc("Substr", 0, (ScriptCmd)&SystemFunctions::Substr),
	
	// Containers
	GHostFunc("NewHashTable", 0, (ScriptCmd)&SystemFunctions::NewHashTable),
	GHostFunc("NewList", 0, (ScriptCmd)&SystemFunctions::NewList),
	GHostFunc("DeleteElement", 0, (ScriptCmd)&SystemFunctions::DeleteElement),

	// File IO
	GHostFunc("ReadTextFile", 0, (ScriptCmd)&SystemFunctions::ReadTextFile),
	GHostFunc("WriteTextFile", 0, (ScriptCmd)&SystemFunctions::WriteTextFile),
	
	// End of list marker
	GHostFunc(0, 0, 0),
};

