#include "Os.h"
#include <process.h>
extern "C" {
	#define _IMPEXP_KERNEL
	#include "lock2.h"
	#include "cache.h"
	#include "fsproto.h"
}

#include <windows.h>

#define MAX_SIZE			64 // mb

class _Device
{
public:
	int Magic;
	HANDLE File;
	HANDLE Map;
	void *Ptr;
	int Refs;
	int Blocks;

	_Device()
	{
		Magic = 0x12345678;
		File = 0;
		Map = 0;
		Ptr = 0;
		Refs = 1;
		Blocks = 1;
	}

	~_Device()
	{
		Empty();
	}

	void Empty()
	{
		if (Ptr)
		{
			UnmapViewOfFile(Ptr);
			Ptr = 0;
		}

		if (Map)
		{
			CloseHandle(Map);
			Map = 0;
		}

		if (File)
		{
			CloseHandle(File);
			File = 0;
		}
	}

	bool IsOk()
	{
		bool Status = this != 0 && Magic == 0x12345678;
		assert(Status);
		return Status;
	}
};

class _Semaphore
{
public:
	int Magic;
	HANDLE Sem;
	int Thread;
	int Count;

	_Semaphore(int c)
	{
		Magic = 0x12340001;
		Sem = CreateSemaphore(0, min(c, 1000), 1000, 0);
		if (! Sem)
		{
			DWORD e = GetLastError();
			_asm int 3
		}

		Thread = find_thread(0);
		Count = c;
	}

	~_Semaphore()
	{
		if (Sem)
		{
			CloseHandle(Sem);
		}
	}

	bool IsOk()
	{
		bool Status = this != 0 && Magic == 0x12340001;
		assert(Status);
		return Status;
	}
};

sem_id create_sem(long thread_count, const char *name)
{
	return (sem_id) new _Semaphore(thread_count);
}

status_t delete_sem(sem_id sem)
{
	_Semaphore *s = (_Semaphore*)sem;
	if (s->IsOk())
	{
		delete s;
		return B_OK;
	}

	return B_ERROR;
}

status_t acquire_sem(sem_id sem)
{
	_Semaphore *s = (_Semaphore*)sem;
	if (s->IsOk() &&
		WaitForSingleObjectEx(s->Sem, INFINITE, false) == WAIT_OBJECT_0)
	{
		s->Count++;
		return B_OK;
	}

	return B_ERROR;
}

status_t acquire_sem_etc(sem_id sem, long count, long flags, double timeout)
{
	_Semaphore *s = (_Semaphore*)sem;
	if (s->IsOk())
	{
		DWORD e = WaitForSingleObjectEx(s->Sem, timeout, false);
		switch (e)
		{
			case WAIT_OBJECT_0:
				s->Count++;
				return B_OK;
			case WAIT_TIMEOUT:
				return B_TIMEOUT;
			case 0xFFFFFFFF:
			{
				DWORD e = GetLastError();
				_asm int 3
				break;
			}
		}
	}
	return B_ERROR;
}

status_t release_sem(sem_id sem)
{
	_Semaphore *s = (_Semaphore*)sem;
	if (s)
	{
		if (!ReleaseSemaphore(s->Sem, 1, 0))
		{
			return B_ERROR;
		}

		s->Count--;
		return B_OK;
	}
	return B_ERROR;
}

status_t release_sem_etc(sem_id sem, long count, long flags)
{
	_Semaphore *s = (_Semaphore*)sem;
	if (s->IsOk())
	{
		if (!ReleaseSemaphore(s->Sem, count, 0))
		{
			return B_ERROR;
		}

		s->Count -= count;
		return B_OK;
	}

	return B_ERROR;
}

status_t set_sem_owner(sem_id sem, team_id team)
{
	_asm int 3
	return 0;
}

thread_id find_thread(const char *name)
{
	if (name)
	{
		_asm int 3
	}
	else
	{
		return GetCurrentThreadId();
	}

	return 0;
}

status_t resume_thread(thread_id thread)
{
	if (thread)
	{
		return ResumeThread((HANDLE)thread) != -1 ? B_OK : B_ERROR;
	}

	return B_ERROR;
}

thread_id spawn_thread(thread_entry func, const char *name, long priority, void *data)
{
	uint ThreadId = 0;
	HANDLE t = (HANDLE) _beginthreadex(	NULL,
										16<<10,
										(unsigned int (__stdcall *)(void *)) func,
										(LPVOID) data,
										CREATE_SUSPENDED,
										&ThreadId);
	return (thread_id)t;
}

long kill_thread(thread_id thread)
{
	if (thread)
	{
		TerminateThread((HANDLE)thread, -1);
		return B_OK;
	}

	return B_ERROR;
}

status_t snooze(bigtime_t microseconds)
{
	Sleep(microseconds / 1000);
	return 0;
}

bigtime_t time(void *p)
{
	return system_time();
}

bigtime_t system_time()
{
	return GetTickCount();
}

int fstat(int fd, struct stat *s)
{
	_Device *d = (_Device*) fd;
	
	BY_HANDLE_FILE_INFORMATION i;
	if (GetFileInformationByHandle(d->File, &i))
	{
		s->st_dev = 0; // device id
		s->st_ino = 0; // serial #
		s->st_mode = i.dwFileAttributes;
		s->st_nlink = 0; // # of links to the file
		s->st_uid = 0; // user if
		s->st_gid = 0; // group id
		s->st_size = ((uint64)i.nFileSizeHigh << 32) | i.nFileSizeLow;
		s->st_atime = ((bigtime_t)i.ftLastAccessTime.dwHighDateTime << 32) | (i.ftLastAccessTime.dwLowDateTime);
		s->st_mtime = ((bigtime_t)i.ftLastWriteTime.dwHighDateTime << 32) | i.ftLastWriteTime.dwLowDateTime;
		s->st_blksize = 4 << 10;
		s->st_ctime = ((bigtime_t)i.ftCreationTime.dwHighDateTime << 32) | i.ftCreationTime.dwLowDateTime;
		s->st_crtime = ((bigtime_t)i.ftCreationTime.dwHighDateTime << 32) | i.ftCreationTime.dwLowDateTime;
		return 0;
	}

	return -1;
}

int open(const char *name, int flags)
{
	_Device *d = new _Device;
	if (d)
	{
		int mode = 0;
		if (flags & O_RDONLY)
		{
			mode = GENERIC_READ;
		}
		else if (flags & O_WRONLY)
		{
			mode = GENERIC_WRITE;
		}
		else if (flags & O_RDWR)
		{
			mode = GENERIC_READ | GENERIC_WRITE;
		}

		d->File = CreateFile(name, mode, 0, 0, OPEN_ALWAYS, FILE_ATTRIBUTE_ARCHIVE | FILE_FLAG_RANDOM_ACCESS, 0);
		if (d->File)
		{
			d->Map = CreateFileMapping(d->File, 0, PAGE_READWRITE, 0, MAX_SIZE << 20, 0);
			if (d->Map)
			{
				d->Ptr = MapViewOfFileEx(d->Map, FILE_MAP_WRITE | FILE_MAP_READ, 0, 0, 0, 0);
				if (!d->Ptr)
				{
					DWORD e = GetLastError();
					_asm int 3
				}
			}
			else
			{
				DWORD e = GetLastError();
				_asm int 3
			}
		}
		else
		{
			DWORD e = GetLastError();
			_asm int 3
		}
	}

	return (int)d;
}

void close(int f)
{
	_Device *d = (_Device*) f;
	if (d->IsOk())
	{
		d->Refs--;
		if (d->Refs < 1)
		{
			d->Empty();
		}
	}
}

size_t read_pos(int fd, off_t offset, uchar *buf, size_t size)
{
	_Device *d = (_Device*)fd;
	DWORD r = 0;
	if (d->IsOk())
	{
		/*
		LONG h = offset >> 32;
		SetFilePointer(d->File, offset, &h, FILE_BEGIN);
		ReadFile(d->File, buf, size, &r, 0);
		*/

		memcpy(buf, (char*)d->Ptr + offset, size);
		r = size;
	}
	return r;
}

size_t write_pos(int fd, off_t offset, const uchar *buf, size_t size)
{
	DWORD w = 0;
	_Device *d = (_Device*)fd;
	if (d->IsOk())
	{
		/*
		LONG h = offset >> 32;
		SetFilePointer(d->File, offset, &h, FILE_BEGIN);
		WriteFile(d->File, buf, size, &w, 0);
		*/

		memcpy((char*)d->Ptr + offset, buf, size);
		w = size;
	}
	return w;
}

int ioctl(int fildes, int request, ...)
{
	_asm int 3
	return 0;
}

uid_t geteuid()
{
	return 0; // super user
}

gid_t getegid()
{
	return 0; // super user
}

int32 atomic_add(vint32 *atomicVariable, int32 addValue)
{
	return InterlockedExchangeAdd((long*)atomicVariable, addValue);
}

int32 atomic_and(vint32 *atomicVariable, int32 andValue)
{
	_asm int 3
	return 0;
}

int32 atomic_or(vint32 *atomicVariable, int32 orValue)
{
	_asm int 3
	return 0;
}

char *__PRETTY_FUNCTION__()
{
	_asm int 3
	return 0;
}

int64 strtoll(char *s, char **end, int base)
{
	_asm int 3
	return 0;
}

uint64 strtoull(char *s, char **end, int base)
{
	_asm int 3
	return 0;
}



////////////////////////////////////////////////////////////////////////////////////
int init_block_cache(int max_blocks, int flags)
{
	_asm int 3
	return 0;
}

void shutdown_block_cache(void)
{
	_asm int 3
}

void force_cache_flush(int dev, int prefer_log_blocks)
{
	_asm int 3
}

int flush_blocks(int dev, off_t bnum, int nblocks)
{
	_asm int 3
	return 0;
}

int flush_device(int dev, int warn_locked)
{
	_asm int 3
	return 0;
}

int init_cache_for_device(int fd, off_t max_blocks)
{
	return B_OK;
}

int remove_cached_device_blocks(int dev, int allow_write)
{
	// _asm int 3
	return 0;
}

void *get_block(int dev, off_t bnum, int bsize)
{
	_Device *d = (_Device*)dev;
	if (d->IsOk() && d->Ptr)
	{
		d->Blocks = max(d->Blocks, bnum + 1);
		return (char*)d->Ptr + (bnum * bsize);
	}
	return 0;
}

void *get_empty_block(int dev, off_t bnum, int bsize)
{
	_Device *d = (_Device*)dev;
	if (d->IsOk())
	{
		return get_block(dev, bnum, bsize);
	}
	return 0;
}

int release_block(int dev, off_t bnum)
{
	return B_OK;
}

int mark_blocks_dirty(int dev, off_t bnum, int nblocks)
{
	_asm int 3
	return B_ERROR;
}

int cached_read(int dev, off_t bnum, void *data, off_t num_blocks, int bsize)
{
	_Device *d = (_Device*)dev;
	if (data && d->IsOk() && d->Ptr)
	{
		d->Blocks = max(d->Blocks, bnum + num_blocks);
		memcpy(data, (char*)d->Ptr + (bnum * bsize), num_blocks * bsize);
		return B_OK;
	}
	return B_ERROR;
}

int cached_write(int dev, off_t bnum, const void *data, off_t num_blocks, int bsize)
{
	_Device *d = (_Device*)dev;
	if (data && d->IsOk() && d->Ptr)
	{
		assert(bnum > 0);

		d->Blocks = max(d->Blocks, bnum + num_blocks);
		memcpy((char*)d->Ptr + (bnum * bsize), data, num_blocks * bsize);
		return B_OK;
	}
	return B_ERROR;
}

int cached_write_locked(int dev, off_t bnum, const void *data, off_t num_blocks, int bsize)
{
	_Device *d = (_Device*)dev;
	if (data && d->IsOk() && d->Ptr)
	{
		assert(bnum > 0);

		d->Blocks = max(d->Blocks, bnum + num_blocks);
		memcpy((char*)d->Ptr + (bnum * bsize), data, num_blocks * bsize);
		return B_OK;
	}
	return B_ERROR;
}

int set_blocks_info(int dev, off_t *blocks, int nblocks, void (*func)(off_t bnum, size_t nblocks, void *arg), void *arg)
{
	_asm int 3
	return B_ERROR;
}

size_t read_phys_blocks (int fd, off_t bnum, void *data, uint num_blocks, int bsize)
{
	_asm int 3
	return 0;
}

size_t write_phys_blocks(int fd, off_t bnum, void *data, uint num_blocks, int bsize)
{
	_asm int 3
	return 0;
}


/////////////////////////////////////////////////////////////////////////////////////
int	new_path(const char *path, char **copy)
{
	_asm int 3
	return 0;
}

void free_path(char *p)
{
	_asm int 3
}

int	notify_listener(int op, nspace_id nsid, vnode_id vnida,	vnode_id vnidb, vnode_id vnidc, const char *name)
{
	return 0;
}

void notify_select_event(selectsync *sync, uint32 ref)
{
	_asm int 3
}

int	send_notification(port_id port, long token, ulong what, long op, nspace_id nsida, nspace_id nsidb, vnode_id vnida, vnode_id vnidb, vnode_id vnidc, const char *name)
{
	_asm int 3
	return 0;
}





#define NODE_DB_SIZE		2048

class _Vnode
{
public:
	nspace_id ns;
	vnode_id id;
	void *data;
	int ref;

	_Vnode()
	{
		ns = 0;
		id = 0;
		data = 0;
		ref = 1;
	}
};

int Nodes = 0;
_Vnode Node[NODE_DB_SIZE];

_Vnode *GetNode(nspace_id nsid, vnode_id vnid)
{
	for (int i=0; i<Nodes; i++)
	{
		if (Node[i].ns == nsid &&
			Node[i].id == vnid)
		{
			return Node + i;
		}
	}

	return 0;
}

int	get_vnode(nspace_id nsid, vnode_id vnid, void **data)
{
	_Vnode *v = GetNode(nsid, vnid);
	if (v)
	{
		*data = v->data;
		return B_OK;
	}

	return B_ERROR;
}

int	put_vnode(nspace_id nsid, vnode_id vnid)
{
	_Vnode *v = GetNode(nsid, vnid);
	if (v)
	{
		return v->ref--;
	}
	return 0;
}

int	new_vnode(nspace_id nsid, vnode_id vnid, void *data)
{
	if (Nodes >= NODE_DB_SIZE)
	{
		return B_ERROR;
	}

	_Vnode *v = Node + Nodes;
	v->ns = nsid;
	v->id = vnid;
	v->data = data;
	v->ref = 1;
	Nodes++;
	return B_OK;
}

int	remove_vnode(nspace_id nsid, vnode_id vnid)
{
	_asm int 3
	return 0;
}

int	unremove_vnode(nspace_id nsid, vnode_id vnid)
{
	_asm int 3
	return 0;
}

int	is_vnode_removed(nspace_id nsid, vnode_id vnid)
{
	_asm int 3
	return 0;
}

