/*hdr
**      FILE:           Jpeg.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           8/9/1998
**      DESCRIPTION:    Jpeg file filter
**
**      Copyright (C) 1998, Matthew Allen
**              fret@memecode.com
*/

#include "Lgi.h"
#if HAS_LIBJPEG

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "GFilterUtils.h"
#include "GScrollBar.h"

#define XMD_H
#undef FAR

//
// 'jpeglib.h' comes from libjpeg available here:
// http://freshmeat.net/projects/libjpeg
//
// If you don't want to build with JPEG support then set
// the define HAS_LIBJPEG to '0' in Lgi.h
//
#include "jpeglib.h"
#include <setjmp.h>

#define IJG_JPEG_ICC_MARKER             JPEG_APP0 + 2
#define IJG_JPEG_APP_ICC_PROFILE_LEN    0xFFFF
#define IJG_MAXIMUM_SEQUENCE_NUMBER     255
#define IJG_SEQUENCE_NUMBER_INDEX       12
#define IJG_NUMBER_OF_MARKERS_INDEX     13
#define IJG_JFIF_ICC_HEADER_LENGTH      14

// JPEG
class LibJpeg : public GLibrary
{
public:
	LibJpeg() :
		#if defined(WIN32) && defined(_DEBUG)
		GLibrary("libjpegd")
		#else
		GLibrary("libjpeg")
		#endif
	{
		#if 0
		char File[256];
		GetModuleFileName(Handle(), File, sizeof(File));
		LgiTrace("%s:%i - JPEG: %s\n", __FILE__, __LINE__, File);
		#endif
	}

	DynFunc1(boolean, jpeg_finish_decompress, j_decompress_ptr, cinfo);
	DynFunc3(JDIMENSION, jpeg_read_scanlines, j_decompress_ptr, cinfo, JSAMPARRAY, scanlines, JDIMENSION, max_lines);
	DynFunc1(boolean, jpeg_start_decompress, j_decompress_ptr, cinfo);
	DynFunc2(int, jpeg_read_header, j_decompress_ptr, cinfo, boolean, require_image);
	DynFunc2(int, jpeg_stdio_src, j_decompress_ptr, cinfo, FILE*, infile);
	DynFunc1(int, jpeg_destroy_decompress, j_decompress_ptr, cinfo);
	DynFunc1(int, jpeg_finish_compress, j_compress_ptr, cinfo);
	DynFunc1(int, jpeg_destroy_compress, j_compress_ptr, cinfo);
	DynFunc3(JDIMENSION, jpeg_write_scanlines, j_compress_ptr, cinfo, JSAMPARRAY, scanlines, JDIMENSION, num_lines);
	DynFunc2(int, jpeg_start_compress, j_compress_ptr, cinfo, boolean, write_all_tables);
	DynFunc1(int, jpeg_set_defaults, j_compress_ptr, cinfo);
	DynFunc1(struct jpeg_error_mgr *, jpeg_std_error, struct jpeg_error_mgr *, err);
	DynFunc2(int, jpeg_stdio_dest, j_compress_ptr, cinfo, FILE *, outfile);
	DynFunc3(int, jpeg_CreateCompress, j_compress_ptr, cinfo, int, version, size_t, structsize);
	DynFunc3(int, jpeg_CreateDecompress, j_decompress_ptr, cinfo, int, version, size_t, structsize);
	DynFunc3(int, jpeg_set_quality, j_compress_ptr, cinfo, int, quality, boolean, force_baseline);

	DynFunc3(int, jpeg_save_markers, j_decompress_ptr, cinfo, int, marker_code, unsigned int, length_limit);

};

class GdcJpeg : public GFilter, public LibJpeg
{
	friend class GJpegOptions;

	bool _Write(GSurface *pDC, char *filename, int Quality);

public:
	GdcJpeg();

	int GetCapabilites() { return FILTER_CAP_READ | FILTER_CAP_WRITE; }
	bool ReadImage(GSurface *pDC);
	bool WriteImage(GSurface *pDC);
};

/////////////////////////////////////////////////////////////////////
#ifdef FILTER_UI
#include "Lgi.h"
#include "../../../i.Mage/Code/resdefs.h"

class GJpegOptions : public GDialog
{
	GScrollBar *Scroll;
	GdcJpeg *Jpeg;
	GSurface *pDC;

	void Update()
	{
		char Temp[256];
		LgiGetTempPath(Temp, sizeof(Temp));
		LgiMakePath(Temp, sizeof(Temp), Temp, "~jpeg.tmp");

		if (Jpeg->_Write(pDC, Temp, Quality))
		{
			int Size = LgiFileSize(Temp);
			if (Size >= 0)
			{
				char s[256];
				LgiFormatSize(s, Size);
				SetCtrlName(IDC_FILE_SIZE, s);
			}

			FileDev->DeleteFile(Temp, false);
		}
	}

public:
	int Quality;

	GJpegOptions(GdcJpeg *jpeg, GSurface *pdc)
	{
		Jpeg = jpeg;
		pDC = pdc;
		GView *Parent = 0;
		Jpeg->Props.Get("parent", (int&)Parent);
		SetParent(Parent);

		Scroll = 0;
		Quality = 50;
		if (LoadFromResource(IDD_JPEG))
		{
			MoveToCenter();

			Scroll = dynamic_cast<GScrollBar*>(FindControl(IDC_QUALITY));
			if (Scroll)
			{
				Scroll->SetPage(5);
				Scroll->SetLimits(0, 100);
				Scroll->Value(Quality);
			}
		}
	}

	void OnCreate()
	{
		Update();
	}

	int OnNotify(GViewI *c, int f)
	{
		switch (c->GetId())
		{
			case IDC_QUALITY:
			{
				Quality = c->Value();
				Update();
				break;
			}
			case IDOK:
			{
				Quality = Scroll ? Scroll->Value() : -1;
				// fall thru
			}
			case IDCANCEL:
			{
				EndModal(c->GetId());
				break;
			}
		}

		return 0;
	}
};
#endif

/////////////////////////////////////////////////////////////////////
class GdcJpegFactory : public GFilterFactory
{
	bool CheckFile(char *File, int Access, uchar *Hint)
	{
		if (Hint)
		{
			if (Hint[6] == 'J' AND
				Hint[7] == 'F' AND
				Hint[8] == 'I' AND
				Hint[9] == 'F')
			{
				return true;
			}
		}

		return (File) ? stristr(File, ".jpeg") != 0 OR
						stristr(File, ".jpg") != 0 : false;
	}

	GFilter *NewObject()
	{
		return NEW(GdcJpeg);
	}

} JpegFactory;

/////////////////////////////////////////////////////////////////////
struct my_error_mgr {
  struct jpeg_error_mgr pub;	/* "public" fields */

  jmp_buf setjmp_buffer;	/* for return to caller */
};

typedef struct my_error_mgr *my_error_ptr;

METHODDEF(void)
my_error_exit (j_common_ptr cinfo)
{
	my_error_ptr myerr = (my_error_ptr) cinfo->err;
	(*cinfo->err->output_message) (cinfo);
	longjmp(myerr->setjmp_buffer, 1);
}

bool IsJfifIccMarker(jpeg_saved_marker_ptr marker)
{
    return	marker->marker == IJG_JPEG_ICC_MARKER AND
			marker->data_length >= IJG_JFIF_ICC_HEADER_LENGTH AND
			GETJOCTET(marker->data[0]) == 0x49 AND
			GETJOCTET(marker->data[1]) == 0x43 AND
			GETJOCTET(marker->data[2]) == 0x43 AND
			GETJOCTET(marker->data[3]) == 0x5F AND
			GETJOCTET(marker->data[4]) == 0x50 AND
			GETJOCTET(marker->data[5]) == 0x52 AND
			GETJOCTET(marker->data[6]) == 0x4F AND
			GETJOCTET(marker->data[7]) == 0x46 AND
			GETJOCTET(marker->data[8]) == 0x49 AND
			GETJOCTET(marker->data[9]) == 0x4C AND
			GETJOCTET(marker->data[10]) == 0x45 AND
			GETJOCTET(marker->data[11]) == 0x0; 
}

bool GetIccProfile(j_decompress_ptr cinfo, uchar **icc_data, uint *icc_datalen)
{
	int num_markers = 0;
	jpeg_saved_marker_ptr markers = cinfo->marker_list;
	JOCTET *local_icc_data = NULL;
	int count = 0;
	uint total_length;
	bool marker_present[IJG_MAXIMUM_SEQUENCE_NUMBER];
	uint data_length[IJG_MAXIMUM_SEQUENCE_NUMBER];
	uint marker_sequence[IJG_MAXIMUM_SEQUENCE_NUMBER];

	if (icc_data == NULL || icc_datalen == NULL)
		return false;

	*icc_data = NULL;
	*icc_datalen = 0;

	ZeroObj(marker_present);
	for (markers = cinfo->marker_list; markers; markers = markers->next)
	{
		if (IsJfifIccMarker(markers))
		{
			int sequence_number = 0;
			if (num_markers == 0)
			{
				num_markers = GETJOCTET(markers->data[IJG_NUMBER_OF_MARKERS_INDEX]);
			}
			else if (num_markers != GETJOCTET(markers->data[IJG_NUMBER_OF_MARKERS_INDEX]))
			{
				return false;
			}

			sequence_number = GETJOCTET(markers->data[IJG_SEQUENCE_NUMBER_INDEX]) - 1;
			if (sequence_number < 0 || sequence_number > num_markers)
			{
				return false;
			}
			if (marker_present[sequence_number])
				return false;

			marker_present[sequence_number] = true;
			data_length[sequence_number] = 
				markers->data_length - IJG_JFIF_ICC_HEADER_LENGTH;
			marker_sequence[sequence_number] = count;
			count++;
	   }

	}

	if (num_markers == 0)
		return false;

	total_length = 0;
	markers = cinfo->marker_list;
	for (count = 0; count < num_markers; count++)
	{
		int	 previous_length = total_length;
		int	 index		   = 0;
		int	 i			   = 0;
		JOCTET  *src_pointer;
		JOCTET  *dst_pointer;

		index = marker_sequence[count];
		if (!(marker_present[count]))
			return false;
		total_length += data_length[count];
		local_icc_data = (JOCTET *)malloc(total_length);
		if (local_icc_data == NULL)
			return false;

		for (i = 0; i < index; i++)
		{
			markers = markers->next;
			if (NOT markers)
			{
				return false;
			}
		}

		dst_pointer = local_icc_data + previous_length;
		src_pointer = markers->data + IJG_JFIF_ICC_HEADER_LENGTH;
		while (data_length[count]--)
		{
			*dst_pointer++ = *src_pointer++;
		}
	}

	*icc_data = local_icc_data;
	*icc_datalen = total_length;

	return true;
}

bool GdcJpeg::ReadImage(GSurface *pDC)
{
	bool Status = false;
	if (NOT IsLoaded())
	{
		Props.Set(LGI_FILTER_ERROR, "libjpeg library isn't installed or wasn't usable.");
		return false;
	}

	if (NOT pDC)
	{
		return false;
	}

	char *filename = NewStr(GetName());
	if (filename)
	{
		if (IsOpen())
		{
			Close();
		}

		struct jpeg_decompress_struct cinfo;
		struct my_error_mgr jerr;

		FILE *infile;
		int row_stride;

		char16 *WideFile = LgiNewUtf8To16(filename);

		#ifdef WIN32
		infile = _wfopen(WideFile, L"rb");
		#else
		infile = fopen(filename, "rb");
		#endif

		DeleteArray(WideFile);
		if (infile)
		{
			int FILE_Size = sizeof(FILE);

			cinfo.err = jpeg_std_error(&jerr.pub);
			jerr.pub.error_exit = my_error_exit;

			if (setjmp(jerr.setjmp_buffer))
			{
				jpeg_destroy_decompress(&cinfo);
				fclose(infile);
				return 0;
			}

			cinfo.client_data = 0;
			jpeg_create_decompress(&cinfo);
			jpeg_stdio_src(&cinfo, infile);

			jpeg_save_markers
			(
				&cinfo,
				IJG_JPEG_ICC_MARKER,
				IJG_JPEG_APP_ICC_PROFILE_LEN
			);
			jpeg_read_header(&cinfo, true);
			
			uchar *icc_data = 0;
			uint icc_len = 0;
			if (GetIccProfile(&cinfo, &icc_data, &icc_len))
			{
				Props.Set(LGI_FILTER_COLOUR_PROF, icc_data, icc_len);
				free(icc_data);
			}

			if (pDC->Create(cinfo.image_width, cinfo.image_height, cinfo.num_components*8))
			{
				// zero out bitmap
				pDC->Colour(0);
				pDC->Rectangle();

				// Progress
				if (Meter)
				{
					Meter->SetDescription("scanlines");
					Meter->SetLimits(0, cinfo.image_height-1);
				}

				// Read
				if (cinfo.num_components == 1)
				{
					// greyscale
					GPalette *Pal = NEW(GPalette(0, 256));
					if (Pal)
					{
						GdcRGB *p = (*Pal)[0];
						if (p)
						{
							for (int i=0; i<256; i++, p++)
							{
								p->R = i;
								p->G = i;
								p->B = i;
							}
						}
						pDC->Palette(Pal);
					}
				}

				// setup decompress
				jpeg_start_decompress(&cinfo);
				row_stride = cinfo.output_width * cinfo.output_components;

				#define Rc 0
				#define Gc (MaxRGB+1)
				#define Bc (MaxRGB+1)*2
				#define DownShift(x) ((int) ((x)+(1L << 15)) >> 16)
				#define UpShifted(x) ((int) ((x)*(1L << 16)+0.5))
				#define MaxRGB                          255

				long  *red   = NEW(long[3*(MaxRGB+1)]);
				long  *green = NEW(long[3*(MaxRGB+1)]);
				long  *blue  = NEW(long[3*(MaxRGB+1)]);
				uchar *range_table = NEW(uchar[3*(MaxRGB+1)]);

				if (red AND
					green AND
					blue AND
					range_table)
				{
					int i;
					for (i=0; i<=MaxRGB; i++)
					{
						range_table[i]=0;
						range_table[i+(MaxRGB+1)]=(uchar) i;
						range_table[i+(MaxRGB+1)*2]=MaxRGB;
					}
					uchar *range_limit=range_table+(MaxRGB+1);

					for (i=0; i<=MaxRGB; i++)
					{
						red[i+Rc]   = UpShifted(1.00000)*i;
						green[i+Rc] = 0;
						blue[i+Rc]  = UpShifted(1.40200*0.5) * ((i << 1)-MaxRGB);

						red[i+Gc]   = UpShifted(1.00000)*i;
						green[i+Gc] = (-UpShifted(0.34414*0.5)) * ((i << 1)-MaxRGB);
						blue[i+Gc]  = (-UpShifted(0.71414*0.5)) * ((i << 1)-MaxRGB);

						red[i+Bc]   = UpShifted(1.00000)*i;
						green[i+Bc] = UpShifted(1.77200*0.5) * ((i << 1)-MaxRGB);
						blue[i+Bc]  = 0;
					}

					class JpegRgb
					{
					public:
						uchar r, g, b;
					};
					
					class JpegXyz
					{
					public:
						uchar x, y, z;
					};
					
					class JpegCmyk
					{
					public:
						uchar c, m, y, k;
					};					
					
					// loop through scanlines
					while (cinfo.output_scanline < cinfo.output_height)
					{
						uchar *Ptr = (*pDC)[cinfo.output_scanline];
						if (Ptr)
						{
							if (jpeg_read_scanlines(&cinfo, &Ptr, 1))
							{
								switch (cinfo.jpeg_color_space)
								{
									case JCS_GRAYSCALE:
									{
										// do nothing
										break;
									}
									case JCS_CMYK:
									{
										/*
										RGB to CMY                         
										Cyan    = 1-Red                    
										Magenta = 1-Green                 
										Yellow  = 1-Blue                  

										CMY to RGB                   
										Red   = 1-Cyan
										Green = 1-Magenta 
										Blue  = 1-Yellow

										CMY to CMYK                             
										Black   = minimum (Cyan,Magenta,Yellow) 
										Cyan    = (Cyan-Black)/(1-Black)                
										Magenta = (Magenta-Black)/(1-Black)     
										Yellow  = (Yellow-Black)/(1-Black)      

										CMYK to CMY
										Cyan    = minimum(1,Cyan*(1-Black)+Black)
										Magenta = minimum(1,Magenta*(1-Black)+Black)
										Yellow  = minimum(1,Yellow*(1-Black)+Black)
										*/

										if (cinfo.num_components == 4)
										{
											Pixel32 *d = (Pixel32*) Ptr;
											
											for (int x=0; x<pDC->X(); x++)
											{
												JpegCmyk *s = (JpegCmyk*) d;
												
												double C = (double) s->c / 255;
												double M = (double) s->m / 255;
												double Y = (double) s->y / 255;
												double K = (double) s->k / 255;

												d->r = (1.0 - min(1, C * (1.0 - K) + K)) * 255;
												d->g = (1.0 - min(1, M * (1.0 - K) + K)) * 255;
												d->b = (1.0 - min(1, Y * (1.0 - K) + K)) * 255;
												d->a = 255;

												d++;
											}
										}
										break;
									}
									case JCS_YCCK: // YCbCrK
									case 10000: // YCbCr
									{
										if (cinfo.num_components == 3)
										{
											JpegXyz *s = (JpegXyz*) Ptr;
											Pixel24 *d = (Pixel24*) Ptr;
											
											for (int x=pDC->X()-1; x>=0; x--)
											{
												/*
												Y =	0.299 * R	+ 0.587 * G	+ 0.114 * B 
												Cb =	-0.169 * R	- 0.331 * G	+ 0.500 * B 
												Cr =	0.500 * R	- 0.419 * G	- 0.081 * B
												*/
												
												int x = s->x;
												int y = s->y;
												int z = s->z;
												
												d->r = range_limit[DownShift(red[x+Rc] + green[y+Rc] + blue[z+Rc])];
												d->g = range_limit[DownShift(red[x+Gc] + green[y+Gc] + blue[z+Gc])];
												d->b = range_limit[DownShift(red[x+Bc] + green[y+Bc] + blue[z+Bc])];
												d = d->Next();
												s++;
											}
										}
										break;
									}
									case JCS_YCbCr:
									case JCS_RGB:
									{
										if (cinfo.num_components == 3)
										{
											Pixel24 *d = (Pixel24*) Ptr;
											JpegRgb *e = (JpegRgb*) Ptr;
											JpegRgb *s = e;
											
											((char*&)d) += Pixel24::Size * (pDC->X()-1);
											s += pDC->X() - 1;
											
											while (s >= e)
											{
												JpegRgb t = *s--;
												
												d->r = t.r;
												d->g = t.g;
												d->b = t.b;
												
												((uchar*&)d) -= Pixel24::Size;
											}
										}
										break;
									}
								}
							}
						}

						if (Meter) Meter->Value(cinfo.output_scanline);
					}
				}

				DeleteArray(red);
				DeleteArray(green);
				DeleteArray(blue);
				DeleteArray(range_table);

				jpeg_finish_decompress(&cinfo);

				Status = true;
			}

			jpeg_destroy_decompress(&cinfo);
			fclose(infile);
		}

		DeleteArray(filename);
	}

	return Status;
}

bool GdcJpeg::WriteImage(GSurface *pDC)
{
	bool Status = FALSE;
	if (NOT IsLoaded())
	{
		Props.Set(LGI_FILTER_ERROR, "libjpeg library isn't installed or wasn't usable.");
		return false;
	}

	bool Ok = true;
	int Quality = 50;
	#ifdef FILTER_UI
	GJpegOptions Dlg(this, pDC);
	if (Ok = (Dlg.DoModal() == IDOK))
	{
		Quality = Dlg.Quality;
	}
	#endif

	if (Ok)
	{
		char *filename = NewStr(GetName());
		if (IsOpen())
		{
			Close();
		}
		FileDev->DeleteFile(filename, false);

		Status = _Write(pDC, filename, Quality);

		DeleteArray(filename);
	}

	return Status;
}

bool GdcJpeg::_Write(GSurface *pDC, char *filename, int Quality)
{
	bool Status = false;
	struct jpeg_compress_struct cinfo;
	struct my_error_mgr jerr;

	FILE *file;
	int row_stride;

	if (pDC AND (file = fopen(filename, "wb")))
	{
		cinfo.err = jpeg_std_error(&jerr.pub);
		jpeg_create_compress(&cinfo);
		jerr.pub.error_exit = my_error_exit;

		if (setjmp(jerr.setjmp_buffer))
		{
			jpeg_destroy_compress(&cinfo);
			fclose(file);
			return 0;
		}

		jpeg_stdio_dest(&cinfo, file);

		bool GreyScale = FALSE;
		GPalette *Pal = pDC->Palette();
		if (pDC->GetBits() == 8)
		{
			if (Pal)
			{
				GreyScale = true;
				for (int i=0; i<Pal->GetSize(); i++)
				{
					GdcRGB *p = (*Pal)[i];
					if (	p AND
						   (p->R != i OR
						p->G != i OR
						p->B != i))
					{
						GreyScale = FALSE;
						break;
					}
				}
			}
			else if (NOT Pal)
			{
				GreyScale = true;
			}
		}

		cinfo.image_width = pDC->X();
		cinfo.image_height = pDC->Y();
		if (GreyScale)
		{
			cinfo.input_components = 1;
			cinfo.in_color_space = JCS_GRAYSCALE;
		}
		else
		{
			cinfo.input_components = 3;
			cinfo.in_color_space = JCS_RGB;
		}

		jpeg_set_defaults(&cinfo);
		
		if (Quality >= 0)
		{
			jpeg_set_quality(&cinfo, Quality, true);
		}

		jpeg_start_compress(&cinfo, true);

		row_stride = pDC->X() * cinfo.input_components;
		uchar *Buffer = NEW(uchar[row_stride]);
		if (Buffer)
		{
			// Progress
			if (Meter)
			{
				Meter->SetDescription("scanlines");
				Meter->SetLimits(0, cinfo.image_height-1);
			}

			// Write
			while (cinfo.next_scanline < cinfo.image_height)
			{
				uchar *d = Buffer;

				switch (pDC->GetBits())
				{
					case 8:
					{
						if (GreyScale)
						{
							memcpy(Buffer, (*pDC)[cinfo.next_scanline], pDC->X());
						}
						else if (Pal)
						{
							GdcRGB *p = (*Pal)[0];
							if (p)
							{
								for (int x=0; x<pDC->X(); x++)
								{
									COLOUR c = pDC->Get(x, cinfo.next_scanline);
									*d++ = p[c].R;
									*d++ = p[c].G;
									*d++ = p[c].B;
								}
							}
						}
						break;
					}
					case 16:
					{
						for (int x=0; x<pDC->X(); x++)
						{
							COLOUR c = pDC->Get(x, cinfo.next_scanline);
							*d++ = (R16(c) << 3) | (R16(c) >> 5);
							*d++ = (G16(c) << 2) | (G16(c) >> 6);
							*d++ = (B16(c) << 3) | (B16(c) >> 5);
						}
						break;
					}
					case 24:
					{
						for (int x=0; x<pDC->X(); x++)
						{
							COLOUR c = pDC->Get(x, cinfo.next_scanline);
							*d++ = R24(c);
							*d++ = G24(c);
							*d++ = B24(c);
						}
						break;
					}
					case 32:
					{
						for (int x=0; x<pDC->X(); x++)
						{
							COLOUR c = pDC->Get(x, cinfo.next_scanline);
							*d++ = R32(c);
							*d++ = G32(c);
							*d++ = B32(c);
						}
						break;
					}
				}

				jpeg_write_scanlines(&cinfo, &Buffer, 1);
				if (Meter) Meter->Value(cinfo.next_scanline);
			}

			DeleteArray(Buffer);
		}

		jpeg_finish_compress(&cinfo);
		jpeg_destroy_compress(&cinfo);
		fclose(file);

		Status = true;
	}

	return Status;
}

GdcJpeg::GdcJpeg()
{
	Props.Set("Type", "JPEG");
	Props.Set("Extension", "JPG,JPEG");
}

#endif
