/*hdr
**	FILE:			GFont.cpp
**	AUTHOR:			Matthew Allen
**	DATE:			5/5/97
**	DESCRIPTION:	Gdc2 Font Support
**
**	Copyright (C) 1997-2002, Matthew Allen
**		fret@memecode.com
*/

#define _WIN32_WINNT	0x500

//////////////////////////////////////////////////////////////////////
// Includes
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "Lgi.h"
#include "GVariant.h"
#include "GFontSelect.h"
#include "GdiLeak.h"

#ifdef FontChange
#undef FontChange
#endif

int _create_style = 0;
int _create_layout = 0;
int _draw_layout = 0;
int64 _last = 0;

//////////////////////////////////////////////////////////////////////
// Helpers
#ifdef WIN32

#ifndef __GNUC__
#include <mbctype.h>
#endif

int WinPointToHeight(int Pt)
{
	int Ht = 0;
	HWND hWnd = GetDesktopWindow();
	HDC hDC = GetDC(hWnd);
	if (hDC)
	{
		Ht = -MulDiv(Pt, GetDeviceCaps(hDC, LOGPIXELSY), 72);
		ReleaseDC(hWnd, hDC);
	}

	return Ht;
}

int WinHeightToPoint(int Ht)
{
	int Pt = 0;
	HWND hWnd = GetDesktopWindow();
	HDC hDC = GetDC(hWnd);
	if (hDC)
	{
		Pt = -MulDiv(Ht, 72, GetDeviceCaps(hDC, LOGPIXELSY));
		ReleaseDC(hWnd, hDC);
	}

	return Pt;
}
#endif

//////////////////////////////////////////////////////////////////////
// Classes
class GTypeFacePrivate
{
public:
	// Type
	char *_Face;			// type face
	int _PtSize;			// point size
	int _Weight;
	bool _Italic;
	bool _Underline;
	char *_CodePage;
	int _Quality;

	// Output
	COLOUR _Fore;
	COLOUR _Back;
	int _TabSize;
	bool _Transparent;
	bool _SubGlyphs;

	// Props
	double _Ascent, _Descent;

	GTypeFacePrivate()
	{
		_Ascent = _Descent = 0;
		_Face = 0;
		_PtSize = 8;
		_Weight = FW_NORMAL;
		_Italic = false;
		_Underline = false;
		_CodePage = NewStr("utf-8");
		_Fore = Rgb24(0, 0, 0);
		_Back = Rgb24(255, 255, 255);
		_TabSize = 32;
		_Transparent = false;
		_SubGlyphs = GFontSystem::Inst()->GetDefaultGlyphSub();
	}

	~GTypeFacePrivate()
	{
		DeleteArray(_Face);
		DeleteArray(_CodePage);
	}
};

GTypeFace::GTypeFace()
{
	d = NEW(GTypeFacePrivate);
}

GTypeFace::~GTypeFace()
{
	DeleteObj(d);
}

// set
void GTypeFace::Face(char *s)
{
	if (s AND
		s != d->_Face AND
		stricmp(s, d->_Face?d->_Face:(char*)"") != 0)
	{
		DeleteArray(d->_Face);
		d->_Face = NewStr(s);
		_OnPropChange(true);
	}
}

void GTypeFace::PointSize(int i)
{
	d->_PtSize = i;
	_OnPropChange(true);
}

void GTypeFace::TabSize(int i)
{
	d->_TabSize = max(i, 8);
	_OnPropChange(false);
}

void GTypeFace::Quality(int i)
{
	d->_Quality = i;
	_OnPropChange(true);
}

void GTypeFace::Fore(COLOUR c)
{
	d->_Fore = c;
	_OnPropChange(false);
}

void GTypeFace::Back(COLOUR c)
{
	d->_Back = c;
	_OnPropChange(false);
}

void GTypeFace::SetWeight(int i)
{
	d->_Weight = i;
	_OnPropChange(true);
}

void GTypeFace::Italic(bool i)
{
	d->_Italic = i;
	_OnPropChange(true);
}

void GTypeFace::Underline(bool i)
{
	d->_Underline = i;
	_OnPropChange(true);
}

void GTypeFace::Transparent(bool i)
{
	d->_Transparent = i;
	_OnPropChange(false);
}

void GTypeFace::Colour(COLOUR Fore, COLOUR Back)
{
	d->_Fore = Fore;
	d->_Back = Back;
	_OnPropChange(false);
}

void GTypeFace::SubGlyphs(bool i)
{
	if (NOT i OR GFontSystem::Inst()->GetGlyphSubSupport())
	{
		d->_SubGlyphs = i;
		_OnPropChange(false);
	}
}

////////////////////////
// get
char *GTypeFace::Face()
{
	return d->_Face;
}

int GTypeFace::PointSize()
{
	return d->_PtSize;
}

int GTypeFace::TabSize()
{
	return d->_TabSize;
}

int GTypeFace::Quality()
{
	return d->_Quality;
}

COLOUR GTypeFace::Fore()
{
	return d->_Fore;
}

COLOUR GTypeFace::Back()
{
	return d->_Back;
}

int GTypeFace::GetWeight()
{
	return d->_Weight;
}

bool GTypeFace::Italic()
{
	return d->_Italic;
}

bool GTypeFace::Underline()
{
	return d->_Underline;
}

bool GTypeFace::Transparent()
{
	return d->_Transparent;
}

bool GTypeFace::SubGlyphs()
{
	return d->_SubGlyphs;
}

double GTypeFace::Ascent()
{
	return d->_Ascent;
}

double GTypeFace::Descent()
{
	return d->_Descent;
}

////////////////////////////////////////////////////////////////////
// GFont class, the implemention
class GFontPrivate
{
public:
	// Data
	OsFont			hFont;
	int				Height;
	bool			Dirty;
	char			*Cp;
	int				Param;
	bool			OwnerUnderline;

	// Glyph substitution
	uchar			*GlyphMap;
	static class GlyphCache *Cache;

	GFontPrivate()
	{
		#if defined BEOS
		hFont = NEW(BFont);
		#elif defined XWIN
		hFont = NEW(XFont);
		#elif defined WIN32
		hFont = 0;
		OwnerUnderline = false;
		#else
		hFont = 0;
		#endif

		GlyphMap = 0;
		Dirty = true;
		Height = 10;
		Cp = 0;
		Param = 0;
	}

	~GFontPrivate()
	{
		#ifdef WIN32
		if (hFont)
		{
			DeleteObject(hFont);
		}
		#elif defined MAC
		if (hFont)
		{
			ATSUDisposeStyle(hFont);
			hFont = 0;
		}
		#else
		DeleteObj(hFont);	
		#endif
		DeleteArray(GlyphMap);
	}
};

GFont::GFont(char *face, int point)
{
	d = NEW(GFontPrivate);
	if (face AND point > 0)
	{
		Create(face, point);
	}
}

GFont::GFont(OsFont Handle)
{
	d = NEW(GFontPrivate);
	GFontType Type;
	if (Type.GetFromRef(Handle))
	{
		Create(&Type);
	}
}

GFont::GFont(GFontType &Type)
{
	d = NEW(GFontPrivate);
	Create(&Type);
}

GFont::GFont(GFont &Fnt)
{
	d = NEW(GFontPrivate);
	*this = Fnt;
}

GFont::~GFont()
{
	DeleteObj(d);
}

uchar *GFont::GetGlyphMap()
{
	return d->GlyphMap;
}

bool GFont::GetOwnerUnderline()
{
	return d->OwnerUnderline;
}

void GFont::_OnPropChange(bool FontChange)
{
	if (FontChange)
	{
		#if defined WIN32
		if (d->hFont)
		{
			if (this == SysFont)
			{
				LgiAssert(0);
			}

			DeleteObject(d->hFont);
			d->hFont = 0;
		}
		#elif defined BEOS
		if (this == SysFont)
		{
			LgiAssert(0);
		}
		else
		{
			Create(0, 0, 0);
		}
		#else
		d->Dirty = true;
		#endif
	}
}

OsFont GFont::Handle()
{
	return d->hFont;
}
int GFont::GetHeight()
{
	return d->Height;
}

bool GFont::IsValid()
{
	bool Status = false;

	// recreate font
	#ifdef WIN32
	if (NOT d->hFont)
	{
		Status = Create(Face(), PointSize());
	}
	#else
	if (d->Dirty)
	{
		Status = Create(Face(), PointSize());
		d->Dirty = false;
	}
	#endif
	else
	{
		Status = true;
	}

	return Status;
}

#ifdef WIN32
#define ENCODING_TABLE_SIZE 8

class type_4_cmap
{
public:
	uint16 format;
	uint16 length;
	uint16 language;
	uint16 seg_count_x_2;
	uint16 search_range;
	uint16 entry_selector;
	uint16 range_shift;
//	uint16 reserved;
	uint16 arrays[1];

	uint16 SegCount()
	{
		return seg_count_x_2 / 2;
	}

	uint16 *GetIdRangeOffset()
	{
		return arrays + 1 + (SegCount()*3);
	}

	uint16 *GetStartCount()
	{
		return arrays + 1 + SegCount();
	}

	uint16 *GetEndCount()
	{
		/* Apparently the reseved spot is not reserved for 
			 the end_count array!? */
		return arrays;
	}
};

class cmap_encoding_subtable
{
public:
	uint16 platform_id;
	uint16 encoding_id;
	uint32 offset;
};

#define INT16_SWAP(i)					(	((i)>>8) | (((i)&0xff)<<8) )
#define INT32_SWAP(i)					(	( ((i)&0x000000ff) << 24) |		\
											( ((i)&0x0000ff00) << 8 ) |		\
											( ((i)&0x00ff0000) >> 8 ) |		\
											( ((i)&0xff000000) >> 24) )
#define MAKE_TT_TABLE_NAME(c1, c2, c3, c4) \
	(((uint32)c4) << 24 | ((uint32)c3) << 16 | ((uint32)c2) << 8 | ((uint32)c1))
#define CMAP							(MAKE_TT_TABLE_NAME('c','m','a','p'))
#define CMAP_HEADER_SIZE				4
#define APPLE_UNICODE_PLATFORM_ID		0
#define MACINTOSH_PLATFORM_ID			1
#define ISO_PLATFORM_ID					2
#define MICROSOFT_PLATFORM_ID			3
#define SYMBOL_ENCODING_ID				0
#define UNICODE_ENCODING_ID				1
#define UCS4_ENCODING_ID				10

type_4_cmap *GetUnicodeTable(HFONT hFont, uint16 &Length)
{
	bool Status = false;
	type_4_cmap *Table = 0;

	HDC hDC = GetDC(0);
	if (hDC)
	{
		HFONT Old = (HFONT)SelectObject(hDC, hFont);

		uint16 Tables = 0;
		uint32 Offset = 0;

		// Get The number of encoding tables, at offset 2
		int32 Res = GetFontData(hDC, CMAP, 2, &Tables, sizeof(uint16));
		if (Res == sizeof (uint16))
		{
			Tables = INT16_SWAP(Tables);
			cmap_encoding_subtable *Sub = (cmap_encoding_subtable*)malloc(ENCODING_TABLE_SIZE*Tables);
			if (Sub)
			{
				Res = GetFontData(hDC, CMAP, CMAP_HEADER_SIZE, Sub, ENCODING_TABLE_SIZE*Tables);
				if (Res == ENCODING_TABLE_SIZE*Tables)
				{
					for (int i = 0; i < Tables; i++)
					{
						Sub[i].platform_id = INT16_SWAP(Sub[i].platform_id);
					Sub[i].encoding_id = INT16_SWAP(Sub[i].encoding_id);
						Sub[i].offset = INT32_SWAP(Sub[i].offset);

						if (Sub[i].platform_id == MICROSOFT_PLATFORM_ID AND
							Sub[i].encoding_id == UNICODE_ENCODING_ID)
						{
							Offset = Sub[i].offset;
						}
					}
				}
				free(Sub);
			}
		}

		if (Offset)
		{
			Length = 0;
			uint Res = GetFontData(hDC, CMAP, Offset + 2, &Length, sizeof(uint16));
			if (Res == sizeof(uint16))
			{
				Length = INT16_SWAP(Length);
				Table = (type_4_cmap*)malloc(Length);
				Res = GetFontData(hDC, CMAP, Offset, Table, Length);
				if (Res == Length)
				{
					Table->format = INT16_SWAP(Table->format);
					Table->length = INT16_SWAP(Table->length);
					Table->language = INT16_SWAP(Table->language);
					Table->seg_count_x_2 = INT16_SWAP(Table->seg_count_x_2);
					Table->search_range = INT16_SWAP(Table->search_range);
					Table->entry_selector = INT16_SWAP(Table->entry_selector);
					Table->range_shift = INT16_SWAP(Table->range_shift);

					if (Table->format == 4)
					{
						uint16 *tbl_end = (uint16 *)((char *)Table + Length);
						uint16 *tbl = Table->arrays;

						while (tbl < tbl_end)
						{
							*tbl = INT16_SWAP(*tbl);
							tbl++;
						}

						Status = true;
					}
				}
			}
		}

		SelectObject(hDC, Old);
		ReleaseDC(0, hDC);
	}

	if (NOT Status)
	{
		free(Table);
		Table = 0;
	}

	return Table;
}
#endif

int GFont::GetParam()
{
	return d->Param;
}

bool GFont::Create(char *face, int height, int Param)
{
	bool FaceChanging = false;

	if (face)
	{
		if (NOT Face() OR strcmp(Face(), face) != 0)
		{
			FaceChanging = true;
		}
		
		Face(face);
	}
	
	if (height > 0)
	{
		GTypeFace::d->_PtSize = height;
	}

	#if defined WIN32

	if (d->hFont)
	{
		DeleteObject(d->hFont);
		d->hFont = 0;
	}

	d->Param = Param;
	HDC hDC = (Param) ? (HDC)Param : GetDC(0);
	int LogPixelsY = GetDeviceCaps(hDC, LOGPIXELSY);
	int Win32Height = -MulDiv(PointSize(), LogPixelsY, 72);

	d->OwnerUnderline = Face() AND
						stricmp(Face(), "Courier New") == 0 AND 
						(PointSize() == 8 OR PointSize() == 9) AND
						GTypeFace::d->_Underline;

	d->hFont = ::CreateFont(Win32Height,
							0,
							0,
							0,
							GTypeFace::d->_Weight,
							GTypeFace::d->_Italic,
							d->OwnerUnderline ? false : GTypeFace::d->_Underline,
							false,
							ANSI_CHARSET,
							OUT_DEFAULT_PRECIS,
							CLIP_DEFAULT_PRECIS,
							GTypeFace::d->_Quality,
							FF_DONTCARE,
							GTypeFace::d->_Face);
	
	if (d->hFont)
	{
		HFONT hFnt = (HFONT) SelectObject(hDC, d->hFont);
		SIZE Size = {0, 0};
		GetTextExtentPoint32(hDC, "Ag", 2, &Size);
		d->Height = Size.cy;

		int Bytes = (MAX_UNICODE + 1) >> 3;
		if (NOT d->GlyphMap)
		{
			d->GlyphMap = NEW(uchar[Bytes]);
		}

		if (d->GlyphMap)
		{
			memset(d->GlyphMap, 0, Bytes);

			int OsVer, OsRev;
			int OsType = LgiGetOs(&OsVer, &OsType);
			if (OsType == LGI_OS_WIN9X OR
				OsVer < 5) // GetFontUnicodeRanges is supported on >= Win2k
			{
				bool HideUnihan = false;

				LgiAssert(sizeof(type_4_cmap)==16);
				uint16 Length = 0;
				type_4_cmap *t = GetUnicodeTable(Handle(), Length);
				if (t)
				{
					uint16 SegCount = t->seg_count_x_2 / 2;
					uint16 *EndCount = t->GetEndCount();
					uint16 *StartCount = t->GetStartCount();
					uint16 *IdRangeOffset = t->GetIdRangeOffset();

					for (int i = 0; i<SegCount; i++)
					{
						if (IdRangeOffset[i] == 0)
						{
							for (uint u = StartCount[i]; u <= EndCount[i]; u++)
							{
								if (HideUnihan AND
									u >= 0x3400 AND
									u <= 0x9FAF)
								{
									// APPROXIMATE
								}
								else
								{
									// EXACT
								}

								if ((u >> 3) < Bytes)
								{
									d->GlyphMap[u>>3] |= 1 << (u & 7);
								}
							}
						}
						else
						{
							uint16 *End = (uint16*) (((char*)t)+Length);
							int IdBytes = (int)End - (int)IdRangeOffset;

							for (uint u = StartCount[i]; u <= EndCount[i] AND IdRangeOffset[i] != 0xffff; u++)
							{
								uint id = *(IdRangeOffset[i]/2 + (u - StartCount[i]) + &IdRangeOffset[i]);
								if (id)
								{
									if (HideUnihan AND
										u >= 0x3400 AND
										u <= 0x9FAF)
									{
										// APPROXIMATE
									}
									else
									{
										// EXACT
									}

									if ((u >> 3) < Bytes)
									{
										d->GlyphMap[u>>3] |= 1 << (u & 7);
									}
								}
							}
						}
					}
				}
				else
				{
					// not a TTF? assume that it can handle 00-ff in the current ansi cp
					/*
					char *Cp = LgiAnsiToLgiCp();
					for (int i=0; i<=0x7f; i++)
					{
						char16 u;
						uchar c = i;
						void *In = &c;
						int Size = sizeof(c);
						LgiBufConvertCp(&u, "ucs-2", sizeof(u), In, Cp, Size);

						if ((u >> 3) < Bytes)
						{
							GlyphMap[u>>3] |= 1 << (u & 7);
						}
					}
					*/
				}
			}
			else
			{
				typedef DWORD (WINAPI *Proc_GetFontUnicodeRanges)(HDC, LPGLYPHSET);
				GLibrary Gdi32("Gdi32");

				Proc_GetFontUnicodeRanges GetFontUnicodeRanges = (Proc_GetFontUnicodeRanges)Gdi32.GetAddress("GetFontUnicodeRanges");
				if (GetFontUnicodeRanges)
				{
					DWORD BufSize = GetFontUnicodeRanges(hDC, 0);
					LPGLYPHSET Set = (LPGLYPHSET) NEW(char[BufSize]);
					if (Set AND GetFontUnicodeRanges(hDC, Set) > 0)
					{
						for (int i=0; i<Set->cRanges; i++)
						{
							WCRANGE *Range = Set->ranges + i;
							for (int n=0; n<Range->cGlyphs; n++)
							{
								int u = Range->wcLow + n;
								if (u >> 3 < Bytes)
								{
									d->GlyphMap[u>>3] |= 1 << (u & 7);
								}
							}
						}
					}

					DeleteArray((char*&)Set);
				}
			}

			if (d->GlyphMap)
			{
				memset(d->GlyphMap, 0xff, 128/8);
			}
		}

		SelectObject(hDC, hFnt);
	}

	if (NOT Param) ReleaseDC(0, hDC);

	return (d->hFont != 0);
	
	#elif defined BEOS

	if (Face())
	{
		d->hFont->SetSize(PointSize());
	
		int f = 0;
		
		if (Bold()) f |= B_BOLD_FACE;
		if (Italic()) f |= B_ITALIC_FACE;
		if (Underline()) f |= B_UNDERSCORE_FACE;
		if (NOT f) f |= B_REGULAR_FACE;
	
		d->hFont->SetFamilyAndFace(Face(), f);
	
		font_height h;
		d->hFont->GetHeight(&h);
		d->Height = h.ascent + h.descent + h.leading + 0.9999;
	}
	
	return true;

	#elif defined XWIN

	if (d->hFont)
	{
		d->hFont->SetPainter((XPainter*)Param);
		d->hFont->SetFamily(Face());
		d->hFont->SetPointSize(PointSize());
		d->hFont->SetBold(Bold());
		d->hFont->SetItalic(Italic());
		d->hFont->SetUnderline(Underline());

		XFontMetrics m(d->hFont);
		d->Height = m.height();

		if (FaceChanging OR NOT d->GlyphMap)
		{
			DeleteArray(d->GlyphMap);
			d->GlyphMap = m.GetCoverage(d->GlyphMap, MAX_UNICODE);
		}

		return true;
	}
	
	#elif defined MAC
	
	if (d->hFont)
	{
		ATSUDisposeStyle(d->hFont);
		d->hFont = 0;
	}
	
	OSStatus e;

	#if 0

	e = MakeThemeATSUIStyle(kThemeApplicationFont, &d->hFont);
	if (e) printf("%s:%i - MakeThemeATSUIStyle failed with %i\n", __FILE__, __LINE__, e);
	else return true;

	#else

	if (Face() AND !(e = ATSUCreateStyle(&d->hFont)))
	{
		_create_style++;
		
		// Lookup ID
		ATSUFontID atsuFont = 0;

		Str255 fontName;
		c2p255(fontName, Face());

		short fontFamily = 0, fontStyle = 0;
		GetFNum(fontName, &fontFamily);

		if (!(e = ATSUFONDtoFontID(fontFamily, fontStyle, &atsuFont)))
		{
			Fixed Size;
			Size = PointSize() << 16;
			Boolean IsBold = Bold();
			Boolean IsItalic = Italic();
			Boolean IsUnder = Underline();
			
			// printf("CreateFont %s,%i bold=%i italic=%i under=%i\n", Face(), PointSize(), IsBold, IsItalic, IsUnder);
			
			// Set style attr
			ATSUAttributeTag Tags[]			= {kATSUFontTag,     kATSUSizeTag, kATSUQDItalicTag, kATSUQDUnderlineTag, kATSUQDBoldfaceTag};
			ATSUAttributeValuePtr Values[]	= {&atsuFont,        &Size,        &IsItalic,        &IsUnder,            &IsBold,            };
			ByteCount Lengths[]				= {sizeof(atsuFont), sizeof(Size), sizeof(IsItalic), sizeof(IsUnder),     sizeof(IsBold)};
			if (!(e = ATSUSetAttributes(d->hFont,
										CountOf(Tags),
										Tags,
										Lengths,
										Values)))
			{
				GDisplayString ds(this, "A");
				d->Height = ds.Y();
				return true;
			}
			else
			{
				printf("%s:%i - Error setting font attr (e=%i)\n", __FILE__, __LINE__, e);
			}
		}
		else
		{
			printf("%s:%i - Error getting font id (e=%i)\n", __FILE__, __LINE__, e);
		}
	}
	else
	{
		printf("%s:%i - Error creating font (e=%i)\n", __FILE__, __LINE__, e);
	}
	
	#endif

	#endif

	return false;
}

char16 *GFont::_ToUnicode(char *In, int &Len)
{
	char16 *WStr = 0;
	
	if (In AND
		Len > 0)
	{
		WStr = (char16*)LgiNewConvertCp(LGI_WideCharset, In, GTypeFace::d->_CodePage, Len);
		if (WStr)
		{
			int l = StrlenW(WStr);
			if (l < Len)
			{
				Len = l;
			}
		}
	}

	return WStr;
}

#if defined WIN32

bool GFont::Create(GFontType *LogFont, int Param)
{
	if (d->hFont)
	{
		DeleteObject(d->hFont);
		d->hFont = 0;
	}

	if (LogFont)
	{
		// set props
		HDC hDC = GetDC(0);
		PointSize(-MulDiv(LogFont->Info.lfHeight, 72, GetDeviceCaps(hDC, LOGPIXELSY)));
		ReleaseDC(0, hDC);

		Face(LogFont->Info.lfFaceName);
		Quality(LogFont->Info.lfQuality);
		Bold(LogFont->Info.lfWeight >= FW_BOLD);
		Italic(LogFont->Info.lfItalic);
		Underline(LogFont->Info.lfUnderline);

		// create the handle
		Create(0, 0, Param);
	}

	return (d->hFont != 0);
}
#elif defined BEOS

bool GFont::Create(GFontType *LogFont, int Param)
{
	bool Status = false;

	if (LogFont)
	{
		Face(LogFont->Info.Face());
		PointSize(LogFont->Info.PointSize());
		Italic(LogFont->Info.Italic());
		Underline(LogFont->Info.Underline());
		Bold(LogFont->Info.Bold());

		Status = true;
	}
	
	return Status;
}

#else
bool GFont::Create(GFontType *LogFont, int Param)
{
	if (LogFont)
	{
		return Create(LogFont->Handle()->Face(), LogFont->Handle()->PointSize());
	}

	return false;
}
#endif

GFont &GFont::operator =(GFont &f)
{
	Face(f.Face());
	PointSize(f.PointSize());
	TabSize(f.TabSize());
	Quality(f.Quality());
	Fore(f.Fore());
	Back(f.Back());
	SetWeight(f.GetWeight());
	Italic(f.Italic());
	Underline(f.Underline());
	Transparent(f.Transparent());
	
	return *this;
}

//////////////////////////////////////////////////////////////////////
#define SubtractPtr(a, b)			(	(((int)(a))-((int)(b))) / sizeof(*a)	)
#define IsTabChar(c)				(c == '\t' OR (c == 0x2192 AND ShowVisibleTab()))

#define DSTR_VIS_TABS				0x01
#define DSTR_APPEND_DOTS			0x02

static OsChar GDisplayStringDots[] = {'.', '.', '.', 0};

class CharInfo
{
public:
	OsChar *Str;
	uint16 Len;
	uint16 X;
	uint8 FontId;
	int8 SizeDelta;

	CharInfo()
	{
		Str = 0;
		Len = 0;
		X = 0;
		FontId = 0;
		SizeDelta = 0;
	}
};

GDisplayString::GDisplayString(GFont *f, char *s, int l, int tabOrigin)
{
	Font = f;
	#ifdef BEOS
	Str = NewStr(s, l);
	#else
	Str = LgiNewUtf8To16(s, l);
	#endif
	Info = 0;
	Blocks = x = y = len = 0;
	TabOrigin = tabOrigin;
	Flags = DSTR_VIS_TABS;
	
	#ifdef MAC
	Hnd = 0;
	ATSUCreateTextLayout(&Hnd);
	_create_layout++;
	#endif
	
	Layout();
}

GDisplayString::GDisplayString(GFont *f, char16 *s, int l, int tabOrigin)
{
	Font = f;
	#ifdef BEOS
	Str = LgiNewUtf16To8(s, l * sizeof(char16));
	#else
	Str = NewStrW(s, l);
	#endif	
	Info = 0;
	Blocks = x = y = len = 0;
	TabOrigin = tabOrigin;
	Flags = DSTR_VIS_TABS;

	#ifdef MAC
	if (Str)
	{
		Hnd = 0;
		if ((len = StrlenW(Str)) > 0)
		{
			OSStatus e = ATSUCreateTextLayout(&Hnd);
			if (e) printf("%s:%i - ATSUCreateTextLayout failed with %i.\n", __FILE__, __LINE__, e);
			_create_layout++;
		}
	}
	#endif
	
	Layout();
}

GDisplayString::~GDisplayString()
{
	#ifdef MAC
	if (Hnd)
		ATSUDisposeTextLayout(Hnd);
	#endif
	
	DeleteArray(Str);
	DeleteArray(Info);
}

void GDisplayString::Layout()
{
	#ifdef MAC
	
	if (Hnd AND Str)
	{
		len = StrlenW(Str);
		
		OSStatus e = ATSUSetTextPointerLocation(Hnd, Str, 0, len, len);
		if (e)
		{
			printf("%s:%i - ATSUSetTextPointerLocation failed with errorcode %i (%S)\n", __FILE__, __LINE__, e, Str);
		}
		else
		{
			e = ATSUSetRunStyle(Hnd, Font->Handle(), 0, len);
			if (e)
			{
				printf("%s:%i - ATSUSetRunStyle failed with errorcode %i (%S)\n", __FILE__, __LINE__, e, Str);
			}
			else
			{
				ATSUTextMeasurement fTextBefore;
				ATSUTextMeasurement fTextAfter;

				#if 1
				{
					ATSTrapezoid t[1];
					ItemCount tlen;
					e = ATSUGetGlyphBounds(Hnd, 0, 0, kATSUFromTextBeginning, kATSUToTextEnd, kATSUseDeviceOrigins, 1, t, &tlen);
					if (e)
					{
						printf("%s:%i - ATSUGetUnjustifiedBounds failed with errorcode %i (%S)\n", __FILE__, __LINE__, e, Str);
					}
					else
					{
						x = (t[0].upperRight.x - t[0].upperLeft.x) >> 16;
						y = (t[0].lowerLeft.y - t[0].upperLeft.y) >> 16;
						
						fAscent = -(t[0].upperLeft.y);
						fDescent = t[0].lowerLeft.y;
						
						#if 0
						printf("str='%p' x=%i y=%i (%g,%g,%g,%g) %i,%i\n", Str, x, y,
							(double)t[0].upperRight.x/0x10000, (double)t[0].upperLeft.x/0x10000,
							(double)t[0].lowerLeft.y/0x10000, (double)t[0].upperLeft.y/0x10000,
							fAscent >> 16, fDescent >> 16);
						#endif
					}
				}
				#else
				e = ATSUGetUnjustifiedBounds(Hnd, kATSUFromTextBeginning, kATSUToTextEnd, &fTextBefore, &fTextAfter, &fAscent, &fDescent);
				if (e)
				{
					printf("%s:%i - ATSUGetUnjustifiedBounds failed with errorcode %i (%S)\n", __FILE__, __LINE__, e, Str);
				}
				else
				{
					x = (fTextAfter - fTextBefore + 0xffff) >> 16;
					y = (fAscent + fDescent) >> 16;
					
					#if 1
					printf("str='%p' x=%i y=%i (%g,%g,%g,%g) %i,%i\n", Str, x, y,
						(double)fTextBefore/0x10000, (double)fTextAfter/0x10000, (double)fAscent/0x10000, (double)fDescent/0x10000,
						fAscent >> 16, fDescent >> 16);
					#endif
				}
				#endif
			
			}
		}

		ATSUSetTransientFontMatching(Hnd, true);
	}
	
	#else
	
	int sx, sy, i = 0;
	if (NOT Font->Handle()) Font->Create();
	y = Font->GetHeight();

	GFontSystem *Sys = GFontSystem::Inst();
	if (Sys AND Str)
	{
		GFont *f = Font;
		OsChar *s;
		bool GlyphSub = Font->SubGlyphs();

		Blocks = 1;
		bool WasTab = IsTabChar(*Str);
		for (s=Str; *s AND *s != '\n'; NextOsChar(s))
		{
			GFont *n = GlyphSub ? Sys->GetGlyph(*s, Font) : Font;
			if (n != f OR (IsTabChar(*s) ^ WasTab))
			{
				Blocks++;
				f = n;
				WasTab = IsTabChar(*s);
			}
		}
		len = ((int)s - (int)Str) / sizeof(OsChar);

		if (*s == '\n')
		{
			*s = 0;
		}

		Info = Blocks ? NEW(CharInfo[Blocks]) : 0;
		if (Info)
		{
			Info[i].Str = Str;

			int TabSize = Font->TabSize() ? Font->TabSize() : 32;

			WasTab = IsTabChar(*Str);
			f = GlyphSub ? Sys->GetGlyph(*Str, Font) : Font;
			if (f AND f != Font)
			{
				f->PointSize(Font->PointSize());
				f->SetWeight(Font->GetWeight());
				if (NOT f->Handle())
					f->Create();
			}

			bool Debug = WasTab;
			
			for (s=Str; true; NextOsChar(s))
			{
				GFont *n = GlyphSub ? Sys->GetGlyph(*s, Font) : Font;
				bool Change = n != f OR (IsTabChar(*s) ^ WasTab) OR NOT *s;

				if (Change)
				{
					// End last segment
					if (n AND n != Font)
					{
						n->PointSize(Font->PointSize());
						n->SetWeight(Font->GetWeight());
						if (NOT n->Handle())
							n->Create();
					}

					Info[i].Len = SubtractPtr(s, Info[i].Str);
					if (Info[i].Len)
					{
						LgiAssert(i < Blocks);

						if (WasTab)
						{
							// Handle tab(s)
							for (int t=0; t<Info[i].Len; t++)
							{
								int Dx = TabSize - ((Info[i].X + x + TabOrigin) % TabSize);
								Info[i].X += Dx;
							}
							x += Info[i].X;
						}
						else
						{
							GFont *m = f;

							if (f AND (f->GetHeight() > Font->GetHeight()))
							{
								Info[i].SizeDelta = -1;
								f->PointSize(Font->PointSize() + Info[i].SizeDelta);
								f->Create();
							}

							if (NOT f)
							{
								// no font, so ? out the chars... as they aren't available anyway
								// printf("Font Cache Miss, Len=%i\n\t", Info[i].Len);
								m = Font;
								for (int n=0; n<Info[i].Len; n++)
								{
									Info[i].Str[n] = '?';
								}
							}
							m->_Measure(sx, sy, Info[i].Str, Info[i].Len);
							x += Info[i].X = sx > 0xffff ? 0xffff : sx;
						}
						Info[i].FontId = NOT f OR Font == f ? 0 : Sys->Lut[Info[i].Str[0]];

						i++;
					}

					f = n;

					// Start next segment
					WasTab = IsTabChar(*s);
					if (i < Blocks)
					{
						Info[i].Str = s;
					}
				}

				if (NOT *s) break;
			}

			Blocks = min(Blocks, i + 1);
			if (Blocks > 0 AND Info[Blocks-1].Len == 0)
			{
				Blocks--;
			}
		}
	}
	
	#endif
}

bool GDisplayString::ShowVisibleTab()
{
	return TestFlag(Flags, DSTR_VIS_TABS);
}

void GDisplayString::ShowVisibleTab(bool i)
{
	if (i)	SetFlag(Flags, DSTR_VIS_TABS);
	else	ClearFlag(Flags, DSTR_VIS_TABS);
}

bool GDisplayString::IsTruncated()
{
	return TestFlag(Flags, DSTR_APPEND_DOTS);
}

void GDisplayString::TruncateWithDots(int Width)
{
	#ifdef MAC
	#else
	
	if (Width < X() + 8)
	{
		int c = CharAt(Width);
		if (c >= 0 AND c < len)
		{
			if (c > 0) c--; // fudge room for dots
			if (c > 0) c--;
			
			Flags |= DSTR_APPEND_DOTS;
			
			if (Info)
			{
				int Width = 0;
				int Pos = 0;
				for (int i=0; i<Blocks; i++)
				{
					if (c >= Pos AND
						c < Pos + Info[i].Len)
					{
						Info[i].Len = c - Pos;
						Info[i].Str[Info[i].Len] = 0;

						GFont *f = Font;
						if (Info[i].FontId)
						{
							GFontSystem *Sys = GFontSystem::Inst();
							f = Sys->Font[Info[i].FontId];
							f->PointSize(Font->PointSize() + Info[i].SizeDelta);
							if (NOT f->Handle())
							{
								f->Create();
							}
						}
						else
						{
							f = Font;
						}
						
						if (f)
						{
							int Sx, Sy;
							f->_Measure(Sx, Sy, Info[i].Str, Info[i].Len);
							Info[i].X = Sx;
							Width += Info[i].X;
						}
						
						Blocks = i + 1;
						
						break;
					}
					
					Pos += Info[i].Len;
					Width += Info[i].X;
				}
				
				int DotsX, DotsY;
				Font->_Measure(DotsX, DotsY, GDisplayStringDots, 3);
				x = Width + DotsX;
			}
		}
	}
	
	#endif
}

int GDisplayString::X()
{
	return x;
}

int GDisplayString::Y()
{
	return y;
}

int GDisplayString::Length()
{
	return len;
}

void GDisplayString::Length(int New)
{
	#ifdef MAC

	#else
	
	if (New < len)
	{
		GFontSystem *Sys = GFontSystem::Inst();
		
		int CurX = 0;
		int CurLen = 0;
		for (int i=0; i<Blocks; i++)
		{
			// Is the cut point in this block?
			if (New >= CurLen AND New < CurLen + Info[i].Len )
			{
				// In this block
				int Offset = New - CurLen;
				Info[i].Len = Offset;
				Info[i].Str[Info[i].Len] = 0;

				// Get the font for this block of characters
				GFont *f = 0;
				if (Info[i].FontId)
				{
					f = Sys->Font[Info[i].FontId];
					f->PointSize(Font->PointSize());
					f->Transparent(Font->Transparent());
					if (NOT f->Handle())
					{
						f->Create();
					}
				}
				else
				{
					f = Font;
				}

				// Chop the current block and re-measure
				int ChoppedX, Unused;
				f->_Measure(ChoppedX, Unused, Info[i].Str, Info[i].Len);
				Info[i].X = ChoppedX;
				x = CurX + Info[i].X;
				Blocks = i+1;
				
				// Leave the loop
				break;
			}
			
			CurX += Info[i].X;
			CurLen += Info[i].Len;
		}
	}
	else
	{
		printf("%s:%i - New>=Len (%i>=%i)\n", __FILE__, __LINE__, New, len);
	}

	#endif
}

void GDisplayString::Size(int *sx, int *sy)
{
	if (sx) *sx = x;
	if (sy) *sy = y;
}

int GDisplayString::CharAt(int Px)
{
	int Status = -1;

	if (Px < 0)
		return 0;
	else if (Px >= x)
		return len;

	#ifdef MAC

	if (Hnd AND Str)
	{
		UniCharArrayOffset Off = 0, Off2 = 0;
		Boolean IsLeading;
		OSStatus e = ATSUPositionToOffset(Hnd, FloatToFixed(Px), FloatToFixed(y / 2), &Off, &IsLeading, &Off2);
		if (e) printf("%s:%i - ATSUPositionToOffset failed with %i, CharAt(%i) x=%i len=%i\n", __FILE__, __LINE__, e, Px, x, len);
		else
		{
			// printf("CharAt(%i) off=%i,%i Leading=%i x=%i len=%i\n", Px, Off, Off2, IsLeading, x, len);
			Status = Off;
		}
	}
	
	#else
	
	GFontSystem *Sys = GFontSystem::Inst();

	if (Info AND Font AND Sys)
	{
		int TabSize = Font->TabSize() ? Font->TabSize() : 32;
		int Cx = 0;
		int Char = 0;

		for (int i=0; i<Blocks AND Status < 0; i++)
		{
			if (Px < Cx)
			{
				Status = Char;
				break;
			}

			if (Px >= Cx AND Px < Cx + Info[i].X)
			{
				// The position is in this block of characters
				if (IsTabChar(Info[i].Str[0]))
				{
					// Search through tab block
					for (int t=0; t<Info[i].Len; t++)
					{
						int TabX = TabSize - (Cx % TabSize);
						if (Px >= Cx AND Px < Cx + TabX)
						{
							Status = Char;
							break;
						}
						Cx += TabX;
						Char++;
					}
				}
				else
				{
					// Get the blocks font
					GFont *f = Font;
					if (Info[i].FontId)
					{
						f = Sys->Font[Info[i].FontId];
						f->Colour(Font->Fore(), Font->Back());
						f->PointSize(Font->PointSize());
						if (NOT f->Handle())
						{
							f->Create();
						}
					}

					// Find the pos
					int Fit = f->_CharAt(Px - Cx, Info[i].Str, Info[i].Len);
					if (Fit >= 0)
					{
						Status = Char + Fit;
					}
					else
					{
						Status = -1;
						goto CharAtError;
					}
					break;
				}
			}
			else
			{
				// Not in this block, skip the whole lot
				Cx += Info[i].X;
				Char += Info[i].Len;
			}
		}

		if (Status < 0)
		{
			Status = Char;
		}
	}

CharAtError:
	#endif

	return Status;
}

void GDisplayString::Draw(GSurface *pDC, int px, int py, GRect *r)
{
	#if defined(BEOS)

	// FIXME
	LgiAssert(0);
	return;

	#elif defined MAC

	int Ox = 0, Oy = 0;
	if (pDC)
		pDC->GetOrigin(Ox, Oy);

	if (NOT Font->Transparent())
	{
		COLOUR Old = pDC->Colour(Font->Back(), 24);
		if (r)
		{
			GRect a = r;
			a.Offset(Ox, Oy);
			pDC->Rectangle(&a);

			// pDC->Colour(Rgb24(0, 0, 255), 24);
			// pDC->Box(r);
		}
		else
		{
			GRect a(px, py, px + x - 1, py + y - 1);
			a.Offset(Ox, Oy);
			pDC->Rectangle(&a);

			// pDC->Colour(Rgb24(0, 255, 0), 24);
			// pDC->Box(r);
		}
		pDC->Colour(Old);
	}
	
	if (Hnd AND pDC AND len > 0)
	{
		OSStatus e;
		OsPainter dc = pDC->Handle();
		ATSUAttributeTag        theTags[1] = {kATSUCGContextTag};
		ByteCount               theSizes[1] = {sizeof(CGContextRef)};
		ATSUAttributeValuePtr   theValues[1] = {&dc};

		e = ATSUSetLayoutControls(Hnd, 1, theTags, theSizes, theValues);
		if (e)
		{
			printf("%s:%i - ATSUSetLayoutControls failed (e=%i)\n", __FILE__, __LINE__, e);
		}
		else
		{
			// Set style attr
			ATSURGBAlphaColor c;
			c.red = (double) R24(Font->Fore()) / 255.0;
			c.green = (double) G24(Font->Fore()) / 255.0;
			c.blue = (double) B24(Font->Fore()) / 255.0;
			c.alpha = 1.0;
			
			ATSUAttributeTag Tags[]			= {kATSURGBAlphaColorTag};
			ATSUAttributeValuePtr Values[]	= {&c};
			ByteCount Lengths[]				= {sizeof(c)};
			
		
			if (!(e = ATSUSetAttributes(Font->Handle(),
										CountOf(Tags),
										Tags,
										Lengths,
										Values)))
			{
				int y = (pDC->Y() - py + Oy);

				if (pDC->IsScreen())
				{
					CGContextSaveGState(pDC->Handle());

					CGContextTranslateCTM (pDC->Handle(), 0, pDC->Y()-1); 
					CGContextScaleCTM (pDC->Handle(), 1.0, -1.0); 

					e = ATSUDrawText(Hnd, kATSUFromTextBeginning, kATSUToTextEnd, (px - Ox) << 16, (y << 16) - fAscent);
					CGContextRestoreGState(pDC->Handle());
				}
				else
				{
					e = ATSUDrawText(Hnd, kATSUFromTextBeginning, kATSUToTextEnd, (px - Ox) << 16, (y << 16) - fAscent);
				}
				if (e)
				{
					printf("%s:%i - ATSUDrawText failed with %i, len=%i, str=%.20S\n", __FILE__, __LINE__, e, len, Str);
				}
				else
				{
					_draw_layout++;
					int64 Now = LgiCurrentTime();
					if (_last == 0)
						_last = Now;
					else if (Now > _last + 1000)
					{
						int ms = Now - _last;
						printf("ms=%i create=%i layout=%i draw=%i\n",
							ms, _create_style, _create_layout, _draw_layout);
						_create_style = 0;
						_create_layout = 0;
						_draw_layout = 0;
						_last = Now;
					}
				}
			}
			else
			{
				printf("%s:%i - Error setting font attr (e=%i)\n", __FILE__, __LINE__, e);
			}
		}
	}
	
	#else
	
	if (Info AND pDC AND Font)
	{
		GFontSystem *Sys = GFontSystem::Inst();
		COLOUR Old = pDC->Colour();
		int TabSize = Font->TabSize() ? Font->TabSize() : 32;
		int Ox = px;

		for (int i=0; i<Blocks; i++)
		{
			GFont *f = 0;

			// Get the font for this block of characters
			if (Info[i].FontId)
			{
				f = Sys->Font[Info[i].FontId];

				#if 1 // true == proper colour
				f->Colour(Font->Fore(), Font->Back());
				#else
				f->Colour(Rgb24(255, 0, 0), Font->Back());
				#endif

				f->PointSize(Font->PointSize() + Info[i].SizeDelta);
				f->Transparent(Font->Transparent());
				f->Underline(Font->Underline());
				if (NOT f->Handle())
				{
					f->Create();
				}
			}
			else
			{
				f = Font;
			}

			if (f)
			{
				GRect b;
				if (r)
				{
					b.x1 = i ? px : r->x1;
					b.y1 = r->y1;
					b.x2 = i < Blocks - 1 ? px + Info[i].X - 1 : r->x2;
					b.y2 = r->y2;
				}
				else
				{
					b.x1 = px;
					b.y1 = py;
					b.x2 = px + Info[i].X - 1;
					b.y2 = py + Y() - 1;
				}
				
				if (b.Valid())
				{
					if (Info[i].Str[0] == '\t')
					{
						// Draw the space occupied by the TAB's
						pDC->Colour(Font->Back(), 24);
						pDC->Rectangle(&b);
					}
					else if (Info[i].Str[0] == 0x2192)
					{
						int X = px;
						for (int n=0; n<Info[i].Len; n++)
						{
							int Dx = TabSize - ((X - Ox + TabOrigin) % TabSize);
							GRect Char(X, b.y1, X + Dx - 1, b.y2);
							f->_Draw(pDC, X, py, Info[i].Str, 1, &Char);
							X += Dx;
						}
					}
					else
					{
						// Draw the character(s)
						f->_Draw(pDC, px, py, Info[i].Str, Info[i].Len, &b);
					}
				}
			}

			// Inc my position
			px += Info[i].X;
		}
		
		if (TestFlag(Flags, DSTR_APPEND_DOTS))
		{
			int Sx, Sy;
			Font->_Measure(Sx, Sy, GDisplayStringDots, 3);

			GRect b;
			if (r)
			{
				b.x1 = px;
				b.y1 = r->y1;
				b.x2 = min(px + Sx - 1, r->x2);
				b.y2 = r->y2;
			}
			else
			{
				b.x1 = px;
				b.y1 = py;
				b.x2 = px + Sx - 1;
				b.y2 = py + Y() - 1;
			}

			Font->_Draw(pDC, px, py, GDisplayStringDots, 3, &b);
		}

		pDC->Colour(Old);
	}
	else if (r AND
			Font AND
			NOT Font->Transparent())
	{
		pDC->Colour(Font->Back(), 24);
		pDC->Rectangle(r);
	}
	
	#endif
}

///////////////////////////////////////////////////////////////////////
GFontType::GFontType(char *face, int pointsize)
{
	#if defined WIN32

	ZeroObj(Info);
	if (face)
	{
		strcpy(Info.lfFaceName, face);
	}
	if (pointsize)
	{
		Info.lfHeight = WinHeightToPoint(pointsize);
	}

	#else

	if (face) Info.Face(face);
	if (pointsize) Info.PointSize(pointsize);

	#endif
}

GFontType::~GFontType()
{
}

char *GFontType::GetFace()
{
	#ifdef WIN32
	return Info.lfFaceName;
	#else
	return Info.Face();
	#endif
}

void GFontType::SetFace(char *Face)
{
	#ifdef WIN32
	if (Face) strcpy(Info.lfFaceName, Face);
	else Info.lfFaceName[0] = 0;
	#else
	Info.Face(Face);
	#endif
}

int GFontType::GetPointSize()
{
	#ifdef WIN32
	return WinHeightToPoint(Info.lfHeight);
	#else
	return Info.PointSize();
	#endif
}

void GFontType::SetPointSize(int PointSize)
{
	#ifdef WIN32
	Info.lfHeight = WinPointToHeight(PointSize);
	#else
	Info.PointSize(PointSize);
	#endif
}

bool GFontType::DoUI(GView *Parent)
{
	bool Status = false;

	#if defined WIN32
	void *i = &Info;
	#else
	char i[128];
	sprintf(i, "%s,%i", Info.Face(), Info.PointSize());
	#endif

	GFontSelect Dlg(Parent, i);
	if (Dlg.DoModal() == IDOK)
	{
		#ifdef WIN32
		Dlg.Serialize(i, true);
		#else
		if (Dlg.Face) Info.Face(Dlg.Face);
		Info.PointSize(Dlg.Size);
		#endif

		Status = true;
	}

	return Status;
}

bool GFontType::GetDescription(char *Str)
{
	if (Str AND GetFace())
	{
		sprintf(Str, "%s, %i pt", GetFace(), GetPointSize());
		return true;
	}

	return false;
}

bool GFontType::Serialize(ObjProperties *Options, char *OptName, bool Write)
{
	bool Status = false;

	if (Options AND OptName)
	{
		#if defined WIN32
		if (Write)
		{
			Status = Options->Set(OptName, &Info, sizeof(Info));
		}
		else
		{
			void *Data = 0;
			int Len = 0;
			if (Options->Get(OptName, Data, Len) AND
				Len == sizeof(Info) AND
				Data)
			{
				memcpy(&Info, Data, Len);
				Status = true;
			}
		}
		#else
		if (Write)
		{
			char Temp[128];
			sprintf(Temp, "%s,%i pt", Info.Face(), Info.PointSize());
			Status = Options->Set(OptName, Temp);
		}
		else
		{
			char *Temp = 0;
			if (Options->Get(OptName, Temp) AND ValidStr(Temp))
			{
				char *OurCopy = NewStr(Temp);
				if (OurCopy)
				{
					char *Comma = strchr(OurCopy, ',');
					if (Comma)
					{
						*Comma++ = 0;
						Info.Face(OurCopy);
						Info.PointSize(atoi(Comma));
						Status = true;
					}

					DeleteArray(OurCopy);
				}
			}
		}
		#endif
	}

	return Status;
}

bool GFontType::Serialize(GDom *Options, char *OptName, bool Write)
{
	bool Status = false;

	if (Options AND OptName)
	{
		GVariant v;
		#if defined WIN32
		if (Write)
		{
			v.SetBinary(sizeof(Info), &Info);
			Status = Options->SetValue(OptName, v);
		}
		else
		{
			if (Options->GetValue(OptName, v))
			{
				if (v.Type == GV_BINARY AND
					v.Value.Binary.Length == sizeof(Info))
				{
					memcpy(&Info, v.Value.Binary.Data, sizeof(Info));
					Status = true;
				}
			}
		}
		#else
		if (Write)
		{
			char Temp[128];
			sprintf(Temp, "%s,%i pt", Info.Face(), Info.PointSize());
			Status = Options->SetValue(OptName, v = Temp);
		}
		else
		{
			if (Options->GetValue(OptName, v) AND ValidStr(v.Str()))
			{
				char *Comma = strchr(v.Str(), ',');
				if (Comma)
				{
					*Comma++ = 0;
					Info.Face(v.Str());
					Info.PointSize(atoi(Comma));
					Status = true;
				}
			}
		}
		#endif
	}

	return Status;
}

bool GFontType::GetConfigFont(char *Tag)
{
	bool Status = false;

	GXmlTag *Font = LgiApp->GetConfig(Tag);
	if (Font)
	{
		// read from config file
		char *f, *p;
		if ((f = Font->GetAttr("Face")) AND (p = Font->GetAttr("Point")))
		{
			SetFace(f);
			SetPointSize(atoi(p));
			Status = true;
		}
	}

	return Status;
}

class GFontTypeCache
{
	char *DefFace;
	int DefSize;
	char Face[64];
	int Size;

public:
	GFontTypeCache(char *defface, int defsize)
	{
		Face[0] = 0;
		Size = -1;
		DefFace = defface;
		DefSize = defsize;
	}
	
	char *GetFace(char *Type = 0)
	{
		#ifdef LINUX
		if (NOT IsInit())
		{
			char f[256];
			int s;
			if (_GetSystemFont(Type, f, sizeof(f), s))
			{
				strsafecpy(Face, f, sizeof(Face));
				Size = s;
			}
			else
			{
				Size = 0;
			}
		}
		#endif
		
		return Face[0] ? Face : DefFace;
	}
	
	int GetSize()
	{
		return Size > 0 ? Size : DefSize;
	}
	
	bool IsInit()
	{
		return Size >= 0;
	}
};

bool GFontType::GetSystemFont(char *Which)
{
	bool Status = false;

	if (Which)
	{
		#if defined WIN32

		// Get the system settings
		NONCLIENTMETRICS info;
		info.cbSize = sizeof(info);
		bool InfoOk = SystemParametersInfo(	SPI_GETNONCLIENTMETRICS,
											sizeof(info),
											&info,
											0);

		// Convert windows font height into points
		int Height = WinHeightToPoint(info.lfMessageFont.lfHeight);

		#elif defined LINUX

		// Define some defaults.. in case the system settings aren't there
		#ifdef __CYGWIN__
		char DefFont[] = "Luxi Sans";
		#else
		char DefFont[] = "Arial";
		#endif
		int DefSize = 10;
		int Offset = 0;
		
		#endif

		if (NOT stricmp(Which, "System"))
		{
			Status = GetConfigFont("Font-System");
			if (NOT Status)
			{
				// read from system
				#if defined WIN32

				if (InfoOk)
				{
					// Copy the font metrics
					memcpy(&Info, &info.lfMessageFont, sizeof(Info));
					Status = true;
				}

				#elif defined BEOS

				// BeOS has no system wide setting so give a valid default
				Info.Face("Swis721 BT");
				Info.PointSize(11);
				Status = true;

				#elif defined LINUX

				static GFontTypeCache Cache(DefFont, DefSize);
				Info.Face(Cache.GetFace("font"));
				Info.PointSize(Cache.GetSize()+Offset);
				Status = true;
			
				#elif defined MAC
				
				Str255 Name;
				SInt16 Size;
				Style St;
				OSStatus e = GetThemeFont(	kThemeSmallSystemFont,
											smSystemScript,
											Name,
											&Size,
											&St);
				if (e) printf("%s:%i - GetThemeFont failed with %i\n", __FILE__, __LINE__, e);
				else
				{
					Info.Face(p2c(Name));
					Info.PointSize(Size);
					Status = true;
					
					printf("System=%s,%i\n", Info.Face(), Size);
				}

				#endif
			}
		}
		else if (NOT stricmp(Which, "Menu"))
		{
			Status = GetConfigFont("Font-Menu");
			if (NOT Status)
			{
				#if defined WIN32

				if (InfoOk)
				{
					// Copy the font metrics
					memcpy(&Info, &info.lfMenuFont, sizeof(Info));
					Status = true;
				}

				#elif defined BEOS

				// BeOS has no system wide setting so give a valid default
				Info.Face("Swis721 BT");
				Info.PointSize(11);
				Status = true;

				#elif defined LINUX

				static GFontTypeCache Cache(DefFont, DefSize);
				Info.Face(Cache.GetFace("menuFont"));
				Info.PointSize(Cache.GetSize()+Offset);
				Status = true;

				#elif defined MAC
				
				Str255 Name;
				SInt16 Size;
				Style St;
				OSStatus e = GetThemeFont(	kThemeMenuItemFont,
											smSystemScript,
											Name,
											&Size,
											&St);
				if (e) printf("%s:%i - GetThemeFont failed with %i\n", __FILE__, __LINE__, e);
				else
				{
					Info.Face(p2c(Name));
					Info.PointSize(Size);
					Status = true;
				}

				#endif
			}
		}
		else if (NOT stricmp(Which, "Caption"))
		{
			Status = GetConfigFont("Font-Caption");
			if (NOT Status)
			{
				#if defined WIN32
				if (InfoOk)
				{
					// Copy the font metrics
					memcpy(&Info, &info.lfCaptionFont, sizeof(Info));
					Status = true;
				}
				
				#elif defined BEOS

				// BeOS has no system wide setting so give a valid default
				Info.Face("Swis721 BT");
				Info.PointSize(11);
				Status = true;

				#elif defined LINUX

				static GFontTypeCache Cache(DefFont, DefSize);
				Info.Face(Cache.GetFace("toolBarFont"));
				Info.PointSize(Cache.GetSize()+Offset);
				Status = true;

				#elif defined MAC
				
				Str255 Name;
				SInt16 Size;
				Style St;
				OSStatus e = GetThemeFont(	kThemeToolbarFont,
											smSystemScript,
											Name,
											&Size,
											&St);
				if (e) printf("%s:%i - GetThemeFont failed with %i\n", __FILE__, __LINE__, e);
				else
				{
					Info.Face(p2c(Name));
					Info.PointSize(Size);
					Status = true;
				}

				#endif
			}
		}
		else if (NOT stricmp(Which, "Status"))
		{
			Status = GetConfigFont("Font-Status");
			if (NOT Status)
			{
				#if defined WIN32
				
				if (InfoOk)
				{
					// Copy the font metrics
					memcpy(&Info, &info.lfStatusFont, sizeof(Info));
					Status = true;
				}
				
				#elif defined BEOS

				// BeOS has no system wide setting so give a valid default
				Info.Face("Swis721 BT");
				Info.PointSize(11);
				Status = true;

				#elif defined LINUX

				static GFontTypeCache Cache(DefFont, DefSize);
				Info.Face(Cache.GetFace("taskbarFont"));
				Info.PointSize(Cache.GetSize()+Offset);
				Status = true;

				#elif defined MAC
				
				Str255 Name;
				SInt16 Size;
				Style St;
				OSStatus e = GetThemeFont(	kThemeToolbarFont,
											smSystemScript,
											Name,
											&Size,
											&St);
				if (e) printf("%s:%i - GetThemeFont failed with %i\n", __FILE__, __LINE__, e);
				else
				{
					Info.Face(p2c(Name));
					Info.PointSize(Size);
					Status = true;
				}

				#endif
			}
		}
		else if (NOT stricmp(Which, "Small"))
		{
			Status = GetConfigFont("Font-Small");
			if (NOT Status)
			{
				#if defined WIN32

				if (InfoOk)
				{
					// Copy the font metrics
					memcpy(&Info, &info.lfSmCaptionFont, sizeof(Info));
					if (LgiGetOs() == LGI_OS_WIN9X AND stricmp(Info.lfFaceName, "ms sans serif") == 0)
					{
						strcpy(Info.lfFaceName, "Arial");
					}

					// Make it a bit smaller than the system font
					Info.lfHeight = WinPointToHeight(WinHeightToPoint(info.lfMessageFont.lfHeight)-1);
					Info.lfWeight = FW_NORMAL;

					Status = true;
				}

				#elif defined BEOS

				// BeOS has no system wide setting so give a valid default
				Info.Face("Swis721 BT");
				Info.PointSize(9);
				Status = true;

				#elif defined LINUX

				static GFontTypeCache Cache(DefFont, DefSize-1);
				Info.Face(Cache.GetFace("toolBarFont"));
				Info.PointSize(Cache.GetSize()+Offset);
				Status = true;
				
				#elif defined MAC
				
				Str255 Name;
				SInt16 Size;
				Style St;
				OSStatus e = GetThemeFont(	kThemeSmallSystemFont,
											smSystemScript,
											Name,
											&Size,
											&St);
				if (e) printf("%s:%i - GetThemeFont failed with %i\n", __FILE__, __LINE__, e);
				else
				{
					Info.Face(p2c(Name));
					Info.PointSize(Size - 2);
					Status = true;
				}

				#endif
			}
		}
		else if (NOT stricmp(Which, "Fixed"))
		{
			Status = GetConfigFont("Font-Fixed");
			if (NOT Status)
			{
				#if defined WIN32

				strcpy(Info.lfFaceName, "Courier New");
				Info.lfHeight = WinPointToHeight(9);
				Status = true;

				#elif defined BEOS

				Info.Face("Courier10 BT");
				Info.PointSize(12);
				Status = true;

				#elif defined LINUX

				static GFontTypeCache Cache(DefFont, DefSize);
				Cache.GetFace("font");
				Info.Face("Courier");
				Info.PointSize(Cache.GetSize()+Offset);
				Status = true;
				
				#elif defined MAC
				
				Info.Face("Courier");
				Info.PointSize(11);
				Status = true;

				#endif
			}
		}
		
		// printf("GetSystemFont(%s) %s,%i\n", Which, Info.Face(), Info.Height());
	}

	return Status;
}

bool GFontType::GetFromRef(OsFont Handle)
{
	#if defined WIN32
	return GetObject(Handle, sizeof(Info), &Info) == sizeof(Info);
	#else
	// FIXME
	return false;
	#endif
}

GFont *GFontType::Create(int Param)
{
	GFont *New = NEW(GFont);
	if (New AND
		NOT New->Create(this, Param))
	{
		DeleteObj(New);
	}
	return New;
}

/////////////////////////////////////////////////////////////////////////////
/*
GUtf8::GUtf8(char *utf, int bytes, bool own)
{
	BufOwn = own;
	BufBytes = bytes;
	if (BufOwn AND utf)
	{
		if (BufBytes < 0)
		{
			for (BufBytes = 0; utf[BufBytes]; BufBytes++);
		}
		Buf = Cur = NEW(char[BufBytes+1]);
		if (Cur)
		{
			memcpy(Cur, utf, BufBytes);
			Cur[BufBytes] = 0;
		}
	}
	else
	{
		Buf = Cur = utf;
	}

	CurChars = -1;
}

GUtf8::~GUtf8()
{
	if (BufOwn)
		DeleteArray(Buf);
}

bool GUtf8::Valid()
{
	return	Buf != NULL AND
			Cur != NULL AND
			Cur >= Buf AND
			(BufBytes >= 0 ? Cur <= Buf + BufBytes : true);
}

int GUtf8::Length()
{
	if (NOT Valid())
		return -1;

	if (BufBytes < 0)
		BufBytes = strlen(Buf);

	return BufBytes;
}

int GUtf8::GetChars()
{
	if (NOT Valid())
		return -1;

	if (CurChars < 0)
	{
		CurChars = 0;

		// CurChars not defined yet, count them.
		char *c = Cur;
		char *e = End();
		int len = GetBytes();
		while (*c AND c < e)
		{
			uint32 ch = LgiUtf8To32((uint8*&)c, len);
			if (ch)
			{
				CurChars++;
			}
			else break;
		}
	}

	return CurChars;
}

int GUtf8::GetBytes()
{
	if (Cur AND Buf)
	{
		char *e = End();
		if (e)
			return e - Cur;
	}

	return 0;
}

uint32 GUtf8::GetChar()
{
	if (Buf AND Cur)
	{
		int len = GetBytes();
		uint8 *c = (uint8*) Cur;
		return LgiUtf8To32(c, len);
	}

	return 0;
}

bool GUtf8::SeekChar(int i)
{
	return false;
}

char16 *GUtf8::ToWide()
{
	return Cur ? LgiNewUtf8To16(Cur, GetBytes()) : 0;
}

char *GUtf8::End()
{
	return Buf ? Buf + Length() : 0;
}


*/