/*hdr
**      FILE:           Mail.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           28/5/98
**      DESCRIPTION:    Mail app
**
**      Copyright (C) 1998, Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <time.h>

#include "Lgi.h"
#include "Mail.h"
#include "GToken.h"
#include "Base64.h"
#include "INetTools.h"
#include "GArray.h"

//////////////////////////////////////////////////////////////////////////////////////////////////
// return true if there are any characters with the 0x80 bit set
bool Is8Bit(char *Text)
{
	for (char *c = Text; *c; c++)
	{
		if (*c & 0x80)
		{
			return true;
		}
	}

	return false;
}

// returns the maximum length of the lines contained in the string
int MaxLineLen(char *Text)
{
	int Max = 0;
	int i = 0;

	for (char *c = Text; *c; c++)
	{
		if (*c == '\r')
		{
			// return
		}
		else if (*c == '\n')
		{
			// eol
			Max = max(i, Max);
			i = 0;
		}
		else
		{
			// normal char
			i++;
		}
	}

	return Max;
}

bool IsDotLined(char *Text)
{
	if (Text)
	{
		for (char *l = Text; l AND *l; )
		{
			if (l[0] == '.')
			{
				if (l[1] == '\n' OR l[1] == 0)
				{
					return true;
				}
			}

			l = strchr(l, '\n');
			if (l) l++;
		}
	}

	return false;
}

char ConvHexToBin(char c)
{
	if (c >= '0' AND c <= '9')
		return c - '0';
	if (c >= 'a' AND c <= 'f')
		return c + 10 - 'a';
	if (c >= 'A' AND c <= 'F')
		return c + 10 - 'A';
	return 0;
}

// Is s a valid non-whitespace string?
bool ValidNonWSStr(char *s)
{
	if (s AND *s)
	{
		while (*s AND strchr(" \r\t\n", *s))
		{
			s++;
		}

		if (*s)
		{
			return true;
		}
	}

	return false;
}

void TokeniseStrList(char *Str, List<char> &Output, char *Delim)
{
	if (Str AND Delim)
	{
		char *s = Str;
		while (*s)
		{
			char *e = s;
			for (; *e; e++)
			{
				if (strchr("\'\"", *e))
				{
					// handle string constant
					char EndChar = *e++;
					while (*e AND *e != EndChar) e++;
				}

				if (strchr(Delim, *e))
				{
					break;
				}
			}

			int Len = (int)e - (int)s;
			if (Len > 0)
			{
				char *Temp = NEW(char[Len+1]);
				if (Temp)
				{
					memcpy(Temp, s, Len);
					Temp[Len] = 0;
					Output.Insert(Temp);
				}
			}

			s = e;
			for (; *s AND strchr(Delim, *s); s++);
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
char *DecodeBase64Str(char *Str, int Len)
{
	if (Str)
	{
		int B64Len = (Len < 0) ? strlen(Str) : Len;
		int BinLen = BufferLen_64ToBin(B64Len);
		char *s = NEW(char[BinLen+1]);
		if (s)
		{
			int Converted = ConvertBase64ToBinary((uchar*)s, BinLen, Str, B64Len);
			s[Converted] = 0;
			DeleteArray(Str);
			Str = s;
		}
	}
	return Str;
}

char *DecodeQuotedPrintableStr(char *Str, int Len)
{
	if (Str)
	{
		if (Len < 0) Len = strlen(Str);
		uchar *s = NEW(uchar[Len+1]);
		if (s)
		{
			char *Out = (char*) s;
			char *Text = Str;

			for (int i=0; i<Len; i++)
			{
				if (Text[i] == '=')
				{
					if (Text[i+1] == '\r' OR Text[i+1] == '\n')
					{
						if (Text[i+1] == '\r') i++;
						if (Text[i+1] == '\n') i++;
					}
					else
					{
						i++;
						char c = ConvHexToBin(Text[i]) << 4;
						i++;
						c |= ConvHexToBin(Text[i]) & 0xF;

						*Out++ = c;
					}
				}
				else
				{
					*Out++ = Text[i];
				}
			}

			*Out++ = 0;
			DeleteArray(Str);
			Str = (char*) s;
		}
	}
	return Str;
}

#ifndef ToUpper
#define ToUpper(c)						( ((c)>='a'AND(c)<='z')?(c)-'a'+'A':(c) )
#endif

char *DecodeRfc2047(char *Str)
{
	if (Str AND strstr(Str, "=?"))
	{
		GBytePipe Temp;
		char *s = Str;
		while (*s)
		{
			// is there a word remaining
			bool Encoded = false;
			char *p = strstr(s, "=?");
			char *First = 0;
			char *Second = 0;
			char *End = 0;
			char *Cp = 0;
			if (p)
			{
				char *CpStart = p + 2;
				
				First = strchr(CpStart, '?');
				if (First)
				{
					Cp = NewStr(CpStart, (int)First-(int)CpStart);
					Second = strchr(First+1, '?');
					if (Second)
					{
						End = strstr(Second+1, "?=");
						Encoded = End != 0;
					}
				}
			}

			if (Encoded)
			{
				int Type = CONTENT_NONE;
				bool StripUnderscores = false;
				if (ToUpper(First[1]) == 'B')
				{
					// Base64 encoding
					Type = CONTENT_BASE64;
				}
				else if (ToUpper(First[1]) == 'Q')
				{
					// Quoted printable
					Type = CONTENT_QUOTED_PRINTABLE;
					StripUnderscores = true;
				}

				if (Type != CONTENT_NONE)
				{
					char *Block = NewStr(Second+1, (int)End-(int)Second-1);
					if (Block)
					{
						switch (Type)
						{
							case CONTENT_BASE64:
								Block = DecodeBase64Str(Block);
								break;
							case CONTENT_QUOTED_PRINTABLE:
								Block = DecodeQuotedPrintableStr(Block);
								break;
						}

						int Len = strlen(Block);
						if (StripUnderscores)
						{
							for (int i=0; i<Len; i++)
							{
								if (Block[i] == '_') Block[i] = ' ';
							}
						}

						Temp.Write((uchar*)s, (int)p-(int)s);
						
						char *Utf8 = (char*)LgiNewConvertCp("utf-8", Block, Cp);
						if (Utf8)
						{
							Temp.Write((uchar*)Utf8, strlen(Utf8));
							DeleteArray(Utf8);
						}
						else
						{
							Temp.Write((uchar*)Block, Len);
						}

						DeleteArray(Block);
					}
					s = End + 2;
				}
			}

			if (NOT Encoded)
			{
				int Len = strlen(s);
				Temp.Write((uchar*) s, Len);
				s += Len;
			}

			DeleteArray(Cp);
		}

		Temp.Write((uchar*) "", 1);
		int Len = Temp.GetSize();
		char *r = NEW(char[Len+1]);
		if (r)
		{
			Temp.Read((uchar*) r, Len);
			r[Len] = 0;
		}

		DeleteArray(Str);
		return r;
	}

	return Str;
}

/* Old Rfc Compliant way
char *EncodeRfc2047(char *Str, char *CodePage)
{
	if (NOT CodePage)
	{
		CodePage = "utf-8";
	}

	if (Str AND
		Is8Bit(Str))
	{
		GStringPipe p;
		char *Ws = " \r\t\n";

		char *n = 0;
		for (char *s=Str; s AND *s; s = n)
		{
			bool Is8Bit = false;

			// Search preceeding whitespace
			for (n = s; strchr(Ws, *n); n++);
			if (n != s)
			{
				// Output whitespace
				p.Push(s, (int)n-(int)s);
				s = n;
			}

			// Find end of word, and check for 8bit chars
			for (n = s; *n; n++)
			{
				if (*n & 0x80)
				{
					Is8Bit = true;
				}

				if (strchr(Ws, *n)) break;
			}

			if (n > s) // fall through if no word
			{
				// Process word
				if (Is8Bit)
				{
					// pick an encoding
					bool Base64 = false;
					char *DestCp = "utf-8";
					int Len = (int)n-(int)s;
					if (stricmp(CodePage, "utf-8") == 0)
					{
						DestCp = LgiDetectCharset(s, Len);
					}

					int Chars = 0;
					for (int i=0; i<Len; i++)
					{
						if (s[i] & 0x80) Chars++;
					}
					if (stristr(DestCp, "utf") OR (((double)Chars/Len)  > 0.4))
					{
						Base64 = true;
					}

					char *Buf = (char*)LgiNewConvertCp(DestCp, s, CodePage, (int)n-(int)s);
					if (Buf)
					{
						// encode the word
						char Temp[512];
						sprintf(Temp, "=?%s?%c?", DestCp, Base64 ? 'B' : 'Q');
						p.Push(Temp);

						if (Base64)
						{
							// Base64
							int Bytes = ConvertBinaryToBase64(Temp, sizeof(Temp), (uchar*)Buf, strlen(Buf));
							p.Push(Temp, Bytes);
						}
						else
						{
							// Quoted printable
							for (char *w = Buf; *w; w++)
							{
								if (*w & 0x80)
								{
									sprintf(Temp, "=%02.2X", (uchar)*w);
									p.Push(Temp);
								}
								else
								{
									p.Push(w, 1);
								}
							}
						}

						p.Push("?=");
						DeleteArray(Buf);
					}
				}
				else
				{
					// No 8bit chars so push plain word
					p.Push(s, (int)n-(int)s);
				}
			}
		}

		DeleteArray(Str);
		Str = p.NewStr();
	}

	return Str;
}
*/

char *EncodeRfc2047(char *Str, char *CodePage, List<char> *CharsetPrefs)
{
	if (NOT CodePage)
	{
		CodePage = "utf-8";
	}

	if (Str AND
		Is8Bit(Str))
	{
		GStringPipe p;

		// pick an encoding
		bool Base64 = false;
		char *DestCp = "utf-8";
		int Len = strlen(Str);;
		if (stricmp(CodePage, "utf-8") == 0)
		{
			DestCp = LgiDetectCharset(Str, Len, CharsetPrefs);
		}

		int Chars = 0;
		for (int i=0; i<Len; i++)
		{
			if (Str[i] & 0x80) Chars++;
		}
		if (stristr(DestCp, "utf") OR (((double)Chars/Len)  > 0.4))
		{
			Base64 = true;
		}

		char *Buf = (char*)LgiNewConvertCp(DestCp, Str, CodePage, Len);
		if (Buf)
		{
			// encode the word
			char Temp[512];
			sprintf(Temp, "=?%s?%c?", DestCp, Base64 ? 'B' : 'Q');
			p.Push(Temp);

			if (Base64)
			{
				// Base64
				int Bytes = ConvertBinaryToBase64(Temp, sizeof(Temp), (uchar*)Buf, strlen(Buf));
				p.Push(Temp, Bytes);
			}
			else
			{
				// Quoted printable
				for (char *w = Buf; *w; w++)
				{
					if (*w == ' ')
					{
						p.Push("_", 1);
					}
					else if (*w & 0x80 OR
							*w == '_' OR
							*w == '?' OR
							*w == '=')
					{
						sprintf(Temp, "=%02.2X", (uchar)*w);
						p.Push(Temp);
					}
					else
					{
						p.Push(w, 1);
					}
				}
			}

			p.Push("?=");
			DeleteArray(Buf);
		}

		DeleteArray(Str);
		Str = p.NewStr();
	}

	return Str;
}

//////////////////////////////////////////////////////////////////////////////
void DeNullText(char *in, int &len)
{
	char *out = in;
	char *end = in + len;
	while (in < end)
	{
		if (*in)
		{
			*out++ = *in;
		}
		else
		{
			len--;
		}
		in++;
	}
}

//////////////////////////////////////////////////////////////////////////////
static char *White = " \t\r\n";
typedef char CharPair[2];

char *RemovePairs(char *Str, CharPair *Pairs)
{
	char *s = Str;
	while (*s AND strchr(White, *s)) s++;

	char *e = s;
	if (e[0])
	{
		while (*e)
		{
			e++;
		}

		while (e > s AND strchr(White, e[-1]))
		{
			*e--;
		}
	}

	for (CharPair *p = Pairs; (*p)[0]; p++)
	{
		if ((*p)[0] == *s AND (*p)[1] == e[-1])
		{
			s++;
			e--;
			if (s < e)
			{
				// reset search
				p = Pairs - 1;
			}
			else break;
		}
	}

	int Len = (int)e - (int)s;
	memmove(Str, s, Len);
	Str[Len] = 0;
	
	return Str;
}

void DecodeAddrName(char *Start, char *&Name, char *&Addr, char *DefaultDomain)
{
	/* Testing code
	char *Input[] =
	{
		"'Matthew Allen (fret)' <fret@memecode.com>",
		"Matthew Allen (fret) <fret@memecode.com>",
		"\"'Matthew Allen'\" <fret@memecode.com>",
		"Matthew Allen",
		"fret@memecode.com",
		"\"<Matthew Allen>\" <fret@memecode.com>",
		"<Matthew Allen> (fret@memecode.com)",
		"Matthew Allen <fret@memecode.com>",
		"\"Matthew, Allen\" (fret@memecode.com)",
		"Matt'hew Allen <fret@memecode.com>",
		0
	};

	char *Name = 0;
	char *Addr = 0;

	for (char **i = Input; *i; i++)
	{
		DecodeAddrName(*i, Name, Addr, "ozemail.com.au");
		printf("N=%-24s A=%-24s\n", Name, Addr);
	}
	*/

	if (Start)
	{
		Name = Addr = 0;
		GArray<char*> Str;

		for (char *c = Start; *c; )
		{
			if (strchr(" \t\r\n", *c))
			{
				// skip whitespace
			}
			else if (strchr("\"'", *c))
			{
				// string delim
				char End = *c++;
				char *s = c;
				for (c; *c AND *c != End; c++)
				{
					if (*c == '\\')
					{
						c++;
					}
				}

				Str.Add(NewStr(s, (int)c-(int)s));
			}
			else if (strchr("<(", *c))
			{
				// brackets
				char Delim = (*c == '<') ? '>' : ')';
				char *End = strchr(c + 1, Delim);
				if (End)
				{
					End++;
				}
				else
				{
					End = c + 1;
					while (*End AND *End != ' ')
						End++;
				}
				Str.Add(NewStr(c, (int)End-(int)c));
				c = End;
			}
			else
			{
				// Some string
				char *s = c;
				for (; *c AND NOT strchr("<(\"", *c); c++);
				Str.Add(NewStr(s, (int)c-(int)s));
				continue;
			}

			if (*c)
			{
				c++;
			}
			else break;
		}

		CharPair Pairs[] =
		{
			{'<', '>'},
			{'(', ')'},
			{'\'', '\''},
			{'\"', '\"'},
			{0, 0},
		};
		
		int i;
		for (i=0; i<Str.Length(); i++)
		{
			if (strchr(Str[i], '@'))
			{
				Addr = RemovePairs(Str[i], Pairs);
				Str.DeleteAt(i, true);
				break;
			}
		}

		GStringPipe n;
		for (i=0; i<Str.Length(); i++)
		{
			char *s;
			if (n.GetSize())
			{
				n.Push(" ");
				s = Str[i];
			}
			else
			{
				s = RemovePairs(Str[i], Pairs);
			}
			n.Push(s);
		}
		Name = n.NewStr();

		if (Name)
		{
			char *In = Name;
			char *Out = Name;
			while (*In)
			{
				if (NOT (In[0] == '\\' AND In[1] == '\"'))
				{
					*Out++ = *In;
				}
				In++;
			}
			*Out = 0;
		}

		Str.DeleteArrays();
	}
}

void StrCopyToEOL(char *d, char *s)
{
	if (d AND s)
	{
		while (*s AND *s != '\r' AND *s != '\n')
		{
			*d++ = *s++;
		}

		*d = 0;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////
MailTransaction::MailTransaction()
{
	Index = -1;
	Flags = 0;
	Delete = false;
	Status = false;
	Oversize = false;
	Stream = 0;
}

MailTransaction::~MailTransaction()
{
}

//////////////////////////////////////////////////////////////////////////////////////////////////
FileDescriptor::FileDescriptor()
{
	Embeded = 0;
	Offset = 0;
	Size = 0;
	Data = 0;
	MimeType = 0;
	Lock = 0;
	OwnEmbeded = false;
}

FileDescriptor::FileDescriptor(GFile *embed, int64 offset, int64 size, char *name)
{
	Embeded = embed;
	Offset = offset;
	Size = size;
	Data = 0;
	MimeType = 0;
	Lock = 0;
	OwnEmbeded = false;

	if (name)
	{
		Name(name);
	}
}

FileDescriptor::FileDescriptor(char *name)
{
	Embeded = 0;
	Offset = 0;
	Size = 0;
	Data = 0;
	MimeType = 0;
	Lock = 0;
	OwnEmbeded = false;

	if (name)
	{
		Name(name);
		if (File.Open(name, O_READ))
		{
			Size = File.GetSize();
			File.Close();
		}
	}
}

FileDescriptor::FileDescriptor(char *data, int64 len)
{
	Embeded = 0;
	Offset = 0;
	MimeType = 0;
	Lock = 0;
	Size = len;
	OwnEmbeded = false;

	Data = data ? NEW(uchar[Size]) : 0;
	if (Data)
	{
		memcpy(Data, data, Size);
	}
}

FileDescriptor::~FileDescriptor()
{
	if (OwnEmbeded)
	{
		DeleteObj(Embeded);
	}

	DeleteArray(MimeType);
	DeleteArray(Data);
}

void FileDescriptor::SetOwnEmbeded(bool i)
{
	OwnEmbeded = i;
}

void FileDescriptor::SetLock(GSemaphore *l)
{
	Lock = l;
}

GSemaphore *FileDescriptor::GetLock()
{
	return Lock;
}

GFile *FileDescriptor::GotoObject()
{
	if (Embeded)
	{
		Embeded->Seek(Offset, SEEK_SET);
		return Embeded;
	}
	else if (Name() AND File.Open(Name(), O_READ))
	{
		return &File;
	}

	return NULL;
}

int FileDescriptor::Sizeof()
{
	return Size;
}

uchar *FileDescriptor::GetData()
{
	return Data;
}

bool FileDescriptor::Decode(char *ContentType,
							char *ContentTransferEncoding,
							char *MimeData,
							int MimeDataLen)
{
	bool Status = false;
	int Content = CONTENT_NONE;

	if (ContentType AND ContentTransferEncoding)
	{
		// Content-Type: application/octet-stream; name="Scribe.opt"
		Content = CONTENT_OCTET_STREAM;

		if (strnistr(ContentTransferEncoding, "base64", 1000))
		{
			Content = CONTENT_BASE64;
		}

		if (strnistr(ContentTransferEncoding, "quoted-printable", 1000))
		{
			Content = CONTENT_QUOTED_PRINTABLE;
		}

		if (Content != CONTENT_NONE);
		{
			char *NameKey = "name";
			char *n = strnistr(ContentType, NameKey, 1000);
			if (n)
			{
				char *Equal = strchr(n, '=');
				if (Equal)
				{
					Equal++;
					while (*Equal AND *Equal == '\"')
					{
						Equal++;
					}

					char *End = strchr(Equal, '\"');
					if (End)
					{
						*End = 0;
					}

					Name(Equal);
					Status = true;
				}
			}
		}
	}

	if (Status AND MimeData AND MimeDataLen > 0 AND Content != CONTENT_NONE)
	{
		Status = false;
		char *Base64 = NEW(char[MimeDataLen]);

		switch (Content)
		{
			case CONTENT_OCTET_STREAM:
			{
				Size = 0;
				DeleteObj(Data);
				Data = NEW(uchar[MimeDataLen]);
				if (Data)
				{
					Size = MimeDataLen;
					memcpy(Data, MimeData, Size);
					Status = true;
				}
				break;
			}
			case CONTENT_QUOTED_PRINTABLE:
			{
				Size = 0;
				DeleteObj(Data);
				Data = NEW(uchar[MimeDataLen+1]);
				if (Data)
				{
					char *Out = (char*) Data;
					for (int i=0; i<MimeDataLen; i++)
					{
						if (MimeData[i] == '=')
						{
							if (MimeData[i+1] == '\r' OR MimeData[i+1] == '\n')
							{
								if (MimeData[i+1] == '\r') i++;
								if (MimeData[i+1] == '\n') i++;
							}
							else
							{
								i++;
								char c = ConvHexToBin(MimeData[i]) << 4;
								i++;
								c |= ConvHexToBin(MimeData[i]) & 0xF;

								Size++;
								*Out++ = c;
							}
						}
						else
						{
							Size++;
							*Out++ = MimeData[i];
						}
					}

					*Out++ = 0;
					Status = true;
				}
				break;
			}
			case CONTENT_BASE64:
			{
				int InCount = 0;
				int OutCount = 0;
				char *In = MimeData;
				char *Out = Base64;
				while (InCount < MimeDataLen)
				{
					if (*In != '\r' AND *In != '\n')
					{
						*Out++ = *In;
						OutCount++;
					}
					InCount++;
					In++;
				}

				Size = BufferLen_64ToBin(OutCount);
				Data = NEW(uchar[Size]);
				if (Data)
				{
					int Converted = ConvertBase64ToBinary(Data, Size, Base64, OutCount);
					Status = Converted <= Size AND Converted >= Size - 3;
					if (Status)
					{
						Size = Converted;
					}
					else
					{
						DeleteArray(Data);
						Size = 0;
					}
				}
				break;
			}
		}
	}
	return Status;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
AddressDescriptor::AddressDescriptor(AddressDescriptor *Copy)
{
	Data = 0;
	Status = Copy ? Copy->Status : false;
	CC = Copy ? Copy->CC : false;
	Addr = Copy ? NewStr(Copy->Addr) : 0;
	Name = Copy ? NewStr(Copy->Name) : 0;
}

AddressDescriptor::~AddressDescriptor()
{
	_Delete();
}

void AddressDescriptor::_Delete()
{
	Data = 0;
	Status = false;
	CC = 0;
	DeleteArray(Name);
	DeleteArray(Addr);
}

void AddressDescriptor::Print(char *Str)
{
	if (Str)
	{
		if (Addr AND Name)
		{
			sprintf(Str, "%s (%s)", Addr, Name);
		}
		else if (Addr)
		{
			strcpy(Str, Addr);
		}
		else if (Name)
		{
			sprintf(Str, "(%s)", Name);
		}
	}
}

/*
int AddressDescriptor::Sizeof()
{
	return	SizeofStr(Name) +
			SizeofStr(Addr);
}

bool AddressDescriptor::Serialize(GFile &f, bool Write)
{
	bool Status = true;
	if (Write)
	{
		WriteStr(f, Name);
		WriteStr(f, Addr);
	}
	else
	{
		DeleteArray(Name);
		Name = ReadStr(f PassDebugArgs);
		DeleteArray(Addr);
		Addr = ReadStr(f PassDebugArgs);
	}
	return Status;
}
*/

//////////////////////////////////////////////////////////////////////////////////////////////////
MailProtocol::MailProtocol()
{
	Socket = 0;
	ServerMsg = 0;
	Buffer[0] = 0;

	LogNotify = 0;
	LogMsg = 0;
	LogData = 0;

	Items = 0;
	Transfer = 0;

	ExtraOutgoingHeaders = 0;
}

MailProtocol::~MailProtocol()
{
	CharsetPrefs.DeleteArrays();
	DeleteArray(ServerMsg);
	DeleteArray(ExtraOutgoingHeaders);
}

void MailProtocol::Log(LogEntry *e)
{
	if (LogNotify AND LogMsg AND e)
	{
		LogNotify->PostEvent(LogMsg, LogData, (int)e);
	}
	else
	{
		DeleteObj(e);
	}
}

void MailProtocol::Log(char *Str, COLOUR c)
{
	if (LogNotify AND LogMsg)
	{
		Log(NEW(LogEntry(Str, c)));
	}
}

bool MailProtocol::Read()
{
	bool Status = false;
	if (Socket)
	{
		Status = Socket->Read(Buffer, sizeof(Buffer), 0) > 0;
	}
	return Status;
}

bool MailProtocol::Write(char *Buf, bool LogWrite)
{
	bool Status = false;
	if (Socket)
	{
		char *p = (Buf) ? Buf : Buffer;
		Status = Socket->Write(p, strlen(p), 0);
		
		if (LogWrite)
		{
			Log(p, MAIL_SEND_COLOUR);
		}
	}
	return Status;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
#define	VERIFY_RET_VAL(Arg)					\
{											\
	if (NOT Arg)							\
	{										\
		Socket = 0;							\
		return false;						\
	}										\
}

#define	VERIFY_ONERR(Arg)					\
{											\
	if (NOT Arg)							\
	{										\
		Socket = 0;							\
		goto CleanUp;						\
	}										\
}

MailSmtp::MailSmtp()
{
	ProgramName = 0;
}

MailSmtp::~MailSmtp()
{
}

bool MailSmtp::Open(GSocketI *S,
					char *RemoteHost,
					char *LocalDomain,
					char *UserName,
					char *Password,
					int Port,
					int Flags)
{
	char Str[256] = "";
	bool Status = false;

	if (RemoteHost)
	{
		strcpy(Str, RemoteHost);
		char *Colon = strchr(Str, ':');
		if (Colon)
		{
			*Colon = 0;
			Colon++;
			Port = atoi(Colon);
		}
		if (Port == 0)
		{
			Port = SMTP_PORT;
		}

		char *Server = TrimStr(Str);
		if (Server)
		{
			Socket = S;
			if (Socket->Open(Server, Port))
			{
				// receive signon message
				VERIFY_RET_VAL(ReadReply("220"));

				if (ValidStr(UserName) AND
					ValidStr(Password))
				{
					GStringPipe Str;

					// Rfc 2554 ESMTP authentication
					SmtpHello:
					sprintf(Buffer, "EHLO %s\r\n", (ValidNonWSStr(LocalDomain)) ? LocalDomain : "default");
					VERIFY_RET_VAL(Write(0, true));
					VERIFY_RET_VAL(ReadReply("250", &Str));

					bool Authed = false;
					bool NoAuthTypes = false;
					bool SupportsStartTLS = false;
					GHashTable TheirAuthTypes;
					GHashTable MyAuthTypes(16);
					MyAuthTypes.Add("PLAIN");
					MyAuthTypes.Add("LOGIN");

					// Look through the response for the auth line
					char *Response = Str.NewStr();
					if (Response)
					{
						GToken Lines(Response, "\n");
						for (int i=0; i<Lines.Length(); i++)
						{
							char *l = Lines[i];
							char *AuthStr = strstr(l, "AUTH");
							if (AuthStr)
							{
								// walk through AUTH types
								GToken Types(AuthStr + 4, " ,;");
								for (int a=0; a<Types.Length(); a++)
								{
									TheirAuthTypes.Add(Types[a]);
								}
							}
							if (stristr(l, "STARTTLS"))
							{
								SupportsStartTLS = true;
							}
						}
						DeleteArray(Response);
					}

					if (SupportsStartTLS AND TestFlag(Flags, MAIL_USE_STARTTLS))
					{
						strcpy(Buffer, "STARTTLS\r\n");
						VERIFY_RET_VAL(Write(0, true));
						VERIFY_RET_VAL(ReadReply("220", &Str));
						if (Socket->SetParameter(GSocket_Protocol, (int)"SSL"))
						{
							Flags &= ~MAIL_USE_STARTTLS;
							goto SmtpHello;
						}
						else
						{
							// SSL init failed... what to do here?
						}
					}

					if (TheirAuthTypes.Length() == 0)
					{
						// No auth types? huh?
						if (TestFlag(Flags, MAIL_USE_AUTH))
						{
							if (TestFlag(Flags, MAIL_USE_PLAIN))
							{
								// Force plain type
								TheirAuthTypes.Add("PLAIN");
							}
							else if (TestFlag(Flags, MAIL_USE_LOGIN))
							{
								// Force login type
								TheirAuthTypes.Add("LOGIN");
							}
							else
							{
								// Oh well, we'll just try all types
								char *a;
								for (bool b=MyAuthTypes.First(&a); b; b=MyAuthTypes.Next(&a))
								{
									TheirAuthTypes.Add(a);
								}
							}
						}
						else
						{						
							NoAuthTypes = true;
							goto NormalLogin;
						}
					}

					// Try all their auth types against our internally support types
					if (TheirAuthTypes.Find("LOGIN"))
					{
						VERIFY_RET_VAL(Write("AUTH LOGIN\r\n", true));
						VERIFY_RET_VAL(ReadReply("334"));

						ZeroObj(Buffer);
						ConvertBinaryToBase64(Buffer, sizeof(Buffer), (uchar*)UserName, strlen(UserName));
						strcat(Buffer, "\r\n");
						VERIFY_RET_VAL(Write(0, true));
						if (ReadReply("334"))
						{
							ZeroObj(Buffer);
							ConvertBinaryToBase64(Buffer, sizeof(Buffer), (uchar*)Password, strlen(Password));
							strcat(Buffer, "\r\n");
							VERIFY_RET_VAL(Write(0, true));
							if (ReadReply("235"))
							{
								Authed = true;
							}
						}
					}
					
					if (NOT Authed AND TheirAuthTypes.Find("PLAIN"))
					{
						uchar Tmp[256];
						ZeroObj(Tmp);
						char *a = (char*)Tmp + 1;
						strcpy(a, UserName);
						a += strlen(UserName) + 1;
						strcpy(a, Password);
						a += strlen(Password);
						int Len = (int)a-(int)Tmp;

						char B64[256];
						ZeroObj(B64);
						ConvertBinaryToBase64(B64, sizeof(B64), Tmp, Len);

						sprintf(Buffer, "AUTH PLAIN %s\r\n", B64);
						VERIFY_RET_VAL(Write(0, true));
						if (ReadReply("235"))
						{
							Authed = true;
						}
					}

					if (NOT Authed)
					{
						DeleteArray(ServerMsg);

						if (NoAuthTypes)
						{
							ServerMsg = NewStr(LgiLoadString(L_ERROR_ESMTP_NO_AUTHS,
															"The server didn't return the authentication methods it supports."));
						}
						else
						{
							strcpy(Buffer, LgiLoadString(L_ERROR_ESMTP_UNSUPPORTED_AUTHS,
														"The server doesn't support any compatible authentication types:\n\t"));
							
							char *a = 0;
							for (void *p = TheirAuthTypes.First(&a); p; p = TheirAuthTypes.Next(&a))
							{
								int Bytes = strlen(Buffer);
								snprintf(Buffer + Bytes, sizeof(Buffer) - Bytes, "%s ", a);
							}
							ServerMsg = NewStr(Buffer);
						}
					}

					Status = Authed;
				}
				else
				{
					NormalLogin:
					// Normal SMTP login
					// send HELO message
					sprintf(Buffer, "HELO %s\r\n", (ValidNonWSStr(LocalDomain)) ? LocalDomain : "default");
					VERIFY_RET_VAL(Write(0, true));
					VERIFY_RET_VAL(ReadReply("250"));

					Status = true;
				}
			}

			DeleteArray(Server);
		}
	}

	return Status;
}

bool MailSmtp::WriteText(char *Str)
{
	// we have to convert all strings to CRLF in here
	bool Status = false;
	if (Str)
	{
		GBytePipe Temp;

		char *Start = Str;
		while (*Str)
		{
			if (*Str == '\n')
			{
				// send a string
				int Size = (int)Str-(int)Start;
				if (Str[-1] == '\r') Size--;
				Temp.Write((uchar*) Start, Size);
				Temp.Write((uchar*) "\r\n", 2);
				Start = Str + 1;
			}

			Str++;
		}

		// send the final string
		int Size = (int)Str-(int)Start;
		if (Str[-1] == '\r') Size--;
		Temp.Write((uchar*) Start, Size);

		Size = Temp.GetSize();
		char *Data = NEW(char[Size]);
		if (Data)
		{
			Temp.Read((uchar*) Data, Size);
			Status = Socket->Write(Data, Size, 0);
			DeleteArray(Data);
		}
	}

	return Status;
}

char *CreateAddressTag(List<AddressDescriptor> &l, int Type, List<char> *CharsetPrefs)
{
	char *Result = 0;
	List<AddressDescriptor> Addr;
	AddressDescriptor *a;
	for (a = l.First(); a; a = l.Next())
	{
		if (a->CC == Type)
		{
			Addr.Insert(a);
		}
	}
	
	if (Addr.Length() > 0)
	{
		GStringPipe StrBuf;
		StrBuf.Push((Type == 0) ? (char*)"To: " : (char*)"Cc: ");

		for (a = Addr.First(); a; )
		{
			AddressDescriptor *NextA = Addr.Next();
			char Buffer[256] = "";

			if (a->Addr AND strchr(a->Addr, ','))
			{
				// Multiple address format
				GToken t(a->Addr, ",");
				for (int i=0; i<t.Length(); i++)
				{
					sprintf(Buffer, "<%s>", t[i]);
					if (i < t.Length()-1) strcat(Buffer, ",\r\n\t");
					StrBuf.Push(Buffer);
					Buffer[0] = 0;
				}
			}
			else if (a->Name)
			{
				// Name and addr
				char *Mem = 0;
				char *Name = a->Name;

				if (Is8Bit(Name))
				{
					Name = Mem = EncodeRfc2047(NewStr(Name), 0, CharsetPrefs);
				}

				sprintf(Buffer, "%s <%s>", Name, a->Addr);

				DeleteArray(Mem);
			}
			else if (a->Addr)
			{
				// Just addr
				sprintf(Buffer, "<%s>", a->Addr);
			}

			if (NextA) strcat(Buffer, ",\r\n\t");
			StrBuf.Push(Buffer);

			a = NextA;
		}

		StrBuf.Push("\r\n");
		Result = StrBuf.NewStr();
	}

	return Result;
}

// This class implements a pipe that writes to a socket
class SocketPipe : public GStringPipe
{
	GSocketI *s;
	MailProtocolProgress *p;

public:
	bool Status;

	SocketPipe(GSocketI *socket, MailProtocolProgress *progress)
	{
		s = socket;
		p = progress;
		Status = true;
	}

	int Read(void *Ptr, int Size, int Flags)
	{
		return false;
	}
	
	int64 SetSize(int64 Size)
	{
		if (p)
		{
			p->Start = LgiCurrentTime();
			p->Range = Size;
			return Size;
		}

		return -1;
	}

	int Write(void *Ptr, int Size, int Flags)
	{
		int w = s->Write((char*)Ptr, Size, 0);

		if (p AND p->Range AND w > 0)
		{
			p->Value += w;
		}

		return w;
	}
};

GStringPipe *MailSmtp::SendStart(List<AddressDescriptor> &To, AddressDescriptor *From)
{
	if (To.First() AND From)
	{
		// send MAIL message
		if (From AND ValidStr(From->Addr))
		{
			sprintf(Buffer, "MAIL FROM: <%s>\r\n", From->Addr);
		}
		else
		{
			ServerMsg = NewStr("No 'from' address in email.");
			return false;
		}

		VERIFY_RET_VAL(Write(0, true));
		VERIFY_RET_VAL(ReadReply("250"));

		// send RCPT message
		bool AddrOk = true;
		Iterator<AddressDescriptor> Recip(&To);
		for (AddressDescriptor *a = Recip.First(); a; a = Recip.Next())
		{
			if (ValidStr(a->Addr))
			{
				GToken Parts(a->Addr, ",");
				for (int p=0; p<Parts.Length(); p++)
				{
					sprintf(Buffer, "RCPT TO: <%s>\r\n", Parts[p]);
					VERIFY_RET_VAL(Write(0, true));
					a->Status = ReadReply("25");
					AddrOk &= a->Status != 0; // at least one address is ok
				}
			}
			else
			{
				LgiTrace("%s:%i - Send Addr wasn't valid\n", __FILE__, __LINE__);
			}
		}

		if (AddrOk)
		{
			// send DATA message
			sprintf(Buffer, "DATA\r\n");
			VERIFY_RET_VAL(Write(0, true));
			VERIFY_RET_VAL(ReadReply("354"));

			return NEW(SocketPipe(Socket, Transfer));
		}
	}

	return 0;
}

bool MailSmtp::SendEnd(GStringPipe *m)
{
	bool Status = false;

	SocketPipe *Msg = dynamic_cast<SocketPipe*>(m);
	if (Msg)
	{
		// send message terminator and receive reply
		if (Msg->Status AND
			Msg->Write((void*)"\r\n.\r\n", 5, 0))
		{
			Status = ReadReply("250");
		}
		// else
		//		just close the connection on them
		//		so nothing gets sent
	}

	DeleteObj(m);

	return Status;
}

bool MailSmtp::Send(MailMessage *Msg, bool Mime)
{
	bool Status = false;

	if (Socket AND Msg)
	{
		GStringPipe *Sink = SendStart(Msg->To, Msg->From);
		if (Sink)
		{
			// setup a gui progress meter to send the email,
			// the length is just a guesstimate as we won't know the exact
			// size until we encode it all, and I don't want it hanging around
			// in memory at once, so we encode and send on the fly.
			int Length =	1024 +
							(Msg->GetBody() ? strlen(Msg->GetBody()) : 0);
			for (FileDescriptor *f=Msg->FileDesc.First(); f; f=Msg->FileDesc.Next())
			{
				Length += f->Sizeof() * 4 / 3;
			}

			// encode and send message for transport
			Msg->Encode(*Sink, 0, this);
			Status = SendEnd(Sink);
		}
	}

	return Status;
}

bool MailSmtp::Close()
{
	if (Socket)
	{
		// send QUIT message
		sprintf(Buffer, "QUIT\r\n");
		VERIFY_RET_VAL(Write(0, true));
		VERIFY_RET_VAL(ReadReply("221"));

		Socket->Close();
		Socket = NULL;

		return true;
	}

	return false;
}

bool MailSmtp::ReadReply(char *Str, GStringPipe *Pipe)
{
	bool Status = false;
	if (Socket AND Str)
	{
		int Pos = 0;
		char *Start = Buffer;
		ZeroObj(Buffer);

		while (Pos < sizeof(Buffer))
		{
			int Len = Socket->Read(Buffer+Pos, sizeof(Buffer)-Pos, 0);
			if (Len > 0)
			{
				char *Eol = strstr(Start, "\r\n");
				while (Eol)
				{
					// wipe EOL chars
					*Eol++ = 0;
					*Eol++ = 0;

					// log it
					LogEntry *Entry = NEW(LogEntry(Start, MAIL_RECEIVE_COLOUR));

					// process
					if (Pipe)
					{
						if (Pipe->GetSize()) Pipe->Push("\n");
						Pipe->Push(Start);
					}

					if (Start[3] == ' ')
					{
						// end of response
						if (NOT strncmp(Start, Str, strlen(Str)))
						{
							Status = true;
						}

						if (Entry)
						{
							// get code
							Start[3] = 0;
							if (atoi(Start) >= 400)
							{
								// error
								Entry->c = MAIL_ERROR_COLOUR;
							}
							Start[3] = ' ';
						}
						Log(Entry);

						// exit loop
						Pos = sizeof(Buffer);
						break;
					}
					else
					{
						Log(Entry);

						// more lines follow
						Start = Eol;
						Eol = strstr(Start, "\r\n");
					}
				}

				Pos += Len;
   			}
			else break;
		}

		if (NOT Status)
		{
			DeleteArray(ServerMsg);
			ServerMsg = NewStr(Buffer);
		}
	}

	return Status;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
class Mail2Folder : public GStringPipe
{
	char File[256];
	GFile F;

public:
	Mail2Folder(char *Path, List<AddressDescriptor> &To)
	{
		do
		{
			char n[32];
			sprintf(n, "%u.mail", LgiRand());
			LgiMakePath(File, sizeof(File), Path, n);
		}
		while (FileExists(File));

		if (F.Open(File, O_WRITE))
		{
			F.Print("Forward-Path: ");

			int i = 0;
			for (AddressDescriptor *a=To.First(); a; a=To.Next(), i++)
			{
				a->Status = true;
				if (i) F.Print(",");
				F.Print("<%s>", a->Addr);
			}

			F.Print("\r\n");
		}
	}

	~Mail2Folder()
	{
		F.Close();
	}

	int Read(void *Buffer, int Size, int Flags = 0)
	{
		return F.Read(Buffer, Size, Flags);
	}

	int Write(void *Buffer, int Size, int Flags = 0)
	{
		return F.Write(Buffer, Size, Flags);
	}
};

class MailPostFolderPrivate
{
public:
	char *Path;

	MailPostFolderPrivate()
	{
		Path = 0;
	}

	~MailPostFolderPrivate()
	{
		DeleteArray(Path);
	}
};

MailSendFolder::MailSendFolder(char *Path)
{
	d = NEW(MailPostFolderPrivate);
	d->Path = NewStr(Path);
}

MailSendFolder::~MailSendFolder()
{
	DeleteObj(d);
}

bool MailSendFolder::Open(GSocketI *S, char *RemoteHost, char *LocalDomain, char *UserName, char *Password, int Port, int Flags)
{
	return DirExists(d->Path);
}

bool MailSendFolder::Close()
{
	return true;
}

GStringPipe *MailSendFolder::SendStart(List<AddressDescriptor> &To, AddressDescriptor *From)
{
	return NEW(Mail2Folder(d->Path, To));
}

bool MailSendFolder::SendEnd(GStringPipe *Sink)
{
	DeleteObj(Sink);
	return true;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
class MailItem
{
public:
	char *File;
	bool Delete;
	
	MailItem(char *f)
	{
		File = NewStr(f);
		Delete = false;
	}

	~MailItem()
	{
		DeleteArray(File);
	}
};

class MailReceiveFolderPrivate
{
public:
	char *Path;
	List<MailItem> Mail;

	MailReceiveFolderPrivate()
	{
		Path = 0;
	}

	~MailReceiveFolderPrivate()
	{
		DeleteArray(Path);
		Mail.DeleteObjects();
	}

	void Empty()
	{
		for (MailItem *m = Mail.First(); m; m = Mail.Next())
		{
			if (m->Delete)
			{
				FileDev->DeleteFile(m->File, false);
			}
		}

		Mail.DeleteObjects();
	}
};

MailReceiveFolder::MailReceiveFolder(char *Path)
{
	d = NEW(MailReceiveFolderPrivate);
	d->Path = NewStr(Path);
}

MailReceiveFolder::~MailReceiveFolder()
{
	DeleteObj(d);
}

bool MailReceiveFolder::Open(GSocketI *S, char *RemoteHost, int Port, char *User, char *Password, char *&Cookie, int Flags)
{
	bool Status = false;

	if (DirExists(d->Path))
	{
		GDirectory *Dir = FileDev->GetDir();
		if (Dir)
		{
			for (bool b = Dir->First(d->Path, "*.mail"); b; b = Dir->Next())
			{
				if (NOT Dir->IsDir())
				{
					char p[300];
					Dir->Path(p, sizeof(p));
					d->Mail.Insert(NEW(MailItem(p)));
				}
			}
			DeleteObj(Dir);
		}
		Status = true;
	}

	return Status;
}

bool MailReceiveFolder::Close()
{
	d->Empty();
	return true;
}

int MailReceiveFolder::GetMessages()
{
	return d->Mail.Length();
}

bool MailReceiveFolder::Receive(GArray<MailTransaction*> &Trans, MailCallbacks *Callbacks)
{
	bool Status = false;

	for (int i=0; i<Trans.Length(); i++)
	{
		MailTransaction *t = Trans[i];
		if (t AND t->Stream)
		{
			t->Status = false;

			MailItem *m = d->Mail[t->Index];
			if (m)
			{
				GFile i;
				if (i.Open(m->File, O_READ))
				{
					GCopyStreamer c;
					if (c.Copy(&i, t->Stream))
					{
						Status = t->Status = true;
					}
				}
			}
		}
	}

	return Status;
}

bool MailReceiveFolder::Delete(int Message)
{
	MailItem *m = d->Mail[Message];
	if (m)
	{
		m->Delete = true;
		return false;
	}

	return false;
}

int MailReceiveFolder::Sizeof(int Message)
{
	MailItem *m = d->Mail[Message];
	if (m)
	{
		return LgiFileSize(m->File);
	}

	return 0;
}

bool MailReceiveFolder::GetUid(int Message, char *Id)
{
	if (Id)
	{
		MailItem *m = d->Mail[Message];
		if (m)
		{
			char *s = strrchr(m->File, DIR_CHAR);
			if (s++)
			{
				char *e = strchr(s, '.');
				if (NOT e) e = s + strlen(s);
				int Len = e - s;
				memcpy(Id, s, Len);
				Id[Len] = 0;

				return true;
			}		
		}
	}

	return false;
}

bool MailReceiveFolder::GetUidList(List<char> &Id)
{
	bool Status = false;

	for (int i=0; i<d->Mail.Length(); i++)
	{
		char Uid[256];
		if (GetUid(i, Uid))
		{
			Status = true;
			Id.Insert(NewStr(Uid));
		}
		else
		{
			Id.DeleteArrays();
			Status = false;
			break;
		}
	}

	return Status;
}

char *MailReceiveFolder::GetHeaders(int Message)
{
	MailItem *m = d->Mail[Message];
	if (m)
	{
		GFile i;
		if (i.Open(m->File, O_READ))
		{
			GStringPipe o;
			GCopyStreamer c;
			GLinePrefix e("", false);
			if (c.Copy(&i, &o, &e))
			{
				return o.NewStr();
			}
		}
	}

	return 0;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
MailPop3::MailPop3()
{
	End = "\r\n.\r\n";
	Marker = End;
	Messages = 0;
}

MailPop3::~MailPop3()
{
}

int MailPop3::GetInt()
{
	char Buf[32];
	char *Start = strchr(Buffer, ' ');
	if (Start)
	{
		Start++;
		char *End = strchr(Start, ' ');
		if (End)
		{
			int Len = (int) (End - Start);
			memcpy(Buf, Start, Len);
			Buf[Len] = 0;
			return atoi(Buf);
		}
	}
	return 0;
}

bool MailPop3::ReadReply(int Timeout)
{
	bool Status = false;
	if (Socket)
	{
		int Now = LgiCurrentTime();
		int Start;
		if (Timeout > 0)
			Start = Now;
		else
			Start = 0;

		int Pos = 0;
		ZeroObj(Buffer);
		do
		{
			int Result = 0;

			if (Timeout > 0)
			{
				// Wait for data to arrive
				do
				{
					if (Socket->Is(GReadable))
					{
						// Got some data...
						Result = Socket->Read(Buffer+Pos, sizeof(Buffer)-Pos, 0);
						break;
					}

					// Not there yet
					LgiSleep(10);
				}
				while (LgiCurrentTime() < Start + Timeout); // Do timeout
			}
			else
			{
				// No timeout so just block...
				Result = Socket->Read(Buffer+Pos, sizeof(Buffer)-Pos, 0);
			}

			if (Result <= 0) // an error?
			{
				// Leave the loop...
				break;
			}

			Pos += Result;
    	}
		while (	NOT strstr(Buffer, "\r\n") AND
				sizeof(Buffer)-Pos > 0);

		Status = (Buffer[0] == '+') AND strstr(Buffer, "\r\n");

		char *Cr = strchr(Buffer, '\r');
		if (Cr) *Cr = 0;
		Log(Buffer, (Status) ? MAIL_RECEIVE_COLOUR : MAIL_ERROR_COLOUR);
		if (Cr) *Cr = '\r';

		if (NOT Status)
		{
			DeleteArray(ServerMsg);
			ServerMsg = NewStr(Buffer);
		}
	}

	return Status;
}

bool MailPop3::Open(GSocketI *S, char *RemoteHost, int Port, char *User, char *Password, char *&Cookie, int Flags)
{
	bool Status = false;
	bool RemoveMail = false;
	char Str[256] = "";
	char *Apop = 0;
	char *Server = 0;

	if (RemoteHost)
	{
		if (Port < 1) Port = POP3_PORT;

		strcpy(Str, RemoteHost);
		char *Colon = strchr(Str, ':');
		if (Colon)
		{
			*Colon = 0;
			Colon++;
			Port = atoi(Colon);
		}

		if (S AND
			User AND
			Password AND
			(Server = TrimStr(Str)))
		{
			ReStartConnection:
			Socket = S;
			if (Socket->Open(Server, Port) AND
				ReadReply())
			{
				bool NoAPOP = Cookie ? stristr(Cookie, "NoAPOP") : 0;
				if (NOT NoAPOP)
				{
					char *s = strchr(Buffer + 3, '<');
					if (s)
					{
						char *e = strchr(s + 1, '>');
						if (e)
						{
							Apop = NewStr(s, e - s + 1);
						}
					}
				}

				// login
				bool Authed = false;
				char *user = (char*) LgiNewConvertCp("iso-8859-1", User, "utf-8");
				char *pass = (char*) LgiNewConvertCp("iso-8859-1", Password, "utf-8");

				if (user AND pass)
				{
					bool SecurityError = false;

					if (TestFlag(Flags, MAIL_USE_STARTTLS))
					{
						strcpy(Buffer, "STARTTLS\r\n");
						VERIFY_RET_VAL(Write(0, true));
						VERIFY_RET_VAL(ReadReply());
						if (Socket->SetParameter(GSocket_Protocol, (int)"SSL"))
						{
							Flags &= ~MAIL_USE_STARTTLS;
						}
						else
						{
							SecurityError = true;
						}
					}

					if (NOT SecurityError AND Apop) // GotKey, not implemented
					{
						// using encrypted password
						unsigned char Digest[16];
						char HexDigest[33];

						// append password
						char Key[256];
						strcpy(Key, Apop);
						strcat(Key, pass);

						ZeroObj(Digest);
						MDStringToDigest(Digest, Key);
						for (int i = 0; i < 16; i++)
							sprintf(HexDigest + (i*2), "%02.2x", Digest[i]);
						HexDigest[32] = 0;

						sprintf(Buffer, "APOP %s %s\r\n", user, HexDigest);
						VERIFY_ONERR(Write(0, true));
						Authed = ReadReply();

						if (NOT Authed)
						{
							DeleteArray(Cookie);
							DeleteArray(Apop);
							Cookie = NewStr("NoAPOP");
							S->Close();
							goto ReStartConnection;
						}
					}

					if (NOT SecurityError AND NOT Authed)
					{
						// have to use non-key method
						sprintf(Buffer, "USER %s\r\n", user);
						VERIFY_ONERR(Write(0, true));
						VERIFY_ONERR(ReadReply());

						sprintf(Buffer, "PASS %s\r\n", pass);
						VERIFY_ONERR(Write(0, false));
						Log("PASS *******", MAIL_SEND_COLOUR);

						Authed = ReadReply();
					}

					DeleteArray(user);
					DeleteArray(pass);
				}

				if (Authed)
				{
					Status = true;

					// see how many messages there are
					VERIFY_ONERR(Write("STAT\r\n", true));
					VERIFY_ONERR(ReadReply());

					Messages = GetInt();
				}
				else
				{
					if (Socket)
					{
						Socket->Close();
						Socket = NULL;
					}
				}
			}
		}
	}

CleanUp:
	DeleteArray(Apop);
	DeleteArray(Server);

	return Status;
}

bool MailPop3::MailIsEnd(char *Ptr, int Len)
{
	for (char *c = Ptr; Len-- > 0; c++)
	{
		if (*c != *Marker)
		{
			Marker = End;
		}

		if (*c == *Marker)
		{
			Marker++;
			if (NOT *Marker)
			{
				return true;
			}
		}
	}

	return false;
}

bool MailPop3::Receive(GArray<MailTransaction*> &Trans, MailCallbacks *Callbacks)
{
	bool Status = false;

	if (Trans.Length() > 0 AND
		Socket)
	{
		for (int n = 0; n<Trans.Length(); n++)
		{
			int Message = Trans[n]->Index;
			GStream *Msg = Trans[n]->Stream;
			if (Msg)
			{
				int Size = 0;
				
				// Transfer is not null when the caller wants info on the bytes comming in
				if (Transfer OR Callbacks)
				{
					// get message size
					sprintf(Buffer, "LIST %i\r\n", Message + 1);
					VERIFY_RET_VAL(Write(0, true));
					VERIFY_RET_VAL(ReadReply());

					char *s = strchr(Buffer, ' ');
					if (s)
					{
						s = strchr(s+1, ' ');
						if (s)
						{
							Size = atoi(s);
						}
					}
				}

				MailSrcStatus Action = DownloadAll;
				int TopLines = 100;

				if (Callbacks AND Callbacks->OnSrc)
				{
					Action = Callbacks->OnSrc(Trans[n], Size, &TopLines, Callbacks->CallbackData);
				}

				if (Action == DownloadAbort)
				{
					break;
				}

				if (Action == DownloadAll OR
					Action == DownloadTop)
				{
					if (Action == DownloadAll)
					{
						sprintf(Buffer, "RETR %i\r\n", Message + 1);
					}
					else
					{
						sprintf(Buffer, "TOP %i %i\r\n", Message + 1, TopLines);
					}
					VERIFY_RET_VAL(Write(0, true));

					GLinePrefix End(".\r\n");

					if (Transfer)
					{
						Transfer->Value = 0;
						Transfer->Range = Size;
						Transfer->Start = LgiCurrentTime();
					}

					// Read status line
					ZeroObj(Buffer);

					int Used = 0;
					bool Ok = false;
					bool Finished = false;
					while (Socket->IsOpen())
					{
						int r = Socket->Read(Buffer+Used, sizeof(Buffer)-Used-1, 0);
						if (r > 0)
						{
							DeNullText(Buffer + Used, r);
							
							if (Transfer)
							{
								Transfer->Value += r;
							}

							char *Eol = strchr(Buffer, '\n');
							if (Eol)
							{
								Eol++;
								Ok = Buffer[0] == '+';
								if (Ok)
								{
									Log(Buffer, MAIL_RECEIVE_COLOUR);	

									// The Buffer was zero'd at the beginning garrenteeing
									// NULL termination
									int Len = strlen(Eol);
									Msg->Write(Eol, Len);

									if (End.IsEnd(Eol, Len) >= 0)
									{
										Status = Trans[n]->Status = true;
										Finished = true;
									}
								}
								else
								{
									Log(Buffer, MAIL_ERROR_COLOUR);	
									Finished = true;								
								}
								break;
							}

							Used += r;
						}
						else break;
					}

					if (NOT Finished)
					{
						if (Ok)
						{
							// Read rest of message
							while (Socket->IsOpen())
							{
								int r = Socket->Read(Buffer, sizeof(Buffer), 0);
								if (r > 0)
								{
									DeNullText(Buffer, r);
									
									if (Transfer)
									{
										Transfer->Value += r;
									}

									if (Msg->Write(Buffer, r) != r)
									{
										LgiAssert(0);
									}

									if (End.IsEnd(Buffer, r) >= 0)
									{
										Status = Trans[n]->Status = true;
										break;
									}
								}
								else
								{
									break;
								}
							}
							
							if (NOT Status)
							{
								LgiTrace("%s:%i - Didn't get end-of-mail marker.\n", __FILE__, __LINE__);
							}
						}
						else
						{
							LgiTrace("%s:%i - Didn't get Ok.\n", __FILE__, __LINE__);
							break;
						}
					}

					if (Callbacks AND Callbacks->OnReceive)
					{
						Callbacks->OnReceive(Trans[n], Callbacks->CallbackData);
					}

					if (Transfer)
					{
						Transfer->Empty();
					}
				}
				else
				{
					Trans[n]->Oversize = Status = true;
				}

				if (Items)
				{
					Items->Value++;
				}
			}
			else
			{
				LgiTrace("%s:%i - No stream.\n", __FILE__, __LINE__);
			}
		}
	}
	else
	{
		LgiTrace("%s:%i - Arg check failed.\n", __FILE__, __LINE__);
	}

	return Status;
}

bool MailPop3::GetSizes(GArray<int> &Sizes)
{
	if (Socket)
	{
		strcpy(Buffer, (char*)"LIST\r\n");
		VERIFY_RET_VAL(Write(0, true));

		char *s = 0;
		if (ReadMultiLineReply(s))
		{
			GToken l(s, "\r\n");
			DeleteArray(s);
			for (int i=0; i<l.Length(); i++)
			{
				char *sp = strrchr(l[i], ' ');
				if (sp)
				{
					Sizes[i] = atoi(sp+1);
				}
			}
		}
	}

	return Sizes.Length() > 0;
}

int MailPop3::Sizeof(int Message)
{
	int Size = 0;
	if (Socket)
	{
		sprintf(Buffer, "LIST %i\r\n", Message + 1);
		VERIFY_RET_VAL(Write(0, true));
		VERIFY_RET_VAL(ReadReply());

		char *s = strchr(Buffer, ' ');
		if (s)
		{
			s = strchr(s+1, ' ');
			if (s)
			{
				Size = atoi(s);
			}
		}
	}

	return Size;
}

bool MailPop3::Delete(int Message)
{
	if (Socket)
	{
		sprintf(Buffer, "DELE %i\r\n", Message + 1);
		VERIFY_RET_VAL(Write(0, true));
		VERIFY_RET_VAL(ReadReply());

		return true;
	}

	return false;
}

bool MailPop3::GetUid(int Index, char *Id)
{
	if (Socket AND Id)
	{
		sprintf(Buffer, "UIDL %i\r\n", Index + 1);
		VERIFY_RET_VAL(Write(0, true));
		VERIFY_RET_VAL(ReadReply());

		char *Space = strchr(Buffer, ' ');
		if (Space)
		{
			Space = strchr(Space+1, ' ');
			if (Space)
			{
				for (char *s = Space+1; *s; s++)
				{
					if (*s == '\r' OR *s == '\n')
					{
						*s = 0;
						break;
					}
				}
				
				strcpy(Id, Space+1);
				return true;
			}
		}
	}

	return false;
}

bool MailPop3::GetUidList(List<char> &Id)
{
	bool Status = false;
	if (Socket)
	{
		char *Str = 0;

		sprintf(Buffer, "UIDL\r\n");
		VERIFY_RET_VAL(Write(0, true));
		VERIFY_RET_VAL(ReadMultiLineReply(Str));
		if (Str)
		{
			Status = true;
			GToken T(Str, "\r\n");
			for (int i=0; i<T.Length(); i++)
			{
				char *s = T[i];
				if (s AND *s != '.')
				{
					char *Space = strchr(s, ' ');
					if (Space++)
					{
						Id.Insert(NewStr(Space));
					}
				}
			}

			DeleteArray(Str);
		}
	}

	return Status;
}

char *MailPop3::GetHeaders(int Message)
{
	char *Header = 0;
	if (Socket)
	{
		sprintf(Buffer, "TOP %i 0\r\n", Message + 1);
		if (Write(0, true))
		{
			ReadMultiLineReply(Header);
		}
	}

	return Header;
}

bool MailPop3::ReadMultiLineReply(char *&Str)
{
	bool Status = false;
	if (Socket)
	{
		GBytePipe Temp;
		int ReadLen = Socket->Read(Buffer, sizeof(Buffer), 0);
		if (ReadLen > 0 AND Buffer[0] == '+')
		{
			// positive response
			char *Eol = strchr(Buffer, '\n');
			if (Eol)
			{
				char *Ptr = Eol + 1;
				ReadLen -= ((int)Ptr-(int)Buffer);
				memmove(Buffer, Ptr, ReadLen);
				Temp.Write((uchar*) Buffer, ReadLen);

				while (	NOT MailIsEnd(Buffer, ReadLen) AND
						NOT Socket->Cancel())
				{
					ReadLen = Socket->Read(Buffer, sizeof(Buffer), 0);
					if (ReadLen > 0)
					{
						Temp.Write((uchar*) Buffer, ReadLen);
					}
					else break;
				}

				int Len = Temp.GetSize();
				Str = NEW(char[Len+1]);
				if (Str)
				{
					Temp.Read((uchar*)Str, Len);
					Str[Len] = 0;
					Status = true;
				}
			}
		}
	}

	return Status;
}

bool MailPop3::Close()
{
	if (Socket)
	{
		// logout
		VERIFY_RET_VAL(Write("QUIT\r\n", true));
		
		// 2 sec timeout, we don't really care about the server's response
		ReadReply(2000);

		Socket->Close();
		Socket = 0;
		Messages = 0;
		return true;
	}
	return false;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
MailMessage::MailMessage()
{
	Private = 0;
	Subject = 0;
	From = 0;
	Reply = 0;
	MessageID = 0;
	InternetHeader = 0;
	Priority = MAIL_PRIORITY_NORMAL;

	Text = 0;
	TextCharset = 0;
	Html = 0;
	HtmlCharset = 0;

	MarkColour = -1;
	DispositionNotificationTo = false;
	Raw = 0;
	References = 0;
}

MailMessage::~MailMessage()
{
	Empty();
	DeleteObj(From);
	DeleteObj(Reply);
	DeleteObj(Raw);
	DeleteObj(Private);
}

void MailMessage::Empty()
{
	DeleteArray(Subject);
	DeleteArray(MessageID);
	DeleteArray(InternetHeader);
	DeleteArray(References);

	DeleteArray(Text);
	DeleteArray(TextCharset);
	DeleteArray(Html);
	DeleteArray(HtmlCharset);

	To.DeleteObjects();
	FileDesc.DeleteObjects();
}

char *MailMessage::GetBody()
{
	return Text;
}

char *MailMessage::GetBodyCharset()
{
	return TextCharset;
}

bool MailMessage::SetBodyCharset(char *Cs, bool Copy)
{
	DeleteArray(TextCharset);
	TextCharset = Copy ? NewStr(Cs) : Cs;
	return true;
}

bool MailMessage::SetBody(char *Txt, int Bytes, bool Copy, char *Cs)
{
	if (Txt != Text)
	{
		DeleteArray(Text);
		Text = Copy ? NewStr(Txt, Bytes) : Txt;
		if (Txt AND NOT Text)
			return false;
	}
	if (Cs != TextCharset)
	{
		DeleteArray(TextCharset);
		if (NOT (TextCharset = NewStr(Cs)))
			return false;
	}
	return true;
}

char *MailMessage::GetHtml()
{
	return Html;
}

char *MailMessage::GetHtmlCharset()
{
	return HtmlCharset;
}

bool MailMessage::SetHtmlCharset(char *Cs, bool Copy)
{
	DeleteArray(HtmlCharset);
	HtmlCharset = Copy ? NewStr(Cs) : Cs;
	return true;
}

bool MailMessage::SetHtml(char *Txt, int Bytes, bool Copy, char *Cs)
{
	if (Txt != Html)
	{
		DeleteArray(Html);
		Html = Copy ? NewStr(Txt, Bytes) : Txt;
		if (Txt AND NOT Html)
			return false;
	}
	if (Cs != HtmlCharset)
	{
		DeleteArray(HtmlCharset);
		if (NOT (HtmlCharset = NewStr(Cs)))
			return false;
	}
	return true;
}

bool MailMessage::EncodeQuotedPrintable(GStream &Out, char *In)
{
	bool Status = true;
	if (In)
	{
		char Buffer[100];
		char *d = Buffer;
		for (char *s = In; Status; )
		{
			if (*s == '\n' OR NOT *s)
			{
				strcpy(d, "\r\n");
				d = Buffer;
				Status &= Out.Write(Buffer, strlen(Buffer)) > 0;
				
				if (NOT Status)
					LgiTrace("%s:%i - Status=%i\n", __FILE__, __LINE__, Status);

				if (NOT *s) break;
			}
			else if (*s & 0x80 OR
					 *s == '.' OR
					 *s == '=')
			{
				sprintf(d, "=%02.2X", (uchar)*s);
				d += 3;
			}
			else if (*s != '\r')
			{
				*d++ = *s;
			}
			s++;

			if ((int)d-(int)Buffer > 73)
			{
				// time for a new line.
				strcpy(d, "=\r\n");
				d = Buffer;
				Status &= Out.Write(Buffer, strlen(Buffer)) > 0;

				if (NOT Status)
					LgiTrace("%s:%i - Status=%i\n", __FILE__, __LINE__, Status);
			}
		}
	}

	return Status;
}

bool MailMessage::EncodeText(GStream &Out, char *In)
{
	bool Status = true;
	char *s;
	for (s = In; s AND *s AND Status; )
	{
		char *NewLine = strchr(s, '\n');
		int Len = (NewLine) ? (int)NewLine-(int)s : strlen(s);
		if (Len > 0 AND s[Len-1] == '\r') Len--;
		if (Len == 1 AND s[0] == '.')
		{
			// this removes the sequence ".\n"
			// which is the END OF MAIL in the SMTP protocol.
			Status &= Out.Write((char*)". ", 2) > 0;

			if (NOT Status)
				LgiTrace("%s:%i - Status=%i\n", __FILE__, __LINE__, Status);
		}
		else
		{
			if (Len)
			{
				int l = Out.Write(s, Len);
				Status &= Len == l;
				// printf("Push('%.*s', %i) = %i, Status=%i\n", Len, s, Len, l, Status);
			}
			
			if (NOT Status)
				LgiTrace("%s:%i - Push(%p,%i)->Status=%i\n", __FILE__, __LINE__, s, Len, Status);
		}
		Status &= Out.Write((char*)"\r\n", 2) > 0;

		if (NOT Status)
			LgiTrace("%s:%i - Status=%i\n", __FILE__, __LINE__, Status);

		s = (NewLine) ? NewLine + 1 : 0;
	}

	return Status;
}

#define SEND_BUF_SIZE		(16 << 10)

class SendBuf : public GStream
{
	GStream *Out;
	int Used;
	uchar Buf[SEND_BUF_SIZE];

public:
	bool Status;

	SendBuf(GStream *out)
	{
		Out = out;
		Used = 0;
		Status = true;
	}

	~SendBuf()
	{
		Flush();
	}

	int64 GetSize() { return Used; }
	int64 SetSize(int64 Size) { return Out->SetSize(Size); }
	int Read(void *Buffer, int Size, int Flags = 0) { return -1; }

	void Flush()
	{
		if (Used > 0)
		{
			int w = Out->Write(Buf, Used, 0);
			if (w < Used)
			{
				Status = false;
			}
			Used = 0;
		}
	}

	int Write(void *Buffer, int Size, int Flags = 0)
	{
		int64 w = 0;

		uchar *Ptr = (uchar*)Buffer;
		while (Ptr AND Size > 0)
		{
			if (Size + Used >= SEND_BUF_SIZE)
			{
				int Chunk = SEND_BUF_SIZE - Used;
				memcpy(Buf + Used, Ptr, Chunk);
				int s = Out->Write(Buf, SEND_BUF_SIZE, 0);
				if (s < SEND_BUF_SIZE)
				{
					return -1;
					break;
				}

				Ptr += Chunk;
				Size -= Chunk;
				w += Chunk;
				Used = 0;
			}
			else
			{
				memcpy(Buf + Used, Ptr, Size);
				Used += Size;
				w += Size;
				Size = 0;
			}
		}

		return w;
	}
};

// Encode the whole email
bool MailMessage::Encode(GStream &Out, GStream *HeadersSink, MailProtocol *Protocol, bool Mime)
{
	GStringPipe p;
	bool Status = EncodeHeaders(p, Protocol, Mime);
	if (Status)
	{
		int Len = p.GetSize();
		char *Headers = p.NewStr();
		if (HeadersSink)
		{
			HeadersSink->Write(Headers, Len);
		}
		else
		{
			InternetHeader = NewStr(Headers);
		}

		if (Headers AND
			Out.Write(Headers, Len))
		{
			SendBuf *Buf = NEW(SendBuf(&Out));
			if (Buf)
			{
				Status = EncodeBody(*Buf, Protocol, Mime);			
				if (Status)
				{
					Buf->Flush();
					Status = Buf->Status;
				}
				else
				{
					LgiTrace("%s:%i - EncodeBody failed.\n", __FILE__, __LINE__);
				}

				DeleteObj(Buf);
			}
		}
		else
		{
			LgiTrace("%s:%i - Headers output failed.\n", __FILE__, __LINE__);
		}
		
		DeleteArray(Headers);
	}
	else
	{
		LgiTrace("%s:%i - EncodeHeaders failed.\n", __FILE__, __LINE__);
	}

	return Status;
}

#include "GDateTime.h"

// This encodes the main headers but not the headers relating to the 
// actual content. Thats done by the ::EncodeBody function.
bool MailMessage::EncodeHeaders(GStream &Out, MailProtocol *Protocol, bool Mime)
{
	bool Status = true;

	// Setup
	char Buffer[1025];

	// Construct date
	char *Weekday[7] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
	char *Month[12] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
	GDateTime Dt;
	int TimeZone = Dt.SystemTimeZone();
	Dt.SetNow();
	sprintf(Buffer,
			"Date: %s, %i %s %i %i:%02.2i:%02.2i %s%2.2d%2.2d\r\n",
			Weekday[Dt.DayOfWeek()],
			Dt.Day(),
			Month[Dt.Month()-1],
			Dt.Year(),
			Dt.Hours(),
			Dt.Minutes(),
			Dt.Seconds(),
			(TimeZone >= 0) ? "+" : "",
			TimeZone / 60,
			abs(TimeZone) % 60);
	Status &= Out.Write(Buffer, strlen(Buffer)) > 0;

	if (Protocol AND Protocol->ProgramName)
	{
		// X-Mailer:
		sprintf(Buffer, "X-Mailer: %s\r\n", Protocol->ProgramName);
		Status &= Out.Write(Buffer, strlen(Buffer)) > 0;
	}

	if (Protocol AND Protocol->ExtraOutgoingHeaders)
	{
		for (char *s=Protocol->ExtraOutgoingHeaders; s AND *s; )
		{
			char *e = s;
			while (*e AND *e != '\r' AND *e != '\n') e++;
			int l = (int)e-(int)s;
			if (l > 0)
			{
				Status &= Out.Write(s, l) > 0;
				Status &= Out.Write((char*)"\r\n", 2) > 0;
			}
			while (*e AND (*e == '\r' OR *e == '\n')) e++;
			s = e;
		}
	}

	if (Priority != MAIL_PRIORITY_NORMAL)
	{
		// X-Priority:
		sprintf(Buffer, "X-Priority: %i\r\n", Priority);
		Status &= Out.Write(Buffer, strlen(Buffer)) > 0;
	}

	if (MarkColour >= 0)
	{
		// X-Color (HTML Colour Ref for email marking)
		sprintf(Buffer,
			"X-Color: #%02.2X%02.2X%02.2X\r\n",
			R24(MarkColour),
			G24(MarkColour),
			B24(MarkColour));
		Status &= Out.Write(Buffer, strlen(Buffer)) > 0;
	}

	// Message-ID:
	if (MessageID)
	{
		for (char *m=MessageID; *m; m++)
		{
			if (*m <= ' ')
			{
				printf("%s:%i - Bad message ID '%s'\n", __FILE__, __LINE__, MessageID);
				return false;
			}			
		}
		
		sprintf(Buffer, "Message-ID: %s\r\n", MessageID);
		Status &= Out.Write(Buffer, strlen(Buffer)) > 0;
	}

	// References:
	if (ValidStr(References))
	{
		sprintf(Buffer, "References: %s\r\n", References);
		Status &= Out.Write(Buffer, strlen(Buffer)) > 0;
	}

	// To:
	char *ToAddr = CreateAddressTag(To, 0, &Protocol->CharsetPrefs);
	if (ToAddr)
	{
		Status &= Out.Write(ToAddr, strlen(ToAddr)) > 0;
		DeleteArray(ToAddr);
	}

	char *CcAddr = CreateAddressTag(To, 1, &Protocol->CharsetPrefs);
	if (CcAddr)
	{
		Status &= Out.Write(CcAddr, strlen(CcAddr)) > 0;
		DeleteArray(CcAddr);
	}

	// From:
	if (From AND From->Addr)
	{
		strcpy(Buffer, "From: ");

		char *Nme = EncodeRfc2047(NewStr(From->Name), 0, &Protocol->CharsetPrefs);
		if (Nme)
		{
			if (strchr(Nme, '\"'))
			{
				sprintf(Buffer + strlen(Buffer), "'%s' ", Nme);
			}
			else
			{
				sprintf(Buffer + strlen(Buffer), "\"%s\" ", Nme);
			}
			DeleteArray(Nme);
		}

		sprintf(Buffer + strlen(Buffer), "<%s>\r\n", From->Addr);
	}
	else
	{
		return false;
	}
	Status &= Out.Write(Buffer, strlen(Buffer)) > 0;

	// Reply-To:
	if (Reply AND
		ValidStr(Reply->Addr))
	{
		strcpy(Buffer, "Reply-To: ");

		char *Nme = EncodeRfc2047(NewStr(Reply->Name), 0, &Protocol->CharsetPrefs);
		if (Nme)
		{
			if (strchr(Nme, '\"'))
			{
				sprintf(Buffer + strlen(Buffer), "'%s' ", Nme);
			}
			else
			{
				sprintf(Buffer + strlen(Buffer), "\"%s\" ", Nme);
			}
			DeleteArray(Nme);
		}

		sprintf(Buffer+strlen(Buffer), "<%s>\r\n", Reply->Addr);

		Status &= Out.Write(Buffer, strlen(Buffer)) > 0;
	}

	// Subject:
	char *Subj = EncodeRfc2047(NewStr(Subject), 0, &Protocol->CharsetPrefs);
	sprintf(Buffer, "Subject: %s\r\n", (Subj) ? Subj : "");
	Status &= Out.Write(Buffer, strlen(Buffer)) > 0;
	DeleteArray(Subj);

	// DispositionNotificationTo
	if (DispositionNotificationTo)
	{
		strcpy(Buffer, "Disposition-Notification-To:");
		char *Nme = EncodeRfc2047(NewStr(From->Name), 0, &Protocol->CharsetPrefs);
		if (Nme)
		{
			sprintf(Buffer + strlen(Buffer), " \"%s\"", Nme);
			DeleteArray(Nme);
		}
		sprintf(Buffer + strlen(Buffer), " <%s>\r\n", From->Addr);
		Status &= Out.Write(Buffer, strlen(Buffer)) > 0;
	}

	return Status;
}

bool MailMessage::EncodeBody(GStream &Out, MailProtocol *Protocol, bool Mime)
{
	bool Status = true;
	char Buffer[1025];

	// LgiTrace("%s:%i - Mime=%i\n", __FILE__, __LINE__, Mime);
	if (Mime)
	{
		bool MultiPart = ((Text ? 1 : 0) + (Html ? 1 : 0) + FileDesc.Length()) > 1;
		bool MultipartAlternate = (Text != 0) AND (Html != 0);
		bool MultipartMixed = FileDesc.Length() > 0;
		int Now = LgiCurrentTime();

		char Separator[256];
		sprintf(Separator, "----=_NextPart_%08.8X.%08.8X", Now, LgiGetCurrentThread());

		sprintf(Buffer, "MIME-Version: 1.0\r\n");
		Status &= Out.Write(Buffer, strlen(Buffer)) > 0;
		
		char StrMultipartAlternative[] = "multipart/alternative";
		// LgiTrace("%s:%i - MultiPart=%i Alt=%i Mix=%i\n", __FILE__, __LINE__, MultiPart, MultipartAlternate, MultipartMixed);
		if (MultiPart)
		{
			char *Type = MultipartMixed ? (char*)"multipart/mixed" : StrMultipartAlternative;
			sprintf(Buffer, "Content-Type: %s;\r\n\tBoundary=\"%s\"\r\n", Type, Separator);
			Status &= Out.Write(Buffer, strlen(Buffer)) > 0;
			// Status &= Out.Write((char*)"Status: RO\r\n\r\n", 14) > 0;
		}

		// LgiTrace("%s:%i - Status=%i\n", __FILE__, __LINE__, Status);
		
		if (Text != 0 OR Html != 0)
		{
			char AlternateBoundry[128] = "";

			if (MultiPart)
			{
				sprintf(Buffer, "\r\n--%s\r\n", Separator);
				Status &= Out.Write(Buffer, strlen(Buffer)) > 0;

				if (MultipartMixed AND MultipartAlternate)
				{
					sprintf(AlternateBoundry, "----=_NextPart_%08.8X.%08.8X", Now++, LgiGetCurrentThread());
					sprintf(Buffer,
							"Content-Type: %s;\r\n\tBoundary=\"%s\"\r\n",
							StrMultipartAlternative,
							AlternateBoundry);
					Status &= Out.Write(Buffer, strlen(Buffer)) > 0;

					sprintf(Buffer, "\r\n--%s\r\n", AlternateBoundry);
					Status &= Out.Write(Buffer, strlen(Buffer)) > 0;
				}
			}

			if (NOT Status)
				LgiTrace("%s:%i - Status=%i\n", __FILE__, __LINE__, Status);

			if (Text)
			{
				char *Cs = 0;
				char *Txt = Text, *Mem = 0;
				
				// Detect charset
				if (NOT TextCharset OR stricmp(TextCharset, "utf-8") == 0)
				{
					Cs = LgiDetectCharset(Text, -1, Protocol ? &Protocol->CharsetPrefs : 0);
					if (Cs)
					{
						Mem = Txt = (char*)LgiNewConvertCp(Cs, Text, "utf-8", -1);
					}
				}
				if (NOT Cs) Cs = TextCharset;				
				
				// Content type
				sprintf(Buffer,
						"Content-Type: text/plain; charset=\"%s\"\r\n",
						Cs ? Cs : "us-ascii");
				Status &= Out.Write(Buffer, strlen(Buffer)) > 0;

				if (NOT Status)
					LgiTrace("%s:%i - Status=%i\n", __FILE__, __LINE__, Status);

				// Transfer encoding
				if (Is8Bit(Txt) OR
					MaxLineLen(Txt) >= 80)
				{
					char QuotPrint[] = "Content-Transfer-Encoding: quoted-printable\r\n\r\n";
					Status &= Out.Write(QuotPrint, strlen(QuotPrint)) > 0;

					if (NOT Status)
						LgiTrace("%s:%i - Status=%i\n", __FILE__, __LINE__, Status);

					Status &= EncodeQuotedPrintable(Out, Txt);
					
					if (NOT Status)
						LgiTrace("%s:%i - Status=%i\n", __FILE__, __LINE__, Status);
				}
				else
				{
					char Cte[] = "Content-Transfer-Encoding: 7bit\r\n\r\n";
					Status &= Out.Write(Cte, strlen(Cte)) > 0;

					if (NOT Status)
						LgiTrace("%s:%i - Status=%i\n", __FILE__, __LINE__, Status);

					Status &= EncodeText(Out, Txt);

					if (NOT Status)
						LgiTrace("%s:%i - Status=%i\n", __FILE__, __LINE__, Status);
				}
				
				DeleteArray(Mem);
			}

			if (NOT Status)
				LgiTrace("%s:%i - Status=%i\n", __FILE__, __LINE__, Status);

			// Break alternate part
			if (AlternateBoundry[0])
			{
				sprintf(Buffer, "\r\n--%s\r\n", AlternateBoundry);
				Status &= Out.Write(Buffer, strlen(Buffer)) > 0;
			}
			else if (MultipartAlternate)
			{
				sprintf(Buffer, "\r\n--%s\r\n", Separator);
				Status &= Out.Write(Buffer, strlen(Buffer)) > 0;
			}

			if (Html)
			{
				// Content type
				sprintf(Buffer,
						"Content-Type: text/html; charset=\"%s\"\r\n",
						TextCharset ? TextCharset : "us-ascii");
				Status &= Out.Write(Buffer, strlen(Buffer)) > 0;

				// Transfer encoding
				if (Is8Bit(Html) OR
					MaxLineLen(Html) >= 80)
				{
					char Qp[] = "Content-Transfer-Encoding: quoted-printable\r\n\r\n";
					Status &= Out.Write(Qp, strlen(Qp)) > 0;
					Status &= EncodeQuotedPrintable(Out, Html);
				}
				else
				{
					char Sb[] = "Content-Transfer-Encoding: 7bit\r\n\r\n";
					Status &= Out.Write(Sb, strlen(Sb)) > 0;
					Status &= EncodeText(Out, Html);
				}
			}

			if (NOT Status)
				LgiTrace("%s:%i - Status=%i\n", __FILE__, __LINE__, Status);

			if (AlternateBoundry[0])
			{
				// End alternate part
				sprintf(Buffer, "\r\n--%s--\r\n", AlternateBoundry);
				Status &= Out.Write(Buffer, strlen(Buffer)) > 0;
			}
		}

		if (NOT Status)
			LgiTrace("%s:%i - Status=%i\n", __FILE__, __LINE__, Status);

		int SizeEst = 1024;
		FileDescriptor *FileDes = FileDesc.First();
		for (; FileDes; FileDes=FileDesc.Next())
		{
			SizeEst += FileDes->Sizeof() * 4 / 3;
		}
		Out.SetSize(SizeEst);

		FileDes = FileDesc.First();
		while (FileDes)
		{
			GFile *F = FileDes->GotoObject();
			int FilePos = 0;
			if (F)
			{
				FilePos = F->GetPos();
			}

			// write a MIME segment for this attachment
			if (MultiPart)
			{
				sprintf(Buffer, "\r\n--%s\r\n", Separator);
				Status &= Out.Write(Buffer, strlen(Buffer)) > 0;
			}

			char FileName[256];
			char *s = FileDes->Name(), *d = FileName;
			if (s)
			{
				while (*s)
				{
					if (*s != '\\')
					{
						*d++ = *s++;
					}
					else
					{
						*d++ = '\\';
						*d++ = '\\';
						s++;
					}
				}
				*d = 0;

				char *FName = EncodeRfc2047(NewStr(FileName), 0, &Protocol->CharsetPrefs);
				sprintf(Buffer,
						"Content-Type: %s; name=\"%s\"\r\n"
						"Content-Disposition: attachment\r\n",
						FileDes->GetMimeType() ? FileDes->GetMimeType() : "application/x-zip-compressed",
						(FName) ? FName : FileName);
				Status &= Out.Write(Buffer, strlen(Buffer)) > 0;
				DeleteArray(FName);

				sprintf(Buffer, "Content-Transfer-Encoding: base64\r\n\r\n");
				Status &= Out.Write(Buffer, strlen(Buffer)) > 0;

				if (NOT Status)
					LgiTrace("%s:%i - Status=%i\n", __FILE__, __LINE__, Status);

				int64 Start = LgiCurrentTime();
				int BufSize = 4 << 10;
				char *InBuf = NEW(char[BufSize]);
				char *OutBuf = NEW(char[BufSize]);
				if (InBuf AND
					OutBuf)
				{
					int InLen = FileDes->Sizeof(); // file remaining to read
					int InUsed = 0;
					int InDone = 0;
					int OutUsed = 0;

					do
					{
						if (InUsed - InDone < 256 AND InLen > 0)
						{
							// Move any bit left over down to the start
							memmove(InBuf, InBuf + InDone, InUsed - InDone);
							InUsed -= InDone;
							InDone = 0;

							// Read in as much data as we can
							int Rmin = min(BufSize-InUsed, InLen);
							int R = -1;

							if (F)
							{
								R = F->Read(InBuf + InUsed, Rmin);
							}
							else if (FileDes->GetData())
							{
								uchar *s = FileDes->GetData();
								memcpy(InBuf + InUsed, s + FilePos, Rmin);
								R = Rmin;
							}

							if (R > 0)
							{
								FilePos += R;
								InUsed += R;
								InLen -= R;
							}
							else break;
						}

						if (OutUsed > BufSize - 256)
						{
							if (Out.Write(OutBuf, OutUsed))
							{
								OutUsed = 0;
							}
							else
							{
								break;
							}
						}

						int OutLen = ConvertBinaryToBase64(	OutBuf + OutUsed,
															76,
															(uchar*)InBuf + InDone,
															InUsed - InDone);
						int In = OutLen * 3 / 4;
						InDone += In;
						OutUsed += OutLen;
						OutBuf[OutUsed++] = '\r';
						OutBuf[OutUsed++] = '\n';
					}
					while (InDone < InUsed);

					if (OutUsed > 0)
					{
						Out.Write(OutBuf, OutUsed);
					}

					#if 0
					double Sec = (double)((int64)LgiCurrentTime() - Start) / 1000.0;
					double Kb = (double)FileDes->Sizeof() / 1024.0;
					LgiTrace("rate: %ikb/s\n", (int)(Kb / Sec));
					#endif
				}
				else
				{
					LgiTrace("%s:%i - Error\n", __FILE__, __LINE__);
				}

				DeleteArray(InBuf);
				DeleteArray(OutBuf);
			}
			else
			{
				LgiTrace("%s:%i - Error\n", __FILE__, __LINE__);
			}

			FileDes = FileDesc.Next();
		}

		if (NOT Status)
			LgiTrace("%s:%i - Status=%i\n", __FILE__, __LINE__, Status);

		if (MultiPart)
		{
			// write final separator
			sprintf(Buffer, "\r\n--%s--\r\n\r\n", Separator);
			Status &= Out.Write(Buffer, strlen(Buffer)) > 0;
		}
	}
	else
	{
		// send content type
		sprintf(Buffer, "Content-Type: text/plain; charset=\"%s\"\r\n", TextCharset ? TextCharset : "us-ascii");
		Status &= Out.Write(Buffer, strlen(Buffer)) > 0;

		if (Is8Bit(Text))
		{
			// send the encoding and a blank line
			char Qp[] = "Content-Transfer-Encoding: quoted-printable\r\n\r\n";
			Status &= Out.Write(Qp, strlen(Qp)) > 0;
			
			// send message text
			Status &= EncodeQuotedPrintable(Out, Text);
		}
		else
		{
			// send a blank line
			Status &= Out.Write((char*)"\r\n", 2) > 0;
			
			// send message text
			Status &= EncodeText(Out, Text);
		}
	}

	if (NOT Status)
		LgiTrace("%s:%i - Status=%i\n", __FILE__, __LINE__, Status);

	return true;
}

////////////////////////////////////////////////////////////////////////////
#ifndef WIN32
#include "AuthNtlm/Ntlm.h"
#else
#include "AuthNtlm\ntlm.h"
#endif

char *DecodeImapString(char *s)
{
	GStringPipe p;
	while (s AND *s)
	{
		if (*s == '&')
		{
			char Escape = *s++;
			
			char *e = s;
			while (*e AND *e != '-')
			{
				e++;
			}

			int Len = e - s;
			if (Len)
			{
				char *Base64 = NEW(char[Len + 4]);
				if (Base64)
				{
					memcpy(Base64, s, Len);
					char *n = Base64 + Len;
					for (int i=Len; i%4; i++)
						*n++ = '=';
					*n++ = 0;

					Len = strlen(Base64);
					int BinLen = BufferLen_64ToBin(Len);
					uint16 *Bin = NEW(uint16[(BinLen/2)+1]);
					if (Bin)
					{
						BinLen = ConvertBase64ToBinary((uchar*)Bin, BinLen, Base64, Len);
						if (BinLen)
						{
							int Chars = BinLen / 2;
							Bin[Chars] = 0;
							for (int i=0; i<Chars; i++)
							{
								Bin[i] = (Bin[i]>>8) | ((Bin[i]&0xff)<<8);
							}

							char *c8 = LgiNewUtf16To8((char16*)Bin, BinLen);
							if (c8)
							{
								p.Push(c8);
								DeleteArray(c8);
							}
						}
						DeleteArray(Bin);
					}
					
					DeleteArray(Base64);
				}
			}
			else
			{
				p.Push(&Escape, 1);
			}

			s = e + 1;
		}
		else
		{
			p.Push(s, 1);
			s++;
		}
	}
	return p.NewStr();
}

char *EncodeImapString(char *s)
{
	GStringPipe p;
	int Len = s ? strlen(s) : 0;
	
	while (s AND *s)
	{
		int c = LgiUtf8To32((uint8*&)s, Len);

		DoNextChar:
		if ((c >= ' ' AND c < 0x80) OR
			c == '\n' OR
			c == '\t' OR
			c == '\r')
		{
			// Literal
			char ch = c;
			p.Push(&ch, 1);
		}
		else
		{
			// Encoded
			GArray<uint16> Str;
			Str[0] = c;
			while (c = LgiUtf8To32((uint8*&)s, Len))
			{
				if ((c >= ' ' AND c < 0x80) OR
					c == '\n' OR
					c == '\t' OR
					c == '\r')
				{
					break;
				}
				else
				{
					Str[Str.Length()] = c;
				}
			}

			for (int i=0; i<Str.Length(); i++)
			{
				Str[i] = (Str[i]>>8) | ((Str[i]&0xff)<<8);
			}
			int BinLen = Str.Length() << 1;
			int BaseLen = BufferLen_BinTo64(BinLen);
			char *Base64 = NEW(char[BaseLen+1]);
			if (Base64)
			{
				int Bytes = ConvertBinaryToBase64(Base64, BaseLen, (uchar*)&Str[0], BinLen);
				while (Bytes > 0 AND Base64[Bytes-1] == '=')
				{
					Base64[Bytes-1] = 0;
					Bytes--;
				}
				Base64[Bytes] = 0;

				p.Print("&%s-", Base64);
				DeleteArray(Base64);
			}

			goto DoNextChar;
		}
	}

	return p.NewStr();
}

void ChopNewLine(char *Str)
{
	char *End = Str+strlen(Str)-1;
	if (*End == '\n')
	{
		*End-- = 0;
	}
	if (*End == '\r')
	{
		*End-- = 0;
	}
}

MailImapFolder::MailImapFolder()
{
	Sep = '/';
	Path = 0;
	NoSelect = false;
	NoInferiors = false;
	Marked = false;
	Exists = -1;
	Recent = -1;
	UnseenIndex = -1;
	Deleted = 0;
}

MailImapFolder::~MailImapFolder()
{
	DeleteArray(Path);
}

char *MailImapFolder::GetPath()
{
	return Path;
}

void MailImapFolder::SetPath(char *s)
{
	char *NewPath = DecodeImapString(s);
	DeleteArray(Path);
	Path = NewPath;
}

char *MailImapFolder::GetName()
{
	if (Path)
	{
		char *s = strrchr(Path, Sep);
		if (s)
		{
			return s + 1;
		}
		else
		{
			return Path;
		}
	}

	return 0;
}

void MailImapFolder::SetName(char *s)
{
	if (s)
	{
		char Buf[256];
		strcpy(Buf, Path?Path:(char*)"");
		DeleteArray(Path);

		char *Last = strrchr(Buf, Sep);
		if (Last)
		{
			strcpy(Last+1, s);
			Path = NewStr(Buf);
		}
		else
		{
			Path = NewStr(s);
		}
	}
}

/////////////////////////////////////////////
class MailIMapPrivate
{
public:
	int NextCmd;
	bool Logging;
	bool ExpungeOnExit;
	char FolderSep;
	char *Current;
	char *Flags;
	bool Mercury_v332;
	GHashTable Capability;
	GStringPipe Untagged;

	MailIMapPrivate()
	{
		FolderSep = '/';
		NextCmd = 1;
		Logging = true;
		ExpungeOnExit = true;
		Current = 0;
		Flags = 0;
		Mercury_v332 = false;
	}

	~MailIMapPrivate()
	{
		DeleteArray(Current);
		DeleteArray(Flags);
	}
};

MailIMap::MailIMap()
{
	d = NEW(MailIMapPrivate);
	Buffer[0] = 0;
}

MailIMap::~MailIMap()
{
	if (Lock())
	{
		ClearDialog();
		ClearUid();
		DeleteObj(d);
	}
}

char MailIMap::GetFolderSep()
{
	return d->FolderSep;
}

char *MailIMap::GetCurrentPath()
{
	return d->Current;
}

bool MailIMap::GetExpungeOnExit()
{
	return d->ExpungeOnExit;
}

void MailIMap::SetExpungeOnExit(bool b)
{
	d->ExpungeOnExit = b; 
}

void MailIMap::ClearUid()
{
	if (Lock())
	{
		Uid.DeleteArrays();
		Unlock();
	}
}

void MailIMap::ClearDialog()
{
	if (Lock())
	{
		Dialog.DeleteArrays();
		Unlock();
	}
}

bool MailIMap::WriteBuf(bool ObsurePass)
{
	if (S)
	{
		if (S->Write(Buf, strlen(Buf), 0) == strlen(Buf))
		{
			char *e = Buf + strlen(Buf) - 1;
			while (e > Buf AND strchr("\r\n", *e))
			{
				*e-- = 0;
			}

			if (ObsurePass)
			{
				char *Sp = strrchr(Buf, ' ');
				if (Sp)
				{
					strcpy(Sp + 1, "********\r\n");
				}
			}
			Log(Buf, MAIL_SEND_COLOUR);

			return true;
		}
	}

	return false;
}

bool MailIMap::ReadResponse(int Cmd, GStringPipe *Out)
{
	bool Done = false;
	bool Status = false;
	if (S)
	{
		GStringPipe p;
		while (NOT Done)
		{
			int r = S->Read(Buffer, sizeof(Buffer));
			if (r > 0)
			{
				p.Push(Buffer, r);
				
				while (p.Pop(Buffer, sizeof(Buffer)))
				{
					// Trim trailing whitespace
					char *e = Buffer + strlen(Buffer) - 1;
					while (e > Buffer AND strchr(" \t\r\n", *e))
						*e-- = 0;

					COLOUR c = MAIL_RECEIVE_COLOUR;

					if (Out)
					{
						Out->Push(Buffer);
						Out->Push("\r\n");
					}
					else
					{
						Dialog.Insert(NewStr(Buffer));
					}

					if (Cmd < 0)
					{
						Status = Done = true;
					}

					char Num[8];
					sprintf(Num, "A%04.4i ", Cmd);
					if (strnicmp(Buffer, Num, 6) == 0)
					{
						Done = true;
						Status = strnicmp(Buffer+6, "OK", 2) == 0;
						if (NOT Status)
						{
							DeleteArray(ServerMsg);
							ServerMsg = NewStr(Buffer+6);
							c = MAIL_ERROR_COLOUR;
						}
					}

					if (d->Logging)
					{
						Log(Buffer, c);
					}
				}
			}
			else break;
		}
	}

	return Status;
}

char *Tok(char *&s)
{
	char *Ret = 0;

	if (*s == '=' OR *s == ',')
	{
		Ret = NewStr(s++, 1);
	}
	else if (*s == '\'' OR *s == '\"')
	{
		char d = *s++;
		char *e = strchr(s, d);
		if (e)
		{
			Ret = NewStr(s, e - s);
			s = e + 1;
		}
	}
	else
	{
		char *e;
		for (e=s; *e AND (isdigit(*e) OR isalpha(*e) OR *e == '-'); e++);
		Ret = NewStr(s, e - s);
		s = e;
	}

	return Ret;
}

void Hex(char *Out, uchar *In, int Len = -1)
{
	if (Out AND In)
	{
		if (Len < 0)
			Len = strlen((char*)In);

		for (int i=0; i<Len; i++)
		{
			sprintf(Out, "%02.2x", *In++);
			Out += 2;
		}
	}
}

#include "GMap.h"

bool MailIMap::Open(GSocketI *s, char *RemoteHost, int Port, char *User, char *Password, char *&Cookie, int Flags)
{
	bool Status = false;

	S = s;
	if (S AND
		ValidStr(RemoteHost) AND
		ValidStr(User) AND
		ValidStr(Password) AND
		Lock())
	{
		// prepare address
		if (Port < 1) Port = IMAP_PORT;

		char Remote[256];
		strcpy(Remote, RemoteHost);
		char *Colon = strchr(Remote, ':');
		if (Colon)
		{
			*Colon++ = 0;
			Port = atoi(Colon);
		}

		// Don't start in SSL mode
		S->SetParameter(GSocket_Protocol, (int)"");

		// connect
		if (S->Open(Remote, Port))
		{
			bool IMAP4Server = false;
			List<char> Auths;

			// check capability
			int CapCmd = d->NextCmd++;
			sprintf(Buf, "A%04.4i CAPABILITY\r\n", CapCmd);
			if (WriteBuf())
			{
				if (ReadResponse(CapCmd))
				{
					for (char *r=Dialog.First(); r; r=Dialog.Next())
					{
						if (strstr(r, "Mercury/32 v3.32"))
						{
							d->Mercury_v332 = true;
						}

						GToken T(r, " ");
						if (T.Length() > 1 AND
							stricmp(T[1], "CAPABILITY") == 0)
						{
							for (int i=2; i<T.Length(); i++)
							{
								if (stricmp(T[i], "IMAP4") == 0 OR
									stricmp(T[i], "IMAP4rev1") == 0)
								{
									IMAP4Server = true;
								}
								if (strnicmp(T[i], "AUTH=", 5) == 0)
								{
									char *Type = T[i] + 5;
									Auths.Insert(NewStr(Type));
								}

								d->Capability.Add(T[i]);
							}
						}
					}

					ClearDialog();
				}
			}

			if (IMAP4Server)
			{
				char *DefaultAuthType = 0;
				if (TestFlag(Flags, MAIL_USE_PLAIN) OR Auths.Length() == 0)
				{
					Auths.Insert(DefaultAuthType = NewStr("PLAIN"));
				}
				if (TestFlag(Flags, MAIL_USE_LOGIN))
				{
					Auths.Insert(DefaultAuthType = NewStr("LOGIN"));
				}
				if (TestFlag(Flags, MAIL_USE_NTLM))
				{
					Auths.Insert(DefaultAuthType = NewStr("NTLM"));
				}

				// SSL
				bool TlsError = false;
				if (TestFlag(Flags, MAIL_SOURCE_STARTTLS))
				{
					int CapCmd = d->NextCmd++;
					sprintf(Buf, "A%04.4i STARTTLS\r\n", CapCmd);
					if (WriteBuf())
					{
						if (ReadResponse(CapCmd))
						{
							TlsError = NOT S->SetParameter(GSocket_Protocol, (int)"SSL");
						}
						else
						{
							TlsError = true;
						}
					}
				}

				// login
				bool LoggedIn = false;
				char AuthTypeStr[256] = "";
				char *AuthType = 0;
				for (AuthType=Auths.First(); NOT TlsError AND AuthType AND NOT LoggedIn; AuthType=Auths.Next())
				{
					if (DefaultAuthType != AuthType)
					{
						// put a string of auth types together
						if (strlen(AuthTypeStr) > 0)
						{
							strcat(AuthTypeStr, ", ");
						}
						strcat(AuthTypeStr, AuthType);
					}

					// do auth
					if (stricmp(AuthType, "NTLM") == 0)
					{
						// NT Lan Man authentication
						
						// Username is in the format: User[@Domain]
						char UserDom[256];
						strcpy(UserDom, User);
						char *Domain = strchr(UserDom, '@');
						if (Domain)
						{
							*Domain++ = 0;
						}
						else
						{
							Domain = UserDom + strlen(UserDom);
						}

						int AuthCmd = d->NextCmd++;
						sprintf(Buf, "A%04.4i AUTHENTICATE NTLM\r\n", AuthCmd);
						if (WriteBuf())
						{
							if (ReadResponse(AuthCmd))
							{
								tSmbNtlmAuthRequest   request;              
								tSmbNtlmAuthChallenge challenge;
								tSmbNtlmAuthResponse  response;
								char tmpstr[32];

								buildSmbNtlmAuthRequest(&request,UserDom,Domain);
								ZeroObj(Buf);
								ConvertBinaryToBase64(Buf, sizeof(Buf), (uchar*) &request, SmbLength(&request));
								strcat(Buf, "\r\n");
								WriteBuf();

								/* read challange data from server, convert from base64 */

								Buf[0] = 0;
								ClearDialog();
								if (ReadResponse())
								{
									/* buffer should contain the string "+ [base 64 data]" */

									char *Line = Dialog.First();
									LgiAssert(Line);
									ChopNewLine(Line);
									ConvertBase64ToBinary((uchar*) &challenge, sizeof(challenge), Line+2, strlen(Line)-2);

									/* prepare response, convert to base64, send to server */

									buildSmbNtlmAuthResponse(&challenge, &response, User, Password);
									ZeroObj(Buf);
									ConvertBinaryToBase64(Buf, sizeof(Buf), (uchar*) &response, SmbLength(&response));
									strcat(Buf, "\r\n");
									WriteBuf();

									/* read line from server, it should be "[seq] OK blah blah blah" */

									LoggedIn = ReadResponse(AuthCmd);
								}
							}
						}

						ClearDialog();
					}
					/*
					else if (stricmp(AuthType, "OTP") == 0)
					{
						// One time password
						int AuthCmd = d->NextCmd++;
						int UserLen = strlen(User) + 1;
						sprintf(Buf, "A%04.4i AUTHENTICATE OTP\r\n", AuthCmd);
						if (WriteBuf(true))
						{
							if (ReadResponse(AuthCmd))
							{
								int asd=0;
							}
						}
					}
					*/
					else if (stricmp(AuthType, "LOGIN") == 0 OR
							 stricmp(AuthType, "OTP") == 0)
					{
						// clear text authentication
						int AuthCmd = d->NextCmd++;
						sprintf(Buf, "A%04.4i LOGIN %s %s\r\n", AuthCmd, User, Password);
						if (WriteBuf(true))
						{
							LoggedIn = ReadResponse(AuthCmd);
						}
					}
					else if (stricmp(AuthType, "PLAIN") == 0)
					{
						// plain auth type
						char s[256];
						char *e = s;
						*e++ = 0;
						strcpy(e, User);
						e += strlen(e);
						e++;
						strcpy(e, Password);
						e += strlen(e);
						*e++ = '\r';
						*e++ = '\n';
						int Len = e - s - 2;					

						int AuthCmd = d->NextCmd++;
						sprintf(Buf, "A%04.4i AUTHENTICATE PLAIN\r\n", AuthCmd);
						if (WriteBuf())
						{
							if (ReadResponse(AuthCmd))
							{
								int b = ConvertBinaryToBase64(Buf, sizeof(Buf), (uchar*)s, Len);
								strcpy(Buf + b, "\r\n");

								if (WriteBuf())
								{
									LoggedIn = ReadResponse(AuthCmd);
								}
							}
						}						
					}
					else if (stricmp(AuthType, "DIGEST-MD5") == 0)
					{
						int AuthCmd = d->NextCmd++;
						sprintf(Buf, "A%04.4i AUTHENTICATE DIGEST-MD5\r\n", AuthCmd);
						if (WriteBuf())
						{
							if (ReadResponse(AuthCmd))
							{
								char *TestCnonce = 0;

								#if 0
								// Test case
								strcpy(Buf, "+ cmVhbG09ImVsd29vZC5pbm5vc29mdC5jb20iLG5vbmNlPSJPQTZNRzl0RVFHbTJoaCIscW9wPSJhdXRoIixhbGdvcml0aG09bWQ1LXNlc3MsY2hhcnNldD11dGYtOA==");
								RemoteHost = "elwood.innosoft.com";
								User = "chris";
								Password = "secret";
								TestCnonce = "OA6MHXh6VqTrRk";
								#endif

								char *In = (char*)Buf;
								if (In[0] == '+' AND In[1] == ' ')
								{
									In += 2;

									uchar Out[2048];
									int b = ConvertBase64ToBinary(Out, sizeof(Out), In, strlen(In));
									Out[b] = 0;
									
									GMap<char*, char*> Map;
									char *s = (char*)Out;
									while (s AND *s)
									{
										char *Var = Tok(s);
										char *Eq = Tok(s);
										char *Val = Tok(s);
										char *Comma = Tok(s);

										if (Var AND Eq AND Val AND strcmp(Eq, "=") == 0)
										{
											Map[Var] = Val;
										}

										DeleteArray(Var);
										DeleteArray(Eq);
										DeleteArray(Val);
										DeleteArray(Comma);
									}

									int32 CnonceI[2] = { LgiRand(), LgiRand() };
									char Cnonce[32];
									if (TestCnonce)
										strcpy(Cnonce, TestCnonce);
									else
										Cnonce[ConvertBinaryToBase64(Cnonce, sizeof(Cnonce), (uchar*)&CnonceI, sizeof(CnonceI))] = 0;
									s = strchr(Cnonce, '=');
									if (s) *s = 0;

									int Nc = 1;
									char *Realm = Map["realm"];
									char DigestUri[256];
									if (Realm)
									{
										sprintf(DigestUri, "imap/%s", Realm);
									}
									else
									{
										sprintf(DigestUri, "imap/%s", RemoteHost);
									}

									GStringPipe p;
									p.Print("username=\"%s\"", User);
									p.Print(",nc=%08.8i", Nc);
									p.Print(",digest-uri=\"%s\"", DigestUri);
									p.Print(",cnonce=\"%s\"", Cnonce);
									char *Nonce = Map["nonce"];
									if (Nonce)
									{
										p.Print(",nonce=\"%s\"", Nonce);
									}
									if (Realm)
									{
										p.Print(",realm=\"%s\"", Realm);
									}
									char *Charset = Map["charset"];
									if (Charset)
									{
										p.Print(",charset=%s", Charset);
									}
									char *Qop = Map["qop"];
									if (Qop)
									{
										p.Print(",qop=%s", Qop);
									}

									// Calculate A1
									char a1[256];
									uchar md5[16];
									sprintf(Buf, "%s:%s:%s", User, Realm ? Realm : (char*)"", Password);
									MDStringToDigest((uchar*)a1, Buf);
									char *Authzid = Map["authzid"];
									char *a1end = a1 + 16;
									if (Authzid)
									{										
										a1end += sprintf(a1end, ":%s:%s:%s", Nonce, Cnonce, Authzid);
									}
									else
									{
										a1end += sprintf(a1end, ":%s:%s", Nonce, Cnonce);
									}
									MDStringToDigest(md5, a1, a1end - a1);
									char a1hex[256];
									Hex(a1hex, (uchar*)md5, sizeof(md5));

									// Calculate 
									char a2[256];
									if (Qop AND (stricmp(Qop, "auth-int") == 0 OR stricmp(Qop, "auth-conf") == 0))
									{
										sprintf(a2, "AUTHENTICATE:%s:00000000000000000000000000000000", DigestUri);
									}
									else
									{
										sprintf(a2, "AUTHENTICATE:%s", DigestUri);
									}
									MDStringToDigest(md5, a2);
									char a2hex[256];
									Hex(a2hex, (uchar*)md5, sizeof(md5));

									// Calculate the final response
									sprintf(Buf, "%s:%s:%08.8i:%s:%s:%s", a1hex, Nonce, Nc, Cnonce, Qop, a2hex);
									MDStringToDigest(md5, Buf);
									Hex(Buf, (uchar*)md5, sizeof(md5));
									p.Print(",response=%s", Buf);
									if (s = p.NewStr())
									{
										int Chars = ConvertBinaryToBase64(Buf, sizeof(Buf) - 4, (uchar*)s, strlen(s));
										LgiAssert(Chars < sizeof(Buf));
										strcpy(Buf + Chars, "\r\n");
										
										if (WriteBuf())
										{
											if (ReadLine(Buf, sizeof(Buf)))
											{
												if (Buf[0] == '+' AND Buf[1] == ' ')
												{
													Log(Buf, MAIL_RECEIVE_COLOUR);
													strcpy(Buf, "\r\n");
													if (WriteBuf())
													{
														LoggedIn = ReadResponse(AuthCmd);
													}
												}
												else
												{
													Log(Buf, MAIL_ERROR_COLOUR);
												}
											}
										}
										
										DeleteArray(s);
									}
								}
							}
						}
					}
					else
					{
						char s[256];
						sprintf(s, "Warning: Unsupport auth type '%s'", AuthType);
						Log(s, MAIL_ERROR_COLOUR);
					}
				}

				if (LoggedIn)
				{
					Status = true;

					// Ask server for it's heirarchy (folder) separator.
					int Cmd = d->NextCmd++;
					sprintf(Buf, "A%04.4i LIST \"\" \"\"\r\n", Cmd);
					if (WriteBuf())
					{
						ClearDialog();
						Buf[0] = 0;
						if (ReadResponse(Cmd))
						{
							for (char *Dlg = Dialog.First(); Dlg; Dlg=Dialog.Next())
							{
								GToken t(Dlg, " ");
								if (t.Length() >= 5)
								{
									if (strcmp(t[0], "*") == 0 AND
										stricmp(t[1], "list") == 0)
									{
										char *Sep = TrimStr(t[3], "\"");
										if (Sep)
										{
											d->FolderSep = *Sep;
											DeleteArray(Sep);
										}
										break;
									}
								}
							}
						}
					}
				}
				else
				{
					char Str[300];
					sprintf(Str,
							"Authentication failed, types available:\n%s",
							(strlen(AuthTypeStr)>0) ? AuthTypeStr : "(none)");
					ServerMsg = NewStr(Str);
				}
			}

			Auths.DeleteArrays();
		}
		
		Unlock();
	}

	return Status;
}

bool MailIMap::Close()
{
	bool Status = false;
	if (S AND Lock())
	{
		if (d->ExpungeOnExit)
		{
			ExpungeFolder(0);
		}
		
		int Cmd = d->NextCmd++;
		sprintf(Buf, "A%04.4i LOGOUT\r\n", Cmd);
		if (WriteBuf())
		{
			Status = true;
		}
		Unlock();
	}
	return Status;
}

bool MailIMap::ServerOption(char *Opt)
{
	return d->Capability.Find(Opt) != 0;
}

char *MailIMap::GetSelectedFolder()
{
	return d->Current;
}

bool MailIMap::SelectFolder(char *Path, MailImapFolder *f)
{
	int Status = 0;

	if (S AND Lock())
	{
		f->UnseenIndex = -1;

		int Cmd = d->NextCmd++;
		char *Enc = EncodeImapString(Path);
		sprintf(Buf, "A%04.4i SELECT \"%s\"\r\n", Cmd, Enc);
		DeleteArray(Enc);
		if (WriteBuf())
		{
			DeleteArray(d->Current);
			ClearDialog();
			if (ReadResponse(Cmd))
			{
				for (char *Dlg = Dialog.First(); Dlg; Dlg=Dialog.Next())
				{
					GToken t(Dlg, " ");
					if (stricmp(t[0], "*") == 0 AND
						t.Length() > 2)
					{
						if (f)
						{
							if (stricmp(t[2], "exists") == 0)
							{
								f->Exists = atoi(t[1]);
							}
							else if (stricmp(t[2], "recent") == 0)
							{
								f->Recent = atoi(t[1]);
							}
							else if (stricmp(t[2], "[UNSEEN") == 0)
							{
								f->UnseenIndex = atoi(t[3]);
							}
						}
					}
				}

				Status = true;
				d->Current = NewStr(Path);
			}
		}
		
		Unlock();
	}

	return Status;
}

int MailIMap::GetMessages(char *Path)
{
	int Status = 0;

	if (S AND Lock())
	{
		MailImapFolder Folder;
		if (SelectFolder(Path, &Folder))
		{
			for (char *d = Dialog.First(); d AND NOT Status; d=Dialog.Next())
			{
				if (strstr(d, "EXISTS"))
				{
					GToken T(d, " ");
					if (T.Length() == 3)
					{
						Status = atoi(T[1]);
					}
				}
			}
		}
		
		Unlock();
	}
	
	return Status;
}

int MailIMap::GetMessages()
{
	return GetMessages("INBOX");
}

// Private
bool MailIMap::GetParts(int Message, GStream &Out, char *Parts, char **Flags)
{
	bool Status = false;

	if (Parts)
	{
		int Cmd = d->NextCmd++;
		sprintf(Buf, "A%04.4i FETCH %i (%s)\r\n", Cmd, Message+1, Parts);
		if (WriteBuf())
		{
			ClearDialog();
			Buf[0] = 0;
			d->Logging = false;

			GStringPipe Buf(4096);
			
			/////////////////////////////////
			// Debug
			int64 Start = LgiCurrentTime();
			/////////////////////////////////
			
			bool ReadOk = ReadResponse(Cmd, &Buf);
			
			/////////////////////////////////
			// Debug
			char Size[64];
			LgiFormatSize(Size, Buf.GetSize());
			// LgiTrace("ReadResponse time %i for %s\n", (int) (LgiCurrentTime() - Start), Size);
			/////////////////////////////////

			if (ReadOk)
			{
				int Chars = 0;
				int DialogItems = Dialog.Length();
				int Pos = 0;

				char *Response = Buf.NewStr();
				if (Response)
				{
					// FIXME..
					// Log(d, MAIL_RECEIVE_COLOUR);

					char *s = Response;
					for (int i=0; i<10 AND s[i]; i++, s++);
					#define Ws(s) while (*s AND strchr(" \t\r\n", *s)) s++

					while (s AND *s)
					{
						Ws(s);
						if (*s == '(' OR *s == ')')
						{
							s++;
						}
						else
						{
							char *Field = LgiTokStr(s);
							if (NOT Field) break;

							if (stricmp(Field, "Flags") == 0)
							{
								char *Start = strchr(s, '(');
								if (Start)
								{
									char *End = strchr(++Start, ')');
									if (End)
									{
										if (Flags)
										{
											*Flags = NewStr(Start, (int)End-(int)Start);
										}
										s = End + 1;
									}
									else break;
								}
								else break;
							}
							else
							{
								char *Start = strchr(s, '{');
								if (Start)
								{
									char *End = strchr(Start, '}');
									if (End)
									{
										*End = 0;
										Chars = atoi(++Start);
										s = End + 1;
										Ws(s);
										
										int i = 0;
										for (; i<Chars AND s[i]; i++);
										Out.Write(s, i);
										Out.Write((char*)"\r\n\r\n", 4);
										s += i;
									}
									else break;
								}
								else break;
							}

							DeleteArray(Field);
						}
					}

					DeleteArray(Response);
				}

				Status = Out.GetSize() > 0;

				char *Last = Dialog.Last();
				if (Last)
				{
					Log(Last, MAIL_RECEIVE_COLOUR);
				}
			}

			d->Logging = true;
		}
	}

	return Status;
}

bool MailIMap::ReceiveFlags(MailMessage *Msg)
{
	bool Status = false;

	if (Msg AND d->Flags)
	{
		MailImapMsg *p = (MailImapMsg*)Msg->Private;
		if (NOT p)
			Msg->Private = p = NEW(MailImapMsg);
		if (p)
		{
			if (stristr(d->Flags, "\\deleted")) p->ImapDeleted = true;
			if (stristr(d->Flags, "\\recent")) p->ImapRecent = true;
			if (stristr(d->Flags, "\\seen")) p->ImapSeen = true;
			Status = true;
		}
	}
	DeleteArray(d->Flags);

	return Status;
}

bool MailIMap::Receive(GArray<MailTransaction*> &Trans, MailCallbacks *Callbacks)
{
	bool Status = false;

	if (Lock())
	{
		for (int i=0; i<Trans.Length(); i++)
		{
			DeleteArray(d->Flags);
			if (Trans[i]->Status = GetParts(Trans[i]->Index, *Trans[i]->Stream, "FLAGS RFC822.HEADER BODY[TEXT]", &d->Flags))
			{
				Status = true;
			}
		}
		Unlock();
	}

	return Status;
}

// Private function
bool MailIMap::ReadLine(char *Buf, int Size)
{
	if (Buf AND S)
	{
		int Pos = 0;
		for (int Len = S->Read(Buf, Size, 0); Len > 0; Len = S->Read(Buf + Pos, Len - Pos, 0))
		{
			Pos += Len;
			Buf[Pos] = 0;

			if (strstr(Buf, "\r\n")) return true;
		}
	}

	return false;
}

bool MailIMap::Send(MailImapFolder *Folder, char *Msg, int Index, bool Seen)
{
	bool Status = false;

	if (Folder AND Msg AND Lock())
	{
		char Text[256] = "";
		if (Seen)
		{
			strcpy(Text, " (\\Seen)");
		}
		
		int Cmd = d->NextCmd++;
		int Len = 0;
		for (char *m = Msg; *m; m++)
		{
			if (*m == '\n') Len += 2;
			else if (*m != '\r') Len++;
		}

		if (Index >= 0)
		{
			// Item already exists... can't change it.
			LgiTrace("%s:%i - index %i already seen...\n", __FILE__, __LINE__, Index);
		}
		else
		{
			// Append on the end of the mailbox
			sprintf(Buf, "A%04.4i APPEND \"%s\"%s {%i}\r\n", Cmd, Folder->GetPath(), Text, Len);
			
			if (WriteBuf() AND
				ReadLine(Text, sizeof(Text)))
			{
				if (Text[0] == '+')
				{
					// Write message literal
					S->Write(Msg, Len, 0);
					S->Write((char*)"\r\n", 2, 0);

					// Read response..
					ClearDialog();
					Status = ReadResponse(Cmd);
				}
			}
		}
		
		Unlock();
	}
	else
	{
		LgiTrace("%s:%i - arg check failed.\n", __FILE__, __LINE__);
	}

	return Status;
}

bool MailIMap::Delete(int Message)
{
	bool Status = false;

	if (S AND Lock())
	{
		int Cmd = d->NextCmd++;
		sprintf(Buf, "A%04.4i STORE %i FLAGS (\\deleted)\r\n", Cmd, Message+1);
		if (WriteBuf())
		{
			ClearDialog();
			Status = ReadResponse(Cmd);
		}
		
		Unlock();
	}

	return Status;
}

int MailIMap::Sizeof(int Message)
{
	int Status = 0;

	if (S AND Lock())
	{
		char *Tag = "RFC822.SIZE";

		int Cmd = d->NextCmd++;
		sprintf(Buf, "A%04.4i FETCH %i (%s)\r\n", Cmd, Message+1, Tag);
		if (WriteBuf())
		{
			ClearDialog();
			Buf[0] = 0;
			if (ReadResponse(Cmd))
			{
				char *d = Dialog.First();
				if (d)
				{
					char *t = strstr(d, Tag);
					if (t)
					{
						t += strlen(Tag) + 1;
						Status = atoi(t);
					}
				}
			}
		}
		
		Unlock();
	}

	return Status;
}

bool MailIMap::GetUid(int Message, char *Id)
{
	bool Status = false;
	
	if (Lock())
	{
		if (FillUidList())
		{
			char *s = Uid.ItemAt(Message);
			if (s AND Id)
			{
				strcpy(Id, s);
				Status = true;
			}
		}
	
		Unlock();
	}

	return Status;
}

bool MailIMap::FillUidList()
{
	bool Status = false;

	if (S AND Lock())
	{
		if (Uid.Length() == 0)
		{
			int Cmd = d->NextCmd++;
			sprintf(Buf, "A%04.4i UID SEARCH ALL\r\n", Cmd);
			if (WriteBuf())
			{
				ClearDialog();
				if (ReadResponse(Cmd))
				{
					for (char *d = Dialog.First(); d AND NOT Status; d=Dialog.Next())
					{
						GToken T(d, " ");
						if (T[1] AND strcmp(T[1], "SEARCH") == 0)
						{
							for (int i=2; i<T.Length(); i++)
							{
								Uid.Insert(NewStr(T[i]));
							}
						}
					}

					Status = true;
				}
			}
		}
		else
		{
			Status = true;
		}
		
		Unlock();
	}

	return Status;
}

bool MailIMap::GetUidList(List<char> &Id)
{
	bool Status = false;

	if (Lock())
	{
		if (FillUidList())
		{
			for (char *s=Uid.First(); s; s=Uid.Next())
			{
				Id.Insert(NewStr(s));
			}

			Status = true;
		}
		Unlock();
	}

	return Status;
}

char *MailIMap::GetHeaders(int Message)
{
	GStringPipe Text;
	if (Lock())
	{
		GetParts(Message, Text, "RFC822.HEADER");
		Unlock();
	}
	
	return Text.NewStr();
}

bool MailIMap::GetFolders(List<MailImapFolder> &Folders, char *Path)
{
	int Status = 0;

	if (S AND Lock())
	{
		int Cmd = d->NextCmd++;
		sprintf(Buf, "A%04.4i LIST \"\" \"*\"\r\n", Cmd, Path ? Path : "");
		if (WriteBuf())
		{
			ClearDialog();
			Buf[0] = 0;
			if (ReadResponse(Cmd))
			{
				for (char *d = Dialog.First(); d; d=Dialog.Next())
				{
					GArray<char*> t;
					char *s;
					while (s = LgiTokStr(d))
					{
						t[t.Length()] = s;
					}
					
					if (t.Length() >= 5)
					{
						if (strcmp(t[0], "*") == 0 AND
							stricmp(t[1], "LIST") == 0)
						{
							char Folder[256];
							strcpy(Folder, t[t.Length()-1]);

							MailImapFolder *f = NEW(MailImapFolder);
							if (f)
							{
								Folders.Insert(f);

								f->Sep = GetFolderSep();

								// Check flags
								f->NoSelect = stristr(t[2], "NoSelect") != 0;
								f->NoInferiors = stristr(t[2], "NoInferiors") != 0;

								// Alloc name
								if (Folder[0] == '\"')
								{
									char *p = TrimStr(Folder, "\"");
									f->Path = DecodeImapString(p);
									DeleteArray(p);
								}
								else
								{
									f->Path = DecodeImapString(Folder);
								}
							}

							Status = true;
						}
					}
				}

				ClearDialog();
			}
		}
		
		Unlock();
	}

	return Status;
}

bool MailIMap::CreateFolder(MailImapFolder *f)
{
	bool Status = false;
	char Dir[2] = { d->FolderSep, 0 };

	if (f AND f->GetPath() AND Lock())
	{
		int Cmd = d->NextCmd++;
		char *Enc = EncodeImapString(f->GetPath());
		sprintf(Buf, "A%04.4i CREATE \"%s\"\r\n", Cmd, Enc);
		DeleteArray(Enc);
		if (WriteBuf())
		{
			ClearDialog();
			Status = ReadResponse(Cmd);
			if (Status)
			{
				char *End = f->Path + strlen(f->Path) - 1;
				if (*End == GetFolderSep())
				{
					f->NoSelect = true;
					*End = 0;
				}
				else
				{
					f->NoInferiors = true;
				}
			}
		}
		
		Unlock();
	}

	return Status;
}

bool MailIMap::DeleteFolder(char *Path)
{
	bool Status = false;

	if (Path AND Lock())
	{
		// Close the current folder if required.
		if (d->Current AND
			stricmp(Path, d->Current) == 0)
		{
			int Cmd = d->NextCmd++;
			sprintf(Buf, "A%04.4i CLOSE\r\n", Cmd);
			if (WriteBuf())
			{
				ClearDialog();
				ReadResponse(Cmd);
				DeleteArray(d->Current);
			}
		}

		// Delete the folder
		int Cmd = d->NextCmd++;
		char *Enc = EncodeImapString(Path);
		sprintf(Buf, "A%04.4i DELETE \"%s\"\r\n", Cmd, Enc);
		DeleteArray(Enc);
		if (WriteBuf())
		{
			ClearDialog();
			Status = ReadResponse(Cmd);
		}
		
		Unlock();
	}

	return Status;
}

bool MailIMap::RenameFolder(char *From, char *To)
{
	bool Status = false;

	if (From AND To AND Lock())
	{
		int Cmd = d->NextCmd++;
		char *f = EncodeImapString(From);
		char *t = EncodeImapString(To);
		sprintf(Buf, "A%04.4i RENAME \"%s\" \"%s\"\r\n", Cmd, f, t);
		DeleteArray(f);
		DeleteArray(t);
		if (WriteBuf())
		{
			ClearDialog();
			Status = ReadResponse(Cmd);
		}
		
		Unlock();
	}

	return Status;
}

bool MailIMap::SetFolderFlags(MailImapFolder *f)
{
	bool Status = false;

	if (f AND Lock())
	{
		int Cmd = d->NextCmd++;
		sprintf(Buf, "A%04.4i RENAME \"%s\" \"%s\"\r\n", Cmd);
		if (WriteBuf())
		{
			ClearDialog();
			Status = ReadResponse(Cmd);
		}
		
		Unlock();
	}

	return Status;
}

bool MailIMap::ExpungeFolder(MailImapFolder *f)
{
	bool Status = false;

	if (Lock())
	{
		if (f)
		{
			if (NOT d->Current OR stricmp(d->Current, f->GetPath()) != 0)
			{
				SelectFolder(f->GetPath(), f);
			}
		}

		int Cmd = d->NextCmd++;
		sprintf(Buf, "A%04.4i EXPUNGE\r\n", Cmd);
		if (WriteBuf())
		{
			ClearDialog();
			Status = ReadResponse(Cmd);
			if (Status AND f)
			{
				f->Exists -= f->Deleted;
				f->Deleted = 0;
			}
		}
		
		Unlock();
	}

	return Status;
}

bool MailIMap::Poll(int *Recent)
{
	bool Status = true;

	if (Lock())
	{
		int Cmd = d->NextCmd++;
		sprintf(Buf, "A%04.4i NOOP\r\n", Cmd);
		if (WriteBuf())
		{
			ClearDialog();
			if (Status = ReadResponse(Cmd))
			{
				if (Recent)
				{
					*Recent = 0;
				}

				for (char *Dlg=Dialog.First(); Dlg; Dlg=Dialog.Next())
				{
					if (Recent AND stristr(Dlg, " RECENT"))
					{
						*Recent = atoi(Dlg + 2);
					}
				}

				if (Recent AND *Recent)
				{
					int Cmd = d->NextCmd++;
					sprintf(Buf, "A%04.4i SEARCH NEW\r\n", Cmd);
					if (WriteBuf())
					{
						ClearDialog();
						if (ReadResponse(Cmd))
						{
						}					
					}
				}
			}
		}
		Unlock();
	}

	return Status;
}

bool MailIMap::OnIdle(int *Exists, int *Recent, GArray<MailImapMsg> *Msg)
{
	bool Status = false;

	if (Lock())
	{
		if (S->IsReadable())
		{
			int Read = S->Read(Buffer, sizeof(Buffer)-1, 0);
			if (Read > 0)
			{
				d->Untagged.Write(Buffer, Read);
			}
		}
		Unlock();
	}

	while (d->Untagged.Pop(Buffer, sizeof(Buffer)))
	{
		Log(Buffer, MAIL_RECEIVE_COLOUR);

		if (Buffer[0] == '*' AND
			Buffer[1] == ' ')
		{
			GToken t(Buffer+2, " \r\n");
			if (t.Length() >= 2)
			{
				if (Msg AND stricmp(t[1], "EXPUNGE") == 0)
				{
					MailImapMsg *f = &(*Msg)[Msg->Length()];
					f->Index = atoi(t[0]) - 1;
					f->ImapExpunged = 1;
					Status = true;
				}
				else if (Msg AND stricmp(t[1], "FETCH") == 0)
				{
					MailImapMsg *f = &(*Msg)[Msg->Length()];
					f->Index = atoi(t[0]) - 1;
					f->ImapDeleted = stristr(Buffer, "\\Deleted") != 0;
					Status = true;
				}
				else if (Exists AND stricmp(t[1], "EXISTS") == 0)
				{
					*Exists = atoi(t[0]);
					Status = true;
				}
				else if (Recent AND stricmp(t[1], "RECENT") == 0)
				{
					*Recent = atoi(t[0]);
					Status = true;
				}
			}
		}
	}

	return Status;
}

