#include <stdarg.h>
#include <ctype.h>
#include <stdio.h>

#include "Lgi.h"
#include "GStream.h"

//////////////////////////////////////////////////////////////////////////////
GMemStream::GMemStream()
{
	Len = Pos = 0;
	Mem = 0;
}

GMemStream::GMemStream(GStreamI *Src, int64 Start, int64 len)
{
	Len = len;
	Pos = 0;
	Mem = 0;

	bool Error = false;
	if (Start >= 0)
	{
		Error = Src->SetPos(Start) != Start;
	}
	if (NOT Error)
	{
		if (Len >= 0)
		{
			if (Mem = NEW(char[Len]))
			{
				int r;
				while (	Pos < Len AND
						(r = Src->Read(Mem + Pos, Len - Pos)) > 0)
				{
					Pos += r;
				}
				Pos = 0;
			}
		}
		else
		{
			char Buf[512];
			GBytePipe p(4 << 10);
			int r;
			while ((r = Src->Read(Buf, sizeof(Buf))) > 0)
			{
				p.Write(Buf, r);
			}

			Len = p.GetSize();
			Mem = (char*)p.New(0);
		}
	}
}

GMemStream::GMemStream(void *mem, int64 len)
{
	Len = len;
	Pos = 0;
	Mem = NEW(char[Len]);
	if (Mem)
	{
		memcpy(Mem, mem, Len);
	}
}

GMemStream::~GMemStream()
{
	Close();
}

int GMemStream::Open(char *Str, int Int)
{
	GFile f;
	if (f.Open(Str, O_READ))
	{
		Close();
		Len = f.GetSize();
		Mem = NEW(char[Len]);
		if (Mem)
		{
			Len = f.Read(Mem, Len);
		}

		return Len > 0;
	}

	return 0;
}

int64 GMemStream::SetSize(int64 Size)
{
	if (Mem)
	{
		int64 Common = min(Len, Size);
		char *NewMem = NEW(char[Size]);
		if (NewMem)
		{
			memcpy(NewMem, Mem, Common);
			DeleteArray(Mem);
			Mem = NewMem;
			Len = Size;
		}
		else return -1;
	}
	else
	{
		Close();
		Len = Size;
		Mem = NEW(char[Len]);
	}

	return Len;
}

bool GMemStream::IsOk()
{
	return Len == 0 OR Mem != 0;
}

int GMemStream::Read(void *Buffer, int Size, int Flags)
{
	int Bytes = 0;
	if (Buffer AND Pos >= 0 AND Pos < Len)
	{
		Bytes = min(Len - Pos, Size);
		memcpy(Buffer, Mem + Pos, Bytes);
		Pos += Bytes;			
	}
	return Bytes;
}

int GMemStream::Write(void *Buffer, int Size, int Flags)
{
	int Bytes = 0;
	if (Buffer AND Pos >= 0 AND Pos < Len)
	{
		Bytes = min(Len - Pos, Size);
		memcpy(Mem + Pos, Buffer, Bytes);
		Pos += Bytes;			
	}
	return Bytes;
}

GStreamI *GMemStream::Clone()
{
	return NEW(GMemStream(Mem, Len));
}

