#include "Lgi.h"
#include "GOptionsFile.h"

GOptionsFile::GOptionsFile(char *BaseName) : GSemaphore("GOptionsFile")
{
	char FullPath[MAX_PATH];
	LgiGetExePath(FullPath, sizeof(FullPath));
	#ifdef WIN32
	char *Last = strrchr(FullPath, '\\');
	if (Last)
	{
		if (stricmp(Last + 1,
			#ifdef _DEBUG
			"Debug"
			#else
			"Release"
			#endif
			) == 0)
		{
			*Last = 0;
		}
	}
	#endif
	char *End = FullPath + strlen(FullPath);
	if (End[-1] != DIR_CHAR)
		*End++ = DIR_CHAR;
	int Used = End - FullPath;
	strsafecpy(End, BaseName ? BaseName : "Options", sizeof(FullPath) - Used);
	if (NOT strchr(End, '.'))
		strcat(End, ".xml");
	File = NewStr(FullPath);

	Tag = NewStr("Options");
	Dirty = false;

	if (Lock())
	{
		_Defaults();
		Unlock();
	}
}

GOptionsFile::~GOptionsFile()
{
	DeleteArray(File);
}

bool GOptionsFile::_OnAccess(bool Start)
{
	if (Start)
	{
		return Lock();
	}

	Unlock();
	return true;
}

bool GOptionsFile::IsValid()
{
	bool Status = false;

	if (Lock())
	{
		if (Attr.Length() > 0 AND
			Children.Length() > 0)
		{
			Status = true;
		}

		Unlock();
	}

	return Status;
}

void GOptionsFile::SetFile(char *f)
{
	DeleteArray(File);
	File = NewStr(f);
	Dirty = true;
}

bool GOptionsFile::Serialize(bool Write)
{
	bool Status = false;

	if (File AND Lock())
	{
		GFile f;
		if (f.Open(File, Write?O_WRITE:O_READ))
		{
			GXmlTree Tree(GXT_PRETTY_WHITESPACE);
			if (Write)
			{
				f.SetSize(0);
				Status = Tree.Write(this, &f);
			}
			else
			{
				Empty();
				if (Status = Tree.Read(this, &f, 0))
				{
					_Defaults();
				}
			}
		}
		Unlock();
	}

	return Status;
}

bool GOptionsFile::DeleteValue(char *Name)
{
	bool Status = false;

	if (Name AND Lock())
	{
		GVariant v;
		SetValue(Name, v);
		Unlock();
	}

	return Status;
}

bool GOptionsFile::CreateTag(char *Name)
{
	bool Status = false;

	if (Name AND Lock())
	{
		Status = GetTag(Name, true) != 0;
		Unlock();
	}

	return Status;
}

bool GOptionsFile::DeleteTag(char *Name)
{
	bool Status = false;

	if (Name AND Lock())
	{
		GXmlTag *t = GetTag(Name);
		if (t)
		{
			t->RemoveTag();
			DeleteObj(t);
			Status = true;
		}
		Unlock();
	}

	return Status;
}

GXmlTag *GOptionsFile::LockTag(char *Name)
{
	GXmlTag *t = 0;

	if (Lock())
	{
		t = GetTag(Name);
		if (NOT t)
		{
			Unlock();
		}
	}

	return t;
}
