#include <stdlib.h>
#include <stdio.h>

#include "Lgi.h"
#include "GVariant.h"
#include "GToken.h"

GVariant::GVariant()
{
	Type = GV_NULL;
	Id = 0;
	Dirty = 0;
	ZeroObj(Value);
}

GVariant::GVariant(int i)
{
	Type = GV_INT32;
	Id = 0;
	Dirty = 0;
	Value.Int = i;
}

GVariant::GVariant(GVariant const &v)
{
	Type = GV_NULL;
	Id = 0;
	Dirty = 0;
	ZeroObj(Value);

	*this = v;
}

GVariant::GVariant(int64 i)
{
	Type = GV_INT64;
	Id = 0;
	Dirty = 0;
	Value.Int64 = i;
}

GVariant::GVariant(double i)
{
	Type = GV_DOUBLE;
	Id = 0;
	Dirty = 0;
	Value.Dbl = i;
}

GVariant::GVariant(char *s)
{
	Type = GV_STRING;
	Id = 0;
	Dirty = 0;
	Value.String = NewStr(s);
}

GVariant::GVariant(void *p)
{
	Type = GV_NULL;
	Id = 0;
	Dirty = 0;
	*this = p;
}

GVariant::GVariant(GDom *p)
{
	Type = GV_NULL;
	Id = 0;
	Dirty = 0;
	*this = p;
}

GVariant::GVariant(GDateTime *d)
{
	Type = GV_NULL;
	Id = 0;
	Dirty = 0;
	*this = d;
}

GVariant::~GVariant()
{
	Empty();
}

GVariant &GVariant::operator =(GDateTime *d)
{
	Empty();

	if (d)
	{
		Type = GV_DATETIME;
		Value.Date = NEW(GDateTime);
		if (Value.Date)
		{
			*Value.Date = *d;
			if (Dirty) *Dirty = true;
		}
	}

	return *this;
}

GVariant &GVariant::operator =(int i)
{
	Empty();
	Type = GV_INT32;
	Value.Int = i;
	if (Dirty) *Dirty = true;

	return *this;
}

GVariant &GVariant::operator =(bool i)
{
	Empty();
	Type = GV_BOOL;
	Value.Bool = i;
	if (Dirty) *Dirty = true;

	return *this;
}

GVariant &GVariant::operator =(int64 i)
{
	Empty();
	Type = GV_INT64;
	Value.Int64 = i;
	if (Dirty) *Dirty = true;

	return *this;
}

GVariant &GVariant::operator =(double i)
{
	Empty();
	Type = GV_DOUBLE;
	Value.Dbl = i;
	if (Dirty) *Dirty = true;

	return *this;
}

GVariant &GVariant::operator =(char *s)
{
	Empty();
	if (s)
	{
		Type = GV_STRING;
		Value.String = NewStr(s);
	}
	if (Dirty) *Dirty = true;

	return *this;
}

GVariant &GVariant::operator =(char16 *s)
{
	Empty();
	if (s)
	{
		Type = GV_STRING;
		Value.String = LgiNewUtf16To8(s);
	}
	if (Dirty) *Dirty = true;

	return *this;
}

GVariant &GVariant::operator =(void *p)
{
	Empty();
	if (p)
	{
		Type = GV_VOID_PTR;
		Value.Ptr = p;
		if (Dirty) *Dirty = true;
	}

	return *this;
}

GVariant &GVariant::operator =(GDom *p)
{
	Empty();
	if (p)
	{
		Type = GV_DOM;
		Value.Dom = p;
		if (Dirty) *Dirty = true;
	}

	return *this;
}

GVariant &GVariant::operator =(class GView *p)
{
	Empty();
	if (p)
	{
		Type = GV_GVIEW;
		Value.Ptr = p;
		if (Dirty) *Dirty = true;
	}

	return *this;
}

GVariant &GVariant::operator =(class GMouse *p)
{
	Empty();
	if (p)
	{
		Type = GV_GMOUSE;
		Value.Ptr = p;
		if (Dirty) *Dirty = true;
	}

	return *this;
}

GVariant &GVariant::operator =(class GKey *p)
{
	Empty();
	if (p)
	{
		Type = GV_GKEY;
		Value.Ptr = p;
		if (Dirty) *Dirty = true;
	}

	return *this;
}

GVariant &GVariant::operator =(GVariant const &i)
{
	Empty();
	Type = i.Type;
	Id = i.Id;

	switch (Type)
	{
		case GV_NULL:
		{
			break;
		}
		case GV_INT32:
		{
			Value.Int = i.Value.Int;
			break;
		}
		case GV_BOOL:
		{
			Value.Bool = i.Value.Bool;
			break;
		}
		case GV_INT64:
		{
			Value.Int64 = i.Value.Int64;
			break;
		}
		case GV_DOUBLE:
		{
			Value.Dbl = i.Value.Dbl;
			break;
		}
		case GV_STRING:
		{
			Value.String = NewStr(((GVariant&)i).Str());
			break;
		}
		case GV_BINARY:
		{
			SetBinary(i.Value.Binary.Length, i.Value.Binary.Data);
			break;
		}
		case GV_LIST:
		{
			SetList(i.Value.Lst);
			break;
		}
		case GV_DOM:
		{
			Value.Dom = i.Value.Dom;
			break;
		}
		case GV_VOID_PTR:
		{
			Value.Ptr = i.Value.Ptr;
			break;
		}
		case GV_DATETIME:
		{
			if (i.Value.Date)
			{
				Value.Date = NEW(GDateTime);
				if (Value.Date)
				{
					*Value.Date = *i.Value.Date;
				}
			}
			break;
		}
		case GV_HASHTABLE:
		{
			if (Value.Hash = NEW(GHashTable))
			{
				char *k;
				for (void *p = i.Value.Hash->First(&k); p; p = i.Value.Hash->Next(&k))
				{
					GVariant *var = (GVariant*)p;
					Value.Hash->Add(k, NEW(GVariant(var)));
				}
			}
			break;
		}
		default:
		{
			LgiAssert(0);
			break;
		}
	}

	if (Dirty) *Dirty = true;

	return *this;
}

bool GVariant::SetBinary(int Len, void *Data)
{
	bool Status = false;

	Empty();

	Type = GV_BINARY;
	Value.Binary.Length = Len;
	Value.Binary.Data = NEW(uchar[Value.Binary.Length]);
	if (Value.Binary.Data)
	{
		if (Data)
		{
			memcpy(Value.Binary.Data, Data, Value.Binary.Length);
		}
		else
		{
			memset(Value.Binary.Data, 0, Value.Binary.Length);
		}

		Status = true;
	}

	if (Dirty) *Dirty = true;

	return Status;
}

bool GVariant::SetList(List<GVariant> *Lst)
{
	Empty();
	Type = GV_LIST;
	Value.Lst = NEW(List<GVariant>);
	if (Value.Lst AND Lst)
	{
		for (GVariant *s=Lst->First(); s; s=Lst->Next())
		{
			GVariant *New = NEW(GVariant);
			if (New)
			{
				*New = *s;
				Value.Lst->Insert(New);
			}
		}
	}

	return Value.Lst != 0;
}

bool GVariant::SetHashTable(GHashTable *Table, bool Copy)
{
	Empty();
	Type = GV_HASHTABLE;

	if (Copy AND Table)
	{
		if (Value.Hash = NEW(GHashTable))
		{
			char *k;
			for (void *p = Table->First(&k); p; p = Table->Next(&k))
			{
				Value.Hash->Add(k, p);
			}
		}
	}
	else
	{
		Value.Hash = Table;
	}

	return Value.Hash != 0;
}

char *GVariant::Str()
{
	return (Type == GV_STRING) ? Value.String : 0;
}

void GVariant::Empty()
{
	switch (Type)
	{
		case GV_STRING:
		{
			DeleteArray(Value.String);
			break;
		}
		case GV_BINARY:
		{
			char *d = (char*) Value.Binary.Data;
			DeleteArray(d);
			Value.Binary.Data = 0;
			break;
		}
		case GV_DATETIME:
		{
			DeleteObj(Value.Date);
			break;
		}
		case GV_LIST:
		{
			if (Value.Lst)
			{
				Value.Lst->DeleteObjects();
				DeleteObj(Value.Lst);
			}
			break;
		}
		case GV_HASHTABLE:
		{
			if (Value.Hash)
			{
				for (GVariant *v = (GVariant*) Value.Hash->First(); v; v = (GVariant*) Value.Hash->Next())
				{
					DeleteObj(v);
				}
				DeleteObj(Value.Hash);
			}
			break;
		}
	}

	Type = GV_NULL;
	ZeroObj(Value);
}

bool GVariant::IsInt()
{
	return Type == GV_INT32 OR Type == GV_INT64;
}

bool GVariant::IsBool()
{
	return Type == GV_BOOL;
}

bool GVariant::IsDouble()
{
	return Type == GV_DOUBLE;
}

bool GVariant::IsString()
{
	return Type == GV_STRING;
}

bool GVariant::IsBinary()
{
	return Type == GV_BINARY;
}

bool GVariant::IsNull()
{
	return Type == GV_NULL;
}

#define IsList() (Type == GV_LIST AND Value.Lst)

GVariant &GVariant::Cast(GVariantType NewType)
{
	if (NewType != Type)
	{
		switch (NewType)
		{
			default:
			{
				// No conversion possible
				break;
			}
			case GV_INT32:
			{
				*this = CastInt32();
				break;
			}
			case GV_INT64:
			{
				*this = CastInt64();
				break;
			}
			case GV_BOOL:
			{
				if (Type == GV_DOUBLE)
				{
					*this = Value.Dbl != 0.0;
				}
				else
				{
					*this = CastInt32() != 0;
				}
				break;
			}
			case GV_DOUBLE:
			{
				*this = CastDouble();
				break;
			}
			case GV_STRING:
			{
				*this = CastString();
				break;
			}
			case GV_DATETIME:
			{
				switch (Type)
				{
					case GV_STRING:
					{
						// String -> GDateTime
						GDateTime *Dt = NEW(GDateTime);
						if (Dt)
						{
							Dt->Set(Value.String);
							Empty();
							Value.Date = Dt;
							Type = NewType;
						}
						break;
					}
					case GV_INT64:
					{
						// Int64 (system date) -> GDateTime
						GDateTime *Dt = NEW(GDateTime);
						if (Dt)
						{
							Dt->Set(Value.Int64);
							Empty();
							Value.Date = Dt;
							Type = NewType;
						}
						break;
					}
					default:
					{
						// No conversion available
						break;
					}
				}
				break;
			}
		}
	}

	return *this;
}

double GVariant::CastDouble()
{
	switch (Type)
	{
		case GV_BOOL:
			return Value.Bool ? 1.0 : 0.0;
		case GV_DOUBLE:
			return Value.Dbl;
		case GV_INT32:
			return (double)Value.Int;
		case GV_INT64:
			return (double)Value.Int64;
		case GV_STRING:
			return Value.String ? atof(Value.String) : 0;
	}
	
	return 0;
}

int32 GVariant::CastInt32()
{
	switch (Type)
	{
		case GV_BOOL:
			return (int32)Value.Bool;
		case GV_DOUBLE:
			return (int32)Value.Dbl;
		case GV_INT32:
			return Value.Int;
		case GV_INT64:
			return Value.Int64;
		case GV_STRING:
			return Value.String ? atoi(Value.String) : 0;
		case GV_DOM:
			return Value.Dom != 0;
	}
	
	return 0;
}

int64 GVariant::CastInt64()
{
	switch (Type)
	{
		case GV_BOOL:
			return (int64)Value.Bool;
		case GV_DOUBLE:
			return (int64)Value.Dbl;
		case GV_INT32:
			return Value.Int;
		case GV_INT64:
			return Value.Int64;
		case GV_STRING:
		{
			if (Value.String)
			{
				#ifdef _MSC_VER
				return _atoi64(Value.String);
				#else
				return atoll(Value.String);
				#endif
			}
			break;
		}
	}
	
	return 0;
}

char *GVariant::CastString()
{
	char i[40];

	switch (Type)
	{
		case GV_INT32:
		{
			sprintf(i, "%i", Value.Int);
			*this = i;
			return Str();
		}
		case GV_DOUBLE:
		{
			sprintf(i, "%f", Value.Dbl);
			*this = i;
			return Str();
		}
		case GV_BOOL:
		{
			sprintf(i, "%i", Value.Bool);
			*this = i;
			return Str();
		}
		case GV_INT64:
		{
			sprintf(i, LGI_PrintfInt64, Value.Int64);
			*this = i;
			return Str();
		}
		case GV_STRING:
		{
			return Value.String;
		}
		case GV_DATETIME:
		{
			if (Value.Date)
			{
				char s[64];
				Value.Date->Get(s);
				*this = s;
				return Str();
			}
			break;
		}
	}
	
	return 0;
}

/////////////////////////////////////////////////////////////////////////////////
GDom *GDom::ResolveObject(char *Var, char *Name, char *Array)
{
	GDom *Object = this;

	// Tokenise the string
	GArray<char*> t;
	for (char *s = Var; s AND *s; )
	{
		char *e = s;
		while (*e AND *e != '.')
		{
			if (*e == '[')
			{
				e++;
				while (*e AND *e != ']')
				{
					if (*e == '\"' OR *e == '\'')
					{
						char d = *e++;
						while (*e AND *e != d)
						{
							e++;
						}
						if (*e == d) e++;
					}
					else e++;
				}
				if (*e == ']') e++;
			}
			else e++;
		}
		
		int Len = e - s;
		if (Len < 1 OR Len > 255)
		{
			Object = 0;
			break;
		}
		
		char Part[256];
		memcpy(Part, s, Len);
		Part[Len] = 0;
		while (strchr(" \t", Part[Len-1]))
		{
			Len--;
		}

		t[t.Length()] = NewStr(Part, Len);
		s = *e ? e + 1 : e;
	}

	// Process elements
	for (int i=0; i<t.Length(); i++)
	{
		char Base[256];
		char Index[256] = "";

		// Parse out object name parts
		char *Obj = t[i];
		if (*Obj == '[')
		{
			Object = 0;
			break;
		}
		
		char *ArrStart = strchr(Obj, '[');
		if (ArrStart)
		{
			*ArrStart++ = 0;
			char *ArrEnd = 0;
			if (strchr("\'\"", *ArrStart))
			{
				ArrEnd = strchr(ArrStart + 1, ArrStart[0]);
				ArrStart++;
			}
			else
			{
				ArrEnd = strchr(ArrStart, ']');
			}
			if (ArrEnd)
			{
				*ArrEnd = 0;
				strcpy(Index, ArrStart);
			}
		}
		strcpy(Base, Obj);

		// Process parts
		if (i == t.Length()-1)
		{
			strcpy(Name, Base);
			strcpy(Array, Index);
			
			char *e = Name + strlen(Name) - 1;
			while (e > Name AND strchr(" \t\r\n", *e))
			{
				*e-- = 0;
			}
		}
		else
		{
			GVariant v;
			if (Index[0])
			{
				if (Object->GetVariant(Base, v, Index))
				{
					if (v.Type == GV_LIST)
					{
						int N = atoi(Index);
						GVariant *Element = v.Value.Lst->ItemAt(N);
						if (Element AND Element->Type == GV_DOM)
						{
							Object = Element->Value.Dom;
						}
						else
						{
							Object = 0;
							goto ResolveDone;
						}
					}
					else if (v.Type == GV_DOM)
					{
						Object = v.Value.Dom;
					}
					else
					{
						Object = 0;
						goto ResolveDone;
					}
				}
				else
				{
					Object = 0;
					goto ResolveDone;
				}
			}
			else
			{
				if (Object->GetVariant(Obj, v) AND
					v.Type == GV_DOM)
				{
					Object = v.Value.Dom;
				}
				else
				{
					Object = 0;
					goto ResolveDone;
				}
			}
		}
	}

ResolveDone:
	t.DeleteArrays();
	return Object;
}

bool GDom::GetValue(char *Var, GVariant &Value)
{
	bool Status = false;

	if (Var)
	{
		if (_OnAccess(true))
		{
			char Name[256] = "", Arr[256] = "";
			GDom *Object = ResolveObject(Var, Name, Arr);
			if (Object)
			{
				if (Name[0] == 0)
				{
					LgiTrace("%s:%i - Warning name parse failed for '%s'\n",
						__FILE__, __LINE__, Var);
				}
				else
				{			
					Status = Object->GetVariant(Name, Value, ValidStr(Arr) ? Arr : 0);
				}
			}

			_OnAccess(false);
		}
	}

	return Status;
}

bool GDom::SetValue(char *Var, GVariant &Value)
{
	bool Status = false;

	if (Var)
	{
		if (_OnAccess(true))
		{
			char Name[256], Arr[256] = "";
			GDom *Object = ResolveObject(Var, Name, Arr);
			if (Object)
			{
				if (Name[0] == 0)
				{
					LgiTrace("%s:%i - Warning name parse failed for '%s'\n",
						__FILE__, __LINE__, Var);
				}
				else
				{			
					Status = Object->SetVariant(Name, Value, ValidStr(Arr) ? Arr : 0);
				}
			}

			_OnAccess(false);
		}
	}

	return Status;
}

