//
//	Cross platform LGI functions
//

#include <time.h>
#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>

#include "Lgi.h"
#include "GToken.h"
#ifdef LINUX
#include "LgiWinManGlue.h"
#endif

#if defined WIN32

	#include <shlobj.h>

#elif defined POSIX

	#include <sys/time.h>
	#include <sys/types.h>
	#include <pwd.h>
	#include <sys/utsname.h>
	#include "GProcess.h"

#elif defined BEOS

	#include <Path.h>

#endif

//////////////////////////////////////////////////////////////////////////
// Misc stuff
#ifdef MAC
bool _get_path_FSRef(FSRef &fs, GArray<char*> &a, int &Len)
{
	HFSUniStr255 Name;
	ZeroObj(Name);
	FSRef Parent;
	OSErr e = FSGetCatalogInfo(&fs,
								kFSCatInfoNone,
								NULL,
								&Name,
								NULL,
								&Parent);
	if (!e)
	{
		char *u = LgiNewUtf16To8((char16*)Name.unicode, Name.length * sizeof(char16));
		if (u)
		{
			Len += strlen(u);
			a.Add(u);
		}
		
		_get_path_FSRef(Parent, a, Len);
		
		return true;
	}

	return false;
}

char *FSRefPath(FSRef &fs)
{
	GArray<char*> a;
	int Len = 0;
	if (_get_path_FSRef(fs, a, Len))
	{
		char *s = NEW(char[Len + a.Length()]);
		if (s)
		{
			char *o = s;
			for (int i=a.Length()-3; i>=0; i--)
			{
				o += sprintf(o, "%c%s", DIR_CHAR, a[i]);
			}
			*o++ = 0;
			return s;
		}
	}
	
	return 0;
}
#endif

bool LgiPostEvent(OsView Wnd, int Event, int a, int b)
{
	#if defined(WIN32)

	return PostMessage(Wnd, Event, a, b);

	#elif defined(LINUX)

	if (Wnd)
	{
		GMessage m;
		m.m() = Event;
		m.a() = a;
		m.b() = b;
		#if defined XWIN
		XApplication::postEvent(Wnd, &m);
		return true;
		#endif
	}
	else
	{
		printf("%s:%i - Warning: LgiPostEvent failed because View=0\n", __FILE__, __LINE__);
	}

	#elif defined(BEOS)
	
	if (Wnd)
	{
		bool Status = false;
		
		if (Wnd->LockLooper())
		{
			BMessage *Msg = NEW(BMessage(Event));
			if (Msg)
			{
				Msg->AddInt32("a", a);
				Msg->AddInt32("b", b);
				BMessenger m(Wnd);
				Status = m.SendMessage(Msg) == B_OK;
				DeleteObj(Msg);
			}
			Wnd->UnlockLooper();
		}
		
		return Status;
	}

	#elif defined(MAC)
	
	EventRef Ev;
	OSStatus e = CreateEvent(NULL,
							kEventClassUser,
							kEventUser,
							0, // EventTime 
							kEventAttributeUserEvent,
							&Ev);
	if (e)
	{
		printf("%s:%i - CreateEvent failed with %i\n", __FILE__, __LINE__, e);
	}
	else
	{
		EventTargetRef t = GetControlEventTarget(Wnd);
		
		e = SetEventParameter(Ev, kEventParamLgiEvent, typeUInt32, sizeof(Event), &Event);
		if (e) printf("%s:%i - error %i\n", __FILE__, __LINE__, e);
		e = SetEventParameter(Ev, kEventParamLgiA, typeUInt32, sizeof(a), &a);
		if (e) printf("%s:%i - error %i\n", __FILE__, __LINE__, e);
		e = SetEventParameter(Ev, kEventParamLgiB, typeUInt32, sizeof(b), &b);
		if (e) printf("%s:%i - error %i\n", __FILE__, __LINE__, e);
		
		// printf("Sending event %i,%i,%i\n", Event, a, b);

		#if 1
		
		e = SetEventParameter(Ev, kEventParamPostTarget, typeEventTargetRef, sizeof(t), &t);
		if (e) printf("%s:%i - error %i\n", __FILE__, __LINE__, e);
		e = PostEventToQueue(GetMainEventQueue(), Ev, kEventPriorityStandard);
		if (e) printf("%s:%i - error %i\n", __FILE__, __LINE__, e);
		
		#else

		e = SendEventToEventTarget(Ev, GetControlEventTarget(Wnd));
		if (e) printf("%s:%i - error %i\n", __FILE__, __LINE__, e);
		
		#endif

		ReleaseEvent(Ev);
	}
	
	#else
	
	#error "Not Impl."
	
	#endif

	return false;
}

void LgiExitApp()
{
	exit(0);
}

//////////////////////////////////////////////////////////////////////////
#ifdef WIN32
bool RegisterActiveXControl(char *Dll)
{
	GLibrary Lib(Dll);
	if (Lib.IsLoaded())
	{
		typedef HRESULT (STDAPICALLTYPE *p_DllRegisterServer)(void);

		p_DllRegisterServer DllRegisterServer = (p_DllRegisterServer)Lib.GetAddress("DllRegisterServer");
		if (DllRegisterServer)
		{
			return DllRegisterServer() == S_OK;
		}
	}

	return false;
}
#endif

//////////////////////////////////////////////////////////////////////////

/// \brief Returns the operating system that Lgi is running on.
/// \sa Returns one of the defines starting with LGI_OS_UNKNOWN in LgiDefs.h
int LgiGetOs
(
	/// Returns the major version of the OS or NULL if you don't care
	int *Ver,
	/// Returns the minor version of the OS or NULL if you don't care
	int *Rev
)
{
	#if defined WIN32

	static int Os = LGI_OS_UNKNOWN;
	static int Version = 0, Revision = 0;

	if (Os == LGI_OS_UNKNOWN)
	{
		OSVERSIONINFO v;
		v.dwOSVersionInfoSize=sizeof(v);
		GetVersionEx(&v);

		Version = v.dwMajorVersion;
		Revision = v.dwMinorVersion;

		Os = (v.dwPlatformId == VER_PLATFORM_WIN32_NT) ? LGI_OS_WINNT : LGI_OS_WIN9X;
	}

	if (Ver) *Ver = Version;
	if (Rev) *Rev = Revision;

	return Os;

	#elif defined BEOS

	if (Ver) *Ver = 5;
	if (Rev) *Rev = 0;
	
	return LGI_OS_BEOS;

	#elif defined LINUX

	if (Ver OR Rev)
	{
		static int r = 0, v = 0;

		if (NOT r AND NOT v)
		{
			#if 0
			
			GStringPipe Out;
			GProcess p;
			if (p.Run("uname", "-r", 0, true, 0, &Out))
			{
				char *o = Out.NewStr();
				if (o)
				{
					char *d = strchr(o, '.');
					if (d)
					{
						*d++ = 0;
						v = atoi(o);
						r = atoi(d);
					}
					DeleteArray(o);
				}
			}
			
			#else
			
			utsname Buf;
			if (NOT uname(&Buf))
			{
				char *Dot = strchr(Buf.release, '.');
				v = atoi(Buf.release);
				if (Dot)
				{
					r = atoi(Dot+1);
				}				
			}
			
			#endif
		}

		if (Ver) *Ver = v;
		if (Rev) *Rev = r;
	}

	return LGI_OS_LINUX;

	#elif defined MAC

	if (Ver OR Rev)
	{
		SInt32 MacVersion;
		if (Gestalt(gestaltSystemVersion, &MacVersion) == noErr)
		{
			char s[10];
			sprintf(s, "%x", MacVersion);
			char *e = s + strlen(s) - 1;
			char r[3] = { e[-1], e[0], 0 };
			e[-1] = 0;
			
			if (Ver) *Ver = atoi(s);
			if (Rev) *Rev = atoi(r);
		}
	}
	
	return LGI_OS_MAC_OS_X;

	#else

	return LGI_OS_UNKNOWN;

	#endif
}

char *LgiGetOsName()
{
	char *Str[] =
	{
		"Unknown",
		"Win32",
		"Win32",
		"BeOS",
		"Linux",
		"AtheOS",
		"MacOS9",
		"MacOSX",
		0,
		0,
		0,
		0
	};

	return Str[LgiGetOs()];
}

#ifdef WIN32
#define RecursiveFileSearch_Wildcard "*.*"
#else // unix'ish OS
#define RecursiveFileSearch_Wildcard "*"
#endif

bool LgiRecursiveFileSearch(char *Root,
							List<char> *Ext,
							List<char> *Files,
							uint64 *Size,
							uint64 *Count,
							RecursiveFileSearch_Callback Callback)
{
	bool Status = false;

	// validate args
	if (NOT Root) return 0;

	// get directory enumerator
	GDirectory *Dir = FileDev->GetDir();
	if (Dir)
	{
		Status = true;

		// enumerate the directory contents
		for (bool Found = Dir->First(Root); Found; Found = Dir->Next())
		{
			char Name[256];
			Dir->Path(Name, sizeof(Name));

			if (Callback)
			{
				if (NOT Callback(Name, Dir))
				{
					continue;
				}
			}

			if (Dir->IsDir())
			{
				// dir
				LgiRecursiveFileSearch(	Name,
										Ext,
										Files,
										Size,
										Count,
										Callback);

			}
			else
			{
				// process file
				bool Match = true; // if no Ext's then default to match
				if (Ext)
				{
					for (char *e = Ext->First(); e; e = Ext->Next())
					{
						char *RawFile = strrchr(Name, DIR_CHAR);
						if (RawFile AND
							(Match = MatchStr(e, RawFile+1)))
						{
							break;
						}
					}
				}

				if (Match)
				{
					// file matched... process:
					if (Files)
					{
						Files->Insert(NewStr(Name));
					}

					if (Size)
					{
						*Size += Dir->GetSize();
					}

					if (Count)
					{
						*Count++;
					}

					Status = true;
				}
			}
		}

		DeleteObj(Dir);
	}

	return Status;
}

#define LGI_TRACE_TO_FILE
// #include <malloc.h>

extern bool _file_debug;
void LgiTrace(char *Msg, ...)
{
	#ifdef _INC_MALLOC
	if (_heapchk() != _HEAPOK)
	{
		return;
	}
	#endif

	if (Msg)
	{
		char Buffer[1025] = "";
		#ifdef LGI_TRACE_TO_FILE
		GFile f;
		// strcpy(Buffer, "C:\\scribe.exe");
		if (LgiGetExeFile(Buffer, sizeof(Buffer)))
		{
			char *Dot = strrchr(Buffer, '.');
			if (Dot AND NOT strchr(Dot, DIR_CHAR))
			{
				strcpy(Dot+1, "txt");
			}
			else
			{
				strcat(Buffer, ".txt");
			}
			f.Open(Buffer, O_WRITE);
		}
		#endif

		va_list Arg;
		va_start(Arg, Msg);
		int Len = vsnprintf(Buffer, sizeof(Buffer)-1, Msg, Arg);
		va_end(Arg);

		#ifdef LGI_TRACE_TO_FILE
		if (f.IsOpen())
		{
			f.Seek(0, SEEK_END);
			f.Write(Buffer, strlen(Buffer));
			f.Close();
		}
		#endif


		#if defined WIN32
		OutputDebugString(Buffer);
		#else
		printf("Trace: %s", Buffer);
		#endif
	}
}

bool LgiTrimDir(char *Path)
{
	if (Path)
	{
		char *p = strrchr(Path, DIR_CHAR);
		if (p)
		{
			*p = 0;
			return true;
		}
	}
	return false;
}

char *LgiMakeRelitivePath(char *Base, char *Path)
{
	GStringPipe Status;

	if (Base AND Path)
	{
		#ifdef WIN32
		bool SameNs = strnicmp(Base, Path, 3) == 0;
		#else
		bool SameNs = true;
		#endif
		if (SameNs)
		{
			GToken b(Base + 1, ":" DIR_STR);
			GToken p(Path + 1, ":" DIR_STR);
			int Same = 0;
			while (b[Same] AND p[Same] AND stricmp(b[Same], p[Same]) == 0)
			{
				Same++;
			}

			for (int i = Same; i<b.Length(); i++)
			{
				Status.Print("..%s", DIR_STR);
			}
			if (Same >= b.Length())
			{
				Status.Print(".%s", DIR_STR);
			}

			for (int n = Same; n<p.Length(); n++)
			{
				if (n > Same)
				{
					Status.Push(DIR_STR);
				}
				Status.Push(p[n]);
			}
		}
	}

	return Status.NewStr();
}

bool LgiIsRelitivePath(char *Path)
{
	if (NOT Path)
		return false;

	if (*Path == '.')
		return true;

	#ifdef WIN32
	if (isalpha(Path[0]) AND Path[1] == ':') // Drive letter path
		return false;
	#endif

	if (*Path == DIR_CHAR)
		return false;

	if (strstr(Path, "://")) // Protocol def
		return false;

	return true; // Correct or not???
}

bool LgiMakePath(char *Str, int StrSize, char *Path, char *File)
{
	if (StrSize <= 4)
	{
		printf("%s:%i - LgiMakeFile buf size=%i?\n", StrSize);
	}

	if (Str AND Path AND File)
	{
		if (Str != Path)
		{
			strsafecpy(Str, Path, StrSize);
		}

		char Dir[] = { '/', '\\', 0 };
		#define EndStr() Str[strlen(Str)-1]
		#define EndDir() if (NOT strchr(Dir, EndStr())) strcat(Str, DIR_STR);

		char *End = Str + strlen(Str) - 1;
		if (strchr(Dir, *End) AND End > Str)
		{
			*End = 0;
		}

		GToken T(File, Dir);
		for (int i=0; i<T.Length(); i++)
		{
			if (stricmp(T[i], ".") == 0)
			{
			}
			else if (stricmp(T[i], "..") == 0)
			{
				LgiTrimDir(Str);
			}
			else
			{
				EndDir();
				strsafecat(Str, T[i], StrSize);
			}
		}
	}
	
	return true;
}

bool LgiGetTempPath(char *Dst, int DstSize)
{
	return LgiGetSystemPath(LSP_TEMP, Dst, DstSize);
}

bool LgiGetSystemPath(LgiSystemPath Which, char *Dst, int DstSize)
{
	bool Status = false;

	#if defined(WIN32) || !defined(CSIDL_LOCAL_APPDATA)
	#define CSIDL_LOCAL_APPDATA					0x001c
	#endif

	if (Dst)
	{
		#ifdef LINUX
		// Ask our window manager add-on if it knows the path
		GLibrary *WmLib = LgiApp->GetWindowManagerLib();
		if (WmLib)
		{
			Proc_LgiWmGetPath WmGetPath = (Proc_LgiWmGetPath) WmLib->GetAddress("LgiWmGetPath");
			if (WmGetPath AND WmGetPath(Which, Dst, DstSize))
			{
				return true;
			}
		}
		#endif
		
		switch (Which)
		{
			case LSP_OS:
			{
				#if defined WIN32

				char p[MAX_PATH];
				if (GetWindowsDirectory(p, sizeof(p)) > 0)
				{
					strcpy(Dst, p);
					Status = true;
				}

				#elif defined MAC
				
				FSRef Ref;
				OSErr e = FSFindFolder(kOnAppropriateDisk,  kSystemFolderType, kDontCreateFolder, &Ref);
				if (e) printf("%s:%i - FSFindFolder failed e=%i\n", __FILE__, __LINE__, e);
				else
				{
					char *u = FSRefPath(Ref);
					if (u)
					{
						strsafecpy(Dst, u, DstSize);
						DeleteArray(u);
						Status = true;
					}
				}

				#else

				strcpy(Dst, "/boot"); // it'll do for now...
				Status = true;

				#endif
				break;
			}
			case LSP_OS_LIB:
			{
				#if defined WIN32

				char p[MAX_PATH];
				if (GetSystemDirectory(p, sizeof(p)) > 0)
				{
					strcpy(Dst, p);
					Status = true;
				}

				
				#elif defined MAC
				
				// Not supported?

				#else

				strcpy(Dst, "/lib"); // it'll do for now...
				Status = true;

				#endif
				break;
			}
			case LSP_TEMP:
			{
				#if defined WIN32

				if (LgiGetOs() == LGI_OS_WIN9X)
				{
					char t[256];
					if (GetTempPath(sizeof(t), t) > 0)
					{
						char *utf = LgiToNativeCp(t);
						if (utf)
						{
							strcpy(Dst, utf);
							DeleteArray(utf);
							Status = true;
						}
					}
				}
				else
				{
					char16 t[256];
					if (GetTempPathW(CountOf(t), t) > 0)
					{
						char *utf = LgiNewUtf16To8(t);
						if (utf)
						{
							strcpy(Dst, utf);
							DeleteArray(utf);
							Status = true;
						}
					}
				}

				#elif defined MAC
				
				FSRef Ref;
				OSErr e = FSFindFolder(kOnAppropriateDisk, kTemporaryFolderType, kDontCreateFolder, &Ref);
				if (e) printf("%s:%i - FSFindFolder failed e=%i\n", __FILE__, __LINE__, e);
				else
				{
					char *u = FSRefPath(Ref);
					if (u)
					{
						strsafecpy(Dst, u, DstSize);
						DeleteArray(u);
						Status = true;
					}
				}

				#else

				strcpy(Dst, "/tmp"); // it'll do for now...
				Status = true;

				#endif
				break;
			}
			case LSP_APP_DATA:
			{
				#if defined WIN32

				char *f = GetWin32Folder(CSIDL_APPDATA);
				if (f)
				{
					strsafecpy(Dst, f, DstSize);
					DeleteArray(f);
					Status = true;
				}
				
				#elif defined MAC
				
				FSRef Ref;
				OSErr e = FSFindFolder(kOnAppropriateDisk, kSystemPreferencesFolderType, kDontCreateFolder, &Ref);
				if (e) printf("%s:%i - FSFindFolder failed e=%i\n", __FILE__, __LINE__, e);
				else
				{
					char *u = FSRefPath(Ref);
					if (u)
					{
						strsafecpy(Dst, u, DstSize);
						DeleteArray(u);
						Status = true;
					}
				}

				#elif defined LINUX

				strcpy(Dst, "/usr");
				Status = true;

				#elif defined BEOS

				strsafecpy(Dst, "/boot/apps", DstSize);
				Status = true;

				#endif
				break;
			}
			case LSP_LOCAL_APP_DATA:
			{
				#if defined WIN32
				char *f = GetWin32Folder(CSIDL_LOCAL_APPDATA);
				if (f)
				{
					strsafecpy(Dst, f, DstSize);
					DeleteArray(f);
					Status = true;
				}

				#elif defined MAC
				
				FSRef Ref;
				OSErr e = FSFindFolder(kOnAppropriateDisk, kPreferencesFolderType, kDontCreateFolder, &Ref);
				if (e) printf("%s:%i - FSFindFolder failed e=%i\n", __FILE__, __LINE__, e);
				else
				{
					char *u = FSRefPath(Ref);
					if (u)
					{
						strsafecpy(Dst, u, DstSize);
						DeleteArray(u);
						Status = true;
					}
				}

				#elif defined LINUX

				strcpy(Dst, "/usr/local");
				Status = true;

				#endif
				break;
			}
			case LSP_DESKTOP:
			{
				#if defined WIN32

				char *f = GetWin32Folder(CSIDL_DESKTOPDIRECTORY);
				if (f)
				{
					strsafecpy(Dst, f, DstSize);
					DeleteArray(f);
					Status = true;
				}

				#elif defined MAC
				
				FSRef Ref;
				OSErr e = FSFindFolder(kOnAppropriateDisk, kDesktopFolderType, kDontCreateFolder, &Ref);
				if (e) printf("%s:%i - FSFindFolder failed e=%i\n", __FILE__, __LINE__, e);
				else
				{
					char *u = FSRefPath(Ref);
					if (u)
					{
						strsafecpy(Dst, u, DstSize);
						DeleteArray(u);
						Status = true;
					}
				}

				#elif defined POSIX

				struct passwd *pw = getpwuid(getuid());
				if (pw)
				{
					#ifdef LINUX
					WindowManager wm = LgiGetWindowManager();
					if (wm == WM_Gnome)
						sprintf(Dst, "%s/.gnome-desktop", pw->pw_dir);
					else
					#endif
						sprintf(Dst, "%s/Desktop", pw->pw_dir);
						
					Status = true;
				}

				#elif defined BEOS

				strsafecpy(Dst, "/boot/home/Desktop", DstSize);
				Status = true;
				
				#endif
				break;
			}
			case LSP_HOME:
			{
				#if defined WIN32

				#ifndef CSIDL_PROFILE
				#define CSIDL_PROFILE 0x0028
				#endif 

				char *f = GetWin32Folder(CSIDL_PROFILE);
				if (f)
				{
					strsafecpy(Dst, f, DstSize);
					DeleteArray(f);
					Status = true;
				}

				#elif defined POSIX

				struct passwd *pw = getpwuid(getuid());
				if (pw)
				{
					strcpy(Dst, pw->pw_dir);
					Status = true;
				}

				#elif defined BEOS

				strsafecpy(Dst, "/boot/home", DstSize);
				Status = true;

				#endif
				break;
			}
			case LSP_EXE:
			{
				if (LgiGetExeFile(Dst, DstSize))
				{
					LgiTrimDir(Dst);
					Status = true;
				}
				break;
			}
			case LSP_TRASH:
			{
				#if defined LINUX
				
				switch (LgiGetWindowManager())
				{
					case WM_Kde:
					{
						static char KdeTrash[256] = "";
						
						if (NOT ValidStr(KdeTrash))
						{
							// Ask KDE where the current trash is...
							GProcess p;
							GStringPipe o;
							if (p.Run("kde-config", "--userpath trash", 0, true, 0, &o))
							{
								char *s = o.NewStr();
								if (s)
								{
									// Store it..
									strsafecpy(KdeTrash, s, sizeof(KdeTrash));
									DeleteArray(s);

									// Clear out any crap at the end...
									char *e = KdeTrash + strlen(KdeTrash) - 1;
									while (e > KdeTrash AND strchr(" \r\n\t/", *e))
									{
										*e-- = 0;
									}
								}
								else
								{
									printf("%s:%i - No output from 'kde-config'.\n", __FILE__, __LINE__);
								}
							}
							else
							{
								printf("%s:%i - Run 'kde-config' failed.\n", __FILE__, __LINE__);
							}
						}
						if (ValidStr(KdeTrash))
						{
							strsafecpy(Dst, KdeTrash, DstSize);
							Status = true;
						}
						break;
					}
					default:
					{
						printf("%s:%i - Unknown window manager.\n", __FILE__, __LINE__);
						break;
					}
				}
				
				#elif defined MAC
				
				FSRef Ref;
				OSErr e = FSFindFolder(kUserDomain, kTrashFolderType, kDontCreateFolder, &Ref);
				if (e) printf("%s:%i - FSFindFolder failed e=%i\n", __FILE__, __LINE__, e);
				else
				{
					char *u = FSRefPath(Ref);
					if (u)
					{
						strsafecpy(Dst, u, DstSize);
						DeleteArray(u);
						Status = true;
					}
				}

				#elif defined WIN32

				// This should work but doesn't... *shrug*
				// char *f = GetWin32Folder(CSIDL_BITBUCKET);

				#else

				#endif
				break;
			}
		}
	}

	return Status;
}

bool LgiGetExeFile(char *Dst, int DstSize)
{
	if (Dst)
	{
		#if defined WIN32

		if (LgiGetOs() == LGI_OS_WIN9X)
		{
			char Exe[256];
			if (GetModuleFileName(NULL, Exe, sizeof(Exe)) > 0)
			{
				char *e = LgiFromNativeCp(Exe);
				if (e)
				{
					strlwr(e);
					strcpy(Dst, e);
					DeleteArray(e);
					return true;
				}
				else
				{
					LgiMsg(0, "LgiFromNativeCp returned 0, ANSI CodePage=%i (%s)", "LgiGetExeFile Error", MB_OK, GetACP(), LgiAnsiToLgiCp());
				}
			}

			char m[256];
			sprintf(m, "GetModuleFileName failed err: %08.8X", GetLastError());
			MessageBox(0, m, "LgiGetExeFile Error", MB_OK);
			LgiExitApp();
		}
		else
		{
			char16 Exe[256];
			if (GetModuleFileNameW(NULL, Exe, sizeof(Exe)) > 0)
			{
				char *e = LgiNewUtf16To8(Exe);
				if (e)
				{
					strcpy(Dst, e);
					DeleteArray(e);
					return true;
				}
			}
		}

		#elif defined BEOS

		app_info Info;
		if (LgiApp->GetAppInfo(&Info) == B_OK)
		{
			BEntry e(&Info.ref);
			BPath p;
			if (e.GetPath(&p) == B_OK)
			{
				strcpy(Dst, p.Path());
				return true;
			}
		}

		#elif defined ATHEOS

		os::Directory AppPath;
		if (AppPath.SetTo("^/.") == 0)
		{
			std::string p;
			if (AppPath.GetPath(&p) == 0)
			{
				sprintf(Dst, "%s%s%s", p.c_str(), DIR_STR, LgiApp->_AppFile);
				return true;
			}
		}

		#elif defined LINUX

		static char ExePathCache[256] = "";
		bool Status = false;

		// this is _REALLY_ lame way to do it... but hey there aren't any
		// other better ways to get the full path of the running executable
		if (NOT ExePathCache[0])
		{
			// First try the self method
			int Len = readlink("/proc/self/exe", ExePathCache, sizeof(ExePathCache));
			Status = FileExists(ExePathCache);
			// printf("readlink=%i Status=%i Exe='%s'\n", Len, Status, ExePathCache);
			if (NOT Status)
			{
				ExePathCache[0] = 0;
				
				// Next try the map file method
				char ProcFile[256];
				sprintf(ProcFile, "/proc/%i/maps", getpid());
				int fd = open(ProcFile, O_RDONLY);

				if (fd >= 0)
				{
					int Len = 16 << 10; // is this enough?
										// no better way of determining the length of proc info?
					char *Buf = NEW(char[Len+1]);
					if (Buf)
					{
						int r = read(fd, Buf, Len);
						Buf[r] = 0;

						char *s = strchr(Buf, '/');
						if (s)
						{
							char *e = strchr(s, '\n');
							if (e)
							{
								*e = 0;
								strcpy(ExePathCache, s);
								Status = true;
							}
						}
						
						DeleteArray(Buf);
					}

					close(fd);
				}
				else
				{
					// Non proc system (like cygwin for example)
					// char Cmd[256];
					// sprintf(Cmd, "ps | grep \"%i\"", getpid());
					
					GStringPipe Ps;
					GProcess p;
					if (p.Run("ps", 0, 0, true, 0, &Ps))
					{
						char *PsOutput = Ps.NewStr();
						if (PsOutput)
						{
							GToken n(PsOutput, "\r\n");
							for (int i=0; NOT Status AND i<n.Length(); i++)
							{
								GToken t(n[i], " \t\r\n");
								if (t.Length() > 7)
								{
									int LinePid = 0;
									for (int i=0; i<t.Length()-1; i++)
									{
										if (isdigit(t[i][0]))
										{
											LinePid = atoi(t[i]);
											break;
										}
									}

									if (LinePid == getpid())
									{
										strcpy(ExePathCache, t[t.Length()-1]);
										Status = true;
									}
								}
							}
							
							DeleteArray(PsOutput);
						}					
					}
				}
			}
		}
		else Status = true;

		if (Status)
		{
			strsafecpy(Dst, ExePathCache, DstSize);
		}
		
		return Status;
		
		#elif defined MAC
		
		bool Status = false;
		ProcessSerialNumber ps;
		OSErr e = GetCurrentProcess(&ps);
		if (!e)
		{
			FSRef fs;
			OSStatus s = GetProcessBundleLocation(&ps, &fs);
			if (!s)
			{
				char *u = FSRefPath(fs);
				if (u)
				{
					strsafecpy(Dst, u, DstSize);
					Status = true;
					DeleteArray(u);
				}
				else
				{
					printf("%s:%i - FSGetCatalogInfo failed (e=%i).\n", __FILE__, __LINE__, e);
				}
			}
			else
			{
				printf("%s:%i - GetProcessBundleLocation failed (e=%i).\n", __FILE__, __LINE__, s);
			}
		}
		else
		{
			printf("%s:%i - GetCurrentProcess failed (e=%i).\n", __FILE__, __LINE__, e);
		}
		
		return Status;
		
		#endif
	}

	return false;
}

bool LgiGetExePath(char *Dst, int DstSize)
{
	return LgiGetSystemPath(LSP_EXE, Dst, DstSize);
}

char *LgiGetExtension(char *File)
{
	if (File)
	{
		char *Dot = strrchr(File, '.');
		if (Dot AND NOT strchr(Dot, DIR_CHAR))
		{
			return Dot + 1;
		}
	}

	return 0;
}

char *LgiFindFile(char *Name)
{
	char *Result = 0;
	
	if (Name)
	{
		char Exe[256];
		LgiGetExePath(Exe, sizeof(Exe));
		
		#ifdef MAC
		GToken p(Exe, "/");
		Exe[0] = 0;
		for (int i=0; i<p.Length(); i++)
		{
			if (strcmp(p[i], "build") == 0)
				break;
			sprintf(Exe+strlen(Exe), "/%s", p[i]);
		}
		#endif

		char *PrefPath[] =
		{
			".",
			"..",
			"Code",
			"Resources",
			"../Code",
			"../Resources",
			"../Debug",
			"../Release",
			"../../Debug",
			"../../Release",
			0
		};

		// Look in prefered paths first...
		for (char **Pref = PrefPath; *Pref; Pref++)
		{
			char Path[256];
			LgiMakePath(Path, sizeof(Path), Exe, *Pref);
			LgiMakePath(Path, sizeof(Path), Path, Name);
			
			if (FileExists(Path))
			{
				return NewStr(Path);
			}

			#ifdef WIN32
			strcat(Path, ".lnk");
			if (ResolveShortcut(Path, Path, sizeof(Path)))
			{
				return NewStr(Path);
			}
			#endif
		}

		// General search fall back...
		List<char> Ext, Files;
		Ext.Insert(Name);
		if (LgiRecursiveFileSearch(Exe, &Ext, &Files) AND
			Files.Length())
		{
			Result = Files.First();
			Files.Delete(Result);
		}
		Files.DeleteArrays();
	}

	return Result;
}

uint64 LgiCurrentTime()
{
	#if defined WIN32

	return GetTickCount(); //timeGetTime()

	#elif defined BEOS

	return system_time() / 1000;

	#elif defined MAC
	
	UnsignedWide t;
	Microseconds(&t);
	uint64 i = ((uint64)t.hi << 32) | t.lo;
	return i / 1000;

	#else

	timeval tv;
	gettimeofday(&tv, 0);
	return (tv.tv_sec * 1000) + (tv.tv_usec / 1000);

	#endif
}

int LgiIsReleaseBuild()
{
	#ifdef _DEBUG
	return 0;
	#else
	return 1;
	#endif
}

void LgiFormatSize(char *Str, uint64 Size)
{
	int64 K = 1024;
	int64 M = K * K;
	int64 G = K * M;

	if (Size == 1)
	{
		strcpy(Str, "1 byte");
	}
	else if (Size < K)
	{
		sprintf(Str, "%u bytes", (int)Size);
	}
	else if (Size < 10 * K)
	{
		double d = (int64)Size;
		sprintf(Str, "%.2f K", d / K);
	}
	else if (Size < M)
	{
		sprintf(Str, "%u K", (int) (Size / K));
	}
	else if (Size < G)
	{
		double d = (int64)Size;
		sprintf(Str, "%.2f M", d / M);
	}
	else
	{
		double d = (int64)Size;
		sprintf(Str, "%.2f G", d / G);
	}
	
// printf("FormatSize " LGI_PrintfInt64 "='%s'\n", Size, Str);
}

char *LgiTokStr(char *&s)
{
	char *Status = 0;

	if (s AND *s)
	{
		// Skip whitespace
		static char Delim[] = ", \t\r\n";
		while (*s AND strchr(Delim, *s)) s++;
		if (*s)
		{
			if (strchr("\'\"", *s))
			{
				char Delim = *s++;
				char *e = strchr(s, Delim);
				if (NOT e)
				{
					// error, no end delimiter
					e = s;
					while (*e) e++;
				}
				Status = NewStr(s, (int)e-(int)s);
				s = *e ? e + 1 : e;
			}
			else
			{
				char *e = s;
				while (*e AND NOT strchr(Delim, *e)) e++;
				Status = NewStr(s, (int)e-(int)s);
				s = e;
			}
		}		
	}
	
	return Status;
}

//////////////////////////////////////////////////////////////////////////////
DoEvery::DoEvery(int p) // p = timeout in ms
{
	Init(p);
}

void DoEvery::Init(int p) // p = timeout in ms
{
	LastTime = LgiCurrentTime();
	if (p > 0)
	{
		Period = p;
	}
}

bool DoEvery::DoNow()
{
	int64 Now = LgiCurrentTime();
	if (LastTime + Period < Now)
	{
		LastTime = Now;
		return true;
	}
	return false;
}

///////////////////////////////////////////////////////////////////////////////////
enum
{
	GFillSolid = 1,
	GFillBitmap = 2,
};

class GViewFillPriv
{
public:
	uint8 Type;
	COLOUR c32;
	GSurface *pDC;

	GViewFillPriv()
	{
		pDC = 0;
	}

	~GViewFillPriv()
	{
		DeleteObj(pDC);
	}
};

GViewFill::GViewFill(COLOUR c, int Bits)
{
	d = NEW(GViewFillPriv);
	d->Type = GFillSolid;
	#ifndef LGI_STATIC
	d->c32 = CBit(32, c, Bits);
	#endif
}

GViewFill::GViewFill(GSurface *dc, GViewFillFlags Flags)
{
	d = NEW(GViewFillPriv);
	d->Type = GFillBitmap;
	#ifndef LGI_STATIC
	if (Flags & GViewFillCopyBitmap)
	{
		d->pDC = NEW(GMemDC(dc));
	}
	else
	{
		d->pDC = dc;
	}
	#endif
}

GViewFill::GViewFill(GViewFill &f)
{
	d = NEW(GViewFillPriv);
	d->Type = f.d->Type;
	if (d->Type == GFillBitmap)
	{
		#ifndef LGI_STATIC
		d->pDC = NEW(GMemDC(f.d->pDC));
		#endif
	}
	else
	{
		d->c32 = f.d->c32;
	}
}

GViewFill::~GViewFill()
{
	DeleteObj(d);
}

COLOUR GViewFill::GetFlatC32()
{
	return d->c32;
}

void GViewFill::Fill(GSurface *pDC, GRect *r, GdcPt2 *Origin)
{
	#ifndef LGI_STATIC
	if (d->Type == GFillSolid)
	{
		pDC->Colour(d->c32, 32);
		pDC->Rectangle(r);
	}
	else if (d->Type == GFillBitmap)
	{
		if (d->pDC)
		{
			GRect a;
			if (NOT r)
			{
				a.ZOff(pDC->X()-1, pDC->Y()-1);
				r = &a;
			}

			for (int y = Origin ? (Origin->y % pDC->Y()) - pDC->Y() : 0; y < r->Y(); y += d->pDC->Y())
			{
				for (int x = Origin ? (Origin->x % pDC->X()) - pDC->X() : 0; x<r->X(); x += d->pDC->X())
				{
					pDC->Blt(r->x1 + x, r->y1 + y, d->pDC);
				}
			}
		}
		else
		{
			LgiAssert(0);
		}
	}
	#endif
}
