#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "GMem.h"
#include "LgiOsDefs.h"
#include "GString.h"
#include "GContainers.h"
#include "LgiCommon.h"

char WhiteSpace[] = " \t\r\n";

// 8 Bit
char *strnchr(char *s, char c, int Len)
{
	if (s)
	{
		for (int i=0; i<Len; i++)
		{
			if (s[i] == c)
			{
				return s + i;
			}
		}
	}

	return 0;
}

char *strnstr(char *a, char *b, int n)
{
	if (a AND b)
	{
		int SLen = strlen(b);
		int DLen = strlen(a);

		while (SLen <= DLen AND n >= SLen)
		{
			int i;
			for (i=0; i<SLen AND a[i] == b[i]; i++);
			if (i == SLen) return a;

			n--;
			DLen--;
			a++;
		}
	}

	return NULL;
}

char *strnistr(char *a, char *b, int n)
{
	if (a AND b)
	{
		int SLen = strlen(b);
		while (n >= SLen)
		{
			int i;
			for (i=0; a[i] AND i<SLen AND tolower(a[i]) == tolower(b[i]); i++);
			if (i == SLen) return a;
			if (a[i] == 0) return 0;

			n--;
			a++;
		}
	}

	return NULL;
}

int strnicmp(char *a, char *b, int i)
{
	int Cmp = -1;
	if (a AND b AND i > 0)
	{
		for (Cmp = 0; i-- AND Cmp == 0; )
		{
			Cmp += tolower(*a) - tolower(*b);
			if (NOT *a OR NOT *b) break;
			a++;
			b++;
		}
	}
	
	return Cmp;
}

char *stristr(char *a, char *b)
{
	if (a AND b)
	{
		while (a AND *a)
		{
			if (tolower(*a) == tolower(*b))
			{
				int i;
				for (i=0; a[i] AND tolower(a[i]) == tolower(b[i]); i++)
					;

				if (b[i] == 0) return a;
			}
			a++;
		}
	}

	return NULL;
}

char *strsafecpy(char *dst, char *src, int len)
{
	char *d = dst;

	if (dst AND src AND len > 0)
	{
		while (*src AND len > 1)
		{
			*dst++ = *src++;
			len--;
		}
		
		*dst++ = 0;
	}
	
	return d;
}

char *strsafecat(char *dst, char *src, int len)
{
	char *d = dst;

	if (dst AND src AND len > 0)
	{
		// Seek past the dst string
		while (*dst AND len > 1)
		{
			dst++;
			len--;
		}
		
		// Copy the source onto the end
		while (*src AND len > 1)
		{
			*dst++ = *src++;
			len--;
		}
		
		*dst++ = 0;
	}
	
	return d;
}

int strcmp(char *a, char *b)
{
	int c = -1;
	if (a AND b)
	{
		c = 0;
		while (NOT c)
		{
			c = *a - *b;
			if (NOT *a OR NOT *b) break;
			a++;
			b++;
		}
	}
	return c;
}

int stricmp(char *a, char *b)
{
	int c = -1;
	if (a AND b)
	{
		c = 0;
		while (NOT c)
		{
			c = tolower(*a) - tolower(*b);
			if (NOT *a OR NOT *b) break;
			a++;
			b++;
		}
	}
	return c;
}

#ifndef WIN32
char *strupr(char *a)
{
	for (char *s = a; s AND *s; s++)
	{
		*s = tolower(*s);
	}

	return a;
}

char *strlwr(char *a)
{
	for (char *s = a; s AND *s; s++)
	{
		*s = tolower(*s);
	}

	return a;
}
#endif

char *TrimStr(char *s, char *Delim)
{
	if (s)
	{
		char *Start = s;
		while (*Start AND strchr(Delim, *Start))
		{
			Start++;
		}

		int StartLen = strlen(Start);
		if (StartLen > 0)
		{
			char *End = Start + strlen(Start) - 1;
			while (*End AND End > Start AND strchr(Delim, *End))
			{
				End--;
			}

			if (*Start)
			{
				int Len = (uint) End - (uint) Start + 1;
				char *n = NEW(char[Len+1]);
				if (n)
				{
					memcpy(n, Start, Len);
					n[Len] = 0;
					return n;
				}
			}
		}
	}

	return NULL;
}

bool ValidStr(char *s)
{
	if (s)
	{
		while (*s)
		{
			if (*s != ' ' AND
				*s != '\t' AND
				*s != '\r' AND
				*s != '\n' AND
				((uchar)*s) != 0xa0)
			{
				return true;
			}

			s++;
		}
	}

	return false;
}

char *NewStr(char *s, int Len)
{
	if (s)
	{
		if (Len < 0) Len = strlen(s);
		char *Ret = NEW(char[Len+1]);
		if (Ret)
		{
			if (Len > 0) memcpy(Ret, s, Len);
			Ret[Len] = 0;
			return Ret;
		}
	}
	return NULL;
}

#ifdef BEOS
#define toupper(c) ( ((c)>='a' AND (c)<='z') ? (c)-'a'+'A' : (c) )
#define tolower(c) ( ((c)>='A' AND (c)<='Z') ? (c)-'A'+'a' : (c) )
#endif

bool MatchStr(char *Template, char *Data)
{
	if (NOT Template)
	{
		return false;
	}

	if (stricmp(Template, (char*)"*") == 0)
	{
		// matches anything
		return true;
	}

	if (NOT Data)
	{
		return false;
	}

	while (*Template AND *Data)
	{
		if (*Template == '*')
		{
			Template++;

			if (*Template)
			{
				char *EndA;
				for (EndA = Template; *EndA AND *EndA!='?' AND *EndA!='*'; EndA++);
				int SegLen = (int)EndA-(int)Template;
				char *Seg = NewStr(Template, SegLen);
				if (NOT Seg) return false;

				// find end of non match
				while (*Data)
				{
					if (strnicmp(Data, Seg, SegLen) == 0)
					{
						break;
					}
					Data++;
				}

				DeleteArray(Seg);

				if (*Data)
				{
					Template += SegLen;
					Data += SegLen;
				}
				else
				{
					// can't find any matching part in Data
					return false;
				}
			}
			else
			{
				// '*' matches everything
				return true;
			}

		}
		else if (*Template == '?')
		{
			Template++;
			Data++;
		}
		else
		{
			if (tolower(*Template) != tolower(*Data))
			{
				return false;
			}

			Template++;
			Data++;
		}
	}

	return	((*Template == 0) OR (stricmp(Template, (char*)"*") == 0)) AND
			(*Data == 0);
}

int htoi(char *a)
{
	int Status = 0;

	for (; a AND *a; a++)
	{
		if (*a >= '0' AND *a <= '9')
		{
			Status <<= 4;
			Status |= *a - '0';
		}
		else if (*a >= 'a' AND *a <= 'f')
		{
			Status <<= 4;
			Status |= *a - 'a' + 10;
		}
		else if (*a >= 'A' AND *a <= 'F')
		{
			Status <<= 4;
			Status |= *a - 'A' + 10;
		}
		else break;
	}

	return Status;
}

int64 htoi64(char *a)
{
	int64 Status = 0;

	for (; a AND *a; a++)
	{
		if (*a >= '0' AND *a <= '9')
		{
			Status <<= 4;
			Status |= *a - '0';
		}
		else if (*a >= 'a' AND *a <= 'f')
		{
			Status <<= 4;
			Status |= *a - 'a' + 10;
		}
		else if (*a >= 'A' AND *a <= 'F')
		{
			Status <<= 4;
			Status |= *a - 'A' + 10;
		}
		else break;
	}

	return Status;
}

//////////////////////////////////////////////////////////////////////////
// UTF-16
#define CompareDefault	(a AND b ? *a - *b : -1)

char16 *StrchrW(char16 *s, char16 c)
{
	if (s)
	{
		while (*s)
		{
			if (*s == c) return s;
			s++;
		}
	}

	return 0;
}

char16 *StrnchrW(char16 *s, char16 c, int Len)
{
	if (s)
	{
		while (*s AND Len > 0)
		{
			if (*s == c)
				return s;
			s++;
			Len--;
		}
	}

	return 0;
}

char16 *StrrchrW(char16 *s, char16 c)
{
	char16 *Last = 0;

	while (s AND *s)
	{
		if (*s == c)
			Last = s;
		s++;
	}

	return Last;
}

char16 *StrstrW(char16 *a, char16 *b)
{
	if (a AND b)
	{
		int Len = StrlenW(b);
		for (char16 *s=a; *s; s++)
		{
			if (*s == *b)
			{
				// check match
				if (StrncmpW(s+1, b+1, Len-1) == 0)
					return s;
			}
		}
	}

	return 0;
}

char16 *StristrW(char16 *a, char16 *b)
{
	if (a AND b)
	{
		int Len = StrlenW(b);
		for (char16 *s=a; *s; s++)
		{
			if (*s == *b)
			{
				// check match
				if (StrnicmpW(s+1, b+1, Len-1) == 0)
					return s;
			}
		}
	}

	return 0;
}

char16 *StrnstrW(char16 *a, char16 *b, int n)
{
	if (a AND b)
	{
		int Len = StrlenW(b);
		for (char16 *s=a; n >= Len AND *s; s++, n--)
		{
			if (*s == *b)
			{
				// check match
				if (StrncmpW(s+1, b+1, Len-1) == 0)
					return s;
			}
		}
	}

	return 0;
}

char16 *StrnistrW(char16 *a, char16 *b, int n)
{
	if (a AND b)
	{
		int Len = StrlenW(b);
		for (char16 *s=a; n >= Len AND *s; s++, n--)
		{
			if (*s == *b)
			{
				// check match
				if (StrnicmpW(s+1, b+1, Len-1) == 0)
					return s;
			}
		}
	}

	return 0;
}

int StrcmpW(char16 *a, char16 *b)
{
	if (a AND b)
	{
		while (true)
		{
			if (NOT *a OR NOT *b OR *a != *b)
				return *a - *b;

			a++;
			b++;
		}

		return 0;
	}

	return -1;
}

int StricmpW(char16 *a, char16 *b)
{
	if (a AND b)
	{
		while (true)
		{
			char16 A = tolower(*a);
			char16 B = tolower(*b);

			if (NOT A OR NOT B OR A != B)
				return A - B;

			a++;
			b++;
		}

		return 0;
	}

	return -1;
}

int StrncmpW(char16 *a, char16 *b, int n)
{
	if (a AND b)
	{
		while (n > 0)
		{
			if (NOT *a OR NOT *b OR *a != *b)
				return *a - *b;

			a++;
			b++;
			n--;
		}

		return 0;
	}

	return -1;
}

int StrnicmpW(char16 *a, char16 *b, int n)
{
	if (a AND b)
	{
		while (n > 0)
		{
			char16 A = tolower(*a);
			char16 B = tolower(*b);

			if (NOT A OR NOT B OR A != B)
				return A - B;

			a++;
			b++;
			n--;
		}

		return 0;
	}

	return -1;
}

char16 *StrcpyW(char16 *a, char16 *b)
{
	if (a AND b)
	{
		do
		{
			*a++ = *b++;
		}
		while (*b);

		*a++ = 0;
	}

	return a;
}

char16 *StrncpyW(char16 *a, char16 *b, int n)
{
	if (a AND b AND n > 0)
	{
		while (*b AND --n > 0)
		{
			*a++ = *b++;
		}

		*a++ = 0; // always NULL terminate
	}

	return a;
}

void StrcatW(char16 *a, char16 *b)
{
	if (a AND b)
	{
		// Seek to end of string
		while (*a)
		{
			*a++;
		}

		// Append 'b'
		while (*b)
		{
			*a++ = *b++;
		}

		*a++ = 0;
	}
}

int StrlenW(char16 *a)
{
	int i = 0;
	if (a)
	{
		while (*a++)
		{
			i++;
		}
	}

	return i;
}

int HtoiW(char16 *a)
{
	int i = 0;
	if (a)
	{
		while (*a)
		{
			if (*a >= '0' AND *a <= '9')
			{
				i <<= 4;
				i |= *a - '0';
			}
			else if (*a >= 'a' AND *a <= 'f')
			{
				i <<= 4;
				i |= *a - 'a' + 10;
			}
			else if (*a >= 'A' AND *a <= 'F')
			{
				i <<= 4;
				i |= *a - 'A' + 10;
			}
			else break;
		}
	}
	return i;
}

char16 *TrimStrW(char16 *s, char16 *Delim)
{
	if (NOT Delim)
	{
		static char16 Def[] = {' ', '\r', '\n', '\t', 0};
		Delim = Def;
	}

	if (s)
	{
		// Leading delim
		while (StrchrW(Delim, *s))
		{
			s++;
		}

		// Trailing delim
		int i = StrlenW(s);
		while (i > 0 AND StrchrW(Delim, s[i-1]))
		{
			i--;
		}

		return NewStrW(s, i);
	}

	return 0;
}

bool MatchStrW(char16 *a, char16 *b)
{
	return 0;
}

char16 *NewStrW(char16 *Str, int l)
{
	char16 *s = 0;
	if (Str)
	{
		if (l < 0) l = StrlenW(Str);
		s = NEW(char16[l+1]);
		if (s)
		{
			memcpy(s, Str, l * sizeof(char16));
			s[l] = 0;
		}
	}
	return s;
}

bool ValidStrW(char16 *s)
{
	if (s)
	{
		for (char16 *c=s; *c; c++)
		{
			if (*c != ' ' AND *c != '\t') return true;
		}
	}

	return false;
}

char *LgiDecodeUri(char *uri, int len)
{
	GStringPipe p;
	if (uri)
	{
		char *end = len >= 0 ? uri + len : 0;
		for (char *s=uri; s AND *s; )
		{
			int Len;
			char *e = s;
			for (Len = 0; *e AND *e != '%' AND (NOT end OR e < end); e++)
				Len++;
			
			p.Push(s, Len);
			if ((NOT end OR e < end) AND *e)
			{
				e++;
				if (e[0] AND e[1])
				{
					char h[3] = { e[0], e[1], 0 };
					char c = htoi(h);
					p.Push(&c, 1);
					e += 2;
					s = e;
				}
				else break;
			}
			else
			{
				break;
			}
		}
	}
	return p.NewStr();
}

char *LgiEncodeUri(char *uri, int len)
{
	GStringPipe p;
	if (uri)
	{
		char *end = len >= 0 ? uri + len : 0;
		for (char *s=uri; s AND *s; )
		{
			int Len;
			char *e = s;
			for
			(
				Len = 0;
				*e
				AND
				(NOT end OR e < end)
				AND
				*e > ' '
				AND
				(uchar)*e < 0x7f
				AND
				strchr("$&+,/:;=?@\'\"<>#%{}|\\^~[]`", *e) == 0;
				e++
			)
			{
				Len++;
			}
			
			p.Push(s, Len);
			if ((NOT end OR e < end) AND *e)
			{
				char h[4];
				sprintf(h, "%%%02.2X", (uchar)*e);
				p.Push(h, 3);
				s = ++e;
			}
			else
			{
				break;
			}
		}
	}
	return p.NewStr();
}

