#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

#include "Lgi.h"
#include "vCard-vCal.h"

char *DeEscape(char *s, bool QuotedPrintable)
{
	char *i = s;
	char *o = s;
	while (*i)
	{
		if (*i == '\\')
		{
			i++;
			switch (*i)
			{
				case 'n':
				case 'N':
					*o++ = '\n';
					break;
				case ',':
				case ';':
				case ':':
					*o++ = *i;
					break;
				default:
					*o++ = '\\';
					i--;
					break;
			}
			
		}
		else if (QuotedPrintable AND *i == '=' AND i[1] AND i[2])
		{
			i++;
			char h[3] = { i[0], i[1], 0};
			*o++ = htoi(h);
			i++;
		}
		else
		{
			*o++ = *i;
		}

		i++;
	}
	*o = 0;
	return s;
}

/////////////////////////////////////////////////////////////
// General IO class
VIo::~VIo()
{
	Object.DeleteObjects();
}

bool VIo::IsBeginEnd()
{
	if (Var.Length() == 2)
	{
		if (stricmp(Var[0], "begin") == 0)
		{
			Object.Insert(NewStr(Var[1]));
			return true;
		}
		else if (stricmp(Var[0], "end") == 0)
		{
			char *o=Object.Last();
			if (o AND stricmp(o, Var[1]) == 0)
			{
				Object.Delete(o);
				DeleteArray(o);
			}
			return true;
		}
	}

	return false;
}

bool VIo::IsObject(char *Obj)
{
	if (Obj)
	{
		for (char *o=Object.Last(); o; o=Object.Prev())
		{
			if (stricmp(Obj, o) == 0)
			{
				return true;
			}
		}
	}

	return false;
}

bool VIo::IsTyped()
{
	Type.Empty();
	Value.Empty();
	Variables.Empty();

	char *Semi = Var[0] ? strchr(Var[0], ';') : 0;
	if (Semi)
	{
		GToken V(Var[0], ";", false);
		*Semi++ = 0;

		if (strchr(Semi, '='))
		{
			for (int i=0; i<V.Length(); i++)
			{
				if (strnicmp(V[i], "type=", 5) == 0)
				{
					GToken T(V[i]+5, ",");
					Type.AppendTokens(&T);
				}
				else if (strnicmp(V[i], "value=", 5) == 0)
				{
					GToken T(V[i]+6, ",");
					Value.AppendTokens(&T);
				}
				else
				{
					char *v = V[i];
					char *e = strchr(v, '=');
					if (e)
					{
						*e++ = 0;
						if (strchr("\'\"", *e))
						{
							char d = *e++;
							char *End = strchr(e, d);
							if (End) *End = 0;
						}
						Variables[v] = e;
					}
				}
			}
		}
		else
		{
			Type.Parse(Semi, ";", false);
		}

		return true;
	}

	return false;
}

bool VIo::IsEncoded()
{
	if (IsType("quoted-printable"))
	{
		if (Var[1])
		{
			char *In = Var[1];
			char *Out = In;
			while (*In)
			{
				if (*In == '=' AND In[1])
				{
					char Hex[3] = {In[1], In[2], 0};
					*Out++ = htoi(Hex);
					In += 3;
				}
				else
				{
					*Out++ = *In++;
				}
			}
			*Out++ = 0;
		}

		return true;
	}

	return false;
}

bool VIo::IsVar(char *a, char *b)
{
	if (Var[0] AND a)
	{
		if (stricmp(a, Var[0]) == 0)
		{
			if (Var[1] AND b)
			{
				return stricmp(Var[1], b) == 0;
			}
			else
			{
				return true;
			}
		}
	}

	return false;
}

bool VIo::IsType(char *type)
{
	for (int t=0; t<Type.Length(); t++)
	{
		if (type AND stricmp(Type[t], type) == 0)
		{
			return true;
		}
	}

	return false;
}

bool VIo::IsValue(char *value)
{
	for (int i=0; i<Value.Length(); i++)
	{
		if (value AND stricmp(Value[i], value) == 0)
		{
			return true;
		}
	}

	return false;
}

bool VIo::ParseDate(GDateTime &Out, char *In)
{
	bool Status = false;

	if (In)
	{
		Out.SetTimeZone(0, false);

		GToken v(In, "T");
		if (v.Length() > 0)
		{
			char *d = v[0];
			if (d AND strlen(d) == 8)
			{
				char Year[5] = {d[0], d[1], d[2], d[3], 0};
				char Month[3] = {d[4], d[5], 0};
				char Day[3] = {d[6], d[7], 0};
				Out.Year(atoi(Year));
				Out.Month(atoi(Month));
				Out.Day(atoi(Day));
				Status = true;
			}

			char *t = v[1];
			if (t AND strlen(t) >= 6)
			{
				char Hour[3] = {t[0], t[1], 0};
				char Minute[3] = {t[2], t[3], 0};
				char Second[3] = {t[4], t[5], 0};
				Out.Hours(atoi(Hour));
				Out.Minutes(atoi(Minute));
				Out.Seconds(atoi(Second));
				Status = true;
			}
		}
	}

	return Status;
}

bool VIo::ParseDuration(GDateTime &Out, int &Sign, char *In)
{
	bool Status = false;

	if (In)
	{
		Sign = 1;
		if (*In == '-')
		{
			Sign = -1;
			In++;
		}

		if (toupper(*In++) == 'P' AND
			toupper(*In++) == 'T')
		{
			while (*In)
			{
				int i = atoi(In);
				while (isdigit(*In)) In++;

				switch (toupper(*In++))
				{
					case 'W':
					{
						Out.Day(Out.Day() + (i * 7));
						break;
					}
					case 'D':
					{
						Out.Day(Out.Day() + i);
						break;
					}
					case 'H':
					{
						Out.Hours(Out.Hours() + i);
						break;
					}
					case 'M':
					{
						Out.Minutes(Out.Minutes() + i);
						break;
					}
					case 'S':
					{
						Out.Seconds(Out.Seconds() + i);
						break;
					}
				}
			}

			Status = true;
		}
	}

	return Status;
}

void VIo::Fold(GStream &o, char *i, int pre_chars)
{
	int x = pre_chars;
	for (char *s=i; s AND *s;)
	{
		if (x >= 74)
		{
			// wrapping
			o.Write(i, (int)s-(int)i);
			o.Write("\r\n\t", 3);
			x = 0;
			i = s;
		}
		else if (*s == '=' OR ((((uint8)*s) & 0x80) != 0))
		{
			// quoted printable
			o.Write(i, (int)s-(int)i);
			o.Print("=%02.2x", (uint8)*s);
			x += 3;
			i = ++s;
		}
		else if (*s == '\n')
		{
			// new line
			o.Write(i, (int)s-(int)i);
			o.Write("\\n", 2);
			x += 2;
			i = ++s;
		}
		else if (*s == '\r')
		{
			o.Write(i, (int)s-(int)i);
			i = ++s;
		}
		else
		{
			s++;
			x++;
		}
	}
	o.Write(i, strlen(i));
}

char *VIo::Unfold(char *In)
{
	if (In)
	{
		GStringPipe p(256);

		for (char *i=In; i AND *i; i++)
		{
			if (*i == '\n')
			{
				if (i[1] AND strchr(" \t", i[1]))
				{
					i++;
				}
				else
				{
					p.Push(i, 1);
				}
			}
			else
			{
				p.Push(i, 1);
			}
		}

		return p.NewStr();
	}

	return 0;
}

char *VIo::UnMultiLine(char *In)
{
	if (In)
	{
		GStringPipe p;
		char *n;
		for (char *i=In; i AND *i; i=n)
		{
			n = stristr(i, "\\n");
			if (n)
			{
				p.Push(i, (int)n-(int)i);
				p.Push("\n");
				n += 2;
			}
			else
			{
				p.Push(i);
			}
		}

		return p.NewStr();
	}

	return 0;
}

/////////////////////////////////////////////////////////////
// VCard class
bool VCard::Import(char *Text)
{
	bool Status = false;

	#if 1

	char *s = Text;
	char *Field = 0;
	char *Type = 0;
	char *Data = 0;
	#define ClearFields() \
		DeleteArray(Field); \
		DeleteArray(Type); \
		DeleteArray(Data)

	int PrefEmail = -1;
	GArray<char*> Emails;

	while (ReadField(s, &Field, &Type, &Data))
	{
		if (stricmp(Field, "begin") == 0 AND
			stricmp(Data, "vcard") == 0)
		{
			while (ReadField(s, &Field, &Type, &Data))
			{
				if (stricmp(Field, "end") == 0 AND
					stricmp(Field, "vcard") == 0)
					break;

				#define IsVar(str) (Field != 0 AND stricmp(Field, str) == 0)
				#define IsType(str) (Types.Find(str) != 0)
				GHashTable Types(0, false);
				GToken t(Type, ",");
				for (int i=0; i<t.Length(); i++)
					Types.Add(t[i]);

				if (IsVar("n"))
				{
					GToken Name(Data, ";", false);
					char *First = Name[1];
					char *Last = Name[0];
					char *Title = Name[3];

					if (First)
					{
						c->SetField(FIELD_FIRST_NAME, First);
						Status = true;
					}

					if (Last)
					{
						c->SetField(FIELD_LAST_NAME, Last);
						Status = true;
					}

					if (Title)
					{
						c->SetField(FIELD_TITLE, Title);
						Status = true;
					}
				}
				else if (IsVar("nickname"))
				{
					c->SetField(FIELD_NICK, Data);
				}
				else if (IsVar("tel"))
				{
					GToken Phone(Data, ";", false);
					for (int p=0; p<Phone.Length(); p++)
					{
						if (IsType("cell"))
						{
							if (IsType("Work"))
							{
								c->SetField(FIELD_WORK_MOBILE, Phone[p]);
							}
							else
							{
								c->SetField(FIELD_HOME_MOBILE, Phone[p]);
							}
						}
						else if (IsType("fax"))
						{
							if (IsType("Work"))
							{
								c->SetField(FIELD_WORK_FAX, Phone[p]);
							}
							else
							{
								c->SetField(FIELD_HOME_FAX, Phone[p]);
							}
						}
						else
						{
							if (IsType("Work"))
							{
								c->SetField(FIELD_WORK_PHONE, Phone[p]);
							}
							else
							{
								c->SetField(FIELD_HOME_PHONE, Phone[p]);
							}
						}
					}
				}
				else if (IsVar("email"))
				{
					if (IsType("pref"))
					{
						PrefEmail = Emails.Length();
					}
					Emails.Add(NewStr(Data));
				}
				else if (IsVar("org"))
				{
					GToken Org(Data, ";", false);
					if (Org[0]) c->SetField(FIELD_COMPANY, Org[0]);
				}
				else if (IsVar("adr"))
				{
					bool IsWork = IsType("work");
					bool IsHome = IsType("home");
					GToken Addr(Data, ";", false);
					if (Addr[2])
					{
						GToken A(Addr[2], "\r\n");
						if (A.Length() > 1)
						{
							c->SetField(IsWork ? FIELD_WORK_STREET : FIELD_HOME_STREET, A[0]);
							if (A[1]) c->SetField(IsWork ? FIELD_WORK_SUBURB : FIELD_HOME_SUBURB, A[1]);
							if (A[2]) c->SetField(IsWork ? FIELD_WORK_COUNTRY : FIELD_HOME_COUNTRY, A[2]);
						}
						else
						{
							c->SetField(IsWork ? FIELD_WORK_STREET : FIELD_HOME_STREET, Addr[2]);
						}
					}

					if (Addr[3]) c->SetField(IsWork ? FIELD_WORK_SUBURB : FIELD_HOME_SUBURB, Addr[3]);
					if (Addr[4]) c->SetField(IsWork ? FIELD_WORK_STATE : FIELD_HOME_STATE, Addr[4]);
					if (Addr[5]) c->SetField(IsWork ? FIELD_WORK_POSTCODE : FIELD_HOME_POSTCODE, Addr[5]);
					if (Addr[6]) c->SetField(IsWork ? FIELD_WORK_COUNTRY : FIELD_HOME_COUNTRY, Addr[6]);
				}
				else if (IsVar("note"))
				{
					c->SetField(FIELD_NOTE, Data);
				}
				else if (IsVar("uid"))
				{
					GToken n(Data, ";", false);
					char *Uid = n[0];
					if (isdigit(*Uid))
					{
						c->SetField(FIELD_UID, atoi(n[0]));
					}
					else
					{
						c->SetField(FIELD_UID, n[0]);
					}
				}
				else if (IsVar("x-perm"))
				{
					int Perms = atoi(Data);
					c->SetField(FIELD_PERMISSIONS, Perms);
				}
				else if (IsVar("url"))
				{
					bool IsWork = IsType("work");
					bool IsHome = IsType("home");
					if (IsWork)
					{
						c->SetField(FIELD_HOME_WEBPAGE, Data);
					}
					else if (IsHome)
					{
						c->SetField(FIELD_WORK_WEBPAGE, Data);
					}
				}
				else if (IsVar("nickname"))
				{
					c->SetField(FIELD_NICK, Data);
				}

				#undef IsVar
				#undef IsType

			}			
		}
	}

	if (Emails.Length())
	{
		if (PrefEmail < 0)
			PrefEmail = 0;

		c->SetField(FIELD_EMAIL_ADDR, Emails[PrefEmail]);
		Emails.DeleteAt(PrefEmail);

		if (Emails.Length())
		{
			GStringPipe p;
			for (int i=0; i<Emails.Length(); i++)
			{
				if (i) p.Print(",%s", Emails[i]);
				else p.Print("%s", Emails[i]);
			}

			char *v = p.NewStr();
			if (v)
			{
				c->SetField(FIELD_ALT_EMAIL, v);
				DeleteArray(v);
			}
		}
	}

	ClearFields();

	#else

	if (Text)
	{
		char *Unfolded = Unfold(Text);
		GToken L(Unfolded, "\r\n");
		if (L.Length() > 0)
		{
			for (int i=0; i<L.Length(); i++)
			{
				char *Line = L[i];
				if (NOT Line) break;
				char *Eol = Line+strlen(Line)-1;
				while (Eol > Line AND strchr(" \t", *Eol)) *Eol-- = 0;

				Var.Empty();

				char *Colon = strchr(L[i], ':');
				if (Colon)
				{
					*Colon++ = 0;
					GToken Temp(L[i], 0);
					Var.AppendTokens(&Temp);
					Temp.Parse(Colon, 0);
					Var.AppendTokens(&Temp);

					if (Var.Length() > 0)
					{
						if (NOT IsBeginEnd() AND
							IsObject("vcard"))
						{
							IsTyped();
							IsEncoded();

							if (IsVar("n"))
							{
								GToken Name(Var[1], ";", false);
								char *First = Name[1];
								char *Last = Name[0];
								char *Title = Name[3];

								if (First)
								{
									c->SetField(FIELD_FIRST_NAME, First);
									Status = true;
								}

								if (Last)
								{
									c->SetField(FIELD_LAST_NAME, Last);
									Status = true;
								}

								if (Title)
								{
									c->SetField(FIELD_TITLE, Title);
									Status = true;
								}
							}
							else if (IsVar("tel"))
							{
								GToken Phone(Var[1], ";", false);
								for (int p=0; p<Phone.Length(); p++)
								{
									if (IsType("cell"))
									{
										if (IsType("Work"))
										{
											c->SetField(FIELD_WORK_MOBILE, Phone[p]);
										}
										else
										{
											c->SetField(FIELD_HOME_MOBILE, Phone[p]);
										}
									}
									else if (IsType("fax"))
									{
										if (IsType("Work"))
										{
											c->SetField(FIELD_WORK_FAX, Phone[p]);
										}
										else
										{
											c->SetField(FIELD_HOME_FAX, Phone[p]);
										}
									}
									else
									{
										if (IsType("Work"))
										{
											c->SetField(FIELD_WORK_PHONE, Phone[p]);
										}
										else
										{
											c->SetField(FIELD_HOME_PHONE, Phone[p]);
										}
									}
								}
							}
							else if (IsVar("email"))
							{
								GToken Email(Var[1], ";", false);
								for (int e=0; e<Email.Length(); e++)
								{
									if (IsType("internet"))
									{
										c->SetField(FIELD_EMAIL_ADDR, Email[e]);
									}
								}
							}
							else if (IsVar("org"))
							{
								GToken Org(Var[1], ";", false);
								if (Org[0]) c->SetField(FIELD_COMPANY, Org[0]);
							}
							else if (IsVar("adr"))
							{
								bool IsWork = IsType("work");
								bool IsHome = IsType("home");
								GToken Addr(Var[1], ";", false);
								if (Addr[2])
								{
									GToken A(Addr[2], "\r\n");
									if (A.Length() > 1)
									{
										c->SetField(IsWork ? FIELD_WORK_STREET : FIELD_HOME_STREET, A[0]);
										if (A[1]) c->SetField(IsWork ? FIELD_WORK_SUBURB : FIELD_HOME_SUBURB, A[1]);
										if (A[2]) c->SetField(IsWork ? FIELD_WORK_COUNTRY : FIELD_HOME_COUNTRY, A[2]);
									}
									else
									{
										c->SetField(IsWork ? FIELD_WORK_STREET : FIELD_HOME_STREET, Addr[2]);
									}
								}

								if (Addr[3]) c->SetField(IsWork ? FIELD_WORK_SUBURB : FIELD_HOME_SUBURB, Addr[3]);
								if (Addr[4]) c->SetField(IsWork ? FIELD_WORK_STATE : FIELD_HOME_STATE, Addr[4]);
								if (Addr[5]) c->SetField(IsWork ? FIELD_WORK_POSTCODE : FIELD_HOME_POSTCODE, Addr[5]);
								if (Addr[6]) c->SetField(IsWork ? FIELD_WORK_COUNTRY : FIELD_HOME_COUNTRY, Addr[6]);
							}
							else if (IsVar("note"))
							{
								c->SetField(FIELD_NOTE, DeEscape(Var[1], true));
							}
							else if (IsVar("uid"))
							{
								GToken n(Var[1], ";", false);
								char *Uid = n[0];
								if (isdigit(*Uid))
								{
									c->SetField(FIELD_UID, atoi(n[0]));
								}
								else
								{
									c->SetField(FIELD_UID, n[0]);
								}
							}
							else if (IsVar("x-perm"))
							{
								int Perms = atoi(Var[1]);
								c->SetField(FIELD_PERMISSIONS, Perms);
							}
							else if (IsVar("url"))
							{
								bool IsWork = IsType("work");
								bool IsHome = IsType("home");
								if (IsWork)
								{
									c->SetField(FIELD_HOME_WEBPAGE, Var[i]);
								}
								else if (IsHome)
								{
									c->SetField(FIELD_WORK_WEBPAGE, Var[i]);
								}
								else
								{
									// ???
								}								
							}
							else if (IsVar("nickname"))
							{
								c->SetField(FIELD_NICK, Var[1]);
							}
						}
					}
				}
			}
		}

		DeleteArray(Unfolded);
	}

	#endif

	return Status;
}

bool VIo::ReadField(char *&s, char **Name, char **Type, char **Data)
{
	bool Status = false;

	DeleteArray(*Name);
	DeleteArray(*Type);
	DeleteArray(*Data);

	if (s AND Name AND Data)
	{
		GStringPipe p(256);

		// Unfold
		while (*s)
		{
			if (*s == '\r')
			{
				// do nothing
			}
			else if (*s == '\n')
			{
				// check for folding
				if (s[1] == ' ' OR s[1] == '\t')
				{
					// skip
					s++;
				}
				else
				{
					// end of field
					s++;
					break;
				}
			}
			else
			{
				p.Write(s, 1);
			}

			s++;
		}

		char *f = p.NewStr();
		if (f)
		{
			GHashTable Mod(0, false);
			char *e = strchr(f, ':');
			if (e)
			{
				*e++ = 0;
				GToken t(f, ";");
				if (t.Length() > 0)
				{
					*Name = NewStr(t[0]);
					for (int i=1; i<t.Length(); i++)
					{
						char *var = t[i];
						char *val = strchr(var, '=');
						if (val)
						{
							*val++ = 0;
							Mod.Add(var, NewStr(val));
						}
					}
				}
			}

			bool QuotedPrintable = false;
			char *Enc = (char*)Mod.Find("encoding");
			if (Enc)
			{
				if (stricmp(Enc, "quoted-printable") == 0)
				{
					QuotedPrintable = true;
				}
				else LgiTrace("%s:%i - Unknown encoding '%s'\n", __FILE__, __LINE__, Enc);
			}

			DeEscape(e, QuotedPrintable);

			char *Charset = (char*)Mod.Find("charset");
			if (Charset)
			{
				*Data = (char*)LgiNewConvertCp("utf-8", e, Charset);
			}
			else
			{
				*Data = NewStr(e);
			}

			Status = *Data != 0;

			if (*Type = (char*)Mod.Find("type"))
			{
				if (NOT Mod.Delete("type"))
				{
					LgiAssert(0);
				}
			}

			for (char *m = (char*)Mod.First(); m; m = (char*)Mod.Next())
			{
				DeleteArray(m);
			}

			DeleteArray(f);
		}
	}

	return Status;
}

void VIo::WriteField(GStream &s, char *Name, char *Type, char *Data)
{
	if (Name AND Data)
	{
		int64 Size = s.GetSize();

		if (Type)
			s.Print("%s;type=%s", Name, Type);
		else
			s.Print("%s", Name);
		
		for (uint8 *c = (uint8*)Data; *c; c++)
		{
			if (*c & 0x80)
			{
				s.Print(";charset=utf-8;encoding=quoted-printable");
				break;
			}
		}

		s.Print(":");
		Fold(s, Data, s.GetSize() - Size);
		s.Print("\r\n");
	}	
}

char *VCard::Export()
{
	GStringPipe o(256);
	char s[512];
	char *Empty = "";

	o.Push("begin:vcard\r\n");
	o.Push("version:3.0\r\n");

	char *First = 0, *Last = 0, *Title = 0;
	c->GetField(FIELD_FIRST_NAME, First);
	c->GetField(FIELD_LAST_NAME, Last);
	c->GetField(FIELD_TITLE, Title);
	if (First OR Last)
	{
		sprintf(s,
				"%s;%s;%s",
				Last?Last:Empty,
				First?First:Empty,
				Title?Title:Empty);
		WriteField(o, "n", 0, s);
	}

	#define OutputTypedField(Field, Name, Type)		\
		{ char *d = 0;								\
		if (c->GetField(Name, d))					\
		{											\
			WriteField(o, Field, Type, d);			\
		} }

	#define OutputField(Field, Name)				\
		{ char *d = 0;								\
		if (c->GetField(Name, d))					\
		{											\
			WriteField(o, Field, 0, d);				\
		} }

	OutputTypedField("tel", FIELD_WORK_PHONE, "Work");
	OutputTypedField("tel", FIELD_WORK_MOBILE, "Work,Cell");
	OutputTypedField("tel", FIELD_WORK_FAX, "Work,Fax");
	OutputTypedField("tel", FIELD_HOME_PHONE, "Home");
	OutputTypedField("tel", FIELD_HOME_MOBILE, "Home,Cell");
	OutputTypedField("tel", FIELD_HOME_FAX, "Home,Fax");
	OutputField("org", FIELD_COMPANY);
	OutputTypedField("email", FIELD_EMAIL_ADDR, "internet,pref");
	char *Alt;
	if (c->GetField(FIELD_ALT_EMAIL, Alt))
	{
		GToken t(Alt, ",");
		for (int i=0; i<t.Length(); i++)
		{
			WriteField(o, "email", "internet", t[i]);
		}
	}

	int Uid;
	if (c->GetField(FIELD_UID, Uid))
	{
		o.Print("UID:%i\r\n", Uid);
	}

	char *Street, *Suburb, *PostCode, *State, *Country;
	Street = Suburb = PostCode = State = Country = 0;
	c->GetField(FIELD_HOME_STREET, Street);
	c->GetField(FIELD_HOME_SUBURB, Suburb);
	c->GetField(FIELD_HOME_POSTCODE, PostCode);
	c->GetField(FIELD_HOME_STATE, State);
	c->GetField(FIELD_HOME_COUNTRY, Country);
	if (Street OR Suburb OR PostCode OR State OR Country)
	{
		snprintf(s, sizeof(s)-1,
				";;%s;%s;%s;%s;%s",
				Street?Street:Empty,
				Suburb?Suburb:Empty,
				State?State:Empty,
				PostCode?PostCode:Empty,
				Country?Country:Empty);
		WriteField(o, "adr", "home", s);
	}

	Street = Suburb = PostCode = State = Country = 0;
	c->GetField(FIELD_WORK_STREET, Street);
	c->GetField(FIELD_WORK_SUBURB, Suburb);
	c->GetField(FIELD_WORK_POSTCODE, PostCode);
	c->GetField(FIELD_WORK_STATE, State);
	c->GetField(FIELD_WORK_COUNTRY, Country);
	if (Street OR Suburb OR PostCode OR State OR Country)
	{
		snprintf(s, sizeof(s)-1,
				";;%s;%s;%s;%s;%s",
				Street?Street:Empty,
				Suburb?Suburb:Empty,
				State?State:Empty,
				PostCode?PostCode:Empty,
				Country?Country:Empty);
		WriteField(o, "adr", "work", s);
	}

	OutputField("X-Perm", FIELD_PERMISSIONS);

	char *Url;
	if (c->GetField(FIELD_HOME_WEBPAGE, Url))
	{
		WriteField(o, "url", "home", Url);
	}
	if (c->GetField(FIELD_WORK_WEBPAGE, Url))
	{
		WriteField(o, "url", "work", Url);
	}
	char *Nick;
	if (c->GetField(FIELD_NICK, Nick))
	{
		WriteField(o, "nickname", 0, Nick);
	}
	char *Note;
	if (c->GetField(FIELD_NOTE, Note))
	{
		WriteField(o, "note", 0, Note);
	}

	o.Push("end:vcard\r\n");

	return o.NewStr();
}

/////////////////////////////////////////////////////////////
// VCal class
bool VCal::Import(char *Text)
{
	bool Status = false;
	if (Text)
	{
		bool SetType = false;
		char *Unfolded = Unfold(Text);
		GToken L(Unfolded, "\r\n");
		if (L.Length() > 0)
		{
			for (int i=0; i<L.Length(); i++)
			{
				char *Line = L[i];
				if (NOT Line) break;
				char *Eol = Line+strlen(Line)-1;
				while (Eol > Line AND strchr(" \t", *Eol)) *Eol-- = 0;

				Var.Empty();

				char *Colon = strchr(L[i], ':');
				if (Colon)
				{
					*Colon++ = 0;
					GToken Temp(L[i], 0);
					Var.AppendTokens(&Temp);
					Temp.Parse(Colon, 0);
					Var.AppendTokens(&Temp);

					if (NOT IsBeginEnd() AND
						IsObject("vcalendar"))
					{
						Status = true;
						IsTyped();
						IsEncoded();

						bool IsAnEvent = IsObject("vevent");
						bool IsAToDo = IsObject("vtodo");
						if (IsAnEvent OR IsAToDo)
						{
							if (NOT SetType)
							{
								SetType = true;

								int Type = IsAToDo ? 1 : 0;
								c->SetField(FIELD_CAL_TYPE, Type);
							}

							if (IsObject("valarm"))
							{
								if (IsVar("trigger"))
								{
									if (IsValue("date-time"))
									{
										// don't really support this...
									}
									else
									{
										c->SetField(FIELD_CAL_REMINDER_ACTION, true);
										c->SetField(FIELD_CAL_REMINDER_TIME, 0); // reasonable default

										GDateTime Dt;
										int Sign = 0;
										if (ParseDuration(Dt, Sign, Var[1]))
										{
											int d = Dt.Day() * 24 * 60;
											int h = Dt.Hours() * 60;
											int m = Dt.Minutes();
											c->SetField(FIELD_CAL_REMINDER_TIME, -(d + h + m)); 
										}
									}
								}
							}
							else
							{
								if (IsVar("dtstart"))
								{
									GDateTime d;
									if (ParseDate(d, Var[1]))
									{
										c->SetField(FIELD_CAL_START_UTC, d);
									}
								}
								else if (IsVar("dtend"))
								{
									GDateTime d;
									if (ParseDate(d, Var[1]))
									{
										c->SetField(FIELD_CAL_END_UTC, d);
									}
								}
								else if (IsVar("summary"))
								{
									c->SetField(FIELD_CAL_SUBJECT, Var[1]);
								}
								else if (IsVar("description"))
								{
									char *Sum = UnMultiLine(Var[1]);
									if (Sum)
									{
										c->SetField(FIELD_CAL_NOTES, Sum);
										DeleteArray(Sum);
									}
								}
								else if (IsVar("location"))
								{
									c->SetField(FIELD_CAL_LOCATION, Var[1]);
								}
								else if (IsVar("uid"))
								{
									char *Uid = Var[1];
									if (isdigit(*Uid))
									{
										c->SetField(FIELD_UID, atoi(Uid));
									}
									else
									{
										c->SetField(FIELD_UID, Uid);
									}
								}
								else if (IsVar("x-showas"))
								{
									if (stricmp(Var[1], "TENTATIVE") == 0)
									{
										c->SetField(FIELD_CAL_SHOW_TIME_AS, 1);
									}
									else if (stricmp(Var[1], "BUSY") == 0)
									{
										c->SetField(FIELD_CAL_SHOW_TIME_AS, 2);
									}
									else if (stricmp(Var[1], "OUT") == 0)
									{
										c->SetField(FIELD_CAL_SHOW_TIME_AS, 3);
									}
									else
									{
										c->SetField(FIELD_CAL_SHOW_TIME_AS, 0);
									}
								}
								else if (IsVar("attendee"))
								{
									char *e = stristr(Var[1], "mailto=");
									if (e)
									{
										e += 7;

										char *Name = Variables["CN"];
										char *Role = Variables["Role"];

										VObject *a = c->CreateAttendee();
										if (a)
										{
											a->SetField(FIELD_ATTENDEE_NAME, Name);
											a->SetField(FIELD_ATTENDEE_EMAIL, e);
											if (stricmp(Role, "CHAIR") == 0)
											{
												a->SetField(FIELD_ATTENDEE_ATTENDENCE, 1);
											}
											else if (stricmp(Role, "REQ-PARTICIPANT") == 0)
											{
												a->SetField(FIELD_ATTENDEE_ATTENDENCE, 2);
											}
											else if (stricmp(Role, "OPT-PARTICIPANT") == 0)
											{
												a->SetField(FIELD_ATTENDEE_ATTENDENCE, 3);
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}

		DeleteArray(Unfolded);
	}

	return Status;
}

char *VCal::Export()
{
	GStringPipe o;
	char s[512];
	int Type = 0;
	c->GetField(FIELD_CAL_TYPE, Type);
	char *TypeStr = Type == 1 ? (char*)"VTODO" : (char*)"VEVENT";

	o.Push("BEGIN:VCALENDAR\r\n");
	o.Push("VERSION:1.0\r\n");

	if (c)
	{
		o.Print("BEGIN:%s\r\n", TypeStr);

		#define OutputStr(Field, Name)						\
		{													\
			char *_s = 0;									\
			if (c->GetField(Field, _s))						\
			{												\
				o.Push(Name ":");							\
				Fold(o, _s);								\
				o.Push("\r\n");								\
			}												\
		}

		OutputStr(FIELD_CAL_SUBJECT, "SUMMARY");
		OutputStr(FIELD_CAL_LOCATION, "LOCATION");
		OutputStr(FIELD_CAL_NOTES, "DESCRIPTION");

		int ShowAs;
		if (c->GetField(FIELD_CAL_SHOW_TIME_AS, ShowAs))
		{
			switch (ShowAs)
			{
				default:
				case 0:
				{
					o.Push("X-SHOWAS:FREE\r\n");
					break;
				}
				case 1:
				{
					o.Push("X-SHOWAS:TENTATIVE\r\n");
					break;
				}
				case 2:
				{
					o.Push("X-SHOWAS:BUSY\r\n");
					break;
				}
				case 3:
				{
					o.Push("X-SHOWAS:OUT\r\n");
					break;
				}
			}
		}

		int Uid;
		if (c->GetField(FIELD_UID, Uid))
		{
			sprintf(s, "UID:%i\r\n", Uid);
			o.Push(s);
		}

		GDateTime Dt;
		if (c->GetField(FIELD_CAL_START_UTC, Dt))
		{
			Dt.ToUtc();
			o.Print("DTSTART:%04.4i%02.2i%02.2iT%02.2i%02.2i%02.2i\r\n",
					Dt.Year(), Dt.Month(), Dt.Day(),
					Dt.Hours(), Dt.Minutes(), Dt.Seconds());
		}
		if (c->GetField(FIELD_CAL_END_UTC, Dt))
		{
			Dt.ToUtc();
			o.Print("DTEND:%04.4i%02.2i%02.2iT%02.2i%02.2i%02.2i\r\n",
					Dt.Year(), Dt.Month(), Dt.Day(),
					Dt.Hours(), Dt.Minutes(), Dt.Seconds());
		}

		int AlarmAction;
		int AlarmTime;
		if (c->GetField(FIELD_CAL_REMINDER_ACTION, AlarmAction) AND
			AlarmAction AND
			c->GetField(FIELD_CAL_REMINDER_TIME, AlarmTime))
		{
			o.Push("BEGIN:VALARM\r\n");
			sprintf(s, "TRIGGER:%sPT%iM\r\n", (AlarmTime<0) ? "-" : "", abs(AlarmTime));
			o.Push(s);
			o.Push("END:VALARM\r\n");
		}

		List<VObject> Attendees;
		if (c->GetAttendees(Attendees))
		{
			for (VObject *a=Attendees.First(); a; a=Attendees.Next())
			{
				// ATTENDEE;CN="Matthew Allen";ROLE=REQ-PARTICIPANT;RSVP=TRUE:MAILTO:matthew@cisra.canon.com.au

				int Attend = 2;
				char *Name = 0;
				char *Email = 0;
				int Response = 0;

				a->GetField(FIELD_ATTENDEE_ATTENDENCE, Attend);
				a->GetField(FIELD_ATTENDEE_NAME, Name);
				a->GetField(FIELD_ATTENDEE_EMAIL, Email);
				a->GetField(FIELD_ATTENDEE_RESPONSE, Response);

				if (Email)
				{
					strcpy(s, "ATTENDEE");
					if (Name)
					{
						sprintf(s+strlen(s), ";CN=\"%s\"", Name);
					}
					switch (Attend)
					{
						case 1:
						{
							strcat(s, ";ROLE=CHAIR");
							break;
						}
						default:
						case 2:
						{
							strcat(s, ";ROLE=REQ-PARTICIPANT");
							break;
						}
						case 3:
						{
							strcat(s, ";ROLE=OPT-PARTICIPANT");
							break;
						}
					}
					if (Email)
					{
						sprintf(s+strlen(s), ":MAILTO=%s", Email);
					}
					strcat(s, "\r\n");
					o.Push(s);
				}				
			}
		}

		o.Print("END:%s\r\n", TypeStr);
	}

	o.Push("END:VCALENDAR\r\n");

	return o.NewStr();
}
