#include "Lgi.h"
#include "GFilterUi.h"
#include "GPath.h"
#include "GEdit.h"

#define IconSize			15
#define TransparentBlk		Rgba32(0, 0, 0, 0)
#define White				Rgba32(255, 255, 255, 255)
#define Black				Rgba32(0, 0, 0, 255)
#define ContentColour		Rgba32(0, 0, 0, 160)

// Field size
#define SIZE_NOT			45
#define SIZE_FIELD_NAME		130
#define SIZE_OP				60
#define SIZE_VALUE			150

enum FilterIcon
{
	IconNewCond,
	IconNewAnd,
	IconNewOr,
	IconDelete,
	IconMoveUp,
	IconMoveDown,
	IconOptions,
	IconDropDown,
	IconBoolFalse,
	IconBoolTrue,
	IconMax,
};

#define IconAlpha	255

static COLOUR IconColour[IconMax] =
{
	Rgba32(192, 192, 192, IconAlpha), // new condition
	Rgba32(118, 160, 230, IconAlpha), // new and
	Rgba32(32, 224, 32, IconAlpha), // new or
	Rgba32(230, 90, 65, IconAlpha), // delete
	Rgba32(192, 192, 192, IconAlpha), // up
	Rgba32(192, 192, 192, IconAlpha), // down
	Rgba32(192, 192, 192, IconAlpha), // options
	Rgba32(192, 192, 192, IconAlpha), // drop down
	Rgba32(255, 192, 0, IconAlpha), // bool false
	Rgba32(255, 192, 0, IconAlpha), // bool true
};

static char *IconName[IconMax];

static char **GetIconNames()
{
	if (NOT IconName[0])
	{
		int i = 0;
		IconName[i++] = LgiLoadString(L_FUI_NEW_CONDITION, "New Condition");
		IconName[i++] = LgiLoadString(L_FUI_NEW_AND, "New And");
		IconName[i++] = LgiLoadString(L_FUI_NEW_OR, "New Or");
		IconName[i++] = LgiLoadString(L_FUI_DELETE, "Delete");
		IconName[i++] = LgiLoadString(L_FUI_MOVE_UP, "Move Up");
		IconName[i++] = LgiLoadString(L_FUI_MOVE_DOWN, "Move Down");
		i++;
		IconName[i++] = LgiLoadString(L_FUI_OPTIONS, "Options");
		i++;
		IconName[i++] = LgiLoadString(L_FUI_NOT, "Not");
	}

	return IconName;
}

void convert_to_nonpremul(GSurface *pDC)
{
	if (pDC AND pDC->GetBits() == 32)
	{
		for (int y=0; y<pDC->Y(); y++)
		{
			Pixel32 *p = (Pixel32*) (*pDC)[y];
			Pixel32 *e = p + pDC->X();

			while (p < e)
			{
				if (p->a > 0 && p->a < 255)
				{
					p->r = (int) p->r * 255 / p->a;
					p->g = (int) p->g * 255 / p->a;
					p->b = (int) p->b * 255 / p->a;
				}

				p++;
			}
		}
	}
}

COLOUR Mix32(COLOUR c1, COLOUR c2, double HowMuch1)
{
	int a = (int)(HowMuch1 * 255);
	int oma = 255 - a;
	int r = ((R32(c1) * a) + (R32(c2) * oma)) / 255;
	int g = ((G32(c1) * a) + (G32(c2) * oma)) / 255;
	int b = ((B32(c1) * a) + (B32(c2) * oma)) / 255;
	return Rgb32(r, g, b);
}

class GFilterViewPrivate
{
public:
	bool ShowLegend;
	GRect Info;
	GArray<GSurface*> Icons;
	FilterUi_Menu Callback;
	int CallbackData;

	void Draw(GSurface *pDC, COLOUR c, FilterIcon Icon)
	{
		// Clear to transparent black
		pDC->Colour(Rgba32(0, 0, 0, 0), 32);
		pDC->Rectangle();

		// Draw the icon backrgound
		double n = pDC->X() - 2;
		GPointF Ctr(n/2, n), Rim(n/2, 0);

		if (1)
		{
			// Shadow
			GPath p;
			double z = ((double)pDC->X()-0.5) / 2;
			p.Circle(z, z, z);
			GSolidBrush b(Rgba32(0, 0, 0, 40));
			p.Fill(pDC, b);
		}

		if (1)
		{
			GPath p;
			p.Circle(n/2, n/2, n/2);

			GBlendStop r[] =
			{
				{0.0, Mix32(c, White, 0.1)},
				{0.3, Mix32(c, White, 0.3)},
				{0.6, c},
				{1.0, Mix32(c, Black, 0.8)},
			};

			GRadialBlendBrush b(Ctr, Rim, CountOf(r), r);
			p.Fill(pDC, b);
		}

		if (1)
		{
			// Draw the highlight
			GPath p;
			p.Circle(n/2, n/2, n/2);
			GBlendStop Highlight[] =
			{
				{0.0, TransparentBlk},
				{0.7, TransparentBlk},
				{0.9, White},
			};

			GLinearBlendBrush b2(Ctr, Rim, CountOf(Highlight), Highlight);
			p.Fill(pDC, b2);
		}

		double k[4] =
		{
			n * 3 / 14,
			n * 3 / 7,
			n * 4 / 7,
			n * 11 / 14
		};

		int Plus[][2] =
		{
			{0, 1}, {0, 2}, {1, 2}, {1, 3}, {2, 3}, {2, 2}, {3, 2},
			{3, 1}, {2, 1}, {2, 0}, {1, 0}, {1, 1}, {0, 1}
		};

		// Draw the icon content
		switch (Icon)
		{
			case IconNewCond:
			case IconBoolTrue:
			{
				GPath p;
				p.Circle(n/2, n/2, n/6);
				GSolidBrush b(ContentColour);
				p.Fill(pDC, b);
				break;
			}
			case IconNewAnd:
			case IconDelete:
			{
				GPath p;
				GPointF Ctr(n/2, n/2);
				for (int i=0; i<CountOf(Plus); i++)
				{
					GPointF pt(k[Plus[i][0]], k[Plus[i][1]]);
					pt = pt - Ctr;
					pt.Rotate(LGI_DegToRad(45));
					pt = pt + Ctr;					
					if (i) p.LineTo(pt.x, pt.y);
					else p.MoveTo(pt.x, pt.y);
				}

				GSolidBrush b(ContentColour);
				p.Fill(pDC, b);
				break;
			}
			case IconNewOr:
			{
				GPath p;
				for (int i=0; i<CountOf(Plus); i++)
				{
					if (i) p.LineTo(k[Plus[i][0]], k[Plus[i][1]]);
					else p.MoveTo(k[Plus[i][0]], k[Plus[i][1]]);
				}
				GSolidBrush b(ContentColour);
				p.Fill(pDC, b);
				break;
			}
			case IconMoveUp:
			case IconMoveDown:
			{
				int Pt[][2] =
				{
					(int)(n/2), (int)(k[0]),
					(int)(k[0]), (int)(k[1]),
					(int)(k[1]), (int)(k[1]),
					(int)(k[1]), (int)(k[3]),
					(int)(k[2]), (int)(k[3]),
					(int)(k[2]), (int)(k[1]),
					(int)(k[3]), (int)(k[1]),
					(int)(n/2), (int)(k[0]),
				};

				GPath p;
				for (int i=0; i<CountOf(Pt); i++)
				{
					GPointF m(Pt[i][0]+0.5, Pt[i][1]+0.5);
					if (Icon == IconMoveDown)
						m.y = k[3] - (m.y - k[0]);
					if (i) p.LineTo(m.x, m.y);
					else p.MoveTo(m.x, m.y);
				}
				GSolidBrush b(ContentColour);
				p.Fill(pDC, b);
				break;
			}
			case IconOptions:
			{
				GPath p;
				for (int i=0; i<3; i++)
				{
					GPointF c(n * (1+i) / 4, n / 2);
					p.Circle(c, n/10);
				}
				
				GSolidBrush b(ContentColour);
				p.Fill(pDC, b);
				break;
			}
			case IconDropDown:
			{
				double x[3] =
				{
					n * 1 / 4,
					n * 2 / 4,
					n * 3 / 4,
				};
				double y[3] =
				{
					n * 2 / 6,
					n * 3 / 6,
					n * 4 / 6,
				};
				GPath p;
				p.MoveTo(x[0], y[0]+0.5);
				p.LineTo(x[1], y[2]+0.5);
				p.LineTo(x[2], y[0]+0.5);
				p.LineTo(x[0], y[0]+0.5);
				
				GSolidBrush b(ContentColour);
				p.Fill(pDC, b);
				break;
			}
		}

		if (1)
		{
			// Draw the outline
			GPath p;
			p.Circle(n/2, n/2, n/2);
			p.Circle(n/2, n/2, n/2 - 1);
			GSolidBrush b(Rgba32(64, 64, 64, 255));
			p.Fill(pDC, b);
		}

		if (1)
		{
			// Make new icons transparent
			switch (Icon)
			{
				case IconNewCond:
				case IconNewAnd:
				case IconNewOr:
				{
					for (int y=0; y<pDC->Y(); y++)
					{
						Pixel32 *p = (Pixel32*) (*pDC)[y];
						Pixel32 *e = p + pDC->X();
						while (p < e)
						{
							p->a = p->a / 2;
							p++;
						}
					}
					break;
				}
			}
		}
	}

	GFilterViewPrivate()
	{
		ShowLegend = true;
		Callback = 0;
		CallbackData = 0;

		for (int i=0; i<IconMax; i++)
		{
			Icons[i] = NEW(GMemDC);
			if (Icons[i] AND Icons[i]->Create(IconSize, IconSize, 32))
			{
				Draw(Icons[i], IconColour[i], (FilterIcon)i);
			}
		}
	}

	~GFilterViewPrivate()
	{
		Icons.DeleteObjects();
	}
};

///////////////////////////////////////////////////////////////////////////////
class GFilterItemPrivate
{
public:
	GFilterNode Node;
	GFilterViewPrivate *Data;
	GRect Btns[IconMax];
	GRect NotBtn;
	GRect FieldBtn, FieldDropBtn;
	GRect OpBtn, OpDropBtn;
	GRect ValueBtn;

	bool Not;
	char *Field, *Op, *Value;
	GEdit *FieldEd, *OpEd, *ValueEd;

	GFilterItemPrivate()
	{
		Not = false;
		Data = 0;
		Node = LNODE_NULL;
		EmptyRects();
		Field = Op = Value = 0;
		FieldEd = OpEd = ValueEd = 0;
	}

	~GFilterItemPrivate()
	{
		DeleteArray(Field);
		DeleteArray(Op);
		DeleteArray(Value);
		DeleteObj(FieldEd);
		DeleteObj(OpEd);
		DeleteObj(ValueEd);
	}

	void EmptyRects()
	{
		for (int i=0; i<CountOf(Btns); i++)
			Btns[i].ZOff(-1, -1);

		FieldBtn.ZOff(-1, -1);
		FieldDropBtn.ZOff(-1, -1);
		OpBtn.ZOff(-1, -1);
		OpDropBtn.ZOff(-1, -1);
		ValueBtn.ZOff(-1, -1);
	}
};

GFilterItem::GFilterItem(GFilterViewPrivate *Data, GFilterNode Node)
{
	d = NEW(GFilterItemPrivate);
	d->Data = Data;
	SetText("Node");
	SetNode(Node);
}

GFilterItem::~GFilterItem()
{
	DeleteObj(d);
}

bool GFilterItem::GetNot()
{
	return d->Not;
}

char *GFilterItem::GetField()
{
	if (d->FieldEd) return d->FieldEd->Name();

	return d->Field;
}

char *GFilterItem::GetOp()
{
	if (d->OpEd) return d->OpEd->Name();

	return d->Op;
}

char *GFilterItem::GetValue()
{
	if (d->ValueEd) return d->ValueEd->Name();

	return d->Value;
}

void GFilterItem::SetNot(bool b)
{
	d->Not = b;
	Update();
}

void GFilterItem::SetField(char *s)
{
	DeleteArray(d->Field);
	d->Field = NewStr(s);
	if (d->FieldEd) d->FieldEd->Name(s);
	else Update();
}

void GFilterItem::SetOp(char *s)
{
	DeleteArray(d->Op);
	d->Op = NewStr(s);
	if (d->OpEd) d->OpEd->Name(s);
	else Update();
}

void GFilterItem::SetValue(char *s)
{
	DeleteArray(d->Value);
	d->Value = NewStr(s);
	if (d->ValueEd) d->ValueEd->Name(s);
	else Update();
}

void GFilterItem::_PourText(GdcPt2 &Size)
{
	Size.y = SysFont->GetHeight() + 10;
	switch (d->Node)
	{
		case LNODE_NEW:
		case LNODE_AND:
		case LNODE_OR:
			Size.x = 120;
			break;
		default:
			Size.x = 126 + SIZE_NOT + SIZE_FIELD_NAME + SIZE_OP + SIZE_VALUE;
			break;
	}
}

#define StartEdit(Rc, Ed, Name) \
	if (NOT d->Ed) \
	{ \
		d->Ed = NEW(GEdit(n++, c.x1 + Pos->x1 + d->Rc.x1, c.y1 + Pos->y1 + d->Rc.y1, \
								d->Rc.X(), d->Rc.Y(), \
								d->Name)); \
		if (d->Ed) d->Ed->Attach(GetTree()); \
	} \
	else \
	{ \
		GRect r = d->Rc; \
		r.Offset(c.x1 + Pos->x1, c.y1 + Pos->y1); \
		d->Ed->SetPos(r); \
	}

#define EndEdit(Rc, Ed, Var) \
	if (d->Ed) \
	{ \
		DeleteArray(d->Var); \
		d->Var = NewStr(d->Ed->Name()); \
		DeleteObj(d->Ed); \
	}

void GFilterItem::_PaintText(GSurface *pDC, COLOUR Fore, COLOUR Back)
{
	GRect *Pos = _GetRect(TreeItemText);

	// Create a memory context
	GMemDC Buf(Pos->X(), Pos->Y(), 32);
	Buf.Colour(LC_WORKSPACE, 24);
	Buf.Rectangle(0);

	// Draw the background
	double ox = 1, oy = 1;
	double r = (double)(Pos->Y() - (oy * 2)) / 2;
	COLOUR BackCol = 0;
	COLOUR Workspace = Rgb24To32(LC_WORKSPACE);
	switch (d->Node)
	{
		case LNODE_NEW:
			BackCol = Rgb24To32(LC_HIGH);
			break;
		case LNODE_COND:
		default:
			BackCol = Rgb24To32(LC_MED);
			break;
		case LNODE_AND:
		case LNODE_OR:
			BackCol = Mix32(IconColour[d->Node], White, 0.3);
			break;
	}

	if (Select())
	{
		GPath p;
		GSolidBrush b(Rgb24To32(LC_SELECTION));
		GRectF PosF(0, 0, Pos->X()-1, Pos->Y());
		p.RoundRect(PosF, PosF.Y()/2);
		p.Fill(&Buf, b);
	}

	{
		GPath p;
		GSolidBrush b(BackCol);
		GRectF PosF(0, 0, Pos->X()-(ox*2)-1, Pos->Y()-(oy*2));
		PosF.Offset(ox, oy);
		p.RoundRect(PosF, r);
		p.Fill(&Buf, b);
	}

	if (d->Node == LNODE_NEW)
	{
		GPath p;
		GSolidBrush b(Workspace);
		GRectF PosF(0, 0, Pos->X()-(ox*2)-3, Pos->Y()-(oy*2)-2);
		PosF.Offset(ox+1, oy+1);
		p.RoundRect(PosF, r);
		p.Fill(&Buf, b);
	}

	// Clear the button rects
	d->EmptyRects();

	// Draw the content
	int IconY = (int)(oy + r) - (IconSize / 2);
	switch (d->Node)
	{
		case LNODE_NEW:
		{
			Buf.Op(GDC_ALPHA);
			int x = Pos->X() - (int)r - IconSize + 4;
			int y = (int)(oy + r) - (IconSize / 2);

			// New OR
			d->Btns[IconNewOr].ZOff(IconSize-1, IconSize-1);
			d->Btns[IconNewOr].Offset(x, y);
			Buf.Blt(x, y, d->Data->Icons[IconNewOr]);
			x -= 2 + IconSize;

			// New AND
			d->Btns[IconNewAnd].ZOff(IconSize-1, IconSize-1);
			d->Btns[IconNewAnd].Offset(x, y);
			Buf.Blt(x, y, d->Data->Icons[IconNewAnd]);
			x -= 2 + IconSize;

			if (NOT IsRoot())
			{
				// New Condition
				d->Btns[IconNewCond].ZOff(IconSize-1, IconSize-1);
				d->Btns[IconNewCond].Offset(x, y);
				Buf.Blt(x, y, d->Data->Icons[IconNewCond]);
			}
			break;
		}
		default:
		{
			Buf.Op(GDC_ALPHA);
			int x = (int)((double)Pos->X() - r - (IconSize / 2) - 2);
			
			// Delete
			d->Btns[IconDelete].ZOff(IconSize-1, IconSize-1);
			d->Btns[IconDelete].Offset(x, IconY);
			Buf.Blt(x, IconY, d->Data->Icons[IconDelete]);
			x -= 2 + IconSize;
			
			if (NOT IsRoot())
			{
				// Move down
				d->Btns[IconMoveDown].ZOff(IconSize-1, IconSize-1);
				d->Btns[IconMoveDown].Offset(x, IconY);
				Buf.Blt(x, IconY, d->Data->Icons[IconMoveDown]);
				x -= 2 + IconSize;
				
				// Move up
				d->Btns[IconMoveUp].ZOff(IconSize-1, IconSize-1);
				d->Btns[IconMoveUp].Offset(x, IconY);
				Buf.Blt(x, IconY, d->Data->Icons[IconMoveUp]);
				x -= 2 + IconSize;
			}

			if (d->Node == LNODE_AND OR
				d->Node == LNODE_OR)
			{
				// Configure
				d->Btns[IconOptions].ZOff(IconSize-1, IconSize-1);
				d->Btns[IconOptions].Offset(x, IconY);
				Buf.Blt(x, IconY, d->Data->Icons[IconOptions]);
				x -= 2 + IconSize;
			}
			break;
		}
	}

	switch (d->Node)
	{
		case LNODE_COND:
		{
			// Layout stuff
			d->FieldBtn.ZOff(SIZE_FIELD_NAME, (int)(r * 2) - 5);
			d->FieldBtn.Offset((int)(ox + r)+SIZE_NOT, (int)oy + 2);
			d->FieldDropBtn.ZOff(IconSize-1, IconSize-1);
			d->FieldDropBtn.Offset(d->FieldBtn.x2 + 1, d->FieldBtn.y1 + ((d->FieldBtn.Y()-d->FieldDropBtn.Y())/2));
			d->NotBtn.ZOff(IconSize-1, IconSize-1);
			d->NotBtn.Offset(r, IconY);
			d->OpBtn.ZOff(SIZE_OP, d->FieldBtn.Y()-1);
			d->OpBtn.Offset(d->FieldDropBtn.x2 + 8, d->FieldBtn.y1);
			d->OpDropBtn.ZOff(IconSize-1, IconSize-1);
			d->OpDropBtn.Offset(d->OpBtn.x2 + 1, d->FieldDropBtn.y1);
			d->ValueBtn.ZOff(SIZE_VALUE, d->FieldBtn.Y()-1);
			d->ValueBtn.Offset(d->OpDropBtn.x2 + 8, d->FieldBtn.y1);

			// Draw stuff
			Buf.Op(GDC_ALPHA);
			Buf.Blt(d->NotBtn.x1, d->NotBtn.y1, d->Data->Icons[d->Not ? IconBoolTrue : IconBoolFalse]);
			Buf.Blt(d->OpDropBtn.x1, d->OpDropBtn.y1, d->Data->Icons[IconDropDown]);
			Buf.Blt(d->FieldDropBtn.x1, d->FieldDropBtn.y1, d->Data->Icons[IconDropDown]);
			Buf.Op(GDC_SET);

			SysFont->Transparent(true);
			SysFont->Colour(LC_TEXT, Back);
			SysFont->Text(&Buf, d->NotBtn.x2 + 3, d->NotBtn.y1, LgiLoadString(L_FUI_NOT, "Not"));

			if (Select())
			{
				GRect c = GetTree()->GetClient();
				int n = 1;

				StartEdit(FieldBtn, FieldEd, Field);
				StartEdit(OpBtn, OpEd, Op);
				StartEdit(ValueBtn, ValueEd, Value);
			}
			else
			{
				EndEdit(FieldBtn, FieldEd, Field);
				EndEdit(OpBtn, OpEd, Op);
				EndEdit(ValueBtn, ValueEd, Value);

				Buf.Colour(LC_WORKSPACE, 24);
				Buf.Rectangle(&d->FieldBtn);
				Buf.Rectangle(&d->OpBtn);
				Buf.Rectangle(&d->ValueBtn);

				int Tx = 6;
				int Ty = 2;

				if (d->Field)
					SysFont->Text(&Buf, d->FieldBtn.x1+Tx, d->FieldBtn.y1+Ty, d->Field, -1, &d->FieldBtn);
				if (d->Op)
					SysFont->Text(&Buf, d->OpBtn.x1+Tx, d->OpBtn.y1+Ty, d->Op, -1, &d->OpBtn);
				if (d->Value)
					SysFont->Text(&Buf, d->ValueBtn.x1+Tx, d->ValueBtn.y1+Ty, d->Value, -1, &d->ValueBtn);
			}
			break;
		}
		case LNODE_NEW:
		{
			SysBold->Transparent(true);
			SysBold->Colour(Rgb24To32(LC_MED), Workspace);
			SysBold->Text(	&Buf,
							(int)r - 3,
							(int)(oy + r) - (SysBold->GetHeight() / 2),
							LgiLoadString(L_FUI_NEW, "New"));
			break;
		}
		case LNODE_AND:
		{
			SysBold->Transparent(true);
			SysBold->Colour(Rgb32To24(IconColour[d->Node]), Rgb32To24(BackCol));
			SysBold->Text(	&Buf,
							(int)r - 3,
							(int)(oy + r) - (SysBold->GetHeight() / 2),
							LgiLoadString(L_FUI_AND, "And"));
			break;
		}
		case LNODE_OR:
		{
			SysBold->Transparent(true);
			SysBold->Colour(Rgb32To24(IconColour[d->Node]), Rgb32To24(BackCol));
			SysBold->Text(	&Buf,
							(int)r - 3, (int)(oy + r) - (SysBold->GetHeight() / 2),
							LgiLoadString(L_FUI_OR, "Or"));
			break;
		}
	}

	// Blt result to the screen
	pDC->Blt(Pos->x1, Pos->y1, &Buf);
}

void GFilterItem::OnMouseClick(GMouse &m)
{
	if (m.Down() AND m.Left())
	{
		GRect *Pos = _GetRect(TreeItemText);
		if (NOT Pos) return;

		int Hit = -1;
		for (int i=0; i<IconMax; i++)
		{
			if (d->Btns[i].Overlap(m.x - Pos->x1, m.y - Pos->y1))
			{
				Hit = i;
				break;
			}
		}

		GRect Rc, Client = GetTree()->GetClient();
		if (d->NotBtn.Overlap(m.x - Pos->x1, m.y - Pos->y1))
		{
			d->Not = NOT d->Not;
			Update();
		}

		if (d->Data->Callback)
		{
			if (d->FieldDropBtn.Overlap(m.x - Pos->x1, m.y - Pos->y1))
			{
				Rc = d->FieldDropBtn;
				Rc.Offset(Pos->x1 + Client.x1, Pos->y1 + Client.y1);
				d->Data->Callback(	(GFilterView*) GetTree(),
									this,
									FMENU_FIELD,
									Rc,
									d->Data->CallbackData);
			}
			else if (d->OpDropBtn.Overlap(m.x - Pos->x1, m.y - Pos->y1))
			{
				Rc = d->OpDropBtn;
				Rc.Offset(Pos->x1 + Client.x1, Pos->y1 + Client.y1);
				d->Data->Callback(	(GFilterView*) GetTree(),
									this,
									FMENU_OP,
									Rc,
									d->Data->CallbackData);
			}
		}

		int Delta = 1;
		switch (Hit)
		{
			case IconMoveUp:
			{
				Delta = -1;
				// Fall thru
			}
			case IconMoveDown:
			{
				GFilterItem *p = dynamic_cast<GFilterItem*>(GetParent());
				if (p)
				{
					GTreeItem *m = this;
					int Count = p->Items.Length();
					int Idx = p->Items.IndexOf(m);
					if (Idx + Delta >= 0 AND Idx + Delta < Count - 1)
					{
						GTreeItem *i = p->Items[Idx + Delta];
						p->Items.Delete(m);
						p->Items.Insert(m, Idx + Delta);
						p->_RePour();
						GetTree()->Invalidate();
					}
				}
				break;
			}
			case IconNewCond:
			{
				SetNode(LNODE_COND);
				break;
			}
			case IconNewAnd:
			{
				SetNode(LNODE_AND);
				break;
			}
			case IconNewOr:
			{
				SetNode(LNODE_OR);
				break;
			}
			case IconDelete:
			{
				if (IsRoot() AND GetTree())
				{
					GetTree()->Insert(NEW(GFilterItem(d->Data)));
				}

				delete this;
				break;
			}
			case IconOptions:
			{
				GSubMenu *s = NEW(GSubMenu);
				if (s)
				{
					s->AppendItem(LgiLoadString(L_FUI_AND, "And"), 1, true);
					s->AppendItem(LgiLoadString(L_FUI_OR, "Or"), 2, true);

					GRect r = d->Btns[IconOptions];
					r.Offset(Pos->x1 + Client.x1, Pos->y1 + Client.y1);
					GdcPt2 p(r.x1, r.y2+1);
					GetTree()->PointToScreen(p);

					int Cmd = s->Float(GetTree(), p.x, p.y, true);
					switch (Cmd)
					{
						case 1:
							d->Node = LNODE_AND;
							Update();
							break;
						case 2:
							d->Node = LNODE_OR;
							Update();
							break;
					}
				}
				break;
			}
		}
	}
}

GFilterNode GFilterItem::GetNode()
{
	return d->Node;
}

void GFilterItem::SetNode(GFilterNode n)
{
	if (d->Node != n)
	{
		if (d->Node == LNODE_NEW AND
			GetParent() AND
			NOT IsRoot())
		{
			GetParent()->Insert(NEW(GFilterItem(d->Data)));
		}

		d->Node = n;
		Update();

		if (d->Node == LNODE_AND OR
			d->Node == LNODE_OR)
		{
			Insert(NEW(GFilterItem(d->Data)));
			Expanded(true);
		}
	}
}

///////////////////////////////////////////////////////////////////////////////
GFilterView::GFilterView(FilterUi_Menu Callback, int Data) : GTree(-1, 0, 0, 100, 100)
{
	d = NEW(GFilterViewPrivate);
	d->Callback = Callback;
	d->CallbackData = Data;
	Insert(NEW(GFilterItem(d)));
}

GFilterView::~GFilterView()
{
	Empty();
	DeleteObj(d);
}

GFilterItem *GFilterView::Create(GFilterNode Node)
{
	return NEW(GFilterItem(d, Node));
}

GRect &GFilterView::GetClient(bool ClientCoods)
{
	static GRect c;

	c = GView::GetClient(ClientCoods);
	if (d->ShowLegend)
	{
		d->Info = c;
		d->Info.y2 = d->Info.y1 + SysFont->GetHeight() + 8;
		c.y1 += d->Info.y2 + 1;
	}
	else
	{
		d->Info.ZOff(-1, -1);
	}
	
	return c;
}

void GFilterView::OnPaint(GSurface *pDC)
{
	GTree::OnPaint(pDC);

	if (d->ShowLegend)
	{
		GMemDC Buf(d->Info.X(), d->Info.Y(), 32);

		Buf.Colour(LC_MED, 24);
		Buf.Rectangle(0, 0, Buf.X()-1, Buf.Y()-2);
		Buf.Colour(LC_LOW, 24);
		Buf.Line(0, Buf.Y()-1, Buf.X()-1, Buf.Y()-1);

		SysFont->Transparent(true);
		SysFont->Colour(LC_TEXT, LC_MED);
		SysBold->Transparent(true);
		SysBold->Colour(LC_TEXT, LC_MED);
		int x = 4, y = 4;

		char *Title = LgiLoadString(L_FUI_LEGEND, "Legend:");
		SysBold->Text(&Buf, x, y, Title);
		x += 8 + SysBold->X(Title);

		char **Ico = GetIconNames();
		for (int i=0; i<IconMax; i++)
		{
			if (Ico[i])
			{
				Buf.Op(GDC_ALPHA);
				Buf.Blt(x, y, d->Icons[i]);
				x += 3 + d->Icons[i]->X();
				Buf.Op(GDC_SET);
				SysFont->Text(&Buf, x, y, IconName[i]);
				x += 12 + SysFont->X(IconName[i]);
			}
		}

		pDC->Blt(d->Info.x1, d->Info.y1, &Buf);
	}
}

