#include <stdlib.h>
#include <stdio.h>

#include "Lgi.h"
#include "GSkinEngine.h"
#include "GRadioGroup.h"

///////////////////////////////////////////////////////////////////////////////////////////
// Radio group
class GRadioGroupPrivate
{
public:
	static int NextId;
	int Val;
	GDisplayString *Txt;

	GRadioGroupPrivate()
	{
		Txt = 0;
	}

	~GRadioGroupPrivate()
	{
		DeleteObj(Txt);
	}

	void Layout(GFont *f, char *s)
	{
		DeleteObj(Txt);
		Txt = NEW(GDisplayString(f, s));
	}
};

int GRadioGroupPrivate::NextId = 10000;

GRadioGroup::GRadioGroup(int id, int x, int y, int cx, int cy, char *name, int Init)
	: ResObject(Res_Group)
{
	d = NEW(GRadioGroupPrivate);
	Name(name);
	GRect r(x, y, x+cx, y+cy);
	SetPos(r);
	SetId(id);
	d->Val = Init;

	#ifdef WIN32
	CreateClass("Lgi.RadioGroup");
	#endif
}

GRadioGroup::~GRadioGroup()
{
	DeleteObj(d);
}

void GRadioGroup::OnAttach()
{
}

int GRadioGroup::OnEvent(GMessage *m)
{
	return GView::OnEvent(m);
}

bool GRadioGroup::Name(char *n)
{
	bool Status = GView::Name(n);
	d->Layout(GetFont(), GObject::Name());
	return Status;
}

bool GRadioGroup::NameW(char16 *n)
{
	bool Status = GView::NameW(n);
	d->Layout(GetFont(), GObject::Name());
	return Status;
}

void GRadioGroup::SetFont(GFont *Fnt, bool OwnIt)
{
	GView::SetFont(Fnt);
	d->Layout(GetFont(), GObject::Name());
	Invalidate();
}

void GRadioGroup::OnCreate()
{
	AttachChildren();
	Value(d->Val);
}

int64 GRadioGroup::Value()
{
	int i=0;
	for (	GViewI *w = Children.First();
			w;
			w = Children.Next())
	{
		GRadioButton *But = dynamic_cast<GRadioButton*>(w);
		if (But)
		{
			if (But->Value())
			{
				d->Val = i;
				break;
			}
			i++;
		}
	}

	return d->Val;
}

void GRadioGroup::Value(int64 Which)
{
	d->Val = Which;

	int i=0;
	for (	GViewI *w = Children.First();
			w;
			w = Children.Next())
	{
		GRadioButton *But = dynamic_cast<GRadioButton*>(w);
		if (But)
		{
			if (i == Which)
			{
				But->Value(true);
				break;
			}
			i++;
		}
	}
}

int GRadioGroup::OnNotify(GViewI *Ctrl, int Flags)
{
	GViewI *n = GetNotify() ? GetNotify() : GetParent();
	if (n)
	{
		if (dynamic_cast<GRadioButton*>(Ctrl))
		{
			return n->OnNotify(this, Flags);
		}
		else
		{
			return n->OnNotify(Ctrl, Flags);
		}
	}
	return 0;
}

void GRadioGroup::OnPaint(GSurface *pDC)
{
	if (GApp::SkinEngine AND
		TestFlag(GApp::SkinEngine->GetFeatures(), GSKIN_GROUP))
	{
		GSkinState State;
		State.pScreen = pDC;
		State.MouseOver = false;
		State.Text = &d->Txt;
		GApp::SkinEngine->OnPaint_GRadioGroup(this, &State);
	}
	else
	{
		pDC->Colour(LC_MED, 24);
		pDC->Rectangle();

		int y = d->Txt ? d->Txt->Y() : 12;
		GRect b(0, y/2, X()-1, Y()-1);
		LgiWideBorder(pDC, b, CHISEL);

		if (d->Txt)
		{
			GRect t;
			t.ZOff(d->Txt->X(), d->Txt->Y());
			t.Offset(6, 0);
			SysFont->Colour(LC_TEXT, LC_MED);
			SysFont->Transparent(false);
			d->Txt->Draw(pDC, t.x1, t.y1, &t);
		}
	}
}

GRadioButton *GRadioGroup::Append(int x, int y, char *name)
{
	GRadioButton *But = NEW(GRadioButton(d->NextId++, x, y, -1, -1, name));
	if (But)
	{
		Children.Insert(But);
	}

	return But;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Radio button
class GRadioButtonPrivate
{
public:
	bool Val;
	bool Over;
	GDisplayString *Txt;

	GRadioButtonPrivate()
	{
		Txt = 0;
	}

	~GRadioButtonPrivate()
	{
		DeleteObj(Txt);
	}

	void Layout(GFont *f, char *s)
	{
		DeleteObj(Txt);
		Txt = NEW(GDisplayString(f, s));
	}
};

GRadioButton::GRadioButton(int id, int x, int y, int cx, int cy, char *name)
	: ResObject(Res_RadioBox)
{
	d = NEW(GRadioButtonPrivate);
	Name(name);
	if (cx < 0 AND d->Txt) cx = d->Txt->X() + 26;
	if (cy < 0 AND d->Txt) cy = d->Txt->Y() + 4;

	GRect r(x, y, x+cx, y+cy);
	SetPos(r);
	SetId(id);
	d->Val = false;
	d->Over = false;
	SetTabStop(true);

	#ifdef WIN32
	CreateClass("Lgi.RadioButton");
	SetDlgCode(GetDlgCode() | DLGC_WANTARROWS);
	#endif
}

GRadioButton::~GRadioButton()
{
	DeleteObj(d);
}

void GRadioButton::OnAttach()
{
}

int GRadioButton::OnEvent(GMessage *m)
{
	return GView::OnEvent(m);
}

bool GRadioButton::Name(char *n)
{
	bool Status = GView::Name(n);
	d->Layout(GetFont(), GObject::Name());
	return Status;
}

bool GRadioButton::NameW(char16 *n)
{
	bool Status = GView::NameW(n);
	d->Layout(GetFont(), GObject::Name());
	return Status;
}

void GRadioButton::SetFont(GFont *Fnt, bool OwnIt)
{
	GView::SetFont(Fnt, OwnIt);
	d->Layout(GetFont(), GObject::Name());
	Invalidate();
}

int64 GRadioButton::Value()
{
	return d->Val;
}

void GRadioButton::Value(int64 i)
{
	if (d->Val != (bool)i)
	{
		if (i)
		{
			// remove the value from the currenly selected radio value
			GViewI *p = GetParent();
			if (p)
			{
				GViewIterator *l = p->IterateViews();
				if (l)
				{
					for (GViewI *c=l->First(); c; c=l->Next())
					{
						if (c != this)
						{
							GRadioButton *b = dynamic_cast<GRadioButton*>(c);
							if (b AND b->d->Val)
							{
								b->d->Val = false;
								b->Invalidate();
							}
						}
					}
					DeleteObj(l);
				}
			}
		}

		d->Val = i;
		Invalidate();

		if (i)
		{
			GViewI *n = GetNotify() ? GetNotify() : GetParent();
			if (n)
			{
				n->OnNotify(this, 0);
			}
		}
	}
}

void GRadioButton::OnMouseClick(GMouse &m)
{
	if (Enabled())
	{
		bool WasCapturing = IsCapturing();

		if (m.Down()) Focus(true);
		Capture(m.Down());
		d->Over = m.Down();
	
		GRect r(0, 0, X()-1, Y()-1);
		if (NOT m.Down() AND
			r.Overlap(m.x, m.y) AND
			WasCapturing)
		{
			Value(true);
		}
		else
		{
			Invalidate();
		}
	}
}

void GRadioButton::OnMouseEnter(GMouse &m)
{
	if (Enabled() AND IsCapturing())
	{
		d->Over = true;
		Invalidate();
	}
}

void GRadioButton::OnMouseExit(GMouse &m)
{
	if (Enabled() AND IsCapturing())
	{
		d->Over = false;
		Invalidate();
	}
}

bool GRadioButton::OnKey(GKey &k)
{
	bool Status = false;
	int Move = 0;

	switch (k.c16)
	{
		case VK_UP:
		case VK_LEFT:
		{
			if (k.Down())
			{
				Move = -1;
			}
			Status = true;
			break;
		}
		case VK_RIGHT:
		case VK_DOWN:
		{
			if (k.Down())
			{
				Move = 1;
			}
			Status = true;
			break;
		}
	}

	if (Move)
	{
		List<GRadioButton> Btns;
		GViewIterator *L = GetParent()->IterateViews();
		if (L)
		{
			for (GViewI *c=L->First(); c; c=L->Next())
			{
				GRadioButton *b = dynamic_cast<GRadioButton*>(c);
				if (b) Btns.Insert(b);
			}
			if (Btns.Length() > 1)
			{
				int Index = Btns.IndexOf(this);
				if (Index >= 0)
				{
					GRadioButton *n = Btns[(Index + Move + Btns.Length()) % Btns.Length()];
					if (n)
					{
						n->Focus(true);
					}
				}
			}
			DeleteObj(L);
		}
	}

	return Status;
}

void GRadioButton::OnFocus(bool f)
{
	Invalidate();
}

void GRadioButton::OnPaint(GSurface *pDC)
{
	if (GApp::SkinEngine AND
		TestFlag(GApp::SkinEngine->GetFeatures(), GSKIN_RADIO))
	{
		GSkinState State;
		State.pScreen = pDC;
		State.MouseOver = d->Over;
		State.Text = &d->Txt;
		GApp::SkinEngine->OnPaint_GRadioButton(this, &State);
	}
	else
	{
		GRect r(0, 0, X()-1, Y()-1);
		GRect c(0, 0, 12, 12);
		pDC->Colour(LC_MED, 24);
		pDC->Rectangle();

		bool e = Enabled();
		if (d->Txt)
		{
			GRect t = r;
			t.x1 = c.x2 + 1;
			
			int Off = e ? 0 : 1;
			SysFont->Colour(e ? LC_TEXT : LC_LIGHT, LC_MED);
			SysFont->Transparent(false);
			d->Txt->Draw(pDC, t.x1 + 5 + Off, t.y1 + Off, &t);

			if (NOT e)
			{
				SysFont->Transparent(true);
				SysFont->Colour(LC_LOW, LC_MED);
				d->Txt->Draw(pDC, t.x1 + 5, t.y1, &t);
			}
		}

		// Draw border
		pDC->Colour(LC_LOW, 24);
		pDC->Line(c.x1+1, c.y1+9, c.x1+1, c.y1+10);
		pDC->Line(c.x1, c.y1+4, c.x1, c.y1+8);
		pDC->Line(c.x1+1, c.y1+2, c.x1+1, c.y1+3);
		pDC->Line(c.x1+2, c.y1+1, c.x1+3, c.y1+1);
		pDC->Line(c.x1+4, c.y1, c.x1+8, c.y1);
		pDC->Line(c.x1+9, c.y1+1, c.x1+10, c.y1+1);

		pDC->Colour(LC_SHADOW, 24);
		pDC->Set(c.x1+2, c.y1+9);
		pDC->Line(c.x1+1, c.y1+4, c.x1+1, c.y1+8);
		pDC->Line(c.x1+2, c.y1+2, c.x1+2, c.y1+3);
		pDC->Set(c.x1+3, c.y1+2);
		pDC->Line(c.x1+4, c.y1+1, c.x1+8, c.y1+1);
		pDC->Set(c.x1+9, c.y1+2);

		pDC->Colour(LC_LIGHT, 24);
		pDC->Line(c.x1+11, c.y1+2, c.x1+11, c.y1+3);
		pDC->Line(c.x1+12, c.y1+4, c.x1+12, c.y1+8);
		pDC->Line(c.x1+11, c.y1+9, c.x1+11, c.y1+10);
		pDC->Line(c.x1+9, c.y1+11, c.x1+10, c.y1+11);
		pDC->Line(c.x1+4, c.y1+12, c.x1+8, c.y1+12);
		pDC->Line(c.x1+2, c.y1+11, c.x1+3, c.y1+11);

		/// Draw center
		pDC->Colour(d->Over OR NOT e ? LC_MED : LC_WORKSPACE, 24);
		pDC->Rectangle(c.x1+2, c.y1+4, c.x1+10, c.y1+8);
		pDC->Box(c.x1+3, c.y1+3, c.x1+9, c.y1+9);
		pDC->Box(c.x1+4, c.y1+2, c.x1+8, c.y1+10);

		// Draw value
		if (d->Val)
		{
			pDC->Colour(e ? LC_TEXT : LC_LOW, 24);
			pDC->Rectangle(c.x1+4, c.y1+5, c.x1+8, c.y1+7);
			pDC->Rectangle(c.x1+5, c.y1+4, c.x1+7, c.y1+8);
		}
	}
}

