/*hdr
**      FILE:           GView.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           23/4/98
**      DESCRIPTION:    Linux GView Implementation
**
**      Copyright (C) 1998-2003, Matthew Allen
**              fret@memecode.com
*/

#include <stdio.h>
#include <stdlib.h>

#include "Lgi.h"
#include "GDragAndDrop.h"
#include "GEdit.h"

#define ADJ_LEFT					1
#define ADJ_RIGHT					2
#define ADJ_UP						3
#define ADJ_DOWN					4

struct X11_INVALIDATE_PARAMS
{
	GView *View;
	GRect Rgn;
	bool Repaint;
};

// CursorData is a bitmap in an array of uint32's. This is generated from a graphics file:
// ./Code/cursors.png
//
// The pixel values are turned into C code by a program called i.Mage:
//	http://www.memecode.com/image.php
//
// Load the graphic into i.Mage and then go Edit->CopyAsCode
// Then paste the text into the CursorData variable at the bottom of this file.
//
// This saves a lot of time finding and loading an external resouce, and even having to
// bundle extra files with your application. Which have a tendancy to get lost along the
// way etc.
extern uint32 CursorData[];
GInlineBmp Cursors =
{
	300, 20, 8, CursorData
};

struct CursorInfo
{
public:
	GRect Pos;
	GdcPt2 HotSpot;
}
CursorMetrics[] =
{
	// up arrow
	{ GRect(0, 0, 8, 15),			GdcPt2(4, 0) },
	// cross hair
	{ GRect(20, 0, 38, 18),		GdcPt2(29, 9) },
	// hourglass
	{ GRect(40, 0, 51, 15),		GdcPt2(45, 8) },
	// I beam
	{ GRect(60, 0, 66, 17),		GdcPt2(63, 8) },
	// N-S arrow
	{ GRect(80, 0, 91, 16),		GdcPt2(85, 8) },
	// E-W arrow
	{ GRect(100, 0, 116, 11),		GdcPt2(108, 5) },
	// NW-SE arrow
	{ GRect(120, 0, 132, 12),		GdcPt2(126, 6) },
	// NE-SW arrow
	{ GRect(140, 0, 152, 12),		GdcPt2(146, 6) },
	// 4 way arrow
	{ GRect(160, 0, 178, 18),		GdcPt2(169, 9) },
	// Blank
	{ GRect(0, 0, 0, 0),			GdcPt2(0, 0) },
	// Vertical split
	{ GRect(180, 0, 197, 16),		GdcPt2(188, 8) },
	// Horizontal split
	{ GRect(200, 0, 216, 17),		GdcPt2(208, 8) },
	// Hand
	{ GRect(220, 0, 233, 13),		GdcPt2(225, 0) },
	// No drop
	{ GRect(240, 0, 258, 18),		GdcPt2(249, 9) },
	// Copy drop
	{ GRect(260, 0, 279, 19),		GdcPt2(260, 0) },
	// Move drop
	{ GRect(280, 0, 299, 19),		GdcPt2(280, 0) },
};

////////////////////////////////////////////////////////////////////////////
GMouse &_lgi_adjust_click_for_window(GMouse &Info, GViewI *Wnd)
{
	static GMouse Temp;
	 
	Temp = Info;
	if (Wnd)
	{
		GdcPt2 Offset(0, 0);
		if (Wnd->WindowVirtualOffset(&Offset))
		{
			Temp.x -= Offset.x;
			Temp.y -= Offset.y;
		}
	}
	
	return Temp;
}

////////////////////////////////////////////////////////////////////////////
class GPulseThread : public GThread
{
	// GView *View;
	OsView Hnd;
	int Length;

public:
	bool Loop;

	GPulseThread(GView *view, int len)
	{
		Loop = true;
		Hnd = view->Handle();
		Length = len;
		DeleteOnExit = true;
		
		Run();
	}
	
	int Main()
	{
		do
		{
			LgiSleep(Length);
			if (Loop)
			{
				LgiPostEvent(Hnd, M_X11_PULSE);
			}
		}
		while (Loop);
	}
};

////////////////////////////////////////////////////////////////////////////
void _lgi_yield()
{
	LgiApp->Run(false);
}

extern int IsAdjacent(GRect &a, GRect &b);
int IsAdjacent(GRect &a, GRect &b)
{
	if (	(a.x1 == b.x2 + 1) AND
		NOT (a.y1 > b.y2 OR a.y2 < b.y1))
	{
		return ADJ_LEFT;
	}

	if (	(a.x2 == b.x1 - 1) AND
		NOT (a.y1 > b.y2 OR a.y2 < b.y1))
	{
		return ADJ_RIGHT;
	}

	if (	(a.y1 == b.y2 + 1) AND
		NOT (a.x1 > b.x2 OR a.x2 < b.x1))
	{
		return ADJ_UP;
	}

	if (	(a.y2 == b.y1 - 1) AND
		NOT (a.x1 > b.x2 OR a.x2 < b.x1))
	{
		return ADJ_DOWN;
	}

	return 0;
}

extern GRect JoinAdjacent(GRect &a, GRect &b, int Adj);
GRect JoinAdjacent(GRect &a, GRect &b, int Adj)
{
	GRect t;

	switch (Adj)
	{
		case ADJ_LEFT:
		case ADJ_RIGHT:
		{
			t.y1 = max(a.y1, b.y1);
			t.y2 = min(a.y2, b.y2);
			t.x1 = min(a.x1, b.x1);
			t.x2 = max(a.x2, b.x2);
			break;
		}
		case ADJ_UP:
		case ADJ_DOWN:
		{
			t.y1 = min(a.y1, b.y1);
			t.y2 = max(a.y2, b.y2);
			t.x1 = max(a.x1, b.x1);
			t.x2 = min(a.x2, b.x2);
			break;
		}
	}

	return t;
}

////////////////////////////////////////////////////////////////////////////
bool LgiIsKeyDown(int Key)
{
	return false;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/*
GFlags::GFlags()
{
	WndFlags = GWF_VISIBLE | GWF_ENABLED;
	_BorderSize = 0;
}

bool GFlags::Enabled()
{
	return TestFlag(WndFlags, GWF_ENABLED);
}

void GFlags::Enabled(bool i)
{
	if (i) SetFlag(WndFlags, GWF_ENABLED);
	else ClearFlag(WndFlags, GWF_ENABLED);
}

bool GFlags::Visible()
{
	return TestFlag(WndFlags, GWF_VISIBLE);
}

void GFlags::Visible(bool i)
{
	if (i) SetFlag(WndFlags, GWF_VISIBLE);
	else ClearFlag(WndFlags, GWF_VISIBLE);
}

bool GFlags::Focus()
{
	return TestFlag(WndFlags, GWF_FOCUS);
}

void GFlags::Focus(bool i)
{
	if (i) SetFlag(WndFlags, GWF_FOCUS);
	else ClearFlag(WndFlags, GWF_FOCUS);
}

bool GFlags::DropTarget()
{
	return TestFlag(WndFlags, GWF_DROP_TARGET);
}

bool GFlags::DropTarget(bool i)
{
	if (i) SetFlag(WndFlags, GWF_DROP_TARGET);
	else ClearFlag(WndFlags, GWF_DROP_TARGET);
}

#define GWF_ALL (GWF_SUNKEN | GWF_FLAT | GWF_RAISED)

bool GFlags::Sunken()
{
	return TestFlag(WndFlags, GWF_SUNKEN);
}

void GFlags::Sunken(bool i)
{
	WndFlags &= ~GWF_ALL;
	if (i) SetFlag(WndFlags, GWF_SUNKEN);
	if (NOT _BorderSize)
	{
	    _BorderSize = i?2:0;
	}
}

bool GFlags::Flat()
{
	return TestFlag(WndFlags, GWF_FLAT);
}

void GFlags::Flat(bool i)
{
	WndFlags &= ~GWF_ALL;
	if (i) SetFlag(WndFlags, GWF_FLAT);
}

bool GFlags::Raised()
{
	return TestFlag(WndFlags, GWF_RAISED);
}

void GFlags::Raised(bool i)
{
	WndFlags &= ~GWF_ALL;
	if (i) SetFlag(WndFlags, GWF_RAISED);
	if (NOT _BorderSize)
	{
    	_BorderSize = i ? 2 : 0;
	}
}
*/

//////////////////////////////////////////////////////////////////////////////
class GViewIter : public GViewIterator
{
	Iterator<GViewI> i;

public:
	GViewIter(GView *v) : i(&v->Children)
	{
	}

	GViewI *First()
	{
		return i.First();
	}

	GViewI *Last()
	{
		return i.Last();
	}

	GViewI *Next()
	{
		return i.Next();
	}

	GViewI *Prev()
	{
		return i.Prev();
	}

	int Length()
	{
		return i.Length();
	}

	int IndexOf(GViewI *view)
	{
		return i.IndexOf(view);
	}

	GViewI *operator [](int Idx)
	{
		return i[Idx];
	}
};

////////////////////////////////////////////////////////////////////////////////////////////////////
class GViewPrivate
{
public:
	static bool		CursorSet;

	GPulseThread	*Pulse;
	GViewI			*ParentI;
	GView			*Parent;
	GViewI			*Notify;
	int				CtrlId;
	GdcPt2			MinimumSize;
	GDragDropTarget	*DropTarget;
	int				CursorId;
	static GView	*LastCursor;
	GFont			*Font;
	bool			FontOwn;
	GView			*Popup;
	GViewFill		*Foreground, *Background;

	#if defined XWIN
	Cursor			XCursor;
	#endif

	GViewPrivate()
	{
		Background = Foreground = 0;
		CursorId = 0;
		Pulse = 0;
		Parent = 0;
		ParentI = 0;
		Notify = 0;
		CtrlId = -1;
		DropTarget = 0;
		Font = 0;
		FontOwn = false;
		Popup = 0;
		
		#if defined XWIN
		XCursor = 0;
		#endif
	}
	
	~GViewPrivate()
	{
		DeleteObj(Background);
		DeleteObj(Foreground);
		DeleteObj(Popup);
	}

	GView *GetParent()
	{
		if (Parent)
			return Parent;
		
		if (ParentI)
			return ParentI->GetGView();

		return 0;
	}
};

bool GViewPrivate::CursorSet = false;
GView *GViewPrivate::LastCursor = 0;
GViewI *GView::_Capturing = 0;
GViewI *GView::_Over = 0;

GView::GView(OsView view)
{
	d = NEW(GViewPrivate);
	_View = view ? view : NEW(DefaultOsView(this));
	_Window = 0;
	_Lock = 0;
	_IsToolBar = false;
	_BorderSize = 0;
	WndFlags = GWF_VISIBLE | GWF_ENABLED;
	Script = 0;
	Pos.ZOff(-1, -1);
}

GView::~GView()
{
	if (_Over == this) _Over = 0;
	if (_Capturing == this) Capture(false);
	_Delete();
	DeleteObj(d);
}

GViewIterator *GView::IterateViews()
{
	return NEW(GViewIter(this));
}

bool GView::AddView(GViewI *v, int Where)
{
	LgiAssert(NOT Children.HasItem(v));
	return Children.Insert(v, Where);
}

bool GView::DelView(GViewI *v)
{
	return Children.Delete(v);
}

bool GView::HasView(GViewI *v)
{
	return Children.HasItem(v);
}

void GView::_Delete()
{
	if (_View)
	{
		if (NOT InThread())
		{
			// If we aren't in the GUI thread the XApplication could be processing events
			// for this window, and if we allow it to be deleted from under the executing
			// code then "Bad Things Happen(TM)". So we lock the application here to make
			// sure it's not in the event handling code. And THEN kill the XWidget's pointer
			// back to the GView after which the link between the objects is broken and
			// it's safe to continue deleting the GView.
			if (XApplication::GetApp()->Lock())
			{
				_View->_SetWndPtr(0);
				XApplication::GetApp()->Unlock();
			}
		}
		else
		{
			_View->_SetWndPtr(0);
		}
	}
	
	SetPulse();
	Pos.ZOff(-1, -1);

	GViewI *c;
	while (c = Children.First())
	{
		if (c->GetParent() != (GViewI*)this)
		{
			printf("Error: ~GView, child not attached correctly: %p(%s) Parent: %p(%s)\n",
				c, c->Name(), c->GetParent(), c->GetParent() ? c->GetParent()->Name() : "");
			Children.Delete(c);
		}

		DeleteObj(c);
	}

	Detach();
	if (_View)
	{
		_View->_SetDeleteMe();
	}
}

GWindow *GView::GetWindow()
{
	if (NOT _Window)
	{
		for (GView *w = dynamic_cast<GView*>(d->Parent); w; w = dynamic_cast<GView*>(w->GetParent()))
		{
			if (w->_Window)
			{
				_Window = w->_Window;
				break;
			}
		}
	}
		
	return dynamic_cast<GWindow*>(_Window);
}

GdcPt2 &GView::GetWindowBorderSize()
{
	static GdcPt2 p;
	return p;
}

bool GView::GetTabStop()
{
	return _View->isTabStop();
}

void GView::SetTabStop(bool i)
{
	_View->isTabStop(i);
}

GViewI *GView::GetParent()
{
	return d->Parent;
}

void GView::SetParent(GViewI *p)
{
	d->Parent = dynamic_cast<GView*>(p);
	d->ParentI = p;
}

void GView::SendNotify(int Data)
{
	GViewI *n = d->Notify ? d->Notify : d->Parent;
	if (n)
	{
		n->OnNotify(this, Data);
	}
}

GViewI *GView::GetNotify()
{
	return d->Notify;
}

void GView::SetNotify(GViewI *p)
{
	d->Notify = p;
}

OsWindow GView::WindowHandle()
{
	return (GetWindow()) ? GetWindow()->Wnd : 0;
}

bool GView::Lock(int TimeOut)
{
	bool Status = true;

	if (_Window AND
		_Window->_Lock)
	{
		if (TimeOut < 0)
		{
			Status = _Window->_Lock->Lock();
		}
		else
		{
			Status = _Window->_Lock->LockWithTimeout(TimeOut);
		}
	}

	return true;
}

void GView::Unlock()
{
	if (_Window AND
		_Window->_Lock)
	{
		_Window->_Lock->Unlock();
	}
}

void GView::OnMouseClick(GMouse &m)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnMouseClick(m);
	}
}

void GView::OnMouseEnter(GMouse &m)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnMouseEnter(m);
	}
}

void GView::OnMouseExit(GMouse &m)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnMouseExit(m);
	}
}

void GView::OnMouseMove(GMouse &m)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnMouseMove(m);
	}
}

void GView::OnMouseWheel(double Lines)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnMouseWheel(Lines);
	}
}

bool GView::OnKey(GKey &k)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnKey(k);
	}
	
	return false;
}

void GView::OnCreate()
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnCreate();
	}
}

void GView::OnDestroy()
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnDestroy();
	}
}

void GView::OnFocus(bool f)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnFocus(f);
	}
}

void GView::OnPulse()
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnPulse();
	}
}

void GView::OnPosChange()
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnPosChange();
	}
}

bool GView::OnRequestClose(bool OsShuttingDown)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnRequestClose(OsShuttingDown);
	}

	return true;
}

int GView::OnHitTest(int x, int y)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnHitTest(x, y);
	}

	return -1;
}

void GView::OnChildrenChanged(GViewI *Wnd, bool Attaching)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnChildrenChanged(Wnd, Attaching);
	}
}

void GView::OnPaint(GSurface *pDC)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnPaint(pDC);
	}
}

int GView::OnNotify(GViewI *Ctrl, int Flags)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnNotify(Ctrl, Flags);
	}
	else if (Ctrl AND d->Parent)
	{
		// default behaviour is just to pass the 
		// notification up to the parent
		return d->Parent->OnNotify(Ctrl, Flags);
	}

	return 0;
}

int GView::OnCommand(int Cmd, int Event, OsView Wnd)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnCommand(Cmd, Event, Wnd);
	}

	return 0;
}

GView *&GView::PopupChild()
{
	return d->Popup;
}

void GView::_Mouse(GMouse &m, bool Move)
{
	if
	(
		NOT _View
		OR
		(
			GetWindow()
			AND
			NOT GetWindow()->HandleViewMouse(this, m)
		)
	)
	{
		return;
	}

	if (_Capturing)
	{
		m.ToScreen();
		m.Target = _Capturing;
		m.ToView();
		
		if (Move)
		{
			GViewI *c = _Capturing;
			d->CursorSet = false;
			c->OnMouseMove(m);
			
			if (NOT d->CursorSet)
			{
				c->SetCursor(LCUR_Normal);
			}
		}
		else
		{
			_Capturing->OnMouseClick(m);
		}
	}
	else
	{
		if (Move)
		{
			/*
			if (d->LastCursor AND
				d->LastCursor != this)
			{
				printf("XUndefineCursor on old view.\n");
				XUndefineCursor(Handle()->XDisplay(), d->LastCursor->Handle()->handle());
				d->LastCursor = 0;
			}
			*/
			
			bool Change = false;
			GViewI *o = WindowFromPoint(m.x, m.y);
			if (_Over != o)
			{
				if (_Over)
				{
					_Over->OnMouseExit(_lgi_adjust_click_for_window(m, _Over));
				}

				_Over = o;

				if (_Over)
				{
					_Over->OnMouseEnter(_lgi_adjust_click_for_window(m, _Over));
				}
			}
		}
			
		GView *Target = dynamic_cast<GView*>(_Over ? _Over : this);
		GRect Client = Target->GView::GetClient();
		if (Target->Raised() OR Target->Sunken())
		{
			Client.Offset(Target->_BorderSize, Target->_BorderSize);
		}
		m = _lgi_adjust_click_for_window(m, Target);
		if (NOT Client.Valid() OR Client.Overlap(m.x, m.y))
		{
			if (Move)
			{
				d->CursorSet = false;
				Target->OnMouseMove(m);
				if (NOT d->CursorSet)
				{
					Target->SetCursor(LCUR_Normal);
				}
			}
			else
			{
				Target->OnMouseClick(m);
			}
		}
	}
}

void GView::OnNcPaint(GSurface *pDC, GRect &r)
{
	int Border = Sunken() OR Raised() ? _BorderSize : 0;
	if (Border == 2)
	{
		LgiWideBorder(pDC, r, Sunken() ? SUNKEN : RAISED);
	}
	else if (Border == 1)
	{
		LgiThinBorder(pDC, r, Sunken() ? SUNKEN : RAISED);
	}
	// r.Offset(Border, Border);
}

void GView::_Paint(GSurface *pDC)
{
	GSurface *Mem = 0;
	if (NOT pDC)
	{
		pDC = Mem = NEW(GScreenDC(_View));
	}
	
	GRect r(0, 0, Pos.X()-1, Pos.Y()-1);

	// Non-Client drawing
	GRect c = r;
	OnNcPaint(pDC, c);
	bool HasEdge = c != r;
	if (HasEdge)
	{
		pDC->Handle()->SetClient(&c);
	}

	// Paint client
	OnPaint(pDC);

	// Paint any virtual children
	int x, y;
	pDC->GetOrigin(x, y);
	GRect Clip = pDC->ClipRgn();

	Iterator<GViewI> ChildList(&Children); // just in case the children access the child list
	for (GViewI *w = ChildList.First(); w; w = ChildList.Next())
	{
		if (NOT w->Handle() AND w->Visible())
		{
			GRect r = w->GetPos();
			GRect c(0, 0, r.X()-1, r.Y()-1);

			// c.Offset(r.x1-x, r.y1-y);
			// c.Intersection(&Clip);
			
			pDC->SetOrigin(x - r.x1, y - r.y1);
			pDC->ClipRgn(&c);
			
			GView *v = dynamic_cast<GView*>(w);
			if (v) v->_Paint(pDC);
			else printf("%s:%i - cast failure\n", __FILE__, __LINE__);
			
			pDC->SetOrigin(x, y);
		}
	}	
	
	// Clean up
	if (HasEdge)
	{
		pDC->Handle()->SetClient(0);
	}	
	DeleteObj(Mem);
}

GRect &GView::GetClient(bool ClientSpace)
{
	int Edge = (Sunken() OR Raised()) ? _BorderSize : 0;

	static GRect c;
	if (ClientSpace)
	{
		c.ZOff(Pos.X() - 1 - (Edge<<1), Pos.Y() - 1 - (Edge<<1));
	}
	else
	{
		c = Pos;
		c.Size(Edge, Edge);
	}
	return c;
}

GRect *GView::FindLargest(GRegion &r)
{
	int Pixels = 0;
	GRect *Best = 0;
	static GRect Final;

	// do initial run through the list to find largest single region
	for (GRect *i = r.First(); i; i = r.Next())
	{
		int Pix = i->X() * i->Y();
		if (Pix > Pixels)
		{
			Pixels = Pix;
			Best = i;
		}
	}

	if (Best)
	{
		Final = *Best;
		Pixels = Final.X() * Final.Y();

		int LastPixels = Pixels;
		GRect LastRgn = Final;
		int ThisPixels = Pixels;
		GRect ThisRgn = Final;

		GRegion TempList;
		for (GRect *i = r.First(); i; i = r.Next())
		{
			TempList.Union(i);
		}
		TempList.Subtract(Best);

		do
		{
			LastPixels = ThisPixels;
			LastRgn = ThisRgn;

			// search for adjoining rectangles that maybe we can add
			for (GRect *i = TempList.First(); i; i = TempList.Next())
			{
				int Adj = IsAdjacent(ThisRgn, *i);
				if (Adj)
				{
					GRect t = JoinAdjacent(ThisRgn, *i, Adj);
					int Pix = t.X() * t.Y();
					if (Pix > ThisPixels)
					{
						ThisPixels = Pix;
						ThisRgn = t;

						TempList.Subtract(i);
					}
				}
			}
		} while (LastPixels < ThisPixels);

		Final = ThisRgn;
	}

	return &Final;
}

GRect *GView::FindSmallestFit(GRegion &r, int Sx, int Sy)
{
	int X = 1000000;
	int Y = 1000000;
	GRect *Best = 0;

	for (GRect *i = r.First(); i; i = r.Next())
	{
		if ((i->X() >= Sx AND i->Y() >= Sy) AND
			(i->X() < X OR i->Y() < Y))
		{
			X = i->X();
			Y = i->Y();
			Best = i;
		}
	}

	return Best;
}

GRect *GView::FindLargestEdge(GRegion &r, int Edge)
{
	GRect *Best = 0;

	for (GRect *i = r.First(); i; i = r.Next())
	{
		if (NOT Best)
		{
			Best = i;
		}

		if ((Edge & GV_EDGE_TOP) AND (i->y1 < Best->y1) OR
			(Edge & GV_EDGE_RIGHT) AND (i->x2 > Best->x2) OR
			(Edge & GV_EDGE_BOTTOM) AND (i->y2 > Best->y2) OR
			(Edge & GV_EDGE_LEFT) AND (i->x1 < Best->x1))
		{
			Best = i;
		}

		if (((Edge & GV_EDGE_TOP) AND (i->y1 == Best->y1) OR
			(Edge & GV_EDGE_BOTTOM) AND (i->y2 == Best->y2)) AND
			(i->X() > Best->X()))
		{
			Best = i;
		}

		if (((Edge & GV_EDGE_RIGHT) AND (i->x2 == Best->x2) OR
			(Edge & GV_EDGE_LEFT) AND (i->x1 == Best->x1)) AND
			(i->Y() > Best->Y()))
		{
			Best = i;
		}
	}

	return Best;
}

int64 GView::GetCtrlValue(int Id)
{
	GViewI *w = FindControl(Id);
	return (w) ? w->Value() : 0;
}

void GView::SetCtrlValue(int Id, int64 i)
{
	GViewI *w = FindControl(Id);
	if (w)
	{
		w->Value(i);
	}
}

char *GView::GetCtrlName(int Id)
{
	GViewI *w = FindControl(Id);
	return (w) ? w->Name() : 0;
}

void GView::SetCtrlName(int Id, char *s)
{
	GViewI *w = FindControl(Id);
	if (w)
	{
		w->Name(s);
	}
}

bool GView::GetCtrlEnabled(int Id)
{	
	GViewI *w = FindControl(Id);
	return (w) ? w->Enabled() : 0;
}

void GView::SetCtrlEnabled(int Id, bool Enabled)
{
	GViewI *w = FindControl(Id);
	if (w)
	{
		w->Enabled(Enabled);
	}
}

bool GView::GetCtrlVisible(int Id)
{
	GViewI *w = FindControl(Id);
	return (w) ? w->Visible() : 0;
}

void GView::SetCtrlVisible(int Id, bool i)
{
	GViewI *w = FindControl(Id);
	if (w)
	{
		w->Visible(i);
	}
}

int GView::GetId()
{
	return d->CtrlId;
}

void GView::SetId(int i)
{
	d->CtrlId = i;
}

bool GView::Sunken()
{
	return TestFlag(WndFlags, GWF_SUNKEN);
}

void GView::Sunken(bool i)
{
	if (i)
	{
		SetFlag(WndFlags, GWF_SUNKEN);
		if (NOT _BorderSize)
		{
			_BorderSize = 2;
		}
	}
	else ClearFlag(WndFlags, GWF_SUNKEN);

	Invalidate();
}

bool GView::Flat()
{
	return TestFlag(WndFlags, GWF_FLAT);
}

void GView::Flat(bool i)
{
	if (i) SetFlag(WndFlags, GWF_FLAT);
	else ClearFlag(WndFlags, GWF_FLAT);

	Invalidate();
}

bool GView::Raised()
{
	return TestFlag(WndFlags, GWF_RAISED);
}

void GView::Raised(bool i)
{
	if (i)
	{
		SetFlag(WndFlags, GWF_RAISED);
		if (NOT _BorderSize)
		{
			_BorderSize = 2;
		}
	}
	else ClearFlag(WndFlags, GWF_RAISED);

	Invalidate();
}

void GView::Quit(bool DontDelete)
{
	if (DontDelete)
	{
		Visible(false);
	}
	else
	{
		_View->_SetDeleteMe();
	}
}

bool GView::AttachChildren()
{
	bool Status = true;

	Iterator<GViewI> It(&Children);
	for (GViewI *c=It.First(); c; c=It.Next())
	{
		if (NOT c->IsAttached())
		{
			Status |= c->Attach(this);
		}
	}

	return Status;
}

bool GView::SetCursor(int CursorId)
{
	d->CursorSet = true;
	GView *Wnd = GetWindow();
	
	if (Wnd AND
		Wnd->Handle() AND
		Wnd->d->CursorId != CursorId)
	{
		Display *Dsp = Wnd->Handle()->XDisplay();

		int Items = CountOf(CursorMetrics);
		if (CursorId > 0 AND CursorId <= Items)
		{
			if (Wnd->d->XCursor)
			{
				XUndefineCursor(Dsp, Wnd->Handle()->handle());
				XFreeCursor(Dsp, Wnd->d->XCursor);
				Wnd->d->XCursor = 0;
			}

			GSurface *Mem = Cursors.Create();
			if (Mem)
			{
				CursorInfo *Metrics = CursorMetrics + CursorId - 1;
				Ximg Src(Metrics->Pos.X(), Metrics->Pos.Y(), 1);
				Ximg Mask(Metrics->Pos.X(), Metrics->Pos.Y(), 1);
				
				for (int y=0; y<Src.Y(); y++)
				{
					for (int x=0; x<Src.X(); x++)
					{
						COLOUR c = Mem->Get(Metrics->Pos.x1 + x, Metrics->Pos.y1 + y);
						Src.Set(x, y, c == 1 ? 1 : 0);
						Mask.Set(x, y, c < 2);
					}
				}
				
				Xpix PSrc(Handle()->handle(), &Src);
				Xpix PMask(Handle()->handle(), &Mask);
				
				XColor Fore, Back;
				ZeroObj(Fore);
				ZeroObj(Back);
				Fore.red = Fore.green = Fore.blue = 0xffff;
				Back.red = Back.green = Back.blue = 0;
				Wnd->d->XCursor = XCreatePixmapCursor
				(
					Dsp,
					PSrc,
					PMask,
					&Fore,
					&Back,
					Metrics->HotSpot.x - Metrics->Pos.x1,
					Metrics->HotSpot.y - Metrics->Pos.y1
				);
				
				if (Wnd->d->XCursor)
				{
					XDefineCursor(Dsp, Wnd->Handle()->handle(), Wnd->d->XCursor);
				}
				
				DeleteObj(Mem);
			}
			else return false;
		}
		else
		{
			if (Wnd->d->XCursor)
			{
				XDefineCursor(Dsp, Wnd->Handle()->handle(), 0);
				XFlush(Dsp);
				
				XFreeCursor(Dsp, Wnd->d->XCursor);
				Wnd->d->XCursor = 0;
			}			
		}
		
		Wnd->d->CursorId = CursorId;
		Wnd->d->LastCursor = this;
		
		return true;
	}
	else
	{
		// printf("%s:%i - Param error"\n", __FILE__, __LINE__);
	}
	
	return false;
}

GFont *GView::GetFont()
{
	return d->Font ? d->Font : SysFont;
}

void GView::SetFont(GFont *Fnt, bool OwnIt)
{
	if (d->FontOwn)
	{
		DeleteObj(d->Font);
	}
	d->Font = Fnt;
	d->FontOwn = OwnIt;
}

bool GView::IsOver(GMouse &m)
{
	return	(m.x >= 0) AND
			(m.y >= 0) AND
			(m.x < Pos.X()) AND
			(m.y < Pos.Y());
}

bool GView::WindowVirtualOffset(GdcPt2 *Offset)
{
	bool Status = false;

	if (Offset)
	{
		Offset->x = 0;
		Offset->y = 0;
		GViewI *Wnd = this;
		while (Wnd)
		{
			GRect r = Wnd->GetPos();
			GViewI *Parent = Wnd->GetParent();
			if (NOT Wnd->Handle())
			{
				Offset->x += r.x1;
				Offset->y += r.y1;
				Wnd = Parent;
				Status = true;
			}
			else
			{
				Wnd = 0;
			}
		}
	}

	return Status;
}

GViewI *GView::WindowFromPoint(int x, int y)
{
	for (GViewI *c = Children.First(); c; c = Children.Next())
	{
		GRect r = c->GetPos();
		GViewI *At = c->WindowFromPoint(x - r.x1, y - r.y1);
		if (At) return At;
	}

	if (x >= 0 AND y >= 0 AND x < Pos.X() AND y < Pos.Y())
	{
		return this;
	}

	return NULL;
}

GdcPt2 GView::GetMinimumSize()
{
	return d->MinimumSize;
}

void GView::SetMinimumSize(GdcPt2 Size)
{
	d->MinimumSize = Size;
}

bool GView::SetPos(GRect &p, bool Repaint)
{
	Pos = p;

	if (_View)
	{
		int o = 0;
		
		GView *Par = d->GetParent();
		if (Par AND (Par->Sunken() OR Par->Raised()))
		{
			o = Par->_BorderSize;
		}
		
		_View->setGeometry( p.x1 + o,
	                        p.y1 + o,
							p.X(),
							p.Y());
		if (Repaint)
		{
			_View->repaint();
		}
	}

	return true;
}

bool GView::Invalidate(GRegion *r, bool Repaint, bool NonClient)
{
	if (r)
	{
		for (int i=0; i<r->Length(); i++)
		{
			bool Last = i == r->Length()-1;
			Invalidate((*r)[i], Last ? Repaint : false, NonClient);
		}
		
		return true;
	}
	
	return false;
}

bool GView::Invalidate(GRect *r, bool Repaint, bool Frame)
{
	if (_View)
	{
		if (InThread())
		{
			GRect Client = Frame ? GView::GetPos() : GView::GetClient(true);
			if (Repaint)
			{
				static bool Repainting = false;
				
				if (NOT Repainting)
				{
					Repainting = true;
					if (r)
					{
						GRect cr = *r;
						cr.Offset(Client.x1, Client.y1);
						_View->repaint(cr.x1, cr.y1, cr.X(), cr.Y());
					}
					else
					{
						_View->repaint();
					}
					Repainting = false;
				}
			}
			else
			{
				if (r)
				{
					GRect cr = *r;
					cr.Offset(Client.x1, Client.y1);
					_View->update(cr.x1, cr.y1, cr.X(), cr.Y());
				}
				else
				{
					_View->update();
				}
			}
		}
		else
		{
			X11_INVALIDATE_PARAMS *p = NEW(X11_INVALIDATE_PARAMS);
			if (p)
			{
				if (r)
				{
					p->Rgn = *r;
				}
				else
				{
					p->Rgn.ZOff(-1, -1);
				}
				p->Repaint = Repaint;
				p->View = this;
				
				PostEvent(M_X11_INVALIDATE, (int)p, 0);
			}
		}
		
		return true;
	}
	else
	{
		GRect Up;
		GViewI *p = this;

		if (r)
		{
			Up = *r;
		}
		else
		{
			Up.Set(0, 0, Pos.X()-1, Pos.Y()-1);
		}

		while (p AND NOT p->Handle())
		{
			Up.Offset(p->GetPos().x1, p->GetPos().y1);
			p = p->GetParent();
		}

		if (p AND p->Handle())
		{
			return p->Invalidate(&Up, Repaint);
		}
	}

	return false;
}

void GView::SetPulse(int Length)
{
	if (d->Pulse)
	{
		d->Pulse->Loop = false;
		d->Pulse = 0;
	}
	
	if (Length > 0)
	{
		d->Pulse = NEW(GPulseThread(this, Length));
	}
}

bool GView::InThread()
{
	int Me = LgiGetCurrentThread();
	return LgiApp->GetGuiThread() == Me;
}

bool GView::PostEvent(int Cmd, int a, int b)
{
	return LgiPostEvent(_View, Cmd, a, b);
}

int GView::OnEvent(GMessage *Msg)
{
	switch (MsgCode(Msg))
	{
		case M_X11_INVALIDATE:
		{
			X11_INVALIDATE_PARAMS *p = (X11_INVALIDATE_PARAMS*)MsgA(Msg);
			if (p AND p->View == this)
			{
				Invalidate(p->Rgn.Valid() ? &p->Rgn : 0, p->Repaint);
				DeleteObj(p);
			}
			break;
		}
		case M_X11_PULSE:
		{
			OnPulse();
			break;
		}
		case M_CHANGE:
		{
			GView *Ctrl = dynamic_cast<GView*>((GView*) MsgA(Msg));
			if (Ctrl)
			{
				return OnNotify(Ctrl, MsgB(Msg));
			}
			break;
		}
		case M_COMMAND:
		{
			int32 a=MsgA(Msg);
			return OnCommand(a&0xFFFF, a>>16, (OsView) MsgB(Msg));
		}
		default:
		{
			break;
		}
	}

	return 0;
}

bool GView::Name(char *n)
{
	bool Status = GObject::Name(n);

	if (_View)
	{
		_View->setText(n);
	}

	return Status;
}

char *GView::Name()
{
	return GObject::Name();
}

bool GView::NameW(char16 *n)
{
	return false;
}

char16 *GView::NameW()
{
	return 0;
}

bool GView::Enabled()
{
	if (_View)
	{
		return _View->isEnabled();
	}

	return TestFlag(WndFlags, GWF_ENABLED);
}

void GView::Enabled(bool e)
{
	if (_View AND (e ^ _View->isEnabled()) )
	{
		_View->setEnabled(e);
		Invalidate();
	}
	
	if (e) SetFlag(WndFlags, GWF_ENABLED);
	else ClearFlag(WndFlags, GWF_ENABLED);
}

GDragDropTarget *&GView::DropTargetPtr()
{
	return d->DropTarget;
}

bool GView::DropTarget()
{
	return TestFlag(WndFlags, GWF_DROP_TARGET);
}

bool GView::DropTarget(bool i)
{
	bool Status = false;
	
	if (i) SetFlag(WndFlags, GWF_DROP_TARGET);
	else ClearFlag(WndFlags, GWF_DROP_TARGET);
	
	GView *Wnd = GetWindow();
	if (Wnd AND Wnd->Handle())
	{
		OsView w = Wnd->Handle();
		
		#if defined XWIN
		if (i)
		{
			// We support XDND
			int Version = 4; // Current version
			XChangeProperty(w->XDisplay(),
							w->handle(),
							XInternAtom(w->XDisplay(), "XdndAware", False),
							XA_ATOM,
							32,
							PropModeReplace,
							(uchar*) &Version,
							1);
		}
		else
		{
			// Remove support for XDND
			XDeleteProperty(w->XDisplay(),
							w->handle(),
							XInternAtom(w->XDisplay(), "XdndAware", False));
		}
		#endif
	}	
}

bool GView::Visible()
{
	return TestFlag(WndFlags, GWF_VISIBLE);
}

void GView::Visible(bool v)
{
	bool Change = TestFlag(WndFlags, GWF_VISIBLE) ^ v;

	if (v) SetFlag(WndFlags, GWF_VISIBLE);
	else ClearFlag(WndFlags, GWF_VISIBLE);

	if (Change AND _View)
	{
		if (v)
		{
			_View->show();
		}
		else
		{
			_View->hide();
		}
	}
}

bool GView::Focus()
{
	return _View ? _View->hasFocus() : false;
}

void GView::Focus(bool f)
{
	if (_View AND (f ^ _View->hasFocus()) )
	{
		if (f) SetFlag(WndFlags, GWF_FOCUS);
		else ClearFlag(WndFlags, GWF_FOCUS);
		
		if (f)
		{
			_View->setFocus();
		}
		else
		{
			_View->clearFocus();
		}
	}
}

GViewI *GView::FindReal(GdcPt2 *Offset)
{
	if (Offset)
	{
		Offset->x = 0;
		Offset->y = 0;
	}

	GViewI *p = d->Parent;
	while (p AND NOT p->Handle())
	{
		if (Offset)
		{
			Offset->x += Pos.x1;
			Offset->y += Pos.y1;
		}

		p = p->GetParent();
	}
	
	if (p AND p->Handle())
	{
		return p;
	}

	return NULL;
}

bool GView::HandleCapture(GView *Wnd, bool c)
{
	bool Status = false;

	if (_View)
	{
		if (c)
		{
			GView *Cap = dynamic_cast<GView*>(_Capturing);
			if (Cap)
			{
				Cap->HandleCapture(NULL, false);
			}

			_View->grabMouse();
			Status = true;
			_Capturing = Wnd;
		}
		else
		{
			if (_Capturing)
			{
				_View->ungrabMouse();
				_Capturing = 0;
			}
		}
	}
	else
	{
		if (d->Parent)
		{
			GView *Cap = dynamic_cast<GView*>(d->Parent);
			Status = Cap ? Cap->HandleCapture(Wnd, c) : false;
		}
	}

	return Status;
}

bool GView::IsCapturing()
{
	return _Capturing == this;
}

bool GView::Capture(bool c)
{
	return HandleCapture(this, c);
}

void GView::PointToScreen(GdcPt2 &p)
{
	GViewI *c = this;

	// Find real parent
	while (NOT c->Handle() AND c->GetParent())
	{
		p.x += Pos.x1;
		p.y += Pos.y1;
		c = c->GetParent();
	}
	
	if (c AND c->Handle())
	{
		OsPoint q(p.x, p.y);
		c->Handle()->mapToGlobal(q);
		p.x = q.x;
		p.y = q.y;
	}
	else
	{
		printf("%s:%i - No real view to map to screen.\n", __FILE__, __LINE__);
	}
}

void GView::PointToView(GdcPt2 &p)
{
	OsPoint q(p.x, p.y);

	if (_View)
	{
		// Real window
		_View->mapFromGlobal(q);
		p.x = q.x;
		p.y = q.y;
	}
	else if (GetParent())
	{
		// Virtual window
		int Sx = 0, Sy = 0;
		GViewI *v;
		// Work out the virutal offset
		for (v = this; v AND NOT v->Handle(); v = v->GetParent())
		{
			Sx += v->GetPos().x1;
			Sy += v->GetPos().y1;
		}
		if (v AND v->Handle())
		{
			// Get the point relitive to the first real parent
			v->PointToView(p);
			
			// Move point back into virtual space
			p.x -= Sx;
			p.y -= Sy;
		}
		else
		{
			printf("%s:%i - No Real view.\n", __FILE__, __LINE__);
		}
	}
}

bool GView::GetMouse(GMouse &m, bool ScreenCoords)
{
	if (_View OR NOT GetParent())
	{
		#if defined XWIN
		Display *Dsp = XWidget::XDisplay();
		Window Root = RootWindow(Dsp, 0), Child;
		int Rx, Ry, Cx, Cy;
		uint Mask;
		XQueryPointer(	Dsp,
						_View ? _View->handle() : Root,
						&Root,
						&Child,
						&Rx, &Ry,
						&Cx, &Cy,
						&Mask);

		if (ScreenCoords)
		{
			m.x = Rx;
			m.y = Ry;
		}
		else
		{
			m.x = Cx;
			m.y = Cy;
		}

		m.Flags = 0;

		if (Mask & Button1Mask)
		{
			m.Left(true);
			m.Down(true);
		}
		if (Mask & Button2Mask)
		{
			m.Middle(true);
			m.Down(true);
		}
		if (Mask & Button3Mask)
		{
			m.Right(true);
			m.Down(true);
		}

		return true;
		#endif
	}
	else if (GetParent())
	{
		bool s = GetParent()->GetMouse(m, ScreenCoords);
		if (s)
		{
			if (NOT ScreenCoords)
			{
				m.x -= Pos.x1;
				m.y -= Pos.y1;
			}
			return true;
		}
	}
	
	return false;
}

bool GView::IsAttached()
{
	return	_View AND
			_View->handle() AND
			_View->parentWidget();
}

bool GView::Attach(GViewI *parent)
{
	bool Status = false;

	SetParent(parent);

	if (_View)
	{
		int o = 0;
		{
			GView *Par = d->GetParent();
			if (Par AND (Par->Sunken() OR Par->Raised()))
			{
				o = Par->_BorderSize;
			}
		}
		OsPoint p(Pos.x1 + o, Pos.y1 + o);

		if (d->Parent)
		{
			// Attech to new parent..
			GWindow *w = dynamic_cast<GWindow*>(d->Parent);
			if (w)
			{
				_View->reparent(w->Wnd, p, TestFlag(WndFlags, GWF_VISIBLE) );
				Status = true;
			}
			else if (d->Parent->Handle())
			{
				_View->reparent(d->Parent->Handle(), p, TestFlag(WndFlags, GWF_VISIBLE) );
				Status = true;
			}

			for (GViewI *w = d->Parent; w; w = w->GetParent())
			{
				if (w->GetWindow())
				{
					_Window = w->GetWindow();
					break;
				}
			}
		}
		else
		{
			Status = true;
		}

		char *n = GObject::Name();
		if (n AND dynamic_cast<GWindow*>(this))
		{
			_View->setText(n);
		}
		
		if (DropTarget())
		{
			DropTarget(true);
		}

		OnCreate();
		OnAttach();
	}

	if (d->Parent AND NOT d->Parent->HasView(this))
	{
		d->Parent->AddView(this);
		d->Parent->OnChildrenChanged(this, true);
	}
	
	return Status;
}

bool GView::Detach()
{
	bool Status = false;

	// Detach view
	if (_View AND _View->parentWidget())
	{
		OnDestroy();

		OsPoint p(0, 0);
		_View->reparent(0, p, false);
	}

	if (d->Parent)
	{
		// Events
		int D = 0;
		while (d->Parent->HasView(this))
		{
			d->Parent->DelView(this);
			D++;
		}

		if (D > 1)
		{
			printf("Error: GView=%p was in %p's list %i times.\n", this, d->Parent, D);
		}

		d->Parent->OnChildrenChanged(this, false);
		d->Parent = 0;
	}

	Status = true;

	return Status;
}

GViewI *GView::FindControl(OsView hCtrl)
{
	if (Handle() == hCtrl)
	{
		return this;
	}

	Iterator<GViewI> List(&Children);
	for (GViewI *c = List.First(); c; c = List.Next())
	{
		GViewI *Ctrl = c->FindControl(hCtrl);
		if (Ctrl)
		{
			return Ctrl;
		}
	}
	return 0;
}

GViewI *GView::FindControl(int Id)
{
	if (GetId() == Id)
	{
		return this;
	}

	Iterator<GViewI> List(&Children);
	for (GViewI *c = List.First(); c; c = List.Next())
	{
		GViewI *Ctrl = c->FindControl(Id);
		if (Ctrl)
		{
			return Ctrl;
		}
	}
	return 0;
}

#define MIN_WINDOW_POS	24

void GView::MoveOnScreen()
{
	GRect Screen(0, 0, GdcD->X()-1, GdcD->Y()-1);
	GRect p = GetPos();
	GRect Old = p;

	if (p.x2 >= Screen.X())
		p.Offset(Screen.X() - p.x2, 0);
	if (p.y2 >= Screen.Y())
		p.Offset(0, Screen.Y() - p.y2);
	if (p.x1 < 0)
		p.Offset(-p.x1, 0);
	if (p.y1 < MIN_WINDOW_POS)
		p.Offset(0, MIN_WINDOW_POS-p.y1);

	if (Old != p)
	{
		SetPos(p);
	}
}

void GView::MoveToCenter()
{
	GRect Screen(0, 0, GdcD->X()-1, GdcD->Y()-1);
	GRect p = GetPos();

	p.Offset(-p.x1, -p.y1);
	p.Offset((Screen.X() - p.X()) / 2, (Screen.Y() - p.Y()) / 2);

	SetPos(p);
}

void GView::MoveToMouse()
{
	GMouse m;
	if (GetMouse(m, true))
	{
		GRect p = GetPos();

		p.Offset(-p.x1, -p.y1);
		p.Offset(m.x-(p.X()/2), m.y-(p.Y()/2));

		SetPos(p);
	}
}

GViewFill *GView::GetForegroundFill()
{
	return d->Foreground;
}

bool GView::SetForegroundFill(GViewFill *f)
{
	if (d->Foreground != f)
	{
		DeleteObj(d->Foreground);
		d->Foreground = f;
		Invalidate();
	}
	
	return true;
}

GViewFill *GView::GetBackgroundFill()
{
	return d->Background;
}

bool GView::SetBackgroundFill(GViewFill *f)
{
	if (d->Background != f)
	{
		DeleteObj(d->Background);
		d->Background = f;
		Invalidate();
	}
	
	return true;
}

///////////////////////////////////////////////////////////////////
bool LgiIsMounted(char *Name)
{
	return false;
}

bool LgiMountVolume(char *Name)
{
	return false;
}

//////////////////////////////////////////////////////////////////////////
List<GViewFactory> *GViewFactory::Factories;

GViewFactory::GViewFactory()
{
	if (NOT Factories) Factories = NEW(List<GViewFactory>);
	if (Factories)
	{
		if (Factories->HasItem(this))
		{
			printf("%s:%i - Factory list already has '%p'\n", __FILE__, __LINE__, this);
		}
		else
		{
			Factories->Insert(this);
		}
	}
}

GViewFactory::~GViewFactory()
{
	if (Factories) Factories->Delete(this);
}

GView *GViewFactory::Create(char *Class, GRect *Pos, char *Text)
{
	if (Factories)
	{
		for (GViewFactory *f=Factories->First(); f; f=Factories->Next())
		{
			GView *v = f->NewView(Class, Pos, Text);
			if (v)
			{
				return v;
			}
		}
	}

	return 0;
}


////////////////////////////////
uint32 CursorData[] = {
0x02020202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020102, 0x02020202, 0x02020202, 0x01010101, 0x01010101, 0x01010101, 0x02020202, 0x02020202, 0x02010101, 0x02010101, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010102, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020102, 0x01020202, 0x02020202, 0x02020202, 0x02020202, 0x01010101, 0x01010101, 0x02020201, 0x02020202, 0x01010101, 0x01010101, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020102, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x02010101, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x02010102, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x01010101, 0x02020202, 0x02020202, 0x02020101, 0x02020202, 
0x02020202, 0x02020202, 0x02020202, 0x02020101, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x00000001, 0x00000000, 0x01000000, 0x02020202, 0x02020202, 0x01000001, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020101, 0x01020202, 0x02020201, 0x02020202, 0x02020202, 0x00000102, 0x00000000, 0x02020201, 0x02020202, 0x00000001, 0x01000000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 
0x02020202, 0x02020202, 0x00010102, 0x00000000, 0x02020101, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x00010001, 0x00010001, 0x01000001, 0x02020202, 0x02020202, 0x00010102, 0x02020101, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x00000000, 0x02020201, 0x02020202, 0x02020202, 0x01020202, 0x02020100, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x00010202, 0x00000000, 0x02020201, 0x02020202, 0x00000001, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x01000000, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 
0x00010202, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x00000001, 0x00000000, 0x02010000, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x01000102, 0x00000100, 0x02010000, 0x02020202, 0x02020202, 0x00010202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x00000000, 0x02020100, 0x02020202, 0x02020202, 0x00010202, 0x02020100, 0x01020202, 0x02010000, 0x02020202, 0x02020202, 0x01020202, 0x00000000, 0x02020201, 0x02020202, 0x00000001, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x00000000, 
0x02020201, 0x02020202, 0x02020202, 0x01020202, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x00000102, 0x01000000, 0x02020202, 0x02020202, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x00000000, 0x00010101, 0x01000000, 0x02020202, 0x00000001, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x00000001, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x00010202, 0x00000001, 0x02020100, 0x02020202, 0x02020202, 0x00010202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x00000102, 0x00000000, 0x02010000, 0x02020202, 0x02020202, 0x00000102, 0x01010100, 0x01010101, 0x01000000, 0x02020202, 0x02020201, 0x00010202, 0x00000000, 0x02020201, 0x02020202, 
0x00000001, 0x02010000, 0x02020202, 0x02020201, 0x02020202, 0x02020202, 0x01010102, 0x01010001, 0x02020101, 0x02020202, 0x02020202, 0x01020201, 0x02010000, 0x02020102, 0x02020202, 0x02020202, 0x01010101, 0x01010100, 0x02020201, 0x02020202, 0x02020202, 0x01000001, 0x02020101, 0x02020202, 0x02020202, 0x00010202, 0x01010000, 0x01020202, 0x00000001, 0x02020201, 0x00000001, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x00000001, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x00000102, 0x01000000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x01020202, 0x00000000, 0x02020201, 0x02020202, 0x02020202, 0x00010202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x01010101, 0x01000001, 0x01010101, 0x02020202, 0x02020202, 0x00000001, 0x00000000, 
0x00000000, 0x00000000, 0x02020201, 0x02020101, 0x00000102, 0x00000100, 0x02020201, 0x02020202, 0x01000001, 0x01000000, 0x01020202, 0x02020201, 0x02020202, 0x02020202, 0x02020201, 0x02010001, 0x02010202, 0x02020202, 0x01020202, 0x01020201, 0x02010000, 0x02010102, 0x02020202, 0x02020202, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x00000001, 0x01010000, 0x02020101, 0x02020202, 0x00000102, 0x01000000, 0x02020202, 0x00000102, 0x02020100, 0x00000001, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x00000001, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x00000102, 0x01000000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x02020201, 0x02020202, 0x02020202, 
0x02020202, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x00000001, 0x00000000, 0x00000000, 0x00000000, 0x02020201, 0x02010001, 0x00000001, 0x00010201, 0x02020201, 0x02020202, 0x02010001, 0x00000001, 0x00010201, 0x02020201, 0x02020202, 0x01020202, 0x02020201, 0x02010001, 0x01010202, 0x02020202, 0x00010202, 0x01020201, 0x02010000, 0x01000102, 0x02020202, 0x02020202, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x02010102, 0x00000001, 0x00000000, 0x02010000, 0x02020202, 0x00000102, 0x00000001, 0x02020201, 0x00010202, 0x02020100, 0x00000001, 0x01000000, 0x02020202, 0x02020202, 0x02020202, 0x00000001, 0x01000000, 0x02020202, 0x02020202, 0x02020202, 0x00000001, 0x00000000, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x02020202, 
0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x00000102, 0x01010100, 0x01010101, 0x01000000, 0x02020202, 0x01000001, 0x01000000, 0x01020202, 0x02020201, 0x02020202, 0x02020101, 0x00000102, 0x00000100, 0x02020201, 0x02020202, 0x00010202, 0x02020201, 0x02010001, 0x00010202, 0x02020201, 0x00000102, 0x01010101, 0x01010000, 0x00000101, 0x02020201, 0x01010101, 0x01010101, 0x01010100, 0x01010101, 0x02020201, 0x01000001, 0x00000001, 0x00000000, 0x02010000, 0x02020202, 0x00000001, 0x00000101, 0x02020100, 0x00010202, 0x02010000, 0x00000001, 0x00000000, 0x02020201, 0x02020202, 0x02020202, 0x00000001, 0x00000000, 0x02020201, 0x02020202, 0x02020202, 0x00000001, 0x00000000, 0x02020201, 
0x02020202, 0x02020202, 0x01010102, 0x01010101, 0x01010001, 0x01010101, 0x02020101, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x02020100, 0x01020202, 0x02010000, 0x02020202, 0x00000001, 0x02010000, 0x02020202, 0x02020201, 0x02020202, 0x02020201, 0x00010202, 0x00000000, 0x02020201, 0x02020202, 0x00000102, 0x01010101, 0x01010001, 0x00010101, 0x02020100, 0x00000001, 0x00000000, 0x00000000, 0x00000000, 0x02020100, 0x00000001, 0x00000000, 0x00000000, 0x00000000, 0x02020201, 0x00000001, 0x00000000, 0x00000000, 0x02010000, 0x02020202, 0x01000001, 0x00010202, 0x02010000, 0x01020202, 0x02010000, 0x00000001, 0x00000000, 0x02020100, 0x02020202, 0x02020202, 
0x00000001, 0x00000000, 0x02020100, 0x02020202, 0x02020202, 0x01010101, 0x01010100, 0x02020201, 0x02020202, 0x02020202, 0x00000001, 0x00000000, 0x00000000, 0x00000000, 0x02010000, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x02020100, 0x01020202, 0x02020100, 0x02020202, 0x00000001, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x00000000, 0x02020201, 0x02020202, 0x00000001, 0x00000000, 0x00000000, 0x00000000, 0x02010000, 0x00000102, 0x01010101, 0x01010000, 0x00000101, 0x02020201, 0x00000001, 0x00000000, 0x00000000, 0x00000000, 0x02020201, 0x00000102, 0x00000000, 0x00000000, 0x02010000, 0x02020202, 0x01000001, 0x01020202, 
0x01000000, 0x01020202, 0x02010000, 0x00000001, 0x00000000, 0x02010000, 0x02020202, 0x02020202, 0x00000001, 0x00000000, 0x02010000, 0x02020202, 0x02020202, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x01010102, 0x01010101, 0x01010001, 0x01010101, 0x02020101, 0x01020202, 0x00000000, 0x02020201, 0x02020202, 0x02020202, 0x00010202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020101, 0x01020202, 0x02020201, 0x02020202, 0x00000001, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x00000000, 0x02020201, 0x02020202, 0x00000102, 0x01010101, 0x01010001, 0x00010101, 0x02020100, 0x00010202, 0x01020201, 0x02010000, 0x01000102, 0x02020202, 0x01010101, 0x01010101, 0x01010100, 0x01010101, 
0x02020201, 0x00010202, 0x00000000, 0x00000000, 0x02010000, 0x02020202, 0x01000001, 0x02020202, 0x00000001, 0x01020201, 0x02010000, 0x00000001, 0x01000000, 0x02010101, 0x02020202, 0x02020202, 0x00000001, 0x01000000, 0x02010101, 0x02020202, 0x02020202, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x00010202, 0x01000001, 0x02020100, 0x02020202, 0x02020202, 0x00010202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x01010101, 0x01000001, 0x01010101, 0x02020202, 0x02020202, 0x02020202, 0x02020102, 0x01020202, 0x02020202, 0x02020202, 0x00000001, 0x01000000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x00000102, 0x00000000, 0x02020201, 0x02020202, 0x00010202, 0x02020201, 0x02010001, 0x00010202, 0x02020201, 0x01020202, 
0x01020201, 0x02010000, 0x02010102, 0x02020202, 0x02020202, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x01020202, 0x00000000, 0x00000000, 0x02010000, 0x02020202, 0x00000001, 0x02020201, 0x00000102, 0x00010100, 0x02010000, 0x00000001, 0x01000001, 0x01020202, 0x01010101, 0x01010101, 0x00000001, 0x01000001, 0x01020202, 0x01010101, 0x01010101, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x01000102, 0x01000001, 0x02010001, 0x02020202, 0x02020202, 0x00010202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x00000102, 0x00000000, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01010101, 0x01010101, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x01010101, 0x01010101, 
0x02020201, 0x02020202, 0x01020202, 0x02020201, 0x02010001, 0x01010202, 0x02020202, 0x02020202, 0x01020201, 0x02010000, 0x02020102, 0x02020202, 0x02020202, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x00000001, 0x00000000, 0x02020100, 0x02020202, 0x00000102, 0x02020201, 0x00010202, 0x00010000, 0x02020100, 0x01000001, 0x01000001, 0x01020202, 0x00000000, 0x01000000, 0x01000001, 0x01000001, 0x01020202, 0x00000000, 0x01000000, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x00010001, 0x00000000, 0x01000100, 0x02020202, 0x02020202, 0x00010202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x00000000, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 
0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020201, 0x02010001, 0x02010202, 0x02020202, 0x02020202, 0x01020202, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x01010101, 0x01010100, 0x02020201, 0x02020202, 0x02020202, 0x00000102, 0x00000000, 0x02020201, 0x02020202, 0x00000102, 0x02020100, 0x01020202, 0x00000000, 0x02020100, 0x02010001, 0x00000102, 0x01020201, 0x01010000, 0x01000001, 0x02010001, 0x00000102, 0x01020201, 0x01010100, 0x01000001, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x00000001, 0x00000000, 0x01000000, 0x02020202, 0x02020202, 0x00010202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x00000000, 
0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01010102, 0x01010001, 0x02020101, 0x02020202, 0x02020202, 0x01020202, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x00000102, 0x01000000, 0x02020202, 0x02020202, 0x02020202, 0x01010202, 0x01010101, 0x02020202, 0x02020202, 0x00010202, 0x01010000, 0x01020202, 0x00000001, 0x02020201, 0x02020101, 0x00000102, 0x01020201, 0x00000100, 0x01000100, 0x02020101, 0x00000102, 0x01020201, 0x01000100, 0x01000100, 0x01020202, 0x02020101, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x01010101, 0x01010101, 0x01010101, 0x02020202, 
0x02020202, 0x00010102, 0x02020101, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x00000000, 0x02020201, 0x02020202, 0x02020202, 0x01020202, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x00000000, 0x00010101, 0x01000000, 0x02020202, 0x02020202, 0x00010202, 0x01020100, 0x00000100, 0x01000000, 0x02020202, 0x00010202, 0x01020100, 0x01000100, 0x01000100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 
0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01000001, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010102, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x01000000, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x02010101, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x00000001, 0x00000000, 0x02010000, 0x02020202, 0x02020202, 0x00010202, 0x01020100, 0x00000100, 0x01000100, 0x02020202, 0x00010202, 0x01020100, 
0x01000100, 0x01000100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010101, 0x02010101, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x00010102, 0x00000000, 0x02020101, 0x02020202, 
0x02020202, 0x01020202, 0x01020201, 0x01010000, 0x01000001, 0x02020202, 0x01020202, 0x01020201, 0x01000100, 0x01000100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020102, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020102, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 
0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x01010101, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x00000000, 0x01000000, 0x02020202, 0x02020202, 0x01020202, 0x00000000, 0x01000000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 
0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x01010101, 0x01010101, 0x02020202, 0x02020202, 0x01020202, 0x01010101, 0x01010101, 
};
