#include <stdio.h>

#include "Lgi.h"
#include "GDragAndDrop.h"
#include "GToken.h"
#include "GPopup.h"
#include "GPanel.h"

///////////////////////////////////////////////////////////////////////
#if defined XWIN

XWindow::XWindow(GWindow *wnd)
{
	Wnd = wnd;
	setBackgroundMode(NoBackground);

	// Setup our protocols
	XSetWMProtocols(XDisplay(), handle(), &XApp()->XA_WmDeleteWindow, 1);
}

XWindow::~XWindow()
{
	if (Wnd)
	{
		Wnd->_View = 0;
		Wnd->_OnViewDelete();
	}
}

void XWindow::_SetWndPtr(void *p)
{
	Wnd = (GWindow*)p;
}

void *XWindow::_GetWndPtr()
{
	return Wnd;
}

void XWindow::SetModal()
{
}

void XWindow::closeEvent(XlibEvent *ce)
{
	if (Wnd->OnRequestClose(false))
	{
		_SetDeleteMe();
	}
}

void XWindow::resizeEvent(XlibEvent *re)
{
	if (Wnd)
	{
		Wnd->GetPos();

		// printf("XWindow('%s')::resizeEvent pos=%s\n", Wnd->Name(), Wnd->GetPos().Describe());
		
		Wnd->OnPosChange();
	}
}

void XWindow::paintEvent(XlibEvent *e)
{
	if (Wnd)
	{
		GScreenDC Dc(this);
		Wnd->_Paint(&Dc);
	}
}

void XWindow::customEvent(XlibEvent *e)
{
	GMessage *m = dynamic_cast<GMessage*>(e);
	if (Wnd AND m)
	{
		Wnd->OnEvent(m);
	}
}

void XWindow::propertyEvent(XlibEvent *e)
{
	char *Prop = XGetAtomName(XDisplay(), e->GetEvent()->xproperty.atom);
	if (Prop)
	{
		if (stricmp(Prop, "_NET_WM_STATE") == 0)
		{
			Atom Type = 0;
			int Format = 0;
			ulong Items = 0, Bytes = 0;
			uchar *Data = 0;
			
			XGetWindowProperty(	XDisplay(),
								handle(),
								e->GetEvent()->xproperty.atom,
								0,
								1024,
								false,
								AnyPropertyType,
								&Type,
								&Format,
								&Items,
								&Bytes,
								&Data);
								
			if (Data AND Type AND Bytes >= sizeof(Atom))
			{
				Atom *a = (Atom*)Data;
				if (*a)
				{
					char *Event = XGetAtomName(XDisplay(), *a);
					if (Event)
					{
						if (strcmp(Event, "_NET_WM_STATE_HIDDEN") == 0)
						{
							Wnd->OnZoom(GZoomMin);
						}
						else if (strstr(Event, "_NET_WM_STATE_MAXIMIZED"))
						{
							Wnd->OnZoom(GZoomMax);
						}
					}
					else if (Wnd)
					{
						Wnd->OnZoom(GZoomNormal);
					}
				}
			}
			
			if (Data)
			{
				XFree(Data);
			}
		}
		
		XFree(Prop);
	}
}

#elif defined YWIN

#include <stdint.h>
#include <sigc++/signal_system.h>
#include <sigc++/func_slot.h>
#include <sigc++/bind.h>

using std::cout;
using std::endl;
using SigC::slot;
using SigC::bind;

class YWindowPrivate
{
public:
	GWindow *This;
	Y::Window *w;
	Y::Canvas *c;
	GRect Pos;
	
	YWindowPrivate()
	{
		Pos.ZOff(-1, -1);
	}
};

YWindow::YWindow(GWindow *t) : YView(t)
{
	d = NEW(YWindowPrivate);
	d->This = t;
	SetHandle(d->w = NEW(Y::Window));
	d->c = NEW(Y::Canvas);
	if (d->c)
	{
		d->w->requestClose.connect (slot (this, &YWindow::OnRequestExit));
		d->c->requestSize(300, 200);
		d->w->setChild(d->c);
		d->c->resize.connect (slot (this, &YWindow::OnDraw));
	}
}

YWindow::~YWindow()
{
	DeleteObj(d);
}

bool YWindow::reparent(void *parent, OsPoint &p, bool vis)
{
	printf("YWindow::reparent\n");
	return true;
}

void YWindow::show()
{
	if (d->w)
	{
		d->w->show();
	}
}

void YWindow::hide()
{
}

void YWindow::OnRequestExit()
{
	if (d->This->OnRequestClose(false))
	{
		printf("delete this\n");
		delete this;
	}
}

void YWindow::OnDraw()
{
	GScreenDC Dc(this, d->c);
	d->This->_Paint(&Dc);
}

GRect &YWindow::geometry()
{
	return d->Pos;
}

void YWindow::setGeometry(int x, int y, int w, int h)
{
	if (d->c AND
		w > 0 AND
		h > 0 AND
		w < 2000 AND
		h < 2000)
	{
		printf("requestSize %i,%i\n", w, h);
		d->c->requestSize(w, h);
		d->Pos.Set(x, y, x + w - 1, y + h - 1);
	}
}

#endif

///////////////////////////////////////////////////////////////////////
class HookInfo
{
public:
	int Flags;
	GView *Target;
};

class GWindowPrivate
{
public:
	int Sx, Sy;
	bool Dynamic;
	GKey LastKey;
	GArray<HookInfo> Hooks;
	bool SnapToEdge;

	GWindowPrivate()
	{
		Sx = Sy = 0;
		Dynamic = true;
		SnapToEdge = false;
		ZeroObj(LastKey);		
	}
	
	int GetHookIndex(GView *Target, bool Create = false)
	{
		for (int i=0; i<Hooks.Length(); i++)
		{
			if (Hooks[i].Target == Target) return i;
		}
		
		if (Create)
		{
			HookInfo *n = &Hooks[Hooks.Length()];
			if (n)
			{
				n->Target = Target;
				n->Flags = 0;
				return Hooks.Length() - 1;
			}
		}
		
		return -1;
	}
};

///////////////////////////////////////////////////////////////////////
#define GWND_CREATE		0x0010000

GWindow::GWindow() :
	GView(0)
{
	d = NEW(GWindowPrivate);
	_QuitOnClose = false;
	Menu = 0;
	_Default = 0;
	_Window = this;
	WndFlags |= GWND_CREATE;
	ClearFlag(WndFlags, GWF_VISIBLE);

    _Lock = NEW(GSemaphore);
    #if defined XWIN
	_View = Wnd = NEW(XWindow(this));
	#elif defined YWIN
	_View = Wnd = NEW(YWindow(this));
	#endif
}

GWindow::~GWindow()
{
	if (LgiApp->AppWnd == this)
	{
		LgiApp->AppWnd = 0;
	}

	DeleteObj(Menu);
	DeleteObj(d);
	DeleteObj(_Lock);
}

bool GWindow::GetSnapToEdge()
{
	return d->SnapToEdge;
}

void GWindow::SetSnapToEdge(bool s)
{
	d->SnapToEdge = s;
}

bool GWindow::Visible()
{
	return GView::Visible();
}

void GWindow::Visible(bool i)
{
	if (NOT GView::Visible() AND i)
	{
		Pour();
	}

	GView::Visible(i);
	OnPosChange();
}

void GWindow::_SetDynamic(bool i)
{
	d->Dynamic = i;
}

void GWindow::_OnViewDelete()
{
	if (d->Dynamic)
	{
		delete this;
	}
}

bool GWindow::Attach(GViewI *p)
{
	bool Status = GView::Attach(p);
	if (Status)
	{
		// Setup default button...
		if (NOT _Default)
		{
			_Default = FindControl(IDOK);
		}
		if (_Default)
		{
			_Default->Invalidate();
		}
	}
	return Status;
}

bool GWindow::OnRequestClose(bool OsShuttingDown)
{
	if (GetQuitOnClose())
	{
		LgiCloseApp();
	}

	return GView::OnRequestClose(OsShuttingDown);
}

bool GWindow::HandleViewMouse(GView *v, GMouse &m)
{
	for (int i=0; i<d->Hooks.Length(); i++)
	{
		if (d->Hooks[i].Flags & GMouseEvents)
		{
			if (NOT d->Hooks[i].Target->OnViewMouse(v, m))
			{
				return false;
			}
		}
	}
	
	return true;
}

#define DEBUG_KEYS			0

/*
	// Any window in a popup always gets the key...
	for (GView *p = v; p; p = p->GetParent())
	{
		GPopup *Popup;
		if (Popup = dynamic_cast<GPopup*>(p))
		{
			Status = v->OnKey(k);
			if (NOT Status)
			{
				if (k.c16 == VK_ESCAPE)
				{
					// Popup window (or child) didn't handle the key, and the user
					// pressed ESC, so do the default thing and close the popup.
					Popup->Cancelled = true;
					Popup->Visible(false);
				}
				else
				{
					#if DEBUG_KEYS
					printf("Popup ignores '%c' down=%i alt=%i ctrl=%i sh=%i\n", k.c16, k.Down(), k.Alt(), k.Ctrl(), k.Shift());
					#endif
					break;
				}
			}
			
			#if DEBUG_KEYS
			printf("Popup ate '%c' down=%i alt=%i ctrl=%i sh=%i\n", k.c16, k.Down(), k.Alt(), k.Ctrl(), k.Shift());
			#endif
			
			goto AllDone;
		}
	}
*/

#define DEBUG_HANDLEVIEWKEY		0

bool GWindow::HandleViewKey(GView *v, GKey &k)
{
	bool Status = false;
	GViewI *Ctrl = 0;
	
	#if DEBUG_HANDLEVIEWKEY
	bool Debug = k.c16 == VK_RETURN;
	if (Debug)
	{
		printf("HandleViewKey=%i ischar=%i %s%s%s%s\n",
			k.c16,
			k.IsChar,
			(char*)(k.Down()?" Down":""),
			(char*)(k.Shift()?" Shift":""),
			(char*)(k.Alt()?" Alt":""),
			(char*)(k.Ctrl()?" Ctrl":""));
	}
	#endif

	// Any window in a popup always gets the key...
	GViewI *p;
	for (p = v->GetParent(); p; p = p->GetParent())
	{
		if (dynamic_cast<GPopup*>(p))
		{
			#if DEBUG_HANDLEVIEWKEY
			if (Debug)
				printf("Sending key to popup\n");
			#endif
			
			return v->OnKey(k);
		}
	}

	// Give key to popups
	if (LgiApp AND
		LgiApp->GetMouseHook() AND
		LgiApp->GetMouseHook()->OnViewKey(v, k))
	{
		goto AllDone;
	}

	// Allow any hooks to see the key...
	for (int i=0; i<d->Hooks.Length(); i++)
	{
		if (d->Hooks[i].Flags & GKeyEvents)
		{
			if (d->Hooks[i].Target->OnViewKey(v, k))
			{
				Status = true;
				
				#if DEBUG_HANDLEVIEWKEY
				if (Debug)
					printf("Hook ate '%c'(%i) down=%i alt=%i ctrl=%i sh=%i\n", k.c16, k.c16, k.Down(), k.Alt(), k.Ctrl(), k.Shift());
				#endif
				
				goto AllDone;
			}
		}
	}

	// Give the key to the window...
	if (v->OnKey(k))
	{
		#if DEBUG_HANDLEVIEWKEY
		if (Debug)
			printf("View ate '%c'(%i) down=%i alt=%i ctrl=%i sh=%i\n",
				k.c16, k.c16, k.Down(), k.Alt(), k.Ctrl(), k.Shift());
		#endif
		
		Status = true;
		goto AllDone;
	}
	
	// Window didn't want the key...
	switch (k.c16)
	{
		case VK_RETURN:
		{
			Ctrl = _Default;
			break;
		}
		case VK_ESCAPE:
		{
			Ctrl = FindControl(IDCANCEL);
			break;
		}
	}

	if (Ctrl AND Ctrl->Enabled())
	{
		if (Ctrl->OnKey(k))
		{
			Status = true;

			#if DEBUG_HANDLEVIEWKEY
			if (Debug)
				printf("Default Button ate '%c'(%i) down=%i alt=%i ctrl=%i sh=%i\n",
					k.c16, k.c16, k.Down(), k.Alt(), k.Ctrl(), k.Shift());
			#endif
			
			goto AllDone;
		}
	}

	if (Menu)
	{
		Status = Menu->OnKey(v, k);
		if (Status)
		{
			#if DEBUG_HANDLEVIEWKEY
			if (Debug)
				printf("Menu ate '%c' down=%i alt=%i ctrl=%i sh=%i\n", k.c16, k.Down(), k.Alt(), k.Ctrl(), k.Shift());
			#endif
		}
	}

AllDone:
	d->LastKey = k;

	return Status;
}


void GWindow::Raise()
{
	if (_View->handle())
	{
		#if defined XWIN
		XWindowChanges c;
		c.stack_mode = XAbove;
		XConfigureWindow(	_View->XDisplay(),
							_View->handle(),
							CWStackMode,
							&c);
		#endif							
	}
}


GWindowZoom GWindow::GetZoom()
{
	#if defined XWIN
	if (Handle())
	{
		Display *Dsp = Handle()->XDisplay();
		Window Win = Handle()->handle();
		Atom Type;
		int Format;
		ulong Items, Bytes;
		uchar *Prop = 0;

		XGetWindowProperty(	Dsp,
							Win,
							XInternAtom(Dsp, "_NET_WM_STATE", false),
							0, 1024,
							false,
							AnyPropertyType,
							&Type,
							&Format,
							&Items,
							&Bytes,
							&Prop);
		if (Prop)
		{
			Atom s[] =
			{
				XInternAtom(Dsp, "_NET_WM_STATE_MAX_V", false),
				XInternAtom(Dsp, "_NET_WM_STATE_MAX_H", false)
			};

			for (int i=0; i<Items; i++)
			{
				Atom a = ((Atom*)Prop)[i];
				if (a == s[0] OR a == s[1])
				{
					XFree(Prop);
					return GZoomMax;
				}
			}

			XFree(Prop);
		}

		XWindowAttributes a;
		ZeroObj(a);
		XGetWindowAttributes(Dsp, Win, &a);
		
		if (a.map_state == IsUnmapped)
		{
			return GZoomMin;
		}
	}
	#endif

	return GZoomNormal;
}

void GWindow::SetZoom(GWindowZoom i)
{
	#if defined XWIN
	Display *Dsp = Handle()->XDisplay();
	Window Win = Handle()->handle();

	switch (i)
	{
		case GZoomMax:
		{
			/*
			_NET_WM_STATE_MAX_V
			_NET_WM_STATE_MAX_H
		    _NET_WM_STATE_STAYS_ON_TOP
            _NET_WM_STATE_MODAL
			*/

			Atom s[] =
			{
				XInternAtom(Dsp, "_NET_WM_STATE_MAX_V", false),
				XInternAtom(Dsp, "_NET_WM_STATE_MAX_H", false)
			};

		    XChangeProperty(Dsp,
							Win,
							XInternAtom(Dsp, "_NET_WM_STATE", false),
							XA_ATOM,
							32,
							PropModeReplace,
							(unsigned char *) s,
							CountOf(s));
			break;
		}
		case GZoomMin:
		{
			XIconifyWindow(Dsp, Win, DefaultScreen(Dsp) );
			break;
		}
		case GZoomNormal:
		{
		    XChangeProperty(Dsp,
							Win,
							XInternAtom(Dsp, "_NET_WM_STATE", false),
							XA_ATOM,
							32,
							PropModeReplace,
							0,
							0);

			XMapRaised(Dsp, Win);
			break;
		}
	}
	#endif
}

GViewI *GWindow::GetDefault()
{
	return _Default;
}

void GWindow::SetDefault(GViewI *v)
{
	if (v AND
		v->GetWindow() == this)
	{
		if (_Default != v)
		{
			GViewI *Old = _Default;
			_Default = v;

			if (Old) Old->Invalidate();
			if (_Default) _Default->Invalidate();
		}
	}
	else
	{
		_Default = 0;
	}
}

bool GWindow::Name(char *n)
{
	bool Status = GObject::Name(n);

	if (Wnd AND _View)
	{
		_View->setText(n);
	}

	return Status;
}

char *GWindow::Name()
{
	return GObject::Name();
}

GRect &GWindow::GetClient(bool ClientSpace)
{
	static GRect r;
	
	r = GView::GetClient(ClientSpace);
	if (Menu AND
		Menu->Handle() AND
		Menu->Handle()->View())
	{
		/*
		#if defined XWIN
		QMenuBar *Bar = dynamic_cast<QMenuBar*>(Menu->Handle());
		if (Bar)
		{
			GRect p = Bar->geometry();
			r.y1 = p.y2 + 1;
		}
		#endif
		*/
		
		GRect p = Menu->Handle()->View()->GetPos();
		r.y1 = p.y2 + 1;
	}
	
	return r;
}

bool GWindow::SerializeState(GDom *Store, char *FieldName, bool Load)
{
	if (NOT Store OR NOT FieldName)
		return false;

	if (Load)
	{
		GVariant v;
		if (Store->GetValue(FieldName, v) AND v.Str())
		{
			GRect Position(0, 0, -1, -1);
			GWindowZoom State = GZoomNormal;

			GToken t(v.Str(), ";");
			for (int i=0; i<t.Length(); i++)
			{
				char *Var = t[i];
				char *Value = strchr(Var, '=');
				if (Value)
				{
					*Value++ = 0;

					if (stricmp(Var, "State") == 0)
						State = (GWindowZoom) atoi(Value);
					else if (stricmp(Var, "Pos") == 0)
						Position.SetStr(Value);
				}
				else return false;
			}
			
			if (Position.Valid())
			{
				SetPos(Position);
			}
			
			SetZoom(State);
		}
		else return false;
	}
	else
	{
		char s[256];
		GWindowZoom State = GetZoom();
		sprintf(s, "State=%i;Pos=%s", State, GetPos().GetStr());

		GVariant v = s;
		if (NOT Store->SetValue(FieldName, v))
			return false;
	}

	return true;
}

GRect &GWindow::GetPos()
{
	if (Wnd AND _View)
	{
		OsRect r = _View->geometry();

		Pos.x1 = r.left;
		Pos.y1 = r.top;
		Pos.x2 = r.right;
		Pos.y2 = r.bottom;

		#if defined XWIN
		XWindowAttributes a;
		ZeroObj(a);
		XGetWindowAttributes(Handle()->XDisplay(), Handle()->handle(), &a);
		if (NOT a.override_redirect)
		{
			int Dx, Dy;
			_View->GetDecorationSize(Dx, Dy);
			Pos.Offset(Dx, Dy);
		}
		#endif
	}

	return Pos;
}

bool GWindow::SetPos(GRect &p, bool Repaint)
{
	Pos = p;

	if (Wnd)
	{
		_View->setGeometry(p.x1, p.y1, p.X(), p.Y());
	}

	return true;
}

void GWindow::OnChildrenChanged(GViewI *Wnd, bool Attaching)
{
	// Force repour
	d->Sx = d->Sy = -1;
}

void GWindow::OnCreate()
{
}

void GWindow::_Paint(GSurface *pDC)
{
	GRect r = GetClient();
	pDC->Handle()->SetClient(&r);
	OnPaint(pDC);
	pDC->Handle()->SetClient(0);
}

void GWindow::OnPaint(GSurface *pDC)
{
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle();
}

void GWindow::OnPosChange()
{
	GView::OnPosChange();

	if (d->Sx != X() OR	d->Sy != Y())
	{
		// printf("Window '%s' Layout: x,y=%i,%i\n", Name(), X(), Y());
		
		Pour();
		d->Sx = X();
		d->Sy = Y();
	}
	else
	{
		// printf("Window '%s' Not Layout: x,y=%i,%i == %i,%i\n", Name(), d->Sx, d->Sy, X(), Y());
	}
}

#define IsTool(v) \
	( \
		dynamic_cast<GView*>(v) \
		AND \
		dynamic_cast<GView*>(v)->_IsToolBar \
	)

void GWindow::Pour()
{
	GRegion Client(GetClient());
	GViewI *MenuView = 0;

	if (Menu AND
		Menu->Handle() AND
		Menu->Handle()->View())
	{
		MenuView = Menu->Handle()->View();
		MenuView->Pour(Client);
		
		GRect *r = &MenuView->GetPos();
		Client.Subtract(r);
		// printf("Menu=%p MenuPos=%s\n", MenuView, r->GetStr());
	}
	else if (Menu)
	{
		printf("%s:%i - Menu has no view!\n", __FILE__, __LINE__);
	}
	
	GRegion Update(Client);
	bool HasTools = false;
	GViewI *v;
	Iterator<GViewI> Lst(&Children);

	// printf("Pour Menu=%p\n", Menu);

	{
		GRegion Tools;
		
		/*
		for (int i=0; i<Client.Length(); i++)
		{
			printf("\tClient[%i]=%s\n", i, Client[i]->GetStr());
		}
		*/
		
		for (v = Lst.First(); v; v = Lst.Next())
		{
			bool IsMenu = MenuView == v;
			if (NOT IsMenu AND IsTool(v))
			{
				GRect OldPos = v->GetPos();
				Update.Union(&OldPos);
				
				// printf("\tv=%p name=%s\n", v, v->Name());

				if (HasTools)
				{
					// 2nd and later toolbars
					if (v->Pour(Tools))
					{
						if (NOT v->Visible())
						{
							v->Visible(true);
						}

						if (OldPos != v->GetPos())
						{
							// position has changed update...
							v->Invalidate();
						}

						// printf("\tNth=%s\n", v->GetPos().GetStr());

						Tools.Subtract(&v->GetPos());
						Update.Subtract(&v->GetPos());
					}
				}
				else
				{
					// First toolbar
					if (v->Pour(Client))
					{
						HasTools = true;

						if (NOT v->Visible())
						{
							v->Visible(true);
						}

						if (OldPos != v->GetPos())
						{
							v->Invalidate();
						}

						GRect Bar(v->GetPos());
						Bar.x2 = GetClient().x2;

						// printf("\tFirst=%s\n", v->GetPos().GetStr());

						Tools = Bar;
						Tools.Subtract(&v->GetPos());
						Client.Subtract(&Bar);
						Update.Subtract(&Bar);
					}
				}
			}
		}
	}

	for (GViewI *v = Lst.First(); v; v = Lst.Next())
	{
		bool IsMenu = MenuView == v;
		if (NOT IsMenu AND NOT IsTool(v))
		{
			GRect OldPos = v->GetPos();
			Update.Union(&OldPos);

			if (v->Pour(Client))
			{
				GRect p = v->GetPos();

				if (NOT v->Visible())
				{
					v->Visible(true);
				}

				v->Invalidate();

				Client.Subtract(&v->GetPos());
				Update.Subtract(&v->GetPos());
			}
			else
			{
				// non-pourable
			}
		}
	}
	
	for (int i=0; i<Update.Length(); i++)
	{
		Invalidate(Update[i]);
	}
}

int GWindow::WillAccept(List<char> &Formats, GdcPt2 Pt, int KeyState)
{
	int Status = DROPEFFECT_NONE;
	
	// printf("First=%p\n", &Formats);
	
	for (char *f=Formats.First(); f; )
	{
		if (stricmp(f, LGI_FileDropFormat) == 0)
		{
			f = Formats.Next();
			Status = DROPEFFECT_COPY;
		}
		else
		{
			Formats.Delete(f);
			DeleteArray(f);
			f = Formats.Current();
		}
	}
	
	return Status;
}

int GWindow::OnDrop(char *Format, GVariant *Data, GdcPt2 Pt, int KeyState)
{
	int Status = DROPEFFECT_NONE;

	if (Format AND Data)
	{
		if (stricmp(Format, LGI_FileDropFormat) == 0)
		{
			GArray<char*> Files;
			if (Data->IsBinary())
			{
				GToken Uri(	(char*)Data->Value.Binary.Data,
							"\r\n,",
							true,
							Data->Value.Binary.Length);
				for (int i=0; i<Uri.Length(); i++)
				{
					char *File = Uri[i];
					if (strnicmp(File, "file:", 5) == 0)
						File += 5;
					
					char *i = File, *o = File;
					while (*i)
					{
						if (i[0] == '%' AND
							i[1] AND
							i[2])
						{
							char h[3] = { i[1], i[2], 0 };
							*o++ = htoi(h);
							i += 3;
						}
						else
						{
							*o++ = *i++;
						}
					}
					*o++ = 0;
					
					if (FileExists(File))
					{
						Files.Add(NewStr(File));
					}
				}
			}
			
			if (Files.Length())
			{
				Status = DROPEFFECT_COPY;
				OnReceiveFiles(Files);
				Files.DeleteArrays();
			}
		}
	}
	
	return Status;
}

int GWindow::OnEvent(GMessage *m)
{
	switch (MsgCode(m))
	{
		case M_CLOSE:
		{
			if (OnRequestClose(false))
			{
				Quit();
				return 0;
			}
			break;
		}
	}

	return GView::OnEvent(m);
}

bool GWindow::RegisterHook(GView *Target, int EventType, int Priority)
{
	bool Status = false;
	
	if (Target AND EventType)
	{
		int i = d->GetHookIndex(Target, true);
		if (i >= 0)
		{
			d->Hooks[i].Flags = EventType;
			Status = true;
		}
	}
	
	return Status;
}

bool GWindow::UnregisterHook(GView *Target)
{
	int i = d->GetHookIndex(Target);
	if (i >= 0)
	{
		d->Hooks.DeleteAt(i);
		return true;
	}
	return false;
}

