// Mac Implementation of General LGI functions
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <ctype.h>

#define _POSIX_TIMERS
#include <time.h>

#include "Lgi.h"
#include "GProcess.h"
#include "GTextLabel.h"
#include "GButton.h"

////////////////////////////////////////////////////////////////
// Local helper functions
bool _lgi_check_file(char *Path)
{
	if (Path)
	{
		if (FileExists(Path))
		{
			// file is there
			return true;
		}
		else
		{
			// shortcut?
			char *e = Path + strlen(Path);
			strcpy(e, ".lnk");
			if (FileExists(Path))
			{
				// resolve shortcut
				char Link[256];
				if (ResolveShortcut(Path, Link, sizeof(Link)))
				{
					// check destination of link
					if (FileExists(Link))
					{
						strcpy(Path, Link);
						return true;
					}
				}
			}
			*e = 0;
		}
	}

	return false;
}

void LgiSleep(int i)
{
	struct timespec request, remain;

	ZeroObj(request);
	ZeroObj(remain);
	
	request.tv_sec = i / 1000;
	request.tv_nsec = (i % 1000) * 1000000;

	while (nanosleep(&request, &remain) == -1)
	{
	    request = remain;
	}
}

char *p2c(unsigned char *s)
{
	if (s)
	{
		s[1+s[0]] = 0;
		return (char*)s + 1;
	}
	return 0;
}

void c2p255(Str255 &d, char *s)
{
	if (s)
	{
		int Len = strlen(s);
		if (Len > 255) Len = 255;
		d[0] = Len;
		for (int i=0; i<Len; i++)
			d[1+i] = s[i];
	}
}

void _lgi_assert(bool b, char *test, char *file, int line)
{
	static bool Asserting = false;

	if (NOT b AND NOT Asserting)
	{
		Asserting = true;

		printf("%s:%i - Assert failed:\n%s\n", file, line, test);

		LgiMsg(0,
				"Assert failed, file: %s, line: %i\n"
				"%s\n",
				"Debug",
				MB_OK,
				file, line, test);

		exit(-1);
	}
}

////////////////////////////////////////////////////////////////////////
// Implementations
GMessage CreateMsg(int m, int a, int b)
{
	static GMessage Msg;
	Msg.m = m;
	Msg.a = a;
	Msg.b = b;
	return Msg;
}

OsView DefaultOsView(GView *v)
{
	return 0;
}

bool LgiGetFileMimeType(char *File, char *Mime, int BufLen)
{
	return LgiApp->GetFileMimeType(File, Mime, BufLen);
}

#include "GToken.h"
bool _GetIniField(char *Grp, char *Field, char *In, char *Out, int OutSize)
{
	if (ValidStr(In))
	{
		bool InGroup = false;
		GToken t(In, "\r\n");
		
		for (int i=0; i<t.Length(); i++)
		{
			char *Line = t[i];
			if (*Line == '[')
			{
				// Reset inside group at the start of each group
				InGroup = false;

				char *e = strchr(++Line, ']');
				if (e)
				{
					*e = 0;
					InGroup = stricmp(Grp, Line) == 0;
				}
			}
			else if (InGroup)
			{
				char *e = strchr(Line, '=');
				if (e)
				{
					// Point v to the data after the equals
					// Leave e pointing to the byte before
					char *v = e-- + 1;

					// Seek e through any whitespace between the equals and the
					// field name
					while (e > Line AND strchr(" \t", *e)) e--;
					
					// Seek v through any whitespace after the equals
					while (*v AND strchr(" \t", *v)) v++;
					
					// Calculate the length of the field
					int flen = (int)e-(int)Line;

					// Check the current field against the input field
					if (strnicmp(Field, Line, flen) == 0)
					{
						while (	*v AND
								strchr("\r\n", *v) == 0 AND
								OutSize > 1) // leave space for NULL
						{
							*Out++ = *v++;
							OutSize--;
						}

						// Add the NULL
						*Out++ = 0;
						
						return true;
					}
				}
			}
		}
	}
	
	return false;
}

bool LgiGetAppsForMimeType(char *Mime, GArray<GAppInfo*> &Apps, int Limit)
{
	bool Status = false;

	if (Mime)
	{
	}

	return Status;
}

bool LgiGetAppForMimeType(char *Mime, char *AppPath, int BufSize)
{
	bool Status = false;
	if (AppPath)
	{
		GArray<GAppInfo*> Apps;
		Status = LgiGetAppsForMimeType(Mime, Apps, 1);
		if (Status)
		{
			strsafecpy(AppPath, Apps[0]->Path, BufSize);
		}
	}
	return Status;
}

int LgiRand(int Limit)
{
	return rand() % Limit;
}

bool LgiPlaySound(char *FileName, int ASync)
{
	return LgiExecute(FileName);
}

OSErr FinderLaunch(long nTargets, FSSpec *targetList) {
  OSErr err;
  AppleEvent theAEvent, theReply;
  AEAddressDesc fndrAddress;
  AEDescList targetListDesc;
  OSType fndrCreator;
  Boolean wasChanged;
  AliasHandle targetAlias;
  long index;
  
    /* verify parameters */
  if ((nTargets == 0) || (targetList == NULL)) return paramErr;
  
    /* set up locals  */
  AECreateDesc(typeNull, NULL, 0, &theAEvent);
  AECreateDesc(typeNull, NULL, 0, &fndrAddress);
  AECreateDesc(typeNull, NULL, 0, &theReply);
  AECreateDesc(typeNull, NULL, 0, &targetListDesc);
  targetAlias = NULL;
  fndrCreator = 'MACS';
  
    /* create an open documents event targeting the finder */
  err = AECreateDesc(typeApplSignature, (Ptr) &fndrCreator,
    sizeof(fndrCreator), &fndrAddress);
        
  if (err == noErr){
        err = AECreateAppleEvent(kCoreEventClass, kAEOpenDocuments,
            &fndrAddress, kAutoGenerateReturnID,
            kAnyTransactionID, &theAEvent);
    }
    
    
  if (err == noErr){  
            /* create the list of files to open */
        err = AECreateList(NULL, 0, false, &targetListDesc);
    }
    
    
  if (err == noErr){
        for ( index=0; index < nTargets; index++) {
            if (targetAlias == NULL){
                err = NewAlias(NULL, (targetList + index), &targetAlias);
            } else {
                err = UpdateAlias(NULL, (targetList + index), targetAlias, &wasChanged);
            }
            
            if (err == noErr){
                HLock((Handle) targetAlias);
                err = AEPutPtr(&targetListDesc, 
                                (index + 1), 
                                typeAlias, 
                                *targetAlias, 
                                GetHandleSize((Handle) targetAlias));
                HUnlock((Handle) targetAlias);
            }
        }
    }
  
    /* add the file list to the apple event */
    if( err == noErr ){
        err = AEPutParamDesc(&theAEvent, keyDirectObject, &targetListDesc);
    }
  if (err == noErr){
            /* send the event to the Finder */
        err = AESend(&theAEvent, &theReply, kAENoReply,
            kAENormalPriority, kAEDefaultTimeout, NULL, NULL);
    }

    /* clean up and leave */
  if (targetAlias != NULL) DisposeHandle((Handle) targetAlias);
  AEDisposeDesc(&targetListDesc);
  AEDisposeDesc(&theAEvent);
  AEDisposeDesc(&fndrAddress);
  AEDisposeDesc(&theReply);
  return err;
}

bool LgiExecute(char *File, char *Args, char *Dir)
{
	bool Status = false;

	if (File)
	{
		FSRef r;
		OSStatus e = FSPathMakeRef((UInt8*)File, &r, NULL);
		if (e) printf("%s:%i - FSPathMakeRef faied with %i\n", __FILE__, __LINE__, e);
		else
		{		
			FSSpec t;
			e = FSGetCatalogInfo(&r, kFSCatInfoNone, 0, 0, &t, 0);
			if (e) printf("%s:%i - NativePathNameToFSSpec faied with %i\n", __FILE__, __LINE__, e);
			else
			{		
				e = FinderLaunch(1, &t);
				if (e) printf("%s:%i - FinderLaunch faied with %i\n", __FILE__, __LINE__, e);
				else Status = true;
			}
		}
	}
	
	return Status;
}

