/*hdr
**      FILE:           GView.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           23/4/98
**      DESCRIPTION:    Linux GView Implementation
**
**      Copyright (C) 1998-2003, Matthew Allen
**              fret@memecode.com
*/

#include <stdio.h>
#include <stdlib.h>

#include "Lgi.h"
#include "GDragAndDrop.h"

#define ADJ_LEFT					1
#define ADJ_RIGHT					2
#define ADJ_UP						3
#define ADJ_DOWN					4

struct X11_INVALIDATE_PARAMS
{
	GView *View;
	GRect Rgn;
	bool Repaint;
};

// CursorData is a bitmap in an array of uint32's. This is generated from a graphics file:
// ./Code/cursors.png
//
// The pixel values are turned into C code by a program called i.Mage:
//	http://www.memecode.com/image.php
//
// Load the graphic into i.Mage and then go Edit->CopyAsCode
// Then paste the text into the CursorData variable at the bottom of this file.
//
// This saves a lot of time finding and loading an external resouce, and even having to
// bundle extra files with your application. Which have a tendancy to get lost along the
// way etc.
extern uint32 CursorData[];
GInlineBmp Cursors =
{
	300, 20, 8, CursorData
};

struct LgiCursor
{
public:
	GRect Pos;
	GdcPt2 HotSpot;
}
CursorMetrics[] =
{
	// up arrow
	{ GRect(0, 0, 8, 15),			GdcPt2(4, 0) },
	// cross hair
	{ GRect(20, 0, 38, 18),		GdcPt2(29, 9) },
	// hourglass
	{ GRect(40, 0, 51, 15),		GdcPt2(45, 8) },
	// I beam
	{ GRect(60, 0, 66, 17),		GdcPt2(63, 8) },
	// N-S arrow
	{ GRect(80, 0, 91, 16),		GdcPt2(85, 8) },
	// E-W arrow
	{ GRect(100, 0, 116, 11),		GdcPt2(108, 5) },
	// NW-SE arrow
	{ GRect(120, 0, 132, 12),		GdcPt2(126, 6) },
	// NE-SW arrow
	{ GRect(140, 0, 152, 12),		GdcPt2(146, 6) },
	// 4 way arrow
	{ GRect(160, 0, 178, 18),		GdcPt2(169, 9) },
	// Blank
	{ GRect(0, 0, 0, 0),			GdcPt2(0, 0) },
	// Vertical split
	{ GRect(180, 0, 197, 16),		GdcPt2(188, 8) },
	// Horizontal split
	{ GRect(200, 0, 216, 17),		GdcPt2(208, 8) },
	// Hand
	{ GRect(220, 0, 233, 13),		GdcPt2(225, 0) },
	// No drop
	{ GRect(240, 0, 258, 18),		GdcPt2(249, 9) },
	// Copy drop
	{ GRect(260, 0, 279, 19),		GdcPt2(260, 0) },
	// Move drop
	{ GRect(280, 0, 299, 19),		GdcPt2(280, 0) },
};

////////////////////////////////////////////////////////////////////////////
HIRect &ConvertRect(GRect &g)
{
	static HIRect r;
	r.origin.x = (float) g.x1;
	r.origin.y = (float) g.y1;
	r.size.width = (float) g.X();
	r.size.height = (float) g.Y();
	return r;
}

GMouse _lgi_adjust_click_for_window(GMouse &Info, GViewI *Wnd)
{
	GMouse Temp = Info;
	
	/*
	if (Wnd AND NOT Wnd->Handle())
	{
		Temp.Target = Wnd;
		int Ox = 0, Oy = 0;
		GViewI *v = Wnd;
		GViewI *Window = v->GetWindow();
		while (v != Window)
		{
			GRect p = v->GetPos();
			GRect c = v->GetClient(false);
			
			Ox += p.x1 + c.x1;
			Oy += p.y1 + c.y1;
			
			if (v->GetParent())
				v = v->GetParent();
			else
				break;
		}

		Temp.x -= Ox;
		Temp.y -= Oy;
	}
	*/
	
	return Temp;
}

////////////////////////////////////////////////////////////////////////////
class GPulseThread : public GThread
{
	GView *View;
	int Length;

public:
	bool Loop;

	GPulseThread(GView *view, int len)
	{
		Loop = true;
		View = view;
		Length = len;
		DeleteOnExit = true;
		
		Run();
	}
	
	int Main()
	{
		do
		{
			LgiSleep(Length);
			if (Loop)
			{
				View->PostEvent(M_PULSE);
			}
		}
		while (Loop);
	}
};

////////////////////////////////////////////////////////////////////////////
void _lgi_yield()
{
	LgiApp->Run(false);
}

extern int IsAdjacent(GRect &a, GRect &b);
int IsAdjacent(GRect &a, GRect &b)
{
	if (	(a.x1 == b.x2 + 1) AND
		NOT (a.y1 > b.y2 OR a.y2 < b.y1))
	{
		return ADJ_LEFT;
	}

	if (	(a.x2 == b.x1 - 1) AND
		NOT (a.y1 > b.y2 OR a.y2 < b.y1))
	{
		return ADJ_RIGHT;
	}

	if (	(a.y1 == b.y2 + 1) AND
		NOT (a.x1 > b.x2 OR a.x2 < b.x1))
	{
		return ADJ_UP;
	}

	if (	(a.y2 == b.y1 - 1) AND
		NOT (a.x1 > b.x2 OR a.x2 < b.x1))
	{
		return ADJ_DOWN;
	}

	return 0;
}

extern GRect JoinAdjacent(GRect &a, GRect &b, int Adj);
GRect JoinAdjacent(GRect &a, GRect &b, int Adj)
{
	GRect t;

	switch (Adj)
	{
		case ADJ_LEFT:
		case ADJ_RIGHT:
		{
			t.y1 = max(a.y1, b.y1);
			t.y2 = min(a.y2, b.y2);
			t.x1 = min(a.x1, b.x1);
			t.x2 = max(a.x2, b.x2);
			break;
		}
		case ADJ_UP:
		case ADJ_DOWN:
		{
			t.y1 = min(a.y1, b.y1);
			t.y2 = max(a.y2, b.y2);
			t.x1 = max(a.x1, b.x1);
			t.x2 = min(a.x2, b.x2);
			break;
		}
	}

	return t;
}

////////////////////////////////////////////////////////////////////////////
bool LgiIsKeyDown(int Key)
{
	return false;
}

///////////////////////////////////////////////////////////////////////////
class GViewIter : public GViewIterator
{
	Iterator<GViewI> i;

public:
	GViewIter(GView *v) : i(&v->Children)
	{
	}

	GViewI *First()
	{
		return i.First();
	}

	GViewI *Last()
	{
		return i.Last();
	}

	GViewI *Next()
	{
		return i.Next();
	}

	GViewI *Prev()
	{
		return i.Prev();
	}

	int Length()
	{
		return i.Length();
	}

	int IndexOf(GViewI *view)
	{
		return i.IndexOf(view);
	}

	GViewI *operator [](int Idx)
	{
		return i[Idx];
	}
};

////////////////////////////////////////////////////////////////////////////////////////////////////
pascal
OSStatus
GViewProc
(
	EventHandlerCallRef nextHandler, 
	EventRef inEvent,
	void * userData
)
{
	OSStatus Status = eventNotHandledErr;

	GView *v = (GView*)userData;
	if (NOT v)
	{
		printf("%s:%i - No GView ptr.\n", __FILE__, __LINE__);
		return 0;
	}
	UInt32 eventClass = GetEventClass( inEvent );
	UInt32 eventKind = GetEventKind( inEvent );
	
	switch (eventKind)
	{
		case kEventControlDraw:
		{
			GScreenDC dc(v);
			v->_Paint(&dc);
			Status = noErr;
			break;
		}
	}
	
	printf("event %i,%i\n", eventClass, eventKind);

	return Status;
}

#define GViewSubClassName			CFSTR("com.memecode.lgiview")

class GViewPrivate
{
public:
	static bool				CursorSet;
	static HIObjectClassRef BaseClass;

	GPulseThread	*Pulse;
	GViewI			*Parent;
	GViewI			*Notify;
	int				CtrlId;
	GdcPt2			MinimumSize;
	GDragDropTarget	*DropTarget;
	int				CursorId;
	static GView	*LastCursor;
	GFont			*Font;
	bool			FontOwn;
	GView			*Popup;
	GViewFill		*Foreground;
	GViewFill		*Background;

	#if defined XWIN
	Cursor			XCursor;
	#endif

	GViewPrivate()
	{
		CursorId = 0;
		Pulse = 0;
		Parent = 0;
		Notify = 0;
		Foreground = Background = 0;
		CtrlId = -1;
		DropTarget = 0;
		Font = 0;
		FontOwn = false;
		Popup = 0;
		
		static bool First = true;
		if (First)
		{
			First = false;
			EventTypeSpec Events[] =
			{
				{ kEventClassHIObject, kEventHIObjectConstruct },
				{ kEventClassHIObject, kEventHIObjectInitialize },
				{ kEventClassHIObject, kEventHIObjectDestruct },
				{ kEventClassControl, kEventControlHitTest },
				{ kEventClassControl, kEventControlTrack },
				{ kEventClassControl, kEventControlDraw }
			};

			OSStatus e = HIObjectRegisterSubclass
				(
					GViewSubClassName,
					0, // base class, 0 = HIObject
					0,
					NewEventHandlerUPP(GViewProc),
					GetEventTypeCount(Events),
					Events,
					0,
					&BaseClass
				);
			if (e)
			{
				printf("%s:%i - HIObjectRegisterSubclass failed (e=%i).\n", __FILE__, __LINE__, e);
			}
		}
	}
	
	~GViewPrivate()
	{
		DeleteObj(Popup);
		DeleteObj(Foreground);
		DeleteObj(Background);
	}
};

bool GViewPrivate::CursorSet = false;
HIObjectClassRef GViewPrivate::BaseClass = 0;

GView *GViewPrivate::LastCursor = 0;
GViewI *GView::_Capturing = 0;
GViewI *GView::_Over = 0;

GView::GView(OsView view)
{
	d = NEW(GViewPrivate);
	_View = 0;
	_Window = 0;
	_Lock = 0;
	_BorderSize = 0;
	Script = 0;
	WndFlags = GWF_VISIBLE | GWF_ENABLED;
	Pos.ZOff(-1, -1);
}

GView::~GView()
{
	if (_Over == this) _Over = 0;
	if (_Capturing == this) Capture(false);
	_Delete();
	DeleteObj(d);
}

GViewIterator *GView::IterateViews()
{
	return NEW(GViewIter(this));
}

bool GView::AddView(GViewI *v, int Where)
{
	LgiAssert(NOT Children.HasItem(v));
	return Children.Insert(v, Where);
}

bool GView::DelView(GViewI *v)
{
	return Children.Delete(v);
}

bool GView::HasView(GViewI *v)
{
	return Children.HasItem(v);
}

void GView::_Delete()
{
	SetPulse();
	Pos.ZOff(-1, -1);

	GViewI *c;
	while (c = Children.First())
	{
		if (c->GetParent() != (GViewI*)this)
		{
			printf("Error: ~GView, child not attached correctly: %p(%s) Parent: %p(%s)\n",
				c, c->Name(),
				c->GetParent(), c->GetParent() ? c->GetParent()->Name() : "");
			Children.Delete(c);
		}

		DeleteObj(c);
	}

	Detach();
}

GWindow *GView::GetWindow()
{
	if (NOT _Window)
	{
		for (GViewI *w = d->Parent; w; w = w->GetParent())
		{
			if (w->GetWindow())
			{
				_Window = w->GetWindow();
				break;
			}
		}
	}
		
	return dynamic_cast<GWindow*>(_Window);
}

bool GView::GetTabStop()
{
	return false;
}

void GView::SetTabStop(bool i)
{
}

GViewI *GView::GetParent()
{
	return d->Parent;
}

void GView::SetParent(GViewI *p)
{
	d->Parent = p;
}

void GView::SendNotify(int Data)
{
	GViewI *n = d->Notify ? d->Notify : d->Parent;
	if (n)
	{
		n->OnNotify(this, Data);
	}
}

GViewI *GView::GetNotify()
{
	return d->Notify;
}

void GView::SetNotify(GViewI *p)
{
	d->Notify = p;
}

OsWindow GView::WindowHandle()
{
	return (GetWindow()) ? GetWindow()->Wnd : 0;
}


bool GView::Lock(int TimeOut)
{
	bool Status = true;

	if (NOT _Window AND d->Parent)
	{
		// Sometimes we have a parent chain but no _Window yet. If something
		// calls GView::Lock when _Window is NULL and then GetWindow, it'll
		// set _Window and the call to GView::Unlock will not match.
		GetWindow();
	}

	if (_Window != 0 AND _Window->_Lock != 0)
	{
		if (TimeOut < 0)
		{
			Status = _Window->_Lock->Lock();
		}
		else
		{
			Status = _Window->_Lock->LockWithTimeout(TimeOut);
		}
	}

	return true;
}

void GView::Unlock()
{
	if (_Window != 0 AND _Window->_Lock != 0)
	{
		_Window->_Lock->Unlock();
	}
}

void GView::OnMouseClick(GMouse &m)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnMouseClick(m);
	}
}

void GView::OnMouseEnter(GMouse &m)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnMouseEnter(m);
	}
}

void GView::OnMouseExit(GMouse &m)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnMouseExit(m);
	}
}

void GView::OnMouseMove(GMouse &m)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnMouseMove(m);
	}
}

void GView::OnMouseWheel(double Lines)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnMouseWheel(Lines);
	}
}

bool GView::OnKey(GKey &k)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnKey(k);
	}
	
	return false;
}

void GView::OnCreate()
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnCreate();
	}
}

void GView::OnDestroy()
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnDestroy();
	}
}

void GView::OnFocus(bool f)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnFocus(f);
	}
}

void GView::OnPulse()
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnPulse();
	}
}

void GView::OnPosChange()
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnPosChange();
	}
}

bool GView::OnRequestClose(bool OsShuttingDown)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnRequestClose(OsShuttingDown);
	}

	return true;
}

int GView::OnHitTest(int x, int y)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnHitTest(x, y);
	}

	return -1;
}

void GView::OnChildrenChanged(GViewI *Wnd, bool Attaching)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnChildrenChanged(Wnd, Attaching);
	}
}

void GView::OnPaint(GSurface *pDC)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnPaint(pDC);
	}
}

int GView::OnNotify(GViewI *Ctrl, int Flags)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnNotify(Ctrl, Flags);
	}
	else if (Ctrl AND d->Parent)
	{
		// default behaviour is just to pass the 
		// notification up to the parent
		return d->Parent->OnNotify(Ctrl, Flags);
	}

	return 0;
}

int GView::OnCommand(int Cmd, int Event, OsView Wnd)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnCommand(Cmd, Event, Wnd);
	}

	return 0;
}

GView *&GView::PopupChild()
{
	return d->Popup;
}

void _DumpHeirarchy(GViewI *i, int Depth = 0)
{
	for (int n=0; n<Depth; n++)
		printf("    ");
	printf("%p: %s(%s) hnd=%p vis=%i\n", i, i->GetClassName(), i->GetPos().GetStr(), i->Handle(), i->Visible());

	GViewIterator *it = i->IterateViews();
	if (it)
	{
		for (GViewI *c=it->First(); c; c=it->Next())
		{
			_DumpHeirarchy(c, Depth+1);
		}
		
		DeleteObj(it);
	}
}

void GView::_Mouse(GMouse &m, bool Move)
{
	#if 0
	if (NOT Move)
	{
		static bool First = true;
		if (First)
		{
			First = false;
			_DumpHeirarchy(GetWindow());
		}
		
		printf("_Mouse %p,%s %i,%i down=%i move=%i\n",
			this, GetClassName(),
			m.x, m.y,
			m.Down(), Move);
	}
	#endif

	GWindow *Wnd = GetWindow();

	if (_Capturing)
	{
		if (_Capturing != m.Target)
		{
			m.ToScreen();
			m.Target = _Capturing;
			m.ToView();
		}
		
		if (Move)
		{
			GViewI *c = _Capturing;
			d->CursorSet = false;
			
			// GMouse a = _lgi_adjust_click_for_window(m, c);
			c->OnMouseMove(m);
			
			if (NOT d->CursorSet)
			{
				c->SetCursor(LCUR_Normal);
			}
		}
		else
		{
			// GMouse a = _lgi_adjust_click_for_window(m, _Capturing);
			if (NOT Wnd OR Wnd->HandleViewMouse(dynamic_cast<GView*>(m.Target), m))
			{
				#if 0
				printf("CaptureClick %p,%s,%s m=%i,%i\n",
					m.Target,
					m.Target->GetClassName(),
					m.Target->GetPos().GetStr(),
					m.x,
					m.y);
				#endif
				m.Target->OnMouseClick(m);
			}
		}
	}
	else
	{
		if (_Over != this)
		{
			if (_Over)
			{
				// GMouse a = _lgi_adjust_click_for_window(m, _Over);
				_Over->OnMouseExit(m);
			}

			_Over = this;

			if (_Over)
			{
				// GMouse a = _lgi_adjust_click_for_window(m, _Over);
				_Over->OnMouseEnter(m);
			}
		}
			
		GView *Target = dynamic_cast<GView*>(_Over ? _Over : this);
		GRect Client = Target->GView::GetClient();
		if (Target->Raised() OR Target->Sunken())
		{
			Client.Offset(Target->_BorderSize, Target->_BorderSize);
		}
		// m = _lgi_adjust_click_for_window(m, Target);
		if (NOT Client.Valid() OR Client.Overlap(m.x, m.y))
		{
			if (Move)
			{
				d->CursorSet = false;
				Target->OnMouseMove(m);
				if (NOT d->CursorSet)
				{
					Target->SetCursor(LCUR_Normal);
				}
			}
			else
			{
				if (NOT Wnd OR Wnd->HandleViewMouse(Target, m))
				{
					#if 0
					printf("NormalClick %p,%s,%s m=%i,%i\n",
						Target,
						Target->GetClassName(),
						Target->GetPos().GetStr(),
						m.x,
						m.y);
					#endif
					Target->OnMouseClick(m);
				}
			}
		}
	}
}

GRect &GView::GetClient(bool ClientSpace)
{
	int Edge = (Sunken() OR Raised()) ? _BorderSize : 0;

	static GRect c;

	c = Pos;
	c.Offset(-c.x1, -c.y1);
	if (ClientSpace)
	{
		c.x2 -= Edge << 1;
		c.y2 -= Edge << 1;
	}
	else
	{
		c.Size(Edge, Edge);
	}
	
	return c;
}

GRect *GView::FindLargest(GRegion &r)
{
	int Pixels = 0;
	GRect *Best = 0;
	static GRect Final;

	// do initial run through the list to find largest single region
	for (GRect *i = r.First(); i; i = r.Next())
	{
		int Pix = i->X() * i->Y();
		if (Pix > Pixels)
		{
			Pixels = Pix;
			Best = i;
		}
	}

	if (Best)
	{
		Final = *Best;
		Pixels = Final.X() * Final.Y();

		int LastPixels = Pixels;
		GRect LastRgn = Final;
		int ThisPixels = Pixels;
		GRect ThisRgn = Final;

		GRegion TempList;
		for (GRect *i = r.First(); i; i = r.Next())
		{
			TempList.Union(i);
		}
		TempList.Subtract(Best);

		do
		{
			LastPixels = ThisPixels;
			LastRgn = ThisRgn;

			// search for adjoining rectangles that maybe we can add
			for (GRect *i = TempList.First(); i; i = TempList.Next())
			{
				int Adj = IsAdjacent(ThisRgn, *i);
				if (Adj)
				{
					GRect t = JoinAdjacent(ThisRgn, *i, Adj);
					int Pix = t.X() * t.Y();
					if (Pix > ThisPixels)
					{
						ThisPixels = Pix;
						ThisRgn = t;

						TempList.Subtract(i);
					}
				}
			}
		} while (LastPixels < ThisPixels);

		Final = ThisRgn;
	}

	return &Final;
}

GRect *GView::FindSmallestFit(GRegion &r, int Sx, int Sy)
{
	int X = 1000000;
	int Y = 1000000;
	GRect *Best = 0;

	for (GRect *i = r.First(); i; i = r.Next())
	{
		if ((i->X() >= Sx AND i->Y() >= Sy) AND
			(i->X() < X OR i->Y() < Y))
		{
			X = i->X();
			Y = i->Y();
			Best = i;
		}
	}

	return Best;
}

GRect *GView::FindLargestEdge(GRegion &r, int Edge)
{
	GRect *Best = 0;

	for (GRect *i = r.First(); i; i = r.Next())
	{
		if (NOT Best)
		{
			Best = i;
		}

		if ((Edge & GV_EDGE_TOP) AND (i->y1 < Best->y1) OR
			(Edge & GV_EDGE_RIGHT) AND (i->x2 > Best->x2) OR
			(Edge & GV_EDGE_BOTTOM) AND (i->y2 > Best->y2) OR
			(Edge & GV_EDGE_LEFT) AND (i->x1 < Best->x1))
		{
			Best = i;
		}

		if (((Edge & GV_EDGE_TOP) AND (i->y1 == Best->y1) OR
			(Edge & GV_EDGE_BOTTOM) AND (i->y2 == Best->y2)) AND
			(i->X() > Best->X()))
		{
			Best = i;
		}

		if (((Edge & GV_EDGE_RIGHT) AND (i->x2 == Best->x2) OR
			(Edge & GV_EDGE_LEFT) AND (i->x1 == Best->x1)) AND
			(i->Y() > Best->Y()))
		{
			Best = i;
		}
	}

	return Best;
}

int64 GView::GetCtrlValue(int Id)
{
	GViewI *w = FindControl(Id);
	return (w) ? w->Value() : 0;
}

void GView::SetCtrlValue(int Id, int64 i)
{
	GViewI *w = FindControl(Id);
	if (w)
	{
		w->Value(i);
	}
}

char *GView::GetCtrlName(int Id)
{
	GViewI *w = FindControl(Id);
	return (w) ? w->Name() : 0;
}

void GView::SetCtrlName(int Id, char *s)
{
	GViewI *w = FindControl(Id);
	if (w)
	{
		w->Name(s);
	}
}

bool GView::GetCtrlEnabled(int Id)
{	
	GViewI *w = FindControl(Id);
	return (w) ? w->Enabled() : 0;
}

void GView::SetCtrlEnabled(int Id, bool Enabled)
{
	GViewI *w = FindControl(Id);
	if (w)
	{
		w->Enabled(Enabled);
	}
}

bool GView::GetCtrlVisible(int Id)
{
	GViewI *w = FindControl(Id);
	return (w) ? w->Visible() : 0;
}

void GView::SetCtrlVisible(int Id, bool i)
{
	GViewI *w = FindControl(Id);
	if (w)
	{
		w->Visible(i);
	}
}

int GView::GetId()
{
	return d->CtrlId;
}

void GView::SetId(int i)
{
	d->CtrlId = i;
}

bool GView::Sunken()
{
	return TestFlag(WndFlags, GWF_SUNKEN);
}

void GView::Sunken(bool i)
{
	if (i) SetFlag(WndFlags, GWF_SUNKEN);
	else ClearFlag(WndFlags, GWF_SUNKEN);

	if (NOT _BorderSize AND i)
	{
		_BorderSize = 2;
	}

	Invalidate();
}

bool GView::Flat()
{
	return TestFlag(WndFlags, GWF_FLAT);
}

void GView::Flat(bool i)
{
	if (i) SetFlag(WndFlags, GWF_FLAT);
	else ClearFlag(WndFlags, GWF_FLAT);
	Invalidate();
}

bool GView::Raised()
{
	return TestFlag(WndFlags, GWF_RAISED);
}

void GView::Raised(bool i)
{
	if (i) SetFlag(WndFlags, GWF_RAISED);
	else ClearFlag(WndFlags, GWF_RAISED);

	if (NOT _BorderSize AND i)
	{
		_BorderSize = 2;
	}

	Invalidate();
}

GViewFill *GView::GetForegroundFill()
{
	return d->Foreground;
}

bool GView::SetForegroundFill(GViewFill *Fill)
{
	if (Fill != d->Foreground)
	{
		DeleteObj(d->Foreground);
		d->Foreground = Fill;
		Invalidate();
	}
	
	return true;
}

GViewFill *GView::GetBackgroundFill()
{
	return d->Background;
}

bool GView::SetBackgroundFill(GViewFill *Fill)
{
	if (Fill != d->Background)
	{
		DeleteObj(d->Background);
		d->Background = Fill;
		Invalidate();
	}
	
	return true;
}

void GView::Quit(bool DontDelete)
{
	if (DontDelete)
	{
		Visible(false);
	}
	else
	{
		//_View->_SetDeleteMe();
	}
}

bool GView::SetCursor(int CursorId)
{
	d->CursorSet = true;
	GView *Wnd = GetWindow();
	
	if (Wnd AND
		Wnd->Handle() AND
		Wnd->d->CursorId != CursorId)
	{
		return true;
	}
	else
	{
		// printf("%s:%i - Param error"\n", __FILE__, __LINE__);
	}
	
	return false;
}

GFont *GView::GetFont()
{
	return d->Font ? d->Font : SysFont;
}

void GView::SetFont(GFont *Fnt, bool OwnIt)
{
	if (d->FontOwn)
	{
		DeleteObj(d->Font);
	}
	d->Font = Fnt;
	d->FontOwn = OwnIt;
}

bool GView::IsOver(GMouse &m)
{
	return	(m.x >= 0) AND
			(m.y >= 0) AND
			(m.x < Pos.X()) AND
			(m.y < Pos.Y());
}

bool GView::WindowVirtualOffset(GdcPt2 *Offset)
{
	bool Status = false;

	if (Offset)
	{
		Offset->x = 0;
		Offset->y = 0;
		
		for (GViewI *Wnd = this; Wnd; Wnd = Wnd->GetParent())
		{
			if (NOT Wnd->Handle())
			{
				GRect r = Wnd->GetPos();
				GViewI *Par = Wnd->GetParent();
				if (Par)
				{
					GRect c = Par->GetClient(false);
					Offset->x += r.x1 + c.x1;
					Offset->y += r.y1 + c.y1;
				}
				else
				{
					Offset->x += r.x1;
					Offset->y += r.y1;
				}

				Status = true;
			}
			else break;
		}
	}

	return Status;
}

int Depth = 0;
GViewI *GView::WindowFromPoint(int x, int y)
{
	// printf("WindowFromPoint %p,%s xy=%i,%i pos=%s\n", this, GetClassName(), x, y, GetPos().GetStr());
	
	Iterator<GViewI> it(&Children);
	for (GViewI *c = it.First(); c; c = it.Next())
	{
		GRect r = c->GetPos();
		
		if (r.Overlap(x, y) AND
			c->Visible())
		{
			GRect cli = c->GetClient(false);

			// printf("Inside %p,%s r=%s", c, c->GetClassName(), r.GetStr());
			// printf(" cli=%s\n", cli.GetStr());

			GViewI *Child = c->WindowFromPoint(x - r.x1 - cli.x1, y - r.y1 - cli.y1);
			if (Child)
			{
				return Child;
			}
			
			return c;
		}
	}

	if (x >= 0 AND y >= 0 AND x < Pos.X() AND y < Pos.Y())
	{
		// printf("Match %p,%s\n", this, GetClassName());
		return this;
	}

	return NULL;
}

GdcPt2 GView::GetMinimumSize()
{
	return d->MinimumSize;
}

void GView::SetMinimumSize(GdcPt2 Size)
{
	d->MinimumSize = Size;
}

bool GView::SetPos(GRect &p, bool Repaint)
{
	Pos = p;

	if (_View)
	{
		int o = 0;
		
		if (d->Parent AND (d->Parent->Sunken() OR d->Parent->Raised()))
		{
			GView *p = dynamic_cast<GView*>(d->Parent);
			if (p)
				o = p->_BorderSize;
		}
		
		MoveControl(_View, Pos.x1 + o, Pos.y1 + o);
		SizeControl(_View, Pos.X(), Pos.Y());
	}

	return true;
}

bool GView::Invalidate(GRegion *r, bool Repaint, bool NonClient)
{
	if (r)
	{
		for (int i=0; i<r->Length(); i++)
		{
			bool Last = i == r->Length()-1;
			Invalidate((*r)[i], Last ? Repaint : false, NonClient);
		}
		
		return true;
	}
	
	return false;
}

bool GView::Invalidate(GRect *r, bool Repaint, bool Frame)
{
	if (_View OR WindowHandle())
	{
		GRect Client = Frame ? GView::GetPos() : GView::GetClient();
		HIViewRef ViewRef = 0;
		if (WindowHandle())
		{
			OSStatus e = HIViewFindByID(HIViewGetRoot(WindowHandle()), kHIViewWindowContentID, &ViewRef);
			if (e)
			{
				LgiTrace("%s:%i - Getting content view failed (e=%i)\n", __FILE__, __LINE__, e);
			}
		}
		else
		{
			ViewRef = _View;
		}
		
		if (Repaint)
		{
			static bool Repainting = false;
			if (NOT Repainting)
			{
				Repainting = true;
				if (r)
				{
					GRect cr = *r;
					cr.Offset(Client.x1, Client.y1);
					
					Rect a = cr;
					OSStatus e = HIViewSetNeedsDisplay(ViewRef, true);
					if (e)
					{
						LgiTrace("%s:%i - HIViewSetNeedsDisplay failed (%i)\n", __FILE__, __LINE__, e);
					}
				}
				else
				{
					Rect a = GetClient();
					OSStatus e = HIViewSetNeedsDisplay(ViewRef, true);
					if (e)
					{
						LgiTrace("%s:%i - HIViewSetNeedsDisplay failed (%i)\n", __FILE__, __LINE__, e);
					}
				}
				Repainting = false;
			}
		}
		else
		{
			if (r)
			{
				GRect cr = *r;
				cr.Offset(Client.x1, Client.y1);
				Rect a = cr;
				OSStatus e = HIViewSetNeedsDisplay(ViewRef, true);
			}
			else
			{
				OSStatus e = HIViewSetNeedsDisplay(ViewRef, true);
			}
		}
		
		return true;
	}
	else
	{
		GRect Up;
		GViewI *p = this;

		if (r)
		{
			Up = *r;
		}
		else
		{
			Up.Set(0, 0, Pos.X()-1, Pos.Y()-1);
		}

		while (p != 0 AND p->Handle() != 0)
		{
			Up.Offset(p->GetPos().x1, p->GetPos().y1);
			p = p->GetParent();
		}

		if (p AND p->Handle())
		{
			LgiAssert(p != this);
				
			return p->Invalidate(&Up, Repaint);
		}
	}

	return false;
}

void GView::SetPulse(int Length)
{
	if (d->Pulse)
	{
		d->Pulse->Loop = false;
		d->Pulse = 0;
	}
	
	if (Length > 0)
	{
		d->Pulse = NEW(GPulseThread(this, Length));
	}
}

bool GView::InThread()
{
	OsThreadId Me = LgiGetCurrentThread();
	return LgiApp->GetGuiThread() == Me;
}

bool GView::PostEvent(int Cmd, int a, int b)
{
	OsView h = 0;
	if (_View)
	{
		h = _View;
	}
	else
	{
		GViewI *r = FindReal();
		if (r) h = r->Handle();
	}
	
	if (h)
		return LgiPostEvent(h, Cmd, a, b);
		
	printf("%s:%i - no handle to post event to.\n", __FILE__, __LINE__);
	return false;
}

int GView::OnEvent(GMessage *Msg)
{
	switch (Msg->m)
	{
		/*
		case M_X11_INVALIDATE:
		{
			X11_INVALIDATE_PARAMS *p = (X11_INVALIDATE_PARAMS*)MsgA(Msg);
			if (p AND p->View == this)
			{
				Invalidate(p->Rgn.Valid() ? &p->Rgn : 0, p->Repaint);
				DeleteObj(p);
			}
			break;
		}
		*/
		case M_PULSE:
		{
			OnPulse();
			break;
		}
		case M_CHANGE:
		{
			GView *Ctrl = dynamic_cast<GView*>((GView*) MsgA(Msg));
			if (Ctrl)
			{
				return OnNotify(Ctrl, MsgB(Msg));
			}
			break;
		}
		case M_COMMAND:
		{
			int32 a=MsgA(Msg);
			return OnCommand(a&0xFFFF, a>>16, (OsView) MsgB(Msg));
		}
		default:
		{
			break;
		}
	}

	return 0;
}

bool GView::Name(char *n)
{
	bool Status = GObject::Name(n);

	if (Status AND _View)
	{
		//_View->setText(n);
	}

	return Status;
}

char *GView::Name()
{
	return GObject::Name();
}

bool GView::NameW(char16 *n)
{
	bool Status = GObject::NameW(n);
	
	if (Status AND _View)
	{
		//_View->setText(n);
	}

	return Status;
}

char16 *GView::NameW()
{
	return GObject::NameW();
}

bool GView::Enabled()
{
	if (_View)
	{
		//return _View->isEnabled();
	}

	return TestFlag(WndFlags, GWF_ENABLED);
}

void GView::Enabled(bool e)
{
	if (e)
	{
		SetFlag(WndFlags, GWF_ENABLED);
	
		if (_View)
		{	
			OSStatus e = EnableControl(_View);
			if (e)
			{
				printf("%s:%i - Error enabling control (%i)\n", __FILE__, __LINE__);
			}
		}
	}
	else
	{
		ClearFlag(WndFlags, GWF_ENABLED);

		if (_View)
		{	
			OSStatus e = DisableControl(_View);
			if (e)
			{
				printf("%s:%i - Error disabling control (%i)\n", __FILE__, __LINE__);
			}
		}
	}
}

GDragDropTarget *&GView::DropTargetPtr()
{
	return d->DropTarget;
}

bool GView::DropTarget()
{
	return TestFlag(WndFlags, GWF_DROP_TARGET);
}

bool GView::DropTarget(bool i)
{
	bool Status = false;
	
	if (i) SetFlag(WndFlags, GWF_DROP_TARGET);
	else ClearFlag(WndFlags, GWF_DROP_TARGET);
	
	GViewI *Wnd = GetWindow();
	if (Wnd AND Wnd->Handle())
	{
		OsView w = Wnd->Handle();
	}	
}

bool GView::Visible()
{
	return TestFlag(WndFlags, GWF_VISIBLE);
}

void GView::Visible(bool v)
{
	bool Change = TestFlag(WndFlags, GWF_VISIBLE) ^ v;

	if (v) SetFlag(WndFlags, GWF_VISIBLE);
	else ClearFlag(WndFlags, GWF_VISIBLE);

	if (Change AND _View != 0)
	{
		OSErr e = SetControlVisibility(_View, v, true);
		if (e)
		{
			printf("%s:%i - SetControlVisibility(%p,%i,1) failed with %i (name=%s)\n", __FILE__, __LINE__, _View, v, e, Name());
		}
	}
}

bool GView::Focus()
{
	if (_View)
	{
		GViewI *Wnd = GetWindow();
		if (Wnd)
		{
			ControlRef Foc;
			OSErr e = GetKeyboardFocus(Wnd->WindowHandle(), &Foc);
			if (e) printf("%s:%i - error getting keyboard focus (%i)\n", __FILE__, __LINE__, e);
			else
			{
				if (Foc == _View)
				{
					SetFlag(WndFlags, GWF_FOCUS);
				}
				else
				{
					ClearFlag(WndFlags, GWF_FOCUS);
				}
			}
		}
		else printf("%s:%i - no window?\n", __FILE__, __LINE__);
	}
	else printf("%s:%i - no view handle?\n", __FILE__, __LINE__);

	return TestFlag(WndFlags, GWF_FOCUS);;
}

void GView::Focus(bool f)
{
	if (_View)
	{		
		GViewI *Wnd = GetWindow();
		if (Wnd)
		{
			if (f)
			{
				OSErr e = SetKeyboardFocus(Wnd ? Wnd->WindowHandle() : 0, _View, 1);
				if (e) printf("%s:%i - error setting keyboard focus (%i)\n", __FILE__, __LINE__, e);
				else
				{
					if (TestFlag(WndFlags, GWF_FOCUS) ^ f)
					{
						// printf("%s:%i - %p->OnFocus(%i)\n", __FILE__, __LINE__, this, f);
						OnFocus(true);
					}
				}
			}
			else
			{
				/*
				OSErr e = ClearKeyboardFocus(Wnd->WindowHandle());
				if (e) printf("%s:%i - error clearing keyboard focus (%i)\n", __FILE__, __LINE__, e);
				*/
			}
		}
		else printf("%s:%i - no window?\n", __FILE__, __LINE__);
	}

	if (f) SetFlag(WndFlags, GWF_FOCUS);
	else ClearFlag(WndFlags, GWF_FOCUS);
}

GViewI *GView::FindReal(GdcPt2 *Offset)
{
	if (Offset)
	{
		Offset->x = 0;
		Offset->y = 0;
	}

	GViewI *p = this;
	while (p AND NOT p->Handle())
	{
		if (Offset)
		{
			Offset->x += p->GetPos().x1;
			Offset->y += p->GetPos().y1;
		}

		p = p->GetParent();
	}
	
	if (p AND p->Handle())
	{
		return p;
	}

	return NULL;
}

bool GView::HandleCapture(GView *Wnd, bool c)
{
	bool Status = false;

	if (_View)
	{
		if (c)
		{
			GView *Cap = dynamic_cast<GView*>(_Capturing);
			if (Cap)
			{
				Cap->HandleCapture(NULL, false);
			}

			// _View->grabMouse();
			Status = true;
			_Capturing = Wnd;
		}
		else
		{
			if (_Capturing)
			{
				// _View->ungrabMouse();
				_Capturing = 0;
			}
		}
	}
	else
	{
		GView *Cap = dynamic_cast<GView*>(d->Parent);
		if (Cap)
		{
			Status = Cap->HandleCapture(Wnd, c);
		}
	}

	return Status;
}

bool GView::IsCapturing()
{
	return _Capturing == this;
}

bool GView::Capture(bool c)
{
	return HandleCapture(this, c);
}

void GView::PointToScreen(GdcPt2 &p)
{
	GViewI *c = this;
	int Ox = 0, Oy = 0;

	// Find offset to window
	while (c AND c != c->GetWindow())
	{
		GRect cli = c->GetClient(false);
		GRect pos = c->GetPos(); 
		Ox += pos.x1 + cli.x1;
		Oy += pos.y1 + cli.y1;
		
		c = c->GetParent();
	}
	
	if (c AND c->WindowHandle())
	{
		Rect r;
		GetWindowBounds(c->WindowHandle(), kWindowContentRgn, &r);
		p.x += r.left + Ox;
		p.y += r.top + Oy;
	}
	else
	{
		printf("%s:%i - No window handle to map to screen. c=%p\n", __FILE__, __LINE__, c);
	}
}

void GView::PointToView(GdcPt2 &p)
{
	GViewI *c = this;
	int Ox = 0, Oy = 0;

	// Find offset to window
	while (c AND c != c->GetWindow())
	{
		GRect cli = c->GetClient(false);
		GRect pos = c->GetPos(); 
		Ox += pos.x1 + cli.x1;
		Oy += pos.y1 + cli.y1;
		c = c->GetParent();
	}
	
	// Apply window position
	if (c AND c->WindowHandle())
	{
		Rect r;
		GetWindowBounds(c->WindowHandle(), kWindowContentRgn, &r);
		p.x -= r.left + Ox;
		p.y -= r.top + Oy;
		
		// printf("Pt2View o=%i,%i r=%i,%i\n", Ox, Oy, (int)r.left, (int)r.top);
	}
	else
	{
		printf("%s:%i - No window handle to map to view. c=%p\n", __FILE__, __LINE__, c);
	}
}

bool GView::GetMouse(GMouse &m, bool ScreenCoords)
{
	if (_View OR NOT GetParent())
	{
		Point p;
		::GetMouse(&p);
		
		if (ScreenCoords)
		{
			m.x = p.h;
			m.y	= p.v;
		}
		else
		{
			GdcPt2 n(p.h, p.v);
			PointToView(n);
			m.x = n.x;
			m.y = n.y;
		}

		m.ViewCoords = NOT ScreenCoords;
		m.Target = this;

		return true;
	}
	else if (GetParent())
	{
		bool s = GetParent()->GetMouse(m, ScreenCoords);
		if (s)
		{
			if (NOT ScreenCoords)
			{
				m.x -= Pos.x1;
				m.y -= Pos.y1;
			}
			return true;
		}
	}
	
	return false;
}

bool GView::AttachChildren()
{
	bool Status = true;

	Iterator<GViewI> It(&Children);
	for (GViewI *c=It.First(); c; c=It.Next())
	{
		if (NOT c->IsAttached())
		{
			Status |= c->Attach(this);
		}
	}

	return Status;
}

bool GView::IsAttached()
{
	if (GetWindow() == this)
		return WindowHandle() != 0;
	else
		return Handle() != 0;
}

void GView::OnNcPaint(GSurface *pDC, GRect &r)
{
	int Border = Sunken() OR Raised() ? _BorderSize : 0;
	if (Border == 2)
	{
		LgiWideBorder(pDC, r, Sunken() ? SUNKEN : RAISED);
	}
	else if (Border == 1)
	{
		LgiThinBorder(pDC, r, Sunken() ? SUNKEN : RAISED);
	}
	// r.Offset(Border, Border);
}

void GView::_Paint(GSurface *pDC)
{
	GSurface *Mem = 0;
	if (NOT pDC)
	{
		pDC = Mem = NEW(GScreenDC(this));
	}
	
	GRect r(0, 0, Pos.X()-1, Pos.Y()-1);

	// Non-Client drawing
	GRect c = r;
	OnNcPaint(pDC, c);
	bool HasEdge = GetParent() AND (c != r);
	if (HasEdge)
	{
		pDC->SetClient(&c);
	}

	// Paint client
	OnPaint(pDC);

	// Paint any virtual children
	int x, y;
	pDC->GetOrigin(x, y);
	GRect Clip = pDC->ClipRgn();

	Iterator<GViewI> ChildList(&Children); // just in case the children access the child list
	for (GViewI *w = ChildList.First(); w; w = ChildList.Next())
	{
		GView *v = dynamic_cast<GView*>(w);
		if (v AND NOT v->Handle() AND v->Visible())
		{
			GRect r = v->GetPos();
			pDC->SetClient(&r);
			// printf("child client=%s name=%s\n", r.GetStr(), v->Name());
			v->_Paint(pDC);
			pDC->SetClient(0);
		}
	}
	
	// Clean up
	if (HasEdge)
	{
		pDC->SetClient(0);
	}	
	DeleteObj(Mem);
}

int VirtualKeyToLgi(UInt32 Virt)
{
	switch (Virt)
	{
		// various
		case 122: return VK_F1;
		case 120: return VK_F2;
		case 99: return VK_F3;
		case 118: return VK_F4;
		case 96: return VK_F5;
		case 97: return VK_F6;
		case 98: return VK_F7;
		case 100: return VK_F8;
		case 101: return VK_F9;
		case 109: return VK_F10;
		case 103: return VK_F11;
		case 110: return VK_APPS;
		case 111: return VK_F12;
		
		case 123: return VK_LEFT;
		case 124: return VK_RIGHT;
		case 125: return VK_DOWN;
		case 126: return VK_UP;

		case 53: return VK_ESCAPE;
		case 51: return VK_BACKSPACE;
		
		// whitespace
		case 36: return '\r';
		case 48: return '\t';
		case 49: return ' ';
		
		// delimiters
		case 50: return '`';
		case 27: return '-';
		case 24: return '=';
		case 42: return '\\';
		
		// digits
		case 18: return '1';
		case 19: return '2';
		case 20: return '3';
		case 21: return '4';
		case 23: return '5';
		case 22: return '6';
		case 26: return '7';
		case 28: return '8';
		case 25: return '9';
		case 29: return '0';
		
		// alpha
		case 0: return 'a';
		case 11: return 'b';
		case 8: return 'c';
		case 2: return 'd';
		case 14: return 'e';
		case 3: return 'f';
		case 5: return 'g';
		case 4: return 'h';
		case 34: return 'i';
		case 38: return 'j';
		case 40: return 'k';
		case 37: return 'l';
		case 46: return 'm';
		case 45: return 'n';
		case 31: return 'o';
		case 35: return 'p';
		case 12: return 'q';
		case 15: return 'r';
		case 1: return 's';
		case 17: return 't';
		case 32: return 'u';
		case 9: return 'v';
		case 13: return 'w';
		case 7: return 'x';
		case 16: return 'y';
		case 6: return 'z';
		
		default:
			printf("%s:%i - unimplemented virt->lgi code mapping: %i\n", __FILE__, __LINE__, Virt);
			break;
	}
	
	return 0;
}

#define GViewThisPtr		'gvtp'

struct WasteOfSpace
{
	HIViewRef Obj;
	GView *v;
};

pascal
OSStatus
CarbonControlProc
(
	EventHandlerCallRef nextHandler, 
	EventRef inEvent,
	void * userData
)
{
	OSStatus Status = eventNotHandledErr;
	UInt32 eventClass = GetEventClass(inEvent);
	UInt32 eventKind = GetEventKind(inEvent);
	
	WasteOfSpace *d = (WasteOfSpace *) userData;
	GView *v = d ? d->v : 0;

	switch (eventClass)
	{
		case kEventClassHIObject:
		{
			switch (eventKind)
			{
				case kEventHIObjectInitialize:
				{
					// CallNextEventHandler(nextHandler, inEvent);
					OSStatus e = GetEventParameter(inEvent, GViewThisPtr, typeVoidPtr, NULL, sizeof(v), NULL, &v);
					if (e)
					{
						printf("%s:%i - No GViewThisPtr prop.\n", __FILE__, __LINE__);
						return eventNotHandledErr;
					}
					
					d->v = v;
					v->_View = (OsView)d->Obj;
					
					int Offset = 0;
					GViewI *Par = v->GetParent();
					if (Par AND (Par->Sunken() OR Par->Raised()))
					{
						GView *View = dynamic_cast<GView*>(Par);
						if (View)
						{
							Offset = View->_BorderSize;
						}
					}
					
					GRect r = v->GetPos();
					MoveControl(v->Handle(), r.x1 + Offset, r.y1 + Offset);
					SizeControl(v->Handle(), r.X(), r.Y());

					v->OnCreate();
					v->OnAttach();
					break;
				}
				case kEventHIObjectConstruct:
				{
					d = NEW(WasteOfSpace);
					if (d)
					{
						d->v = 0;

						OSStatus e = GetEventParameter(	inEvent,
														kEventParamHIObjectInstance,
														typeHIObjectRef,
														NULL,
														sizeof(HIObjectRef),
														NULL,
														&d->Obj);
						if (e)
						{
							printf("%s:%i - failed to get view ptr (e=%i).\n", __FILE__, __LINE__, e);
							d->Obj = 0;
						}

						Status = SetEventParameter(inEvent, kEventParamHIObjectInstance, typeVoidPtr, sizeof(d), &d); 
					}
					break;
				}
				case kEventHIObjectDestruct:
				{
					DeleteObj(d);
					Status = noErr;
					break;
				}
			}
			break;
		}		
		case kEventClassControl:
		{
			if (NOT v)
			{
				LgiTrace("%s:%i - No view ptr.\n", __FILE__, __LINE__);
				return Status;
			}

			switch (eventKind)
			{
				case kEventControlBoundsChanged:
				{
					v->OnPosChange();
					break;
				}
				case kEventControlInitialize:
				{
					UInt32 features = 0;
					Status = GetEventParameter(inEvent, kEventParamControlFeatures, typeUInt32, NULL, sizeof(features), NULL, &features);
					if (Status == noErr)
						features |= kControlSupportsEmbedding;
					else
						features = kControlSupportsEmbedding;

					Status = SetEventParameter(inEvent, kEventParamControlFeatures, typeUInt32, sizeof features, &features);
					break;
				}
		  		case kEventControlDraw:
				{
					CGContextRef Ctx;
					Status = GetEventParameter(	inEvent,
												kEventParamCGContextRef,
												typeCGContextRef,
												NULL,
												sizeof(CGContextRef),
												NULL,
												&Ctx);
					if (NOT Status)
					{
						GScreenDC dc(v, Ctx);
						v->_Paint(&dc);
					}
					break;
				}
				case kEventControlHitTest:
				{
					HIPoint  pt;
					HIRect  bounds;

					Status = noErr;

					// the point parameter is in view-local coords.
					OSStatus status = GetEventParameter(inEvent, kEventParamMouseLocation, typeHIPoint, NULL, sizeof(pt), NULL, &pt);
					HIViewGetBounds(v->Handle(), &bounds);
					if (CGRectContainsPoint(bounds, pt))
					{
						ControlPartCode part = kControlButtonPart;
						status = SetEventParameter(inEvent, kEventParamControlPart, typeControlPartCode, sizeof(part), &part);
					}
					break;
				}
				case kEventControlHiliteChanged:
				{
					HIViewSetNeedsDisplay(v->Handle(), true);
					break;
				}
				case kEventControlGetFocusPart:
				{
					printf("kEventControlGetFocusPart ??\n");
					break;
				}
				case kEventControlSetFocusPart:
				{
					ControlPartCode Part;
					OSStatus e = GetEventParameter(inEvent, kEventParamControlPart, typeControlPartCode, NULL, sizeof(Part), NULL, &Part);
					if (e) printf("%s:%i - error (%i)\n", __FILE__, __LINE__, e);
					
					bool f = Part != kControlFocusNoPart;
					if (TestFlag(v->WndFlags, GWF_FOCUS) ^ f)
					{
						// printf("%s:%i - %p->OnFocus(%i)\n", __FILE__, __LINE__, v, f);
						v->OnFocus(f);
						if (f) SetFlag(v->WndFlags, GWF_FOCUS);
						else ClearFlag(v->WndFlags, GWF_FOCUS);
					}
					
					Status = noErr;
					break;
				}
			}
			break;
		}
		case kEventClassKeyboard:
		{
			/*
			kEventClassKeyboard quick reference:
			
				kEventRawKeyDown
				kEventRawKeyRepeat
				kEventRawKeyUp
				kEventRawKeyModifiersChanged
				kEventHotKeyPressed
				kEventHotKeyReleased
			*/

			switch (eventKind)
			{
				case kEventRawKeyDown:
				{
					UInt32		key = 0;
					UInt32		mods = 0;

					OSStatus e = GetEventParameter(inEvent, kEventParamKeyCode, typeUInt32, NULL, sizeof(UInt32), NULL, &key);
					if (e) printf("%s:%i - error %i\n", __FILE__, __LINE__, e);
					e = GetEventParameter(inEvent, kEventParamKeyModifiers, typeUInt32, NULL, sizeof(mods), NULL, &mods);
					if (e) printf("%s:%i - error %i\n", __FILE__, __LINE__, e);
					
					GKey k;
					if (k.c16 = VirtualKeyToLgi(key))
					{
						k.Down(false);
						if (mods & 0x200) k.Shift(true);
						if (mods & 0x1000) k.Ctrl(true);
						if (mods & 0x800) k.Alt(true);

						if (k.c16 == VK_APPS AND v->Focus())
						{
							// printf("RawDown key=%i sh=%i,alt=%i,ctrl=%i v=%p\n", k.c16, k.Shift(), k.Alt(), k.Ctrl(), v->Handle());

							GWindow *Wnd = v->GetWindow();
							if (Wnd) Wnd->HandleViewKey(v, k);
							else v->OnKey(k);
						}
					}
					
					// Status = noErr;
					break;
				}
				case kEventRawKeyRepeat:
				{
					break;
				}
				case kEventRawKeyUp:
				{
					UInt32		key = 0;
					UInt32		mods = 0;

					OSStatus e = GetEventParameter(inEvent, kEventParamKeyCode, typeUInt32, NULL, sizeof(UInt32), NULL, &key);
					if (e) printf("%s:%i - error %i\n", __FILE__, __LINE__, e);
					e = GetEventParameter(inEvent, kEventParamKeyModifiers, typeUInt32, NULL, sizeof(mods), NULL, &mods);
					if (e) printf("%s:%i - error %i\n", __FILE__, __LINE__, e);
					
					GKey k;
					if (k.c16 = VirtualKeyToLgi(key))
					{
						k.Down(false);
						if (mods & 0x200) k.Shift(true);
						if (mods & 0x1000) k.Ctrl(true);
						if (mods & 0x800) k.Alt(true);

						// printf("RawUp key=%i sh=%i,alt=%i,ctrl=%i\n", k.c16, k.Shift(), k.Alt(), k.Ctrl());
						
						GWindow *Wnd = v->GetWindow();
						if (Wnd) Wnd->HandleViewKey(v, k);
						else v->OnKey(k);
					}
					
					Status = noErr;
					break;
				}
			}
			break;
		}
		case kEventClassTextInput:
		{
			/*
			kEventClassTextInput quick reference:
				
				kEventTextInputUpdateActiveInputArea
				kEventTextInputUnicodeForKeyEvent
				kEventTextInputOffsetToPos
				kEventTextInputPosToOffset
				kEventTextInputShowHideBottomWindow
				kEventTextInputGetSelectedText
				kEventTextInputUnicodeText
				kEventTextInputFilterText
			*/
		
			switch (eventKind)
			{
				case kEventTextInputUnicodeForKeyEvent:
				{
					UniChar		*text;
					UInt32		actualSize;
					EventRef	src;
					UInt32		mods = 0;
					UInt32		key = 0;
	 
					OSStatus e = GetEventParameter(inEvent, kEventParamTextInputSendText, typeUnicodeText, NULL, 0, &actualSize, NULL);
					text = new UniChar[actualSize+1];
					e = GetEventParameter(inEvent, kEventParamTextInputSendText, typeUnicodeText, NULL, actualSize, NULL, text);
					e = GetEventParameter(inEvent, kEventParamTextInputSendKeyboardEvent, typeEventRef, NULL, sizeof(src), NULL, &src);
					if (e) printf("%s:%i - error %i\n", __FILE__, __LINE__, e);
					else
					{
						e = GetEventParameter(src, kEventParamKeyModifiers, typeUInt32, NULL, sizeof(mods), NULL, &mods);
						if (e) printf("%s:%i - error %i\n", __FILE__, __LINE__, e);
						e = GetEventParameter(src, kEventParamKeyCode, typeUInt32, NULL, sizeof(UInt32), NULL, &key);
						if (e) printf("%s:%i - error %i\n", __FILE__, __LINE__, e);
					}
					
					GKey k;

					char16 *utf = (char16*) text;
					int size = actualSize;
					k.c16 = LgiUtf16To32(utf, size);
					if (k.c16 == 16)
					{
						int c = VirtualKeyToLgi(key);
						if (c) k.c16 = c;
					}
					
					k.IsChar =	(mods & 0x100) == 0 AND
								(
									k.c16 >= ' ' OR
									k.c16 == VK_RETURN OR
									k.c16 == VK_TAB OR
									k.c16 == VK_BACKSPACE
								);
					k.Down(true);
					if (mods & 0x200) k.Shift(true);
					if (mods & 0x1000) k.Ctrl(true);
					if (mods & 0x800) k.Alt(true);
					if (mods & 0x100) k.System(true);

					// printf("Key Txt=%i Len=%i Modifiers=%x\n", k.c16, actualSize, mods);

					GWindow *Wnd = v->GetWindow();
					if (Wnd) Wnd->HandleViewKey(v, k);
					else v->OnKey(k);
					
					DeleteArray(text);
					Status = noErr;
					break;
				}
			}
			break;
		}
		case kEventClassUser:
		{
			if (eventKind == kEventUser)
			{
				GMessage m;
				
				OSStatus status = GetEventParameter(inEvent, kEventParamLgiEvent, typeUInt32, NULL, sizeof(UInt32), NULL, &m.m);
				status = GetEventParameter(inEvent, kEventParamLgiA, typeUInt32, NULL, sizeof(UInt32), NULL, &m.a);
				status = GetEventParameter(inEvent, kEventParamLgiB, typeUInt32, NULL, sizeof(UInt32), NULL, &m.b);
				
				v->OnEvent(&m);
				
				Status = noErr;
			}
			break;
		}
		default:
		{
			printf("%s:%i - Unhandled event %i,%i\n", __FILE__, __LINE__, eventClass, eventKind);
			break;
		}
	}
	
	return Status;
}

HIObjectClassRef ViewClass = 0;

bool GView::_Attach(GViewI *parent)
{
	bool Status = false;

	d->Parent = parent;

	_Window = d->Parent ? d->Parent->GetWindow() : 0;
	if (_Window)
	{
		_Lock = _Window->_Lock;
	}
	
	if (d->Parent AND d->Parent->IsAttached())
	{
		GWindow *w = GetWindow();
		if (w)
		{
		
			OSStatus e;
			CFStringRef Class = CFSTR("com.memecode.LgiView");
			if (NOT ViewClass)
			{
				static EventTypeSpec LgiViewEvents[] =
				{
					{ kEventClassHIObject, kEventHIObjectConstruct },
					{ kEventClassHIObject, kEventHIObjectInitialize },
					{ kEventClassHIObject, kEventHIObjectDestruct },
					{ kEventClassControl, kEventControlInitialize },
					{ kEventClassControl, kEventControlHitTest },
					{ kEventClassControl, kEventControlHiliteChanged },
					{ kEventClassControl, kEventControlDraw },
					// { kEventClassControl, kEventControlGetFocusPart },
					{ kEventClassControl, kEventControlSetFocusPart },
					{ kEventClassControl, kEventControlBoundsChanged },
					{ kEventClassTextInput, kEventTextInputUnicodeForKeyEvent },
					{ kEventClassKeyboard, kEventRawKeyUp },
					{ kEventClassKeyboard, kEventRawKeyDown },
					{ kEventClassUser, kEventUser }
				};

				e = HIObjectRegisterSubclass(	Class,
												kHIViewClassID,
												0,
												CarbonControlProc,
												GetEventTypeCount(LgiViewEvents),
												LgiViewEvents,
												0,
												&ViewClass);
				if (e)
				{
					LgiTrace("%s:%i - HIObjectRegisterSubclass failed with '%i'.\n", __FILE__, __LINE__, e);
					LgiExitApp();
				}
			}
			
			if (ViewClass)
			{
				EventRef Ev;
				e = CreateEvent(NULL,
								kEventClassHIObject,
								kEventHIObjectInitialize,
								GetCurrentEventTime(),
								0,
								&Ev);
				GView *Myself = this;
				e = SetEventParameter(Ev, GViewThisPtr, typeVoidPtr, sizeof(Myself), &Myself); 
													
				e = HIObjectCreate(Class, Ev, (HIObjectRef*)&_View);
				if (NOT e AND _View != 0)
				{
					// Set the view id
					SetControlCommandID(_View, GetId());
					
					// Get the parent viewref
					HIViewRef Par = d->Parent->Handle();
					if (NOT Par)
					{
						// This view will connect to a top-level window
						WindowRef Wnd = d->Parent->WindowHandle();
						if (Wnd)
						{
							HIViewRef ViewRef = HIViewGetRoot(Wnd);
							if (ViewRef)
							{
								e = HIViewFindByID(ViewRef, kHIViewWindowContentID, &Par);
								if (e)
								{
									LgiTrace("%s:%i - HIViewFindByID(%p,%p) failed with '%i'.\n", __FILE__, __LINE__, Par, _View, e);
								}
							}
							else
							{
								LgiTrace("%s:%i - HIViewGetRoot failed.\n", __FILE__, __LINE__);
							}
						}
						else
						{
							LgiTrace("%s:%i - No window handle.\n", __FILE__, __LINE__);
						}
					}
					
					if (Par)
					{
						// Attach the view to the parent view
						e = HIViewAddSubview(Par, _View);
						if (e)
						{
							LgiTrace("%s:%i - HIViewAddSubview(%p,%p) failed with '%i' (name=%s).\n", __FILE__, __LINE__, Par, _View, e, Name());
						}
						else
						{						
							// Move the view into position
							// Set flags and show it
							if (NOT TestFlag(WndFlags, GWF_ENABLED))
							{
								DisableControl(_View);
							}
							if (TestFlag(WndFlags, GWF_VISIBLE))
							{
								ShowControl(_View);
							}
							
							Status = true;
						}
					}
					else
					{
						printf("%s:%i - No parent to attach to\n", __FILE__, __LINE__);
					}
				}
				else
				{
					printf("%s:%i - HIObjectCreate failed (%i)\n", __FILE__, __LINE__, e);
				}
			}
		}
		else
		{
			printf("%s:%i - No window to attach to.\n", __FILE__, __LINE__);
		}
	}
	else
	{
		// Virtual attach
		Status = true;
	}

	return Status;
}

bool GView::Attach(GViewI *parent)
{
	if (_Attach(parent))
	{
		if (d->Parent AND NOT d->Parent->HasView(this))
		{
			d->Parent->AddView(this);
			d->Parent->OnChildrenChanged(this, true);
		}
		
		return true;
	}
	
	return false;
}

bool GView::Detach()
{
	bool Status = false;

	// Detach view
	if (d->Parent)
	{
		// Delete the OS's handle
		OSStatus e = HIViewRemoveFromSuperview(_View);
		// if (e) printf("%s:%i - HIViewRemoveFromSuperview failed with %i\n", __FILE__, __LINE__, e);
		
		// Events
		int D = 0;
		while (d->Parent->HasView(this))
		{
			d->Parent->DelView(this);
			D++;
		}

		if (D > 1)
		{
			printf("Error: GView=%p was in %p's list %i times.\n", this, d->Parent, D);
		}

		d->Parent->OnChildrenChanged(this, false);
		d->Parent = 0;
	}

	Status = true;

	return Status;
}
 
GViewI *GView::FindControl(OsView hCtrl)
{
	if (Handle() == hCtrl)
	{
		return this;
	}

	Iterator<GViewI> List(&Children);
	for (GViewI *c = List.First(); c; c = List.Next())
	{
		GViewI *Ctrl = c->FindControl(hCtrl);
		if (Ctrl)
		{
			return Ctrl;
		}
	}
	return 0;
}

GViewI *GView::FindControl(int Id)
{
	if (GetId() == Id)
	{
		return this;
	}

	Iterator<GViewI> List(&Children);
	for (GViewI *c = List.First(); c; c = List.Next())
	{
		GViewI *Ctrl = c->FindControl(Id);
		if (Ctrl)
		{
			return Ctrl;
		}
	}
	return 0;
}

#define MIN_WINDOW_POS	24

void GView::MoveOnScreen()
{
	GRect Screen(0, 0, GdcD->X()-1, GdcD->Y()-1);
	GRect p = GetPos();
	GRect Old = p;

	if (p.x2 >= Screen.X())
		p.Offset(Screen.X() - p.x2, 0);
	if (p.y2 >= Screen.Y())
		p.Offset(0, Screen.Y() - p.y2);
	if (p.x1 < 0)
		p.Offset(-p.x1, 0);
	if (p.y1 < MIN_WINDOW_POS)
		p.Offset(0, MIN_WINDOW_POS-p.y1);

	if (Old != p)
	{
		SetPos(p);
	}
}

GdcPt2 &GView::GetWindowBorderSize()
{
	static GdcPt2 p;
	return p;
}

void GView::MoveToCenter()
{
	GRect Screen(0, 0, GdcD->X()-1, GdcD->Y()-1);
	GRect p = GetPos();

	p.Offset(-p.x1, -p.y1);
	p.Offset((Screen.X() - p.X()) / 2, (Screen.Y() - p.Y()) / 2);

	SetPos(p);
}

void GView::MoveToMouse()
{
	GMouse m;
	if (GetMouse(m, true))
	{
		GRect p = GetPos();

		p.Offset(-p.x1, -p.y1);
		p.Offset(m.x-(p.X()/2), m.y-(p.Y()/2));

		SetPos(p);
	}
}

///////////////////////////////////////////////////////////////////
bool LgiIsMounted(char *Name)
{
	return false;
}

bool LgiMountVolume(char *Name)
{
	return false;
}

//////////////////////////////////////////////////////////////////////////
List<GViewFactory> *GViewFactory::Factories;

GViewFactory::GViewFactory()
{
	if (NOT Factories) Factories = NEW(List<GViewFactory>);
	if (Factories)
	{
		if (Factories->HasItem(this))
		{
			printf("%s:%i - Factory list already has '%p'\n", __FILE__, __LINE__, this);
		}
		else
		{
			Factories->Insert(this);
		}
	}
}

GViewFactory::~GViewFactory()
{
	if (Factories) Factories->Delete(this);
}

GView *GViewFactory::Create(char *Class, GRect *Pos, char *Text)
{
	if (Factories)
	{
		for (GViewFactory *f=Factories->First(); f; f=Factories->Next())
		{
			GView *v = f->NewView(Class, Pos, Text);
			if (v)
			{
				return v;
			}
		}
	}

	return 0;
}


////////////////////////////////
uint32 CursorData[] = {
0x02020202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020102, 0x02020202, 0x02020202, 0x01010101, 0x01010101, 0x01010101, 0x02020202, 0x02020202, 0x02010101, 0x02010101, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010102, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020102, 0x01020202, 0x02020202, 0x02020202, 0x02020202, 0x01010101, 0x01010101, 0x02020201, 0x02020202, 0x01010101, 0x01010101, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020102, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x02010101, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x02010102, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x01010101, 0x02020202, 0x02020202, 0x02020101, 0x02020202, 
0x02020202, 0x02020202, 0x02020202, 0x02020101, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x00000001, 0x00000000, 0x01000000, 0x02020202, 0x02020202, 0x01000001, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020101, 0x01020202, 0x02020201, 0x02020202, 0x02020202, 0x00000102, 0x00000000, 0x02020201, 0x02020202, 0x00000001, 0x01000000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 
0x02020202, 0x02020202, 0x00010102, 0x00000000, 0x02020101, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x00010001, 0x00010001, 0x01000001, 0x02020202, 0x02020202, 0x00010102, 0x02020101, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x00000000, 0x02020201, 0x02020202, 0x02020202, 0x01020202, 0x02020100, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x00010202, 0x00000000, 0x02020201, 0x02020202, 0x00000001, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x01000000, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 
0x00010202, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x00000001, 0x00000000, 0x02010000, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x01000102, 0x00000100, 0x02010000, 0x02020202, 0x02020202, 0x00010202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x00000000, 0x02020100, 0x02020202, 0x02020202, 0x00010202, 0x02020100, 0x01020202, 0x02010000, 0x02020202, 0x02020202, 0x01020202, 0x00000000, 0x02020201, 0x02020202, 0x00000001, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x00000000, 
0x02020201, 0x02020202, 0x02020202, 0x01020202, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x00000102, 0x01000000, 0x02020202, 0x02020202, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x00000000, 0x00010101, 0x01000000, 0x02020202, 0x00000001, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x00000001, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x00010202, 0x00000001, 0x02020100, 0x02020202, 0x02020202, 0x00010202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x00000102, 0x00000000, 0x02010000, 0x02020202, 0x02020202, 0x00000102, 0x01010100, 0x01010101, 0x01000000, 0x02020202, 0x02020201, 0x00010202, 0x00000000, 0x02020201, 0x02020202, 
0x00000001, 0x02010000, 0x02020202, 0x02020201, 0x02020202, 0x02020202, 0x01010102, 0x01010001, 0x02020101, 0x02020202, 0x02020202, 0x01020201, 0x02010000, 0x02020102, 0x02020202, 0x02020202, 0x01010101, 0x01010100, 0x02020201, 0x02020202, 0x02020202, 0x01000001, 0x02020101, 0x02020202, 0x02020202, 0x00010202, 0x01010000, 0x01020202, 0x00000001, 0x02020201, 0x00000001, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x00000001, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x00000102, 0x01000000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x01020202, 0x00000000, 0x02020201, 0x02020202, 0x02020202, 0x00010202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x01010101, 0x01000001, 0x01010101, 0x02020202, 0x02020202, 0x00000001, 0x00000000, 
0x00000000, 0x00000000, 0x02020201, 0x02020101, 0x00000102, 0x00000100, 0x02020201, 0x02020202, 0x01000001, 0x01000000, 0x01020202, 0x02020201, 0x02020202, 0x02020202, 0x02020201, 0x02010001, 0x02010202, 0x02020202, 0x01020202, 0x01020201, 0x02010000, 0x02010102, 0x02020202, 0x02020202, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x00000001, 0x01010000, 0x02020101, 0x02020202, 0x00000102, 0x01000000, 0x02020202, 0x00000102, 0x02020100, 0x00000001, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x00000001, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x00000102, 0x01000000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x02020201, 0x02020202, 0x02020202, 
0x02020202, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x00000001, 0x00000000, 0x00000000, 0x00000000, 0x02020201, 0x02010001, 0x00000001, 0x00010201, 0x02020201, 0x02020202, 0x02010001, 0x00000001, 0x00010201, 0x02020201, 0x02020202, 0x01020202, 0x02020201, 0x02010001, 0x01010202, 0x02020202, 0x00010202, 0x01020201, 0x02010000, 0x01000102, 0x02020202, 0x02020202, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x02010102, 0x00000001, 0x00000000, 0x02010000, 0x02020202, 0x00000102, 0x00000001, 0x02020201, 0x00010202, 0x02020100, 0x00000001, 0x01000000, 0x02020202, 0x02020202, 0x02020202, 0x00000001, 0x01000000, 0x02020202, 0x02020202, 0x02020202, 0x00000001, 0x00000000, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x02020202, 
0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x00000102, 0x01010100, 0x01010101, 0x01000000, 0x02020202, 0x01000001, 0x01000000, 0x01020202, 0x02020201, 0x02020202, 0x02020101, 0x00000102, 0x00000100, 0x02020201, 0x02020202, 0x00010202, 0x02020201, 0x02010001, 0x00010202, 0x02020201, 0x00000102, 0x01010101, 0x01010000, 0x00000101, 0x02020201, 0x01010101, 0x01010101, 0x01010100, 0x01010101, 0x02020201, 0x01000001, 0x00000001, 0x00000000, 0x02010000, 0x02020202, 0x00000001, 0x00000101, 0x02020100, 0x00010202, 0x02010000, 0x00000001, 0x00000000, 0x02020201, 0x02020202, 0x02020202, 0x00000001, 0x00000000, 0x02020201, 0x02020202, 0x02020202, 0x00000001, 0x00000000, 0x02020201, 
0x02020202, 0x02020202, 0x01010102, 0x01010101, 0x01010001, 0x01010101, 0x02020101, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x02020100, 0x01020202, 0x02010000, 0x02020202, 0x00000001, 0x02010000, 0x02020202, 0x02020201, 0x02020202, 0x02020201, 0x00010202, 0x00000000, 0x02020201, 0x02020202, 0x00000102, 0x01010101, 0x01010001, 0x00010101, 0x02020100, 0x00000001, 0x00000000, 0x00000000, 0x00000000, 0x02020100, 0x00000001, 0x00000000, 0x00000000, 0x00000000, 0x02020201, 0x00000001, 0x00000000, 0x00000000, 0x02010000, 0x02020202, 0x01000001, 0x00010202, 0x02010000, 0x01020202, 0x02010000, 0x00000001, 0x00000000, 0x02020100, 0x02020202, 0x02020202, 
0x00000001, 0x00000000, 0x02020100, 0x02020202, 0x02020202, 0x01010101, 0x01010100, 0x02020201, 0x02020202, 0x02020202, 0x00000001, 0x00000000, 0x00000000, 0x00000000, 0x02010000, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x02020100, 0x01020202, 0x02020100, 0x02020202, 0x00000001, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x00000000, 0x02020201, 0x02020202, 0x00000001, 0x00000000, 0x00000000, 0x00000000, 0x02010000, 0x00000102, 0x01010101, 0x01010000, 0x00000101, 0x02020201, 0x00000001, 0x00000000, 0x00000000, 0x00000000, 0x02020201, 0x00000102, 0x00000000, 0x00000000, 0x02010000, 0x02020202, 0x01000001, 0x01020202, 
0x01000000, 0x01020202, 0x02010000, 0x00000001, 0x00000000, 0x02010000, 0x02020202, 0x02020202, 0x00000001, 0x00000000, 0x02010000, 0x02020202, 0x02020202, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x01010102, 0x01010101, 0x01010001, 0x01010101, 0x02020101, 0x01020202, 0x00000000, 0x02020201, 0x02020202, 0x02020202, 0x00010202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020101, 0x01020202, 0x02020201, 0x02020202, 0x00000001, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x00000000, 0x02020201, 0x02020202, 0x00000102, 0x01010101, 0x01010001, 0x00010101, 0x02020100, 0x00010202, 0x01020201, 0x02010000, 0x01000102, 0x02020202, 0x01010101, 0x01010101, 0x01010100, 0x01010101, 
0x02020201, 0x00010202, 0x00000000, 0x00000000, 0x02010000, 0x02020202, 0x01000001, 0x02020202, 0x00000001, 0x01020201, 0x02010000, 0x00000001, 0x01000000, 0x02010101, 0x02020202, 0x02020202, 0x00000001, 0x01000000, 0x02010101, 0x02020202, 0x02020202, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x00010202, 0x01000001, 0x02020100, 0x02020202, 0x02020202, 0x00010202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x01010101, 0x01000001, 0x01010101, 0x02020202, 0x02020202, 0x02020202, 0x02020102, 0x01020202, 0x02020202, 0x02020202, 0x00000001, 0x01000000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x00000102, 0x00000000, 0x02020201, 0x02020202, 0x00010202, 0x02020201, 0x02010001, 0x00010202, 0x02020201, 0x01020202, 
0x01020201, 0x02010000, 0x02010102, 0x02020202, 0x02020202, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x01020202, 0x00000000, 0x00000000, 0x02010000, 0x02020202, 0x00000001, 0x02020201, 0x00000102, 0x00010100, 0x02010000, 0x00000001, 0x01000001, 0x01020202, 0x01010101, 0x01010101, 0x00000001, 0x01000001, 0x01020202, 0x01010101, 0x01010101, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x01000102, 0x01000001, 0x02010001, 0x02020202, 0x02020202, 0x00010202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x00000102, 0x00000000, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01010101, 0x01010101, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x01010101, 0x01010101, 
0x02020201, 0x02020202, 0x01020202, 0x02020201, 0x02010001, 0x01010202, 0x02020202, 0x02020202, 0x01020201, 0x02010000, 0x02020102, 0x02020202, 0x02020202, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x00000001, 0x00000000, 0x02020100, 0x02020202, 0x00000102, 0x02020201, 0x00010202, 0x00010000, 0x02020100, 0x01000001, 0x01000001, 0x01020202, 0x00000000, 0x01000000, 0x01000001, 0x01000001, 0x01020202, 0x00000000, 0x01000000, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x00010001, 0x00000000, 0x01000100, 0x02020202, 0x02020202, 0x00010202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x00000000, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 
0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020201, 0x02010001, 0x02010202, 0x02020202, 0x02020202, 0x01020202, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x01010101, 0x01010100, 0x02020201, 0x02020202, 0x02020202, 0x00000102, 0x00000000, 0x02020201, 0x02020202, 0x00000102, 0x02020100, 0x01020202, 0x00000000, 0x02020100, 0x02010001, 0x00000102, 0x01020201, 0x01010000, 0x01000001, 0x02010001, 0x00000102, 0x01020201, 0x01010100, 0x01000001, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x00000001, 0x00000000, 0x01000000, 0x02020202, 0x02020202, 0x00010202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x00000000, 
0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01010102, 0x01010001, 0x02020101, 0x02020202, 0x02020202, 0x01020202, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x00000102, 0x01000000, 0x02020202, 0x02020202, 0x02020202, 0x01010202, 0x01010101, 0x02020202, 0x02020202, 0x00010202, 0x01010000, 0x01020202, 0x00000001, 0x02020201, 0x02020101, 0x00000102, 0x01020201, 0x00000100, 0x01000100, 0x02020101, 0x00000102, 0x01020201, 0x01000100, 0x01000100, 0x01020202, 0x02020101, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x01010101, 0x01010101, 0x01010101, 0x02020202, 
0x02020202, 0x00010102, 0x02020101, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01000001, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x00000000, 0x02020201, 0x02020202, 0x02020202, 0x01020202, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x00010202, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x00000000, 0x00010101, 0x01000000, 0x02020202, 0x02020202, 0x00010202, 0x01020100, 0x00000100, 0x01000000, 0x02020202, 0x00010202, 0x01020100, 0x01000100, 0x01000100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 
0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01000001, 0x02010000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010102, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x01000000, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x02010101, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x02020100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x00000001, 0x00000000, 0x02010000, 0x02020202, 0x02020202, 0x00010202, 0x01020100, 0x00000100, 0x01000100, 0x02020202, 0x00010202, 0x01020100, 
0x01000100, 0x01000100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010101, 0x02010101, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02010001, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020201, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x00010102, 0x00000000, 0x02020101, 0x02020202, 
0x02020202, 0x01020202, 0x01020201, 0x01010000, 0x01000001, 0x02020202, 0x01020202, 0x01020201, 0x01000100, 0x01000100, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020102, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020102, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 
0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x01010101, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x00000000, 0x01000000, 0x02020202, 0x02020202, 0x01020202, 0x00000000, 0x01000000, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 
0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x02020202, 0x01020202, 0x01010101, 0x01010101, 0x02020202, 0x02020202, 0x01020202, 0x01010101, 0x01010101, 
};
