#include <stdio.h>

#include "Lgi.h"
#include "GDragAndDrop.h"
#include "GToken.h"
#include "GPopup.h"

///////////////////////////////////////////////////////////////////////
class HookInfo
{
public:
	int Flags;
	GView *Target;
};

class GWindowPrivate
{
public:
	int Sx, Sy;
	bool Dynamic;
	GKey LastKey;
	GArray<HookInfo> Hooks;
	bool SnapToEdge;
	EventHandlerUPP EventUPP;
	GDialog *ChildDlg;

	GWindowPrivate()
	{
		ChildDlg = 0;
		EventUPP = 0;
		Sx = Sy = 0;
		Dynamic = true;
		SnapToEdge = false;
		ZeroObj(LastKey);		
	}
	
	~GWindowPrivate()
	{
		if (EventUPP)
		{
			DisposeEventHandlerUPP(EventUPP);
		}
	}
	
	int GetHookIndex(GView *Target, bool Create = false)
	{
		for (int i=0; i<Hooks.Length(); i++)
		{
			if (Hooks[i].Target == Target) return i;
		}
		
		if (Create)
		{
			HookInfo *n = &Hooks[Hooks.Length()];
			if (n)
			{
				n->Target = Target;
				n->Flags = 0;
				return Hooks.Length() - 1;
			}
		}
		
		return -1;
	}
};

///////////////////////////////////////////////////////////////////////
#define GWND_CREATE		0x0010000

GWindow::GWindow() :
	GView(0)
{
	d = NEW(GWindowPrivate);
	_QuitOnClose = false;
	Wnd = 0;
	Menu = 0;
	_Default = 0;
	_Window = this;
	WndFlags |= GWND_CREATE;
	GView::Visible(false);

    _Lock = NEW(GSemaphore);

	GRect pos(0, 50, 200, 100);
	Rect r = pos;
	
	OSStatus e = CreateNewWindow
		(
			kDocumentWindowClass,
			kWindowStandardDocumentAttributes |
				kWindowStandardHandlerAttribute |
				kWindowCompositingAttribute |
				kWindowLiveResizeAttribute,
			&r,
			&Wnd
		);
	if (e)
	{
		printf("%s:%i - CreateNewWindow failed (e=%i).\n", __FILE__, __LINE__, e); 
	}
}

GWindow::~GWindow()
{
	if (LgiApp->AppWnd == this)
	{
		LgiApp->AppWnd = 0;
	}

	if (Wnd)
	{
		DisposeWindow(Wnd);
		Wnd = 0;
	}

	DeleteObj(Menu);
	DeleteObj(d);
	DeleteObj(_Lock);
}

void GWindow::SetChildDialog(GDialog *Dlg)
{
	d->ChildDlg = Dlg;
}

bool GWindow::GetSnapToEdge()
{
	return d->SnapToEdge;
}

void GWindow::SetSnapToEdge(bool s)
{
	d->SnapToEdge = s;
}

bool GWindow::Visible()
{
	if (Wnd)
	{
		return IsWindowVisible(Wnd);
	}
	
	return false;
}

void GWindow::Visible(bool i)
{
	if (Wnd)
	{
		if (i)
		{
			ShowWindow(Wnd);
		}
		else
		{
			HideWindow(Wnd);
		}
		
		OnPosChange();
	}
}

void GWindow::_SetDynamic(bool i)
{
	d->Dynamic = i;
}

void GWindow::_OnViewDelete()
{
	if (d->Dynamic)
	{
		delete this;
	}
}

bool GWindow::PostEvent(int Event, int a, int b)
{
	bool Status = false;

	if (Wnd)
	{
		EventRef Ev;
		OSStatus e = CreateEvent(NULL,
								kEventClassUser,
								kEventUser,
								0, // EventTime 
								kEventAttributeUserEvent,
								&Ev);
		if (e)
		{
			printf("%s:%i - CreateEvent failed with %i\n", __FILE__, __LINE__, e);
		}
		else
		{
			EventTargetRef t = GetWindowEventTarget(Wnd);
			
			e = SetEventParameter(Ev, kEventParamLgiEvent, typeUInt32, sizeof(Event), &Event);
			if (e) printf("%s:%i - error %i\n", __FILE__, __LINE__, e);
			e = SetEventParameter(Ev, kEventParamLgiA, typeUInt32, sizeof(a), &a);
			if (e) printf("%s:%i - error %i\n", __FILE__, __LINE__, e);
			e = SetEventParameter(Ev, kEventParamLgiB, typeUInt32, sizeof(b), &b);
			if (e) printf("%s:%i - error %i\n", __FILE__, __LINE__, e);
			
			// printf("Sending event to window %i,%i,%i\n", Event, a, b);

			#if 1
			
			e = SetEventParameter(Ev, kEventParamPostTarget, typeEventTargetRef, sizeof(t), &t);
			if (e) printf("%s:%i - error %i\n", __FILE__, __LINE__, e);
			e = PostEventToQueue(GetMainEventQueue(), Ev, kEventPriorityStandard);
			if (e) printf("%s:%i - error %i\n", __FILE__, __LINE__, e);
			Status = e == 0;
			
			#else

			e = SendEventToEventTarget(Ev, GetControlEventTarget(Wnd));
			if (e) printf("%s:%i - error %i\n", __FILE__, __LINE__, e);
			
			#endif

			ReleaseEvent(Ev);
		}
	}
	
	return Status;
}

pascal OSStatus LgiWindowProc(EventHandlerCallRef inHandlerCallRef, EventRef inEvent, void *inUserData)
{
	OSStatus result = eventNotHandledErr;
	
	GView *v = (GView*)inUserData;
	if (NOT v) return result;
	
	UInt32 eventClass = GetEventClass( inEvent );
	UInt32 eventKind = GetEventKind( inEvent );

	switch (eventClass)
	{
		case kEventClassCommand:
		{
			if (eventKind == kEventProcessCommand)
			{
				GWindow *w = v->GetWindow();
				if (w)
				{
					HICommand command;
					GetEventParameter(inEvent, kEventParamDirectObject, typeHICommand, NULL, sizeof(command), NULL, &command);
					if (command.commandID != kHICommandSelectWindow)
					{
						extern int *ReturnFloatCommand;
						if (ReturnFloatCommand)
						{
							*ReturnFloatCommand = command.commandID;
						}
						else if (command.commandID == kHICommandQuit)
						{
							if (v->OnRequestClose(false))
							{
								DeleteObj(v);
							}
						}
						else
						{
							w->OnCommand(command.commandID, 0, 0);
						}
					}
				}
				result = noErr;
			}
			break;
		}
		case kEventClassWindow:
		{
			switch (eventKind)
			{
				case kEventWindowDispose:
				{
					v->OnDestroy();
					result = noErr;
					break;
				}
				case kEventWindowClose:
				{
					if (v->OnRequestClose(false))
					{
						DeleteObj(v);
					}
					
					result = noErr;
					break;
				}
				/*
				case kEventWindowShown:
				{
					printf("kEventWindowShown\n");
					CGrafPtr port = GetWindowPort(v->WindowHandle());
					CGContextRef Ctx;
					OSStatus e = QDBeginCGContext(port, &Ctx);
					if (e)
					{
						LgiTrace("%s:%i - QDBeginCGContext failed (%i)\n", __FILE__, __LINE__, e);
					}
					else
					{
						GScreenDC s(v->WindowHandle(), Ctx);
						v->_Paint(&s);
						CGContextSynchronize(Ctx);
						QDEndCGContext(port, &Ctx);
						result = noErr;
						
						printf("paint %s\n", v->GetClient().GetStr());
					}
					break;
				}
				*/
				case kEventWindowBoundsChanged:
				{
					v->GetPos();
					v->Invalidate();
					v->OnPosChange();
					result = noErr;
					break;
				}
			}
			break;
		}
		case kEventClassMouse:
		{
			switch (eventKind)
			{
				case kEventMouseDown:
				{
					GWindow *Wnd = dynamic_cast<GWindow*>(v->GetWindow());
					if (Wnd AND NOT Wnd->d->ChildDlg)
					{
						OsWindow Hnd = Wnd->WindowHandle();
						if (NOT IsWindowActive(Hnd))
						{
							ProcessSerialNumber Psn;
							GetCurrentProcess(&Psn);
							SetFrontProcess(&Psn);
							
							SelectWindow(Hnd);
						}
					}
					// Fall thru
				}
				case kEventMouseUp:
				{
					UInt32		modifierKeys = 0;
					UInt32		Clicks = 0;
					Point		Pt;
					UInt16		Btn = 0;
					Rect		Client, Grow;
					
					GetWindowBounds(v->WindowHandle(), kWindowContentRgn, &Client);
					GetWindowBounds(v->WindowHandle(), kWindowGrowRgn, &Grow);
					
					// the point parameter is in view-local coords.
					OSStatus status = GetEventParameter (inEvent, kEventParamMouseLocation, typeQDPoint, NULL, sizeof(Point), NULL, &Pt);
					//if (status) printf("error(%i) getting kEventParamMouseLocation\n", status);
					status = GetEventParameter(inEvent, kEventParamKeyModifiers, typeUInt32, NULL, sizeof(modifierKeys), NULL, &modifierKeys);
					//if (status) printf("error(%i) getting kEventParamKeyModifiers\n", status);
					status = GetEventParameter(inEvent, kEventParamClickCount, typeUInt32, NULL, sizeof(Clicks), NULL, &Clicks);
					//if (status) printf("error(%i) getting kEventParamClickCount\n", status);
					status = GetEventParameter(inEvent, kEventParamMouseButton, typeMouseButton, NULL, sizeof(Btn), NULL, &Btn);
					//if (status) printf("error(%i) getting kEventParamMouseButton\n", status);

					#define InRect(r, x, y) \
						( (x >= r.left) && (y >= r.top) && (x <= r.right) && (y <= r.bottom) )
					if (InRect(Grow, Pt.h, Pt.v))
					{
						// printf("In resize box\n");
						break;				
					}
					if (NOT InRect(Client, Pt.h, Pt.v))
					{
						// printf("Not in client\n");
						break;				
					}
					
					GMouse m;
					m.ViewCoords = false;
					m.x = Pt.h; // - Client.left;
					m.y = Pt.v; // - Client.top;
					m.Shift(modifierKeys & 0x200);
					m.Alt(modifierKeys & 0x800);
					m.Ctrl(modifierKeys & 0x1000);
					m.Down(eventKind == kEventMouseDown);
					m.Double(m.Down() AND Clicks > 1);
					if (Btn == kEventMouseButtonPrimary) m.Left(true);
					else if (Btn == kEventMouseButtonSecondary) m.Right(true);
					else if (Btn == kEventMouseButtonTertiary) m.Middle(true);
					
					#if 0
					printf("Client=%i,%i,%i,%i Pt=%i,%i\n",
						Client.left, Client.top, Client.right, Client.bottom,
						(int)Pt.h, (int)Pt.v);
					#endif
					
					#if 0
					printf("click %i,%i down=%i, left=%i right=%i middle=%i, ctrl=%i alt=%i shift=%i Double=%i\n",
						m.x, m.y,
						m.Down(), m.Left(), m.Right(), m.Middle(),
						m.Ctrl(), m.Alt(), m.Shift(), m.Double());
					#endif

					int Cx = m.x - Client.left;
					int Cy = m.y - Client.top;
					if (m.Target = v->WindowFromPoint(Cx, Cy))
					{
						m.ToView();
						
						GView *v = m.Target->GetGView();
						if (v)
						{
							v->_Mouse(m, false);
						}
						else printf("%s:%i - Not a GView\n", __FILE__, __LINE__);
					}
					else printf("%s:%i - No target window for mouse event.\n", __FILE__, __LINE__);
					
					result = noErr;
					break;
				}
				case kEventMouseMoved:
				case kEventMouseDragged:
				{
					UInt32		modifierKeys;
					Point		Pt;
					Rect		Client;
					
					GetWindowBounds(v->WindowHandle(), kWindowContentRgn, &Client);

					// the point parameter is in view-local coords.
					OSStatus status = GetEventParameter(inEvent, kEventParamMouseLocation, typeQDPoint, NULL, sizeof(Point), NULL, &Pt);
					if (status) printf("error(%i) getting kEventParamMouseLocation\n", status);
					status = GetEventParameter(inEvent, kEventParamKeyModifiers, typeUInt32, NULL, sizeof(modifierKeys), NULL, &modifierKeys);
					if (status) printf("error(%i) getting kEventParamKeyModifiers\n", status);
					
					GMouse m;
					m.ViewCoords = false;
					m.x = Pt.h;
					m.y = Pt.v;
					m.Shift(modifierKeys & 0x200);
					m.Alt(modifierKeys & 0x800);
					m.Ctrl(modifierKeys & 0x1000);
					m.Down(eventKind == kEventMouseDragged);
					if (m.Down())
					{
						UInt32 Btn = GetCurrentEventButtonState();
						if (Btn == kEventMouseButtonPrimary) m.Left(true);
						else if (Btn == kEventMouseButtonSecondary) m.Right(true);
						else if (Btn == kEventMouseButtonTertiary) m.Middle(true);
					}
					
					#if 0
					printf("move %i,%i down=%i, left=%i right=%i middle=%i, ctrl=%i alt=%i shift=%i Double=%i\n",
						m.x, m.y,
						m.Down(), m.Left(), m.Right(), m.Middle(),
						m.Ctrl(), m.Alt(), m.Shift(), m.Double());
					#endif

					#if 1
					if (m.Target = v->WindowFromPoint(m.x - Client.left, m.y - Client.top))
					{
						m.ToView();

						GView *v = m.Target->GetGView();
						if (v)
						{
							v->_Mouse(m, true);
						}
					}
					#endif
					break;
				}
				case kEventMouseWheelMoved:
				{
					UInt32		modifierKeys;
					Point		Pt;
					Rect		Client;
					SInt32		Delta;
					GViewI		*Target;

					GetWindowBounds(v->WindowHandle(), kWindowContentRgn, &Client);
					
					OSStatus status = GetEventParameter(inEvent, kEventParamMouseLocation, typeQDPoint, NULL, sizeof(Point), NULL, &Pt);
					if (status) printf("error(%i) getting kEventParamMouseLocation\n", status);
					status = GetEventParameter(inEvent, kEventParamKeyModifiers, typeUInt32, NULL, sizeof(modifierKeys), NULL, &modifierKeys);
					if (status) printf("error(%i) getting kEventParamKeyModifiers\n", status);
					status = GetEventParameter(inEvent, kEventParamMouseWheelDelta, typeSInt32, NULL, sizeof(Delta), NULL, &Delta);
					if (status) printf("error(%i) getting kEventParamMouseWheelDelta\n", status);
					
					if (Target = v->WindowFromPoint(Pt.h - Client.left, Pt.v - Client.top))
					{
						GView *v = Target->GetGView();
						if (v)
						{
							v->OnMouseWheel(Delta * -3);
						}
						else printf("%s:%i - no GView\n", __FILE__, __LINE__);
					}
					else printf("%s:%i - no target\n", __FILE__, __LINE__);
					break;
				}
			}
			break;
		}
		case kEventClassUser:
		{
			if (eventKind == kEventUser)
			{
				GMessage m;
				
				OSStatus status = GetEventParameter(inEvent, kEventParamLgiEvent, typeUInt32, NULL, sizeof(UInt32), NULL, &m.m);
				status = GetEventParameter(inEvent, kEventParamLgiA, typeUInt32, NULL, sizeof(UInt32), NULL, &m.a);
				status = GetEventParameter(inEvent, kEventParamLgiB, typeUInt32, NULL, sizeof(UInt32), NULL, &m.b);
				
				v->OnEvent(&m);
				
				result = noErr;
			}
			break;
		}
	}
	
	return result;
}

pascal OSStatus LgiRootCtrlProc(EventHandlerCallRef inHandlerCallRef, EventRef inEvent, void *inUserData)
{
	OSStatus result = eventNotHandledErr;
	
	GView *v = (GView*)inUserData;
	if (NOT v)
	{
		LgiTrace("%s:%i - no gview\n", __FILE__, __LINE__);
		return result;
	}
	
	UInt32 eventClass = GetEventClass( inEvent );
	UInt32 eventKind = GetEventKind( inEvent );

	switch (eventClass)
	{
		case kEventClassControl:
		{
			switch (eventKind)
			{
				case kEventControlDraw:
				{
					CGContextRef Ctx = 0;
					result = GetEventParameter (inEvent,  
												kEventParamCGContextRef, 
												typeCGContextRef, 
												NULL, 
												sizeof(CGContextRef),
												NULL,
												&Ctx);
					if (Ctx)
					{
						GScreenDC s(v->GetWindow(), Ctx);
						v->_Paint(&s);
					}
					else
					{
						LgiTrace("%s:%i - No context.\n", __FILE__, __LINE__);
					}
					break;
				}
			}
			break;
		}
	}
	
	return result;
}

bool GWindow::Attach(GViewI *p)
{
	bool Status = false;
	
	if (Wnd)
	{
		if (GObject::Name())
			Name(GObject::Name());
		
		EventTypeSpec	WndEvents[] =
		{
			{ kEventClassCommand, kEventProcessCommand },
			{ kEventClassWindow, kEventWindowClose },
			{ kEventClassWindow, kEventWindowInit },
			{ kEventClassWindow, kEventWindowDispose },
			{ kEventClassWindow, kEventWindowBoundsChanged },
			{ kEventClassWindow, kEventWindowShown },
			{ kEventClassMouse, kEventMouseDown },
			{ kEventClassMouse, kEventMouseUp },
			{ kEventClassMouse, kEventMouseMoved },
			{ kEventClassMouse, kEventMouseDragged },
			{ kEventClassMouse, kEventMouseWheelMoved },
			{ kEventClassUser, kEventUser }

		};
		
		EventHandlerRef Handler = 0;
		OSStatus e = InstallWindowEventHandler(	Wnd,
												d->EventUPP = NewEventHandlerUPP(LgiWindowProc),
												GetEventTypeCount(WndEvents),
												WndEvents,
												(void*)this,
												&Handler);
		if (e)
		{
			LgiTrace("%s:%i - InstallEventHandler failed (%i)\n", __FILE__, __LINE__, e);
		}

		HIViewRef Content = 0;
		e = HIViewFindByID(HIViewGetRoot(Wnd), kHIViewWindowContentID, &Content);
		if (Content)
		{
			EventTypeSpec	CtrlEvents[] =
			{
				{ kEventClassControl, kEventControlDraw },
			};
			EventHandlerRef CtrlHandler = 0;
			e = InstallEventHandler(GetControlEventTarget(Content),
									NewEventHandlerUPP(LgiRootCtrlProc),
									GetEventTypeCount(CtrlEvents),
									CtrlEvents,
									(void*)this,
									&CtrlHandler);
			if (e)
			{
				LgiTrace("%s:%i - InstallEventHandler failed (%i)\n", __FILE__, __LINE__, e);
			}
		}

		Status = true;
		
		// Setup default button...
		if (NOT _Default)
		{
			_Default = FindControl(IDOK);
			if (_Default)
			{
				_Default->Invalidate();
			}
		}

		OnCreate();
		OnAttach();
		OnPosChange();
	}
	
	return Status;
}

bool GWindow::OnRequestClose(bool OsShuttingDown)
{
	if (GetQuitOnClose())
	{
		LgiCloseApp();
	}

	return GView::OnRequestClose(OsShuttingDown);
}

bool GWindow::HandleViewMouse(GView *v, GMouse &m)
{
	for (int i=0; i<d->Hooks.Length(); i++)
	{
		if (d->Hooks[i].Flags & GMouseEvents)
		{
			if (NOT d->Hooks[i].Target->OnViewMouse(v, m))
			{
				return false;
			}
		}
	}
	
	return true;
}

#define DEBUG_KEYS			0

/*
	// Any window in a popup always gets the key...
	for (GView *p = v; p; p = p->GetParent())
	{
		GPopup *Popup;
		if (Popup = dynamic_cast<GPopup*>(p))
		{
			Status = v->OnKey(k);
			if (NOT Status)
			{
				if (k.c16 == VK_ESCAPE)
				{
					// Popup window (or child) didn't handle the key, and the user
					// pressed ESC, so do the default thing and close the popup.
					Popup->Cancelled = true;
					Popup->Visible(false);
				}
				else
				{
					#if DEBUG_KEYS
					printf("Popup ignores '%c' down=%i alt=%i ctrl=%i sh=%i\n", k.c16, k.Down(), k.Alt(), k.Ctrl(), k.Shift());
					#endif
					break;
				}
			}
			
			#if DEBUG_KEYS
			printf("Popup ate '%c' down=%i alt=%i ctrl=%i sh=%i\n", k.c16, k.Down(), k.Alt(), k.Ctrl(), k.Shift());
			#endif
			
			goto AllDone;
		}
	}
*/

bool GWindow::HandleViewKey(GView *v, GKey &k)
{
	bool Status = false;
	GViewI *Ctrl = 0;

	// Give key to popups
	if (LgiApp AND
		LgiApp->GetMouseHook() AND
		LgiApp->GetMouseHook()->OnViewKey(v, k))
	{
		goto AllDone;
	}

	// Allow any hooks to see the key...
	for (int i=0; i<d->Hooks.Length(); i++)
	{
		if (d->Hooks[i].Flags & GKeyEvents)
		{
			if (d->Hooks[i].Target->OnViewKey(v, k))
			{
				Status = true;
				
				#if DEBUG_KEYS
				printf("Hook ate '%c'(%i) down=%i alt=%i ctrl=%i sh=%i\n", k.c16, k.c16, k.Down(), k.Alt(), k.Ctrl(), k.Shift());
				#endif
				
				goto AllDone;
			}
		}
	}

	// Give the key to the window...
	if (v->OnKey(k))
	{
		#if DEBUG_KEYS
		printf("View ate '%c'(%i) down=%i alt=%i ctrl=%i sh=%i\n",
			k.c16, k.c16, k.Down(), k.Alt(), k.Ctrl(), k.Shift());
		#endif
		
		Status = true;
		goto AllDone;
	}
	
	// Window didn't want the key...
	switch (k.c16)
	{
		case VK_RETURN:
		{
			Ctrl = _Default;
			break;
		}
		case VK_ESCAPE:
		{
			Ctrl = FindControl(IDCANCEL);
			break;
		}
	}

	if (Ctrl AND Ctrl->Enabled())
	{
		if (Ctrl->OnKey(k))
		{
			Status = true;

			#if DEBUG_KEYS
			printf("Default Button ate '%c'(%i) down=%i alt=%i ctrl=%i sh=%i\n",
				k.c16, k.c16, k.Down(), k.Alt(), k.Ctrl(), k.Shift());
			#endif
			
			goto AllDone;
		}
	}

	if (Menu)
	{
		Status = Menu->OnKey(v, k);
		if (Status)
		{
			#if DEBUG_KEYS
			printf("Menu ate '%c' down=%i alt=%i ctrl=%i sh=%i\n", k.c16, k.Down(), k.Alt(), k.Ctrl(), k.Shift());
			#endif
		}
	}

AllDone:
	d->LastKey = k;

	return Status;
}


void GWindow::Raise()
{
	if (Wnd)
	{
		BringToFront(Wnd);
	}
}

GWindowZoom GWindow::GetZoom()
{
	return GZoomNormal;
}

void GWindow::SetZoom(GWindowZoom i)
{
}

GViewI *GWindow::GetDefault()
{
	return _Default;
}

void GWindow::SetDefault(GViewI *v)
{
	if (v AND
		v->GetWindow() == (GViewI*)this)
	{
		if (_Default != v)
		{
			GViewI *Old = _Default;
			_Default = v;

			if (Old) Old->Invalidate();
			if (_Default) _Default->Invalidate();
		}
	}
	else
	{
		_Default = 0;
	}
}

bool GWindow::Name(char *n)
{
	bool Status = GObject::Name(n);

	if (Wnd)
	{	
		CFStringRef s = CFStringCreateWithBytes(NULL, (UInt8*)n, strlen(n), kCFStringEncodingUTF8, false);
		if (s)
		{
			OSStatus e = SetWindowTitleWithCFString(Wnd, s);
			if (e)
			{
				printf("%s:%i - SetWindowTitleWithCFString failed (e=%i)\n", __FILE__, __LINE__, e);
			}
			else
			{
				Status = true;
			}
			
			CFRelease(s);
		}
		else
		{
			printf("%s:%i - CFStringCreateWithBytes failed.\n", __FILE__, __LINE__);
		}
	}

	return Status;
}

char *GWindow::Name()
{
	return GObject::Name();
}

GRect &GWindow::GetClient(bool ClientSpace)
{
	if (Wnd)
	{
		static GRect c;
		Rect r;
		OSStatus e = GetWindowBounds(Wnd, kWindowContentRgn, &r);
		if (NOT e)
		{
			c = r;
			c.Offset(-c.x1, -c.y1);
			return c;
		}
		else
		{
			printf("%s:%i - GetWindowBounds failed\n", __FILE__, __LINE__);
		}
	}
	
	return GView::GetClient(ClientSpace);
}

bool GWindow::SerializeState(GDom *Store, char *FieldName, bool Load)
{
	if (NOT Store OR NOT FieldName)
		return false;

	if (Load)
	{
		GVariant v;
		if (Store->GetValue(FieldName, v) AND v.Str())
		{
			GRect Position(0, 0, -1, -1);
			GWindowZoom State = GZoomNormal;

			GToken t(v.Str(), ";");
			for (int i=0; i<t.Length(); i++)
			{
				char *Var = t[i];
				char *Value = strchr(Var, '=');
				if (Value)
				{
					*Value++ = 0;

					if (stricmp(Var, "State") == 0)
						State = (GWindowZoom) atoi(Value);
					else if (stricmp(Var, "Pos") == 0)
						Position.SetStr(Value);
				}
				else return false;
			}
			
			if (Position.Valid())
			{
				SetPos(Position);
			}
			
			SetZoom(State);
		}
		else return false;
	}
	else
	{
		char s[256];
		GWindowZoom State = GetZoom();
		sprintf(s, "State=%i;Pos=%s", State, GetPos().GetStr());

		GVariant v = s;
		if (NOT Store->SetValue(FieldName, v))
			return false;
	}

	return true;
}

GRect &GWindow::GetPos()
{
	if (Wnd)
	{
		Rect r;
		OSStatus e = GetWindowBounds(Wnd, kWindowGlobalPortRgn, &r);
		if (!e)
		{
			Pos = r;
		}
		else
		{
			printf("%s:%i - GetWindowBounds failed (e=%i)\n", __FILE__, __LINE__, e);
		}
	}

	return Pos;
}

bool GWindow::SetPos(GRect &p, bool Repaint)
{
	Pos = p;

	if (Wnd)
	{
		MoveWindow(Wnd, p.x1, p.y1, false);
		SizeWindow(Wnd, p.X(), p.Y(), true);
	}

	return true;
}

void GWindow::OnChildrenChanged(GViewI *Wnd, bool Attaching)
{
	// Force repour
	d->Sx = d->Sy = -1;
}

void GWindow::OnCreate()
{
}

void GWindow::OnPaint(GSurface *pDC)
{
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle();
}

void GWindow::OnPosChange()
{
	GView::OnPosChange();

	if (d->Sx != X() OR	d->Sy != Y())
	{
		// printf("Window '%s' Layout: x,y=%i,%i\n", Name(), X(), Y());
		
		Pour();
		d->Sx = X();
		d->Sy = Y();
	}
	else
	{
		// printf("Window '%s' Not Layout: x,y=%i,%i == %i,%i\n", Name(), d->Sx, d->Sy, X(), Y());
	}
}

void GWindow::Pour()
{
	GRect r = GetClient();
	
	//printf("::Pour r=%s\n", r.GetStr());

	GRegion Client(r);
	GRegion Update;

	Iterator<GViewI> Lst(&Children);
	for (GViewI *w = Lst.First(); w; w = Lst.Next())
	{
		GRect OldPos = w->GetPos();
		Update.Union(&OldPos);

		if (w->Pour(Client))
		{
			GRect p = w->GetPos();

			//printf("::Pour p=%s\n", p.GetStr());

			if (NOT w->Visible())
			{
				w->Visible(true);
			}

			w->Invalidate();

			Client.Subtract(&w->GetPos());
			Update.Subtract(&w->GetPos());
		}
		else
		{
			// non-pourable
			// printf("::Pour non pourable\n");
		}
	}
}

int GWindow::WillAccept(List<char> &Formats, GdcPt2 Pt, int KeyState)
{
	int Status = DROPEFFECT_NONE;
	
	for (char *f=Formats.First(); f; )
	{
		if (stricmp(f, LGI_FileDropFormat) == 0)
		{
			f = Formats.Next();
			Status = DROPEFFECT_COPY;
		}
		else
		{
			Formats.Delete(f);
			DeleteArray(f);
			f = Formats.Current();
		}
	}
	
	return Status;
}

int GWindow::OnDrop(char *Format, GVariant *Data, GdcPt2 Pt, int KeyState)
{
	int Status = DROPEFFECT_NONE;

	if (Format AND Data)
	{
		if (stricmp(Format, LGI_FileDropFormat) == 0)
		{
			GArray<char*> Files;
			if (Data->IsBinary())
			{
				GToken Uri(	(char*)Data->Value.Binary.Data,
							"\r\n,",
							true,
							Data->Value.Binary.Length);
				for (int i=0; i<Uri.Length(); i++)
				{
					char *File = Uri[i];
					if (strnicmp(File, "file:", 5) == 0)
						File += 5;
					
					char *i = File, *o = File;
					while (*i)
					{
						if (i[0] == '%' AND
							i[1] AND
							i[2])
						{
							char h[3] = { i[1], i[2], 0 };
							*o++ = htoi(h);
							i += 3;
						}
						else
						{
							*o++ = *i++;
						}
					}
					*o++ = 0;
					
					if (FileExists(File))
					{
						Files.Add(NewStr(File));
					}
				}
			}
			
			if (Files.Length())
			{
				Status = DROPEFFECT_COPY;
				OnReceiveFiles(Files);
				Files.DeleteArrays();
			}
		}
	}
	
	return Status;
}

int GWindow::OnEvent(GMessage *m)
{
	switch (MsgCode(m))
	{
		case M_CLOSE:
		{
			if (OnRequestClose(false))
			{
				Quit();
				return 0;
			}
			break;
		}
	}

	return GView::OnEvent(m);
}

bool GWindow::RegisterHook(GView *Target, int EventType, int Priority)
{
	bool Status = false;
	
	if (Target AND EventType)
	{
		int i = d->GetHookIndex(Target, true);
		if (i >= 0)
		{
			d->Hooks[i].Flags = EventType;
			Status = true;
		}
	}
	
	return Status;
}

bool GWindow::UnregisterHook(GView *Target)
{
	int i = d->GetHookIndex(Target);
	if (i >= 0)
	{
		d->Hooks.DeleteAt(i);
		return true;
	}
	return false;
}

