#include <stdio.h>

#include "Lgi.h"
#include <ole2.h>
#include <commctrl.h>
#include <time.h>
#include "GSkinEngine.h"

static GApp *TheApp = 0;
GApp *GApp::ObjInstance()
{
	return TheApp;
}

class GAppPrivate
{
public:
	// Common
	GXmlTag *Config;
	GFileSystem *FileSystem;
	GdcDevice *GdcSystem;
	OsAppArguments Args;
	GLibrary *SkinLib;
	OsThreadId GuiThread;

	// Win32
	bool QuitReceived;
	List<GWin32Class> Classes;

	GAppPrivate()
	{
		QuitReceived = false;
		SkinLib = 0;
		GuiThread = LgiGetCurrentThread();
	}

	~GAppPrivate()
	{
		DeleteObj(SkinLib);
	}
};

/////////////////////////////////////////////////////////////////////////////
LONG __stdcall _ExceptionFilter_Redir(LPEXCEPTION_POINTERS e)
{
	if (LgiApp)
	{
		return LgiApp->_ExceptionFilter(e);
	}

	return 0;
}

/////////////////////////////////////////////////////////////////////////////
GSkinEngine *GApp::SkinEngine = 0;
GMouseHook *GApp::MouseHook = 0;
GMouseHook *GApp::GetMouseHook()
{
	return MouseHook;
}

/////////////////////////////////////////////////////////////////////////////
extern int MouseRollMsg;

bool GApp::Win9x = LgiGetOs() == LGI_OS_WIN9X;

GApp::GApp(char *MimeType, OsAppArguments &AppArgs, GAppArguments *ObjArgs)
{
int64 Time = LgiCurrentTime();
#define DumpTime(str) /* \
	{ int64 n = LgiCurrentTime(); \
	LgiTrace("%s=%ims\n", str, (int)(n-Time)); \
	Time = n; } */

	// Sanity Checks
	LgiAssert(sizeof(int8) == 1);
	LgiAssert(sizeof(uint8) == 1);

	LgiAssert(sizeof(int16) == 2);
	LgiAssert(sizeof(uint16) == 2);

	LgiAssert(sizeof(int32) == 4);
	LgiAssert(sizeof(uint32) == 4);

	LgiAssert(sizeof(int64) == 8);
	LgiAssert(sizeof(uint64) == 8);

	// Setup exception handler
    InitializeCriticalSection(&StackTraceSync);
	SetUnhandledExceptionFilter(_ExceptionFilter_Redir);

DumpTime("initcrit");

	// Private data
	d = NEW(GAppPrivate);

DumpTime("priv");

	// Initialize windows dll's
	OleInitialize(NULL);
	CoInitialize(NULL);
	InitCommonControls();

DumpTime("init common ctrls");

	// Setup LGI Sub-systems
	GFontSystem::Inst();

DumpTime("font sys");

	d->FileSystem = NEW(GFileSystem);

DumpTime("file sys");

	d->GdcSystem = NEW(GdcDevice);

DumpTime("gdc");

	// Vars...
	d->Config = 0;
	AppWnd = 0;
	SetAppArgs(AppArgs);

	// GApp instance
	LgiAssert(TheApp == 0);
	TheApp = this;

DumpTime("vars");

	// System font
	SystemNormal = 0;
	SystemBold = 0;
	GFontType SysFontType;
	if (SysFontType.GetSystemFont("System"))
	{
		SystemNormal = SysFontType.Create();
		if (SystemNormal)
		{
			// Force load
			SystemNormal->Y("A");
		}

		SystemBold = SysFontType.Create();
		if (SystemBold)
		{
			SystemBold->Bold(true);
			SystemBold->Y("A");
		}
	}

	if (NOT SystemNormal)
	{
		LgiMsg(0, "Error: Couldn't create system font.", "Lgi Error: GApp::GApp", MB_OK);
		LgiExitApp();
	}

DumpTime("fonts");

	// Other vars and init
	hNormalCursor = LoadCursor(NULL, IDC_ARROW);
	LgiRandomize(LgiCurrentTime()*LgiGetCurrentThread());
	MouseRollMsg = RegisterWindowMessage("MSWHEEL_ROLLMSG");

DumpTime("cursor/rand/msg");

	LgiInitColours();

DumpTime("colours");

	// Setup mouse hook
	MouseHook = NEW(GMouseHook);

DumpTime("ms hook");

	if
	(
		(
			NOT ObjArgs
			OR
			NOT ObjArgs->NoSkin
		)
		AND
		NOT GetOption("noskin")
	)
	{
		#ifdef SKIN_MAGIC

		char *SkinFile = LgiFindFile("PC_Angel_2.smf");
		if (NOT SkinFile)
		{
			SkinFile = NewStr("C:\\Public\\SoftThinks\\Skinning\\PC_Angel_2.smf");
		}

		bool SkinInit = InitSkinMagicLib(AppArgs.hInstance,	"Dialog", 0, 0);
		if (SkinInit)
		{
			if (LoadSkinFile(SkinFile))
			{
				SetDialogSkin("Dialog");
			}
			else
			{
				LgiMsg(0, "LoadSkinFile failed.\n", "Error");
			}
		}
		else
		{
			LgiMsg(0, "InitSkinMagicLib failed.\n", "Error");
		}

		DeleteArray(SkinFile);

		#else

		// Load library
		char SkinLibName[] =
		#ifdef __GNUC__
			"lib"
		#endif
			"lgiskin"
		#if _MSC_VER >= 1300
			"-7"
		#endif
		#ifdef _DEBUG
			"d"
		#endif
			;
		
		d->SkinLib = NEW(GLibrary(SkinLibName));
		if (d->SkinLib)
		{
			if (d->SkinLib->IsLoaded())
			{
				Proc_CreateSkinEngine CreateSkinEngine =
					(Proc_CreateSkinEngine)d->SkinLib->GetAddress(LgiSkinEntryPoint);
				if (CreateSkinEngine)
				{
					SkinEngine = CreateSkinEngine(this);
				}
			}
			else
			{
				DeleteObj(d->SkinLib);
				printf("Failed to load '%s'\n", SkinLibName);
			}
		}

		#endif
	}

DumpTime("skin");
}

GApp::~GApp()
{
	#ifdef SKIN_MAGIC
	
	ExitSkinMagicLib();
	
	#endif

	DeleteObj(AppWnd);
	DeleteObj(SystemNormal);
	DeleteObj(SystemBold);
	DeleteObj(MouseHook);

	LgiResources::ResourceContainers.DeleteObjects();

	TheApp = 0;
	DeleteObj(SkinEngine);

	DeleteObj(GFontSystem::Me);
	DeleteObj(d->FileSystem);
	DeleteObj(d->GdcSystem);
	DeleteObj(d->Config);
	d->Classes.DeleteObjects();

	CoUninitialize();
	OleUninitialize();

	DeleteObj(d);
	DeleteCriticalSection(&StackTraceSync);
}

bool GApp::IsOk()
{
	bool Status =	(this != 0) AND
					(d != 0) AND
					(d->FileSystem != 0) AND
					(d->GdcSystem != 0);
	LgiAssert(Status);
	return Status;
}

OsThreadId GApp::GetGuiThread()
{
	return d->GuiThread;
}

List<GWin32Class> *GApp::GetClasses()
{
	return IsOk() ? &d->Classes : 0;
}

OsAppArguments *GApp::GetAppArgs()
{
	return IsOk() ? &d->Args : 0;
}

GXmlTag *GApp::GetConfig(char *Tag)
{
	if (IsOk() AND NOT d->Config)
	{
		char File[] = "lgi.conf";
		char Path[256];
		if (LgiGetExePath(Path, sizeof(Path)))
		{
			if (Path[strlen(Path)-1] != DIR_CHAR) strcat(Path, DIR_STR);
			strcat(Path, File);

			/*
			if (NOT FileExists(Path))
			{
				char *f = LgiFindFile(File);
				if (f)
				{
					strcpy(Path, f);
					DeleteArray(f);
				}
			}
			*/

			if (FileExists(Path))
			{
				d->Config = NEW(GXmlTag("Config"));
				if (d->Config)
				{
					GFile f;
					if (f.Open(Path, O_READ))
					{
						GXmlTree t;
						t.Read(d->Config, &f, 0);
					}
				}
			}
		}

		if (NOT d->Config)
		{
			d->Config = NEW(GXmlTag("Options"));
		}
	}

	if (Tag AND d->Config)
	{
		return d->Config->GetTag(Tag);
	}

	return 0;
}

void GApp::SetConfig(GXmlTag *Tag)
{
	if (IsOk() AND Tag)
	{
		GXmlTag *Old = GetConfig(Tag->Tag);
		if (Old)
		{
			Old->RemoveTag();
			DeleteObj(Old);
		}

		if (NOT d->Config)
		{
			GetConfig(0);
		}
		if (d->Config)
		{
			d->Config->InsertTag(Tag);
		}
	}
}

char *GApp::GetArgumentAt(int n)
{
	if (d->Args.lpCmdLine)
	{
		char16 *s = d->Args.lpCmdLine;
		for (int i=0; i<=n; i++)
		{
			char16 *e = 0;
			while (*s AND strchr(WhiteSpace, *s)) s++;
			if (*s == '\'' OR *s == '\"')
			{
				char16 Delim = *s++;
				e = StrchrW(s, Delim);
			}
			else
			{
				for (e = s; *e AND NOT strchr(WhiteSpace, *e); e++)
					;
			}

			if (i == n)
			{
				return LgiNewUtf16To8(s, (e - s) * sizeof(char16));
			}

			s = *e ? e + 1 : e;
		}
	}

	return 0;
}

bool GApp::GetOption(char *Option, GArray<char> &Buf)
{
	if (NOT ValidStr(Option))
	{
		return false;
	}

	char16 *c = d->Args.lpCmdLine;
	char16 *Opt = LgiNewUtf8To16(Option);
	int OptLen = StrlenW(Opt);

	while (c AND *c)
	{
		if (*c == '/' OR *c == '-')
		{
			c++;
			if (StrnicmpW(c, Opt, OptLen) == 0)
			{
				c += OptLen;
				if (*c)
				{
					// skip leading whitespace
					while (*c AND strchr(WhiteSpace, *c))
					{
						c++;
					}

					// write str out if they want it
					char16 End = (*c == '\'' OR *c == '\"') ? *c++ : ' ';
					char16 *e = StrchrW(c, End);
					if (NOT e) e = c + StrlenW(c);
					char *Arg = LgiNewUtf16To8(c, (int)e-(int)c);
					if (Arg)
					{
						int ArgLen = strlen(Arg);
						Buf.Length(ArgLen + 1);
						strcpy(&Buf[0], Arg);
						DeleteArray(Arg);
					}
				}

				// yeah we got the option
				DeleteArray(Opt);
				return true;
			}
		}

		c = StrchrW(c, ' ');
		if (c) c++;
	}

	DeleteArray(Opt);
	return false;
}

bool GApp::GetOption(char *Option, char *Dest, int DestLen)
{
	GArray<char> Buf;
	if (GetOption(Option, Buf))
	{
		if (Dest)
		{
			strsafecpy(Dest, &Buf[0], DestLen);
		}
		return true;
	}
	return false;
}

#include "ginput.h"

void GApp::SetAppArgs(OsAppArguments &AppArgs)
{
	if (AppArgs.nCmdShow)
	{
		d->Args.nCmdShow = AppArgs.nCmdShow;
	}
	
	if (AppArgs.hInstance)
	{
		d->Args.hInstance = AppArgs.hInstance;
	}

	d->Args.Pid = GetCurrentProcessId();
	
	d->Args.lpCmdLine = AppArgs.lpCmdLine;
}

void GApp::OnCommandLine()
{
	char WhiteSpace[] = " \r\n\t";
	char *CmdLine = LgiNewUtf16To8(d->Args.lpCmdLine);

	if (ValidStr(CmdLine))
	{
		GArray<char*> Files;
		
		char *Delim = "\'\"";
		char *s;
		for (s = CmdLine; *s; )
		{
			// skip ws
			while (*s AND strchr(WhiteSpace, *s)) s++;

			// read to end of token
			char *e = s;
			if (strchr(Delim, *s))
			{
				char Delim = *s++;
				e = strchr(s, Delim);
			}
			else
			{
				for (; *e AND NOT strchr(WhiteSpace, *e); e++);
			}

			char *Arg = NewStr(s, (int)e-(int)s);
			if (Arg)
			{
				if (FileExists(Arg))
				{
					Files.Add(Arg);
				}
				else
				{
					DeleteArray(Arg);
				}
			}

			// next
			s = (*e) ? e + 1 : e;
		}

		// call app
		if (Files.Length() > 0)
		{
			OnReceiveFiles(Files);
		}

		// clear up
		Files.DeleteArrays();
	}

	DeleteArray(CmdLine);
}

void GApp::OnReceiveFiles(GArray<char*> &Files)
{
	if (AppWnd)
	{
		AppWnd->OnReceiveFiles(Files);
	}
}

int32 GApp::GetMetric(int Metric)
{
	int32 Status = 0;

	switch (Metric)
	{
		case LGI_MET_DECOR_X:
		{
			Status = GetSystemMetrics(SM_CXFRAME) * 2;
			break;
		}
		case LGI_MET_DECOR_Y:
		{
			Status = GetSystemMetrics(SM_CYFRAME) * 2;
			Status += GetSystemMetrics(SM_CYCAPTION);
			break;
		}
		case LGI_MET_DECOR_CAPTION:
		{
			Status = GetSystemMetrics(SM_CYCAPTION);
			break;
		}
		case LGI_MET_MENU:
		{
			Status = GetSystemMetrics(SM_CYMENU);
			break;
		}
	}

	return Status;
}

GViewI *GApp::GetFocus()
{
	HWND h = ::GetFocus();
	if (h)
	{
		return GWindowFromHandle(h);
	}

	return 0;
}

HINSTANCE GApp::GetInstance()
{
	if (this AND IsOk())
	{
		return d->Args.hInstance;
	}

	return (HINSTANCE)GetCurrentProcess();
}

OsProcessId GApp::GetProcessId()
{
	if (this)
	{
		return IsOk() ? d->Args.Pid : 0;
	}

	return GetCurrentProcessId();
}

int GApp::GetShow()
{
	return IsOk() ? d->Args.nCmdShow : 0;
}

class GWnd
{
public:
	GSemaphore *GetLock(GWindow *w)
	{
		return w->_Lock;
	}
};

bool GApp::Run(bool Loop)
{
	MSG Msg;
	bool status = true;
	ZeroObj(Msg);

	if (Loop)
	{
		OnCommandLine();
	
		while (	NOT d->QuitReceived AND
				GetMessage(&Msg, NULL, 0, 0) > 0)
		{
			#ifdef _DEBUG
			int64 Last = LgiCurrentTime();
			#endif
			
			TranslateMessage(&Msg);
			DispatchMessage(&Msg);

			#ifdef _DEBUG
			int64 Now = LgiCurrentTime();
			if (Now - Last > 1000)
			{
				// _asm int 3
				LgiTrace("%s:%i - Msg Loop Blocked: %i ms (Msg: %i)\n",
					__FILE__, __LINE__,
					(int) (Now - Last), Msg.message);
			}
			#endif
		}
	}
	else
	{
		while (	PeekMessage(&Msg, NULL, 0, 0, PM_REMOVE) AND
				Msg.message != WM_QUIT)
		{
			TranslateMessage(&Msg);
			DispatchMessage(&Msg);
		}

		if (Msg.message == WM_QUIT)
		{
			d->QuitReceived = true;
		}
	}

	return Msg.message != WM_QUIT;
}

void GApp::Exit(int Code)
{
	if (Code)
	{
		exit(0);
	}
	else
	{
		PostQuitMessage(Code);
	}
}

bool GApp::GetFileMimeType(char *File, char *Mime, int BufLen)
{
	if (File AND Mime AND BufLen > 0)
	{
		
	}

	return 0;
}
