// Win32 Implementation of General LGI functions
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "Lgi.h"

////////////////////////////////////////////////////////////////
// Local helper functions
bool _lgi_check_file(char *Path)
{
	if (Path)
	{
		if (FileExists(Path))
		{
			// file is there
			return true;
		}
		else
		{
			// shortcut?
			char *e = Path + strlen(Path);
			strcpy(e, ".lnk");
			if (FileExists(Path))
			{
				// resolve shortcut
				char Link[256];
				if (ResolveShortcut(Path, Link, sizeof(Link)))
				{
					// check destination of link
					if (FileExists(Link))
					{
						strcpy(Path, Link);
						return true;
					}
				}
			}
			*e = 0;
		}
	}

	return false;
}

////////////////////////////////////////////////////////////////
// Implementations
void LgiSleep(DWORD i)
{
	::Sleep(i);
}

bool LgiGetFileMimeType(char *File, char *Mime, int BufLen)
{
	bool Status = false;

	if (File AND Mime)
	{
		char *Dot = strrchr(File, '.');
		if (Dot)
		{
			GRegKey Key("HKEY_CLASSES_ROOT\\%s", Dot);
			if (Key.IsOk())
			{
				char *Ct = Key.GetStr("Content Type");
				if (Ct AND
					NOT stricmp(Dot, ".dsw") == 0 AND
					NOT stricmp(Dot, ".dsp") == 0)
				{
					strsafecpy(Mime, Ct, BufLen);
					Status = true;
				}
				else
				{
					// Search mime type DB.
					GRegKey Db("HKEY_LOCAL_MACHINE\\SOFTWARE\\Classes\\MIME\\Database\\Content Type");
					List<char> Sub;
					Db.GetKeyNames(Sub);
					for (char *k = Sub.First(); k; k = Sub.Next())
					{
						GRegKey Type("HKEY_LOCAL_MACHINE\\SOFTWARE\\Classes\\MIME\\Database\\Content Type\\%s", k);
						char *Ext = Type.GetStr("Extension");
						if (Ext AND stricmp(Ext, Dot) == 0)
						{
							strsafecpy(Mime, k, BufLen);
							Status = true;
							break;
						}
					}

					Sub.DeleteArrays();

					if (NOT Status)
					{
						// This is a hack to get around file types without a MIME database entry
						// but do have a .ext entry. LgiGetAppsForMimeType knows about the hack too.
						snprintf(Mime, BufLen, "application/%s", Dot);
						Status = true;
					}
				}
			}
		}

		if (NOT Status)
		{
			// no extension?
			// no registry entry for file type?
			strsafecpy(Mime, "application/octet-stream", BufLen);
			Status = true;
		}
	}

	return Status;
}

bool _GetApps_Add(GArray<GAppInfo*> &Apps, char *In)
{
	char *Path = LgiTokStr(In);
	if (Path)
	{
		GStringPipe p;

		char *RootVar = "%SystemRoot%";
		char *SysRoot = strstr(Path, RootVar);
		if (SysRoot)
		{
			// correct path for variables
			char Temp[256];
			GetWindowsDirectory(Temp, sizeof(Temp));
			if (Temp[strlen(Temp)-1] != DIR_CHAR) strcat(Temp, DIR_STR);
			p.Push(Temp);

			char *End = SysRoot + strlen(RootVar);
			p.Push(*End == DIR_CHAR ? End + 1 : End);
		}
		else
		{
			p.Push(Path);
		}
		DeleteArray(Path);

		GAppInfo *a = NEW(GAppInfo);
		if (a)
		{
			Apps[Apps.Length()] = a;
			a->Path = p.NewStr();
			if (a->Path)
			{
				char e[256];
				char *d = strrchr(a->Path, DIR_CHAR);
				if (d) strcpy(e, d + 1);
				else strcpy(e, a->Path);
				d = strchr(e, '.');
				if (d) *d = 0;
				e[0] = toupper(e[0]);
				a->Name = NewStr(e);
				if (ValidStr(a->Name))
				{
					bool AllCaps = true;
					for (char *s=a->Name; *s; s++)
					{
						if (islower(*s))
						{
							AllCaps = false;
							break;
						}
					}
					if (AllCaps)
					{
						strlwr(a->Name + 1);
					}
				}
			}

			return true;
		}
	}

	return false;
}

bool LgiGetAppsForMimeType(char *Mime, GArray<GAppInfo*> &Apps, int Limit)
{
	bool Status = false;

	if (Mime)
	{
		if (stricmp(Mime, "application/email") == 0)
		{
			// get email app
			HKEY hKey;
			if (RegOpenKey(HKEY_CLASSES_ROOT, "mailto\\shell\\open\\command", &hKey) == ERROR_SUCCESS)
			{
				char Str[256];
				DWORD StrLen = sizeof(Str), Type = 0;

				// get app path
				if (RegQueryValueEx(hKey, 0, 0, &Type, (uchar*)Str, &StrLen) == ERROR_SUCCESS)
				{
					Status = _GetApps_Add(Apps, Str);
				}

				RegCloseKey(hKey);
			}
		}
		else
		{
			// get generic app
			char Ext[64] = "";
			char *EndPart = strrchr(Mime, '/');
			if (EndPart AND EndPart[1] == '.')
			{
				strsafecpy(Ext, EndPart, sizeof(Ext));

				// This is a hack to get around file types without a MIME database entry
				// but do have a .ext entry. LgiGetFileMimeType knows about the hack too.
				GRegKey ExtEntry("HKEY_CLASSES_ROOT\\%s", Ext + 1);
				char *Name = ExtEntry.GetStr();
				if (Name)
				{
					GRegKey Edit("HKEY_CLASSES_ROOT\\%s\\shell\\edit\\command", Name);
					char *App = Edit.GetStr();
					if (App)
					{
						Status = _GetApps_Add(Apps, App);
					}
					else
					{
						GRegKey Open("HKEY_CLASSES_ROOT\\%s\\shell\\open\\command", Name);
						char *App = Open.GetStr();
						if (App)
						{
							Status = _GetApps_Add(Apps, App);
						}
					}
				}
			}
			else
			{
				// This branch gets the list of application available to edit/open the file

				// Map the MIME type to a .ext
				GRegKey MimeEntry("HKEY_CLASSES_ROOT\\MIME\\Database\\Content Type\\%s", Mime);
				char *e = MimeEntry.GetStr("Extension");
				if (e)
					strsafecpy(Ext, e, sizeof(Ext));
				if (Ext[0])
				{
					// Get list of "Open With" apps
					GRegKey Other("HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\FileExts\\%s\\OpenWithList", Ext);
					char *Mru = NewStr(Other.GetStr("MRUList"));
					if (Mru)
					{
						for (char *a = Mru; *a; a++)
						{
							char Str[] = {*a, 0};
							char *Application = Other.GetStr(Str);
							if (Application)
							{
								GRegKey Shell("HKEY_CLASSES_ROOT\\Applications\\%s\\shell", Application);
								List<char> Keys;
								if (Shell.GetKeyNames(Keys))
								{
									GRegKey First("HKEY_CLASSES_ROOT\\Applications\\%s\\shell\\%s\\command", Application, Keys.First());
									char *Path;
									if (Path = First.GetStr())
									{
										Status |= _GetApps_Add(Apps, Path);
									}
								}
								Keys.DeleteArrays();
							}
						}

						DeleteArray(Mru);
					}

					if (NOT Status)
					{
						// Explorers file extensions
						GRegKey FileExt("HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\FileExts\\%s", Ext);
						char *Application;
						if (Application = FileExt.GetStr("Application"))
						{
							GRegKey App("HKEY_CLASSES_ROOT\\Applications\\%s\\shell\\open\\command", Application);
							char *Path;
							if (Path = App.GetStr())
							{
								Status = _GetApps_Add(Apps, Path);
							}
						}
					}

					if (NOT Status)
					{
						// get classes location
						GRegKey ExtEntry("HKEY_CLASSES_ROOT\\%s", Ext);
						GRegKey TypeEntry("HKEY_CLASSES_ROOT\\%s\\shell\\open\\command", ExtEntry.GetStr());
						char *Path = TypeEntry.GetStr();
						if (Path)
						{
							char *c = Path;
							char *Part = LgiTokStr(c);
							if (Part)
							{
								char AppPath[256];
								snprintf(AppPath, sizeof(AppPath), "\"%s\"", Part);
								Status = _GetApps_Add(Apps, AppPath);

								DeleteArray(Part);
							}
							else
							{
								Status = _GetApps_Add(Apps, Path);
							}
						}
					}
				}
			}
		}
	}

	return Status;
}

bool LgiGetAppForMimeType(char *Mime, char *AppPath, int BufSize)
{
	bool Status = false;
	if (AppPath)
	{
		GArray<GAppInfo*> Apps;
		Status = LgiGetAppsForMimeType(Mime, Apps, 1);
		if (Status)
		{
			strsafecpy(AppPath, Apps[0]->Path, BufSize);
			Apps.DeleteObjects();
		}
	}
	return Status;
}

int LgiRand(int i)
{
	return (rand() % i);
}

bool LgiPlaySound(char *FileName, int Flags)
{
	bool Status = false;

	HMODULE hDll = LoadLibrary("winmm.dll");
	if (hDll)
	{
		typedef BOOL (__stdcall *Proc_sndPlaySound)(LPCSTR pszSound, UINT fuSound);
		Proc_sndPlaySound psndPlaySound = (Proc_sndPlaySound)GetProcAddress(hDll, "sndPlaySoundA");

		if (psndPlaySound)
		{
			if (LgiGetOs() == LGI_OS_WIN9X)
			{
				// async broken on 98?
				Flags = 0;
			}

			Status = psndPlaySound(FileName, Flags);
		}

		FreeLibrary(hDll);
	}

	return Status;
}

bool LgiExecute(char *File, char *Arguments, char *Dir)
{
	int Status = 0;
	
	if (File)
	{
		if (LgiGetOs() == LGI_OS_WIN9X)
		{
			char *f = LgiToNativeCp(File);
			char *a = LgiToNativeCp(Arguments);
			char *d = LgiToNativeCp(Dir);
			if (f)
			{
				Status = (int) ShellExecute(NULL, "open", f, a, d, 5);
			}
			DeleteArray(f);
			DeleteArray(a);
			DeleteArray(d);
		}
		else
		{
			char16 *f = LgiNewUtf8To16(File);
			char16 *a = LgiNewUtf8To16(Arguments);
			char16 *d = LgiNewUtf8To16(Dir);
			if (f)
			{
				Status = (int) ShellExecuteW(NULL, L"open", f, a, d, 5);
			}
			DeleteArray(f);
			DeleteArray(a);
			DeleteArray(d);
		}
	}
	
	return Status > 32;
}

////////////////////////////////////////////////////////////////////////////////////
HKEY GetRootKey(char *s)
{
	HKEY Root = 0;

	#define TestKey(Name) \
		if (strncmp(s, #Name, strlen(#Name)) == 0) \
		{ \
			Root = Name; \
		}

	TestKey(HKEY_CLASSES_ROOT)
	else TestKey(HKEY_CURRENT_CONFIG)
	else TestKey(HKEY_CURRENT_USER)
	else TestKey(HKEY_LOCAL_MACHINE)
	else TestKey(HKEY_USERS)

	#undef TestKey

	return Root;
}

GRegKey::GRegKey(char *Key, ...)
{
	char Buffer[1025];

	k = 0;
	s[0] = 0;
	Root = (HKEY)-1;

	va_list Arg;
	va_start(Arg, Key);
	vsprintf(Buffer, Key, Arg);
	va_end(Arg);
	KeyName = NewStr(Buffer);

	if (KeyName)
	{
		int Len = 0;
		char *SubKey = 0;

		#define TestKey(Name) if (strncmp(KeyName, #Name, Len = strlen(#Name)) == 0) { Root = Name; SubKey = KeyName[Len] ? KeyName + Len + 1 : 0; }
		TestKey(HKEY_CLASSES_ROOT)
		else TestKey(HKEY_CURRENT_CONFIG)
		else TestKey(HKEY_CURRENT_USER)
		else TestKey(HKEY_LOCAL_MACHINE)
		else TestKey(HKEY_USERS)
		else return;

		RegOpenKeyEx(Root, SubKey, 0, KEY_ALL_ACCESS, &k);
	}
}

GRegKey::~GRegKey()
{
	if (k) RegCloseKey(k);
	DeleteArray(KeyName);
}

bool GRegKey::IsOk()
{
	return k != 0;
}

bool GRegKey::Create()
{
	bool Status = false;

	if (NOT k AND KeyName)
	{
		char *Sub = strchr(KeyName, '\\');
		if (Sub)
		{
			Status = RegCreateKey(Root, Sub+1, &k) == ERROR_SUCCESS;
		}
	}

	return Status;
}

char *GRegKey::Name()
{
	return KeyName;
}

bool GRegKey::DeleteValue(char *Name)
{
	if (k)
	{
		if (RegDeleteValue(k, Name) == ERROR_SUCCESS)
		{
			return true;
		}
		else
		{
			DWORD Err = GetLastError();
			int asd=0;
		}
	}

	return false;
}

bool GRegKey::DeleteKey()
{
	bool Status = false;

	if (k)
	{
		char *n = strchr(KeyName, '\\');
		if (n++)
		{
			RegCloseKey(k);
			k = 0;

			HKEY Root = GetRootKey(KeyName);
			Status = RegDeleteKey(Root, n) == ERROR_SUCCESS;
			DeleteArray(KeyName);
		}
	}

	return false;
}

char *GRegKey::GetStr(char *Name)
{
	DWORD Size = sizeof(s), Type;
	if (k AND RegQueryValueEx(k, Name, 0, &Type, (uchar*)s, &Size) == ERROR_SUCCESS)
	{
		return s;
	}

	return 0;		
}

bool GRegKey::SetStr(char *Name, char *Value)
{
	return k AND RegSetValueEx(k, Name, 0, REG_SZ, (uchar*)Value, strlen(Value)) == ERROR_SUCCESS;
}

int GRegKey::GetInt(char *Name)
{
	int i = 0;
	DWORD Size = sizeof(i), Type;
	if (k AND RegQueryValueEx(k, Name, 0, &Type, (uchar*)&i, &Size) != ERROR_SUCCESS)
	{
		i = 0;
	}

	return i;
}

bool GRegKey::SetInt(char *Name, int Value)
{
	return k AND RegSetValueEx(k, Name, 0, REG_DWORD, (uchar*)Value, sizeof(Value)) == ERROR_SUCCESS;
}

bool GRegKey::GetBinary(char *Name, void *&Ptr, int &Len)
{
	DWORD Size = 0, Type;
	if (k AND RegQueryValueEx(k, Name, 0, &Type, 0, &Size) == ERROR_SUCCESS)
	{
		Len = Size;
		Ptr = NEW(uchar[Len]);
		return RegQueryValueEx(k, Name, 0, &Type, (uchar*)Ptr, &Size) == ERROR_SUCCESS;
	}

	return false;
}

bool GRegKey::SetBinary(char *Name, void *Ptr, int Len)
{
	return false;
}

bool GRegKey::GetKeyNames(List<char> &n)
{
	FILETIME t;
	char Buf[256];
	DWORD Size = sizeof(Buf), i = 0;
	while (RegEnumKeyEx(k, i++, Buf, &Size, 0, 0, 0, &t) == ERROR_SUCCESS)
	{
		n.Insert(NewStr(Buf));
		Size = sizeof(Buf);
	}
	return n.First() != 0;
}

bool GRegKey::GetValueNames(List<char> &n)
{
	char Buf[256];
	DWORD Type, Size = sizeof(Buf), i = 0;
	while (RegEnumValue(k, i++, Buf, &Size, 0, &Type, 0, 0) == ERROR_SUCCESS)
	{
		n.Insert(NewStr(Buf));
		Size = sizeof(Buf);
	}
	return n.First() != 0;
}

//////////////////////////////////////////////////////////////////////////////////////
char *GetWin32Folder(int Id)
{
	GLibrary Shell("Shell32");
	
	if (LgiGetOs() == LGI_OS_WIN9X)
	{
		char ap[MAX_PATH] = "";
		pSHGetSpecialFolderPathA a = (pSHGetSpecialFolderPathA) Shell.GetAddress("SHGetSpecialFolderPathA");

		if (a AND a(0, ap, Id, false))
		{
			if (ValidStr(ap))
			{
				char *s = LgiFromNativeCp(ap);

				return s;
			}
		}
	}
	else
	{
		char16 wp[MAX_PATH] = { 0 };
		pSHGetSpecialFolderPathW w = (pSHGetSpecialFolderPathW) Shell.GetAddress("SHGetSpecialFolderPathW");
		if (w AND w(0, wp, Id, false))
		{
			if (ValidStrW(wp))
			{
				return LgiNewUtf16To8(wp);
			}
		}
		else
		{
			DWORD e = GetLastError();
			int asd=0;
		}
	}

	return 0;
}