#include "Lgi.h"
#include "Ftp.h"
#include "GTabView.h"
#include "GCombo.h"
#include "GCheckBox.h"
#include "GEdit.h"

char *TempStr = "(temporary)";

ConnectDlg::ConnectDlg(AppWnd *parent, List<FtpAcc> *accounts)
{
	SetParent(App = parent);
	Accounts = accounts;
	Account = 0;
	Index = 0;
	Cur = 0;

	Map(OPT_RemoteHost, IDC_SERVER, GV_STRING);
	Map(OPT_Port, IDC_PORT, GV_INT32);
	Map(OPT_UserName, IDC_USERNAME, GV_STRING);
	// Map(OPT_Password, IDC_PASSWORD, GV_STRING);
	Map(OPT_Anonymous, IDC_ANONYMOUS, GV_BOOL);
	Map(OPT_LocalDir, IDC_LOCAL, GV_STRING);
	Map(OPT_RemoteDir, IDC_REMOTE, GV_STRING);
	Map(OPT_UseActive, IDC_USE_ACTIVE, GV_BOOL);
	Map(OPT_NormalList, IDC_SIMPLE_LIST, GV_BOOL);
	Map(OPT_SFtp, IDC_SFTP, GV_BOOL);

	if (parent)
	{
		AppOpts.Map(OPT_Socks5Server, IDC_SOCKS5_SERVER, GV_STRING);
		AppOpts.Map(OPT_Socks5UserName, IDC_SOCKS5_USERNAME, GV_STRING);
		AppOpts.Map(OPT_Socks5Password, IDC_SOCKS5_PASSWORD, GV_STRING);
	}

	if (LoadFromResource(IDD_CONNECT))
	{
		GTabView *Tab = dynamic_cast<GTabView*>(FindControl(IDC_TAB));
		if (Tab)
		{
			MoveToCenter();

			Cur = Temp = NEW(FtpAcc);
			if (Temp)
			{
				GVariant v = TempStr;
				Temp->Tag->SetValue(OPT_AccountName, v);
				v = 21;
				Temp->Tag->SetValue(OPT_Port, v);

				Accounts->Insert(Temp, 0);
			}

			AccountCombo = dynamic_cast<GCombo*>(FindControl(IDC_ACCOUNT));
			LoadAccountList();

			// Data -> Controls
			Convert(Cur->Tag, this, true);
			AppOpts.Convert(App->GetOptions(), this, true);

			DoModal();
		}
	}
}

ConnectDlg::~ConnectDlg()
{
	Accounts->Delete(Temp);
	DeleteObj(Temp);
}

FtpAcc *ConnectDlg::GetAccount(char *Name)
{
	if (Name)
	{
		for (FtpAcc *a = Accounts->First(); a; a = Accounts->Next())
		{
			GVariant n;
			if (a->Tag->GetValue(OPT_AccountName, n) AND
				n.Str() AND
				stricmp(Name, n.Str()) == 0)
			{
				return a;
			}
		}
	}
	return 0;
}

void ConnectDlg::LoadAccountList()
{
	if (AccountCombo AND Accounts)
	{
		while (AccountCombo->Delete(0));

		for (FtpAcc *a = Accounts->First(); a; )
		{
			GVariant n;
			if (a->Tag->GetValue(OPT_AccountName, n) AND n.Str())
			{
				AccountCombo->Insert(n.Str());
				a = Accounts->Next();
			}
			else
			{
				Accounts->Delete(a);
				DeleteObj(a);
				a = Accounts->Current();
			}
		}

		AccountCombo->Value(Index);
	}
}

void ConnectDlg::Serialize(bool Write)
{
	if (Write)
	{
		GCheckBox *Anon = dynamic_cast<GCheckBox*>(FindControl(IDC_ANONYMOUS));

		// check for URL and decode
		char *Protocol = "ftp://";

		char *Server = NewStr(GetCtrlName(IDC_SERVER));
		if (Server AND strnicmp(Server, Protocol, strlen(Protocol)) == 0)
		{
			// ftp:// URL
			char *s = Server + strlen(Protocol);
			
			// User/Password
			char *At = strchr(s, '@');
			if (At)
			{
				*At = 0;
				char *Colon = strchr(s, ':');
				if (Colon)
				{
					*Colon = 0;
					SetCtrlName(IDC_USERNAME, s);
					SetCtrlName(IDC_PASSWORD, Colon+1);
				}
				s = At + 1;

				if (Anon)
				{
					Anon->Value(false);
				}
			}
			else if (Anon)
			{
				Anon->Value(true);
			}

			// Remote path
			char *Dir = strchr(s, '/');
			if (Dir)
			{
				// Chop off any file name
				char *LastDir = 0;
				for (char *End = Dir+strlen(Dir)-1; End>Dir; End--)
				{
					if (*End == '/')
					{
						LastDir = End;
						break;
					}
				}
				if (LastDir AND strchr(LastDir, '.'))
				{
					// er it's a file name... choppy chop
					*LastDir = 0;
				}

				SetCtrlName(IDC_REMOTE, Dir);
				*Dir++ = 0;
			}

			// Port?
			char *Colon = strchr(s, ':');
			if (Colon)
			{
				*Colon++ = 0;
				SetCtrlName(IDC_PORT, Colon);
			}

			// All thats left is the remote host
			SetCtrlName(IDC_SERVER, s);
		}
		DeleteArray(Server);

		// Check for user / anon
		if (Anon AND NOT ValidStr(GetCtrlName(IDC_USERNAME)))
		{
			Anon->Value(true);
		}

		// Controls -> Data
		Convert(Cur->Tag, this, false);

		char *p = ObsurePsw(GetCtrlName(IDC_PASSWORD), true);
		if (p)
		{
			GVariant Pwd = p;
			Cur->Tag->SetValue(OPT_Password, Pwd);
			DeleteArray(p);
		}
	}
	else
	{
		// Data -> Controls
		Convert(Cur->Tag, this, true);

		GVariant Pwd;
		if (Cur->Tag->GetValue(OPT_Password, Pwd))
		{
			char *p = ObsurePsw(Pwd.Str(), false);
			if (p)
			{
				SetCtrlName(IDC_PASSWORD, p);
				DeleteArray(p);
			}
		}
	}
}

int ConnectDlg::OnNotify(GViewI *Ctrl, int Flags)
{
	switch (Ctrl->GetId())
	{
		case IDC_USERNAME:
		{
			SetCtrlValue(IDC_ANONYMOUS, NOT ValidStr(Ctrl->Name()));
			break;
		}
		case IDC_ACCOUNT:
		{
			if (AccountCombo)
			{
				int i = AccountCombo->Value();
				if (i >= 0 AND Index != i)
				{
					Index = i;
					FtpAcc *Acc = Accounts->ItemAt(Index);
					if (Acc)
					{
						Serialize(true);
						Cur = Acc;
						Serialize(false);
						AccountCombo->Value(Index);
					}
				}
				// else user is typing something
			}
			break;
		}
		case IDC_ADD:
		{
			if (AccountCombo)
			{
				// Controls -> Data
				Serialize(true);

				FtpAcc *Acc = 0;
				if (Cur)
				{
					// save the temp into the list
					Acc = NEW(FtpAcc);
					if (Acc)
					{
						// Work out the account name
						char AName[256] = "";
						GInput Dlg(this, AName, LgiLoadString(IDS_NEW_ACCOUNT_NAME), AppName);
						if (Dlg.DoModal() AND Dlg.Str)
						{
							strcpy(AName, Dlg.Str);
						}
						else return 0;
						
						GVariant v(AName);
						Acc->Tag->SetValue(OPT_AccountName, v);

						// new current record
						// Options = Acc;
						Accounts->Insert(Cur = Acc);
						Index = Accounts->IndexOf(Acc);

						// delete keys out of temp... starts blank
						Temp->Tag->Attr.Empty();
						v = TempStr;
						Temp->Tag->SetValue(OPT_AccountName, v);
						v = 21;
						Temp->Tag->SetValue(OPT_Port, v);

						// load the data back into the UI
						// Data -> Controls
						Serialize(false);
					}
				}

				LoadAccountList();
			}
			break;
		}
		case IDC_DELETE:
		{
			if (AccountCombo AND Index > 0)
			{
				Cur->Tag->RemoveTag();
				Accounts->Delete(Cur);
				Cur = Accounts->ItemAt(--Index);
				Serialize(false); // Data -> Controls
				LoadAccountList();
			}
			break;
		}
		case IDC_BROWSE:
		{
			GEdit *LocalDir = dynamic_cast<GEdit*>(FindControl(IDC_LOCAL));
			if (LocalDir)
			{
				GFileSelect Select;
	
				Select.Name(LocalDir->Name());
				Select.Parent(this);
				if (Select.OpenFolder())
				{
					LocalDir->Name(Select.Name());
				}
			}
			break;
		}
		case IDOK:
		{
			// Save connect fields
			Serialize(true);
			Account = Cur;

			// Save options
			AppOpts.Convert(App->GetOptions(), this, false);

			EndModal(1);
			break;
		}
		case IDCANCEL:
		{
			// Save UI to xml
			Convert(Cur->Tag, this, false);
			AppOpts.Convert(App->GetOptions(), this, false);

			EndModal(0);
			break;
		}
	}
	return 0;
}

