#include "Lgi.h"
#include "Ftp.h"
#include "GEdit.h"

DirListItem::DirListItem(IFtpEntry *entry, GFileSystemView *view)
{
	_UserData = this;
	SizeStr[0] = 0;
	Entry = entry;
	View = view;
}

DirListItem::~DirListItem()
{
	DeleteObj(Entry);
}

bool DirListItem::GetSel(List<DirListItem> &e)
{
	List<GListItem> s;
	if (Parent->GetSelection(s))
	{
		for (GListItem *i=s.First(); i; i=s.Next())
		{
			DirListItem *d = dynamic_cast<DirListItem*>(i);
			if (d)
			{
				e.Insert(d);
			}
		}
	}
	
	return e.First() != 0;
}

bool DirListItem::OnKey(GKey &k)
{
	if (View)
	{
		switch (k.c16)
		{
			case VK_DELETE:
			{
				if (k.Down())
				{
					int Index = Parent->Value();
					GList *l = Parent;
					GFileSystemView *v = View;

					List<DirListItem> e;
					if (GetSel(e))
					{
						for (DirListItem *d=e.First(); d; d=e.Next())
						{
							l->Remove(d);
							v->Delete(d->Entry);
							DeleteObj(d);
						}

						l->Value(Index);
					}
				}
				return true;
				break;
			}
			case VK_RETURN:
			{
				if (k.Down() AND k.IsChar)
				{
					if (Entry->IsDir())
					{
						char Path[300];
						LgiMakePath(Path, sizeof(Path), View->Edit->Name(), Entry->Name);
						View->Cd(Path);
					}
				}
				return true;
				break;
			}
			case VK_BACKSPACE:
			{
				if (k.Down() AND k.IsChar)
				{
					View->Cd();
				}
				return true;
				break;
			}
		}
	}

	return false;
}

void DirListItem::OnMouseClick(GMouse &m)
{
	if (View)
	{
		View->OnClick(Entry, &m);
	}
}

bool DirListItem::OnBeginDrag(GMouse &m)
{
	if (View AND
		View->IsLocal())
	{
		Drag(Parent, DROPEFFECT_COPY);
	}

	return false;
}

bool DirListItem::GetData(GVariant *Data, char *Format)
{
	if (Data AND Format AND View)
	{
		char Path[256];
		List<char> Files;
		GMouse m;
		
		List<GListItem> Sel;
		if (Parent->GetSelection(Sel))
		{
			for (GListItem *i=Sel.First(); i; i=Sel.Next())
			{
				DirListItem *di = dynamic_cast<DirListItem *>(i);
				if (di)
				{
					View->GetDir(Path, di->Entry->Name);
					Files.Insert(NewStr(Path));
				}
			}

			Parent->GetMouse(m, true);

			bool Ret = CreateFileDrop(Data, m, Files);

			Files.DeleteArrays();

			return Ret;
		}
	}
	return false;
}

bool DirListItem::GetFormats(List<char> &Formats)
{
	Formats.Insert(NewStr(LGI_FileDropFormat));
	return Formats.First() != 0;
}

int DirListItem::GetImage(int Flags)
{
	int Ico = 1;

	if (Entry)
	{
		if (Entry->Attributes & IFTP_SYM_LINK)
		{
			Ico = 2;
		}
		else
		{
			Ico = Entry->IsDir() ? 0 : 1;
		}
		
		if (Entry->Attributes & IFTP_HIDDEN)
		{
			Ico += 3;
		}
	}

	return Ico;
}

char *DirListItem::GetText(int i)
{
	if (NOT Entry) return 0;

	switch (i)
	{
		case 0:
		{
			if (Entry->Name) return Entry->Name;
			break;
		}
		case 1:
		{
			if (Entry->IsDir())
			{
				return "<dir>";
			}
			else if (Entry->Name)
			{
				char *Dot = strrchr(Entry->Name, '.');
				return (Dot) ? Dot + 1 : LgiLoadString(IDS_NONE);
			}
			break;
		}
		case 2:
		{
			LgiFormatSize(SizeStr, Entry->Size);
			return SizeStr;
			break;
		}
		case 3:
		{
			static char s[64];
			Entry->Date.Get(s);
			return s;
			break;
		}
	}

	return 0;
}

