#include "Ftp.h"
#include "GEdit.h"
#include "GCombo.h"
#include "GButton.h"
#include "GToken.h"

GFileSystemView::GFileSystemView(AppWnd *wnd, GImageList *ImageList, bool Local)
{
	Wnd = wnd;
	Opposite = 0;
	Worker = 0;
	Listing = true;
	GRect r(0, 0, 100, 120);
	SetPos(r);
	Lst = 0;
	CurFolder = 0;
	Dir = 0;
	Edit = 0;
	Lst = 0;
	SortCol = -1;
	SortAscend = true;
	Mount = 0;
	#ifdef WIN32
	WndFlags |= WS_CLIPCHILDREN;
	#endif

	Children.Insert(Edit = NEW(GEdit(GFSV_DIR, 0, 0, 100, 20, "")));
	Children.Insert(Lst = NEW(FtpList(GFSV_LIST, 0, 20, 100, 100)));
	if (Lst)
	{
		Lst->View = this;
		Lst->Sunken(true);
		Lst->AddColumn(LgiLoadString(IDS_NAME), 240);
		Lst->AddColumn(LgiLoadString(IDS_TYPE), 60);
		Lst->AddColumn(LgiLoadString(IDS_SIZE), 70);
		Lst->AddColumn(LgiLoadString(IDS_DATE), 140);
		Lst->SetImageList(ImageList, false);
	}

	Commands = LgiLoadToolbar(this, "_pane.gif");
	if (Commands)
	{
		Commands->Raised(false);
		Commands->TextLabels(true);
		Commands->Customizable(Wnd->GetOptions(), IsLocal()?(char*)"LocalTools":(char*)"RemoteTools");
		Children.Insert(Commands);

		CmdCreateDir.ToolButton = Commands->AppendButton(LgiLoadString(IDS_CREATE_DIR), IDM_CREATE_DIR, TBT_PUSH, true, 0);
		CmdDelete.ToolButton =  Commands->AppendButton(LgiLoadString(IDS_DELETE), IDM_DELETE, TBT_PUSH, true, 1);
		CmdRenameFile.ToolButton =  Commands->AppendButton(LgiLoadString(IDS_RENAME), IDM_RENAME, TBT_PUSH, true, 2);
		CmdSetPerms.ToolButton = Commands->AppendButton(LgiLoadString(IDS_PERMS), IDM_SETPERMS, TBT_PUSH, true, 3);
		if (Local)
		{
			CmdExecute.ToolButton = Commands->AppendButton(LgiLoadString(IDS_EXECUTE_FILE), IDM_EXECUTE, TBT_PUSH, true, 4);
			CmdViewText.ToolButton = Commands->AppendButton(LgiLoadString(IDS_VIEW_TEXT), IDM_VIEW_TEXT, TBT_PUSH, true, 5);
		}

		Commands->AppendSeparator();

		CmdTransfer.ToolButton = Commands->AppendButton(LgiLoadString(IDS_TRANSFER), IDM_UPLOAD, TBT_PUSH, true, 6);
		Commands->AppendSeparator();

		CmdRefresh.ToolButton = Commands->AppendButton(LgiLoadString(IDS_LIST), IDM_LIST, TBT_PUSH, true, 7);
		CmdCdup.ToolButton = Commands->AppendButton(LgiLoadString(IDS_CDUP), IDM_UP, TBT_PUSH, true, 8);
	}
}

GFileSystemView::~GFileSystemView()
{
	DeleteArray(CurFolder);
}

int FtpItemCompare(GListItem *a, GListItem *b, int Data)
{
	DirListItem *A = (DirListItem*) a->_UserData;
	DirListItem *B = (DirListItem*) b->_UserData;
	GFileSystemView *View = (GFileSystemView*)Data;
	int Col = View->SortCol;
	int Status = 0;

	if (A AND B)
	{
		NameCompare:
		switch (Col)
		{
			case -1:
			{
				if (A->Entry->IsDir() AND
					NOT B->Entry->IsDir())
				{
					return -1;
				}
				if (NOT A->Entry->IsDir() AND
					B->Entry->IsDir())
				{
					return 1;
				}

				char *Sa = A->GetText(0);
				char *Sb = B->GetText(0);
				if (Sa AND Sb)
				{
					Status = stricmp(Sa, Sb);
				}
				break;
			}
			case 0:
			case 1:
			{
				char *Sa = A->GetText(Col);
				char *Sb = B->GetText(Col);
				if (Sa AND Sb)
				{
					Status = stricmp(Sa, Sb);
				}
				break;
			}
			case 2:
			{
				Status = A->Entry->Size - B->Entry->Size;
				break;
			}
			case 3:
			{
				Status = A->Entry->Date.Compare(&B->Entry->Date);
				break;
			}
			default:
			{
				LgiAssert(0);
				break;
			}
		}

		if (Status == 0 AND Col > 0)
		{
			Col = 0;
			goto NameCompare;
		}
	}

	return View->SortAscend ? Status : -Status;
}

void GFileSystemView::OnColumnClick(int Col, GMouse &m)
{
	if (SortCol < 0)
	{
		SortCol = Col;
		SortAscend = true;
	}
	else if (Col == SortCol)
	{
		if (Col <= 0)
		{
			if (SortAscend)
			{
				SortAscend = false;
			}
			else if (NOT SortAscend)
			{
				SortCol = -1;
				SortAscend = true;
			}
		}
		else
		{
			SortAscend = NOT SortAscend;
		}
	}
	else
	{
		SortCol = Col;
		SortAscend = true;
	}

	Lst->Sort(FtpItemCompare, (int)this);

	for (int i=0; i<Lst->GetColumns(); i++)
	{
		if (i == SortCol)
		{
			Lst->ColumnAt(i)->Mark(SortAscend ? GLI_MARK_DOWN_ARROW : GLI_MARK_UP_ARROW);
		}
		else
		{
			Lst->ColumnAt(i)->Mark(GLI_MARK_NONE);
		}
	}
}

void GFileSystemView::PostCmd(FtpCmd *Cmd)
{
	if (Cmd)
	{
		if (NOT Cmd->ForceRemote AND
			Cmd->View->IsCmdLocal(Cmd->Cmd))
		{
			DoCmd(Cmd);
		}
		else
		{
			Worker->PostCmd(Cmd);
		}
	}
}

bool GFileSystemView::Attach(GViewI *p)
{
	bool Status = GView::Attach(p);
	if (Status)
	{
		AttachChildren();
	}
	return Status;
}

bool GFileSystemView::Pour(GRegion &r)
{
	GRect p = GetPos();
	r.Offset(-p.x1, -p.y1);

	if (Commands)
	{
		Commands->Pour(r);
		r.Subtract(&Commands->GetPos());
	}

	GRect *Rgn = FindLargest(r);
	if (Rgn)
	{
		GRect e(Rgn->x1, Rgn->y1, Rgn->x2, Rgn->y1 + 19);
		
		if (Mount)
		{
			GRect m = e;
			e.x1 += Mount->X();
			m.x2 = e.x1 - 1;
			Mount->SetPos(m);
			r.Subtract(&m);
		}

		if (Dir)
		{
			GRect d = e;
			e.x2 -= Dir->X();
			d.x1 = e.x2 + 1;
			Dir->SetPos(d);
			r.Subtract(&d);
		}

		if (Edit)
		{
			Edit->SetPos(e);
			r.Subtract(&e);
		}
	}

	Rgn = FindLargest(r);
	if (Rgn AND Lst)
	{
		Lst->SetPos(*Rgn);
		r.Subtract(Rgn);
	}

	return true;
}

void GFileSystemView::OnPaint(GSurface *pDC)
{
	GRect a = GetPos();
	a.Offset(-a.x1, -a.y1);
	GRegion r(a);

	for (GViewI *c = Children.First(); c; c = Children.Next())
	{
		r.Subtract(&c->GetPos());
	}

	pDC->Colour(LC_MED, 24);
	for (GRect *b = r.First(); b; b = r.Next())
	{
		pDC->Rectangle(b);
	}
}

void GFileSystemView::GetMenuCommands(IFtpEntry *e, GSubMenu *RClick)
{
	if (RClick)
	{
		RClick->AppendItem(LgiLoadString(IDS_CHANGE_DIR), IDM_CD, e->IsDir() OR TestFlag(e->Attributes, IFTP_SYM_LINK));
		RClick->AppendItem(LgiLoadString(IDS_TRANSFER), IDM_TRANSFER, true);
		RClick->AppendItem(LgiLoadString(IDS_TRANSFER_MASK), IDM_TRANSFER_MASK, true);
	}
}

void GFileSystemView::OnMenuCommand(IFtpEntry *e, int Cmd)
{
	char *Mask = 0;
	
	switch (Cmd)
	{
		case IDM_CD:
		{
			if (Edit)
			{
				char p[300];
				LgiMakePath(p, sizeof(p), Edit->Name(), e->Name);
				
				FtpCmd *Cmd = NEW(FtpCmd(p));
				if (Cmd)
				{
					
					Cmd->Cmd = TASK_Cd;
					Cmd->Entry = NEW(IFtpEntry(e));
					Cmd->View = this;
					PostCmd(Cmd);
				}
			}
			break;
		}
		case IDM_TRANSFER_MASK:
		{
			GInput Dlg(this, "", LgiLoadString(IDS_ENTER_MASKS), AppName);
			if (Dlg.DoModal() == IDOK)
			{
				Mask = NewStr(Dlg.Str);
			}
			else
			{
				break;
			}
			// fall thru
		}
		case IDM_TRANSFER:
		{
			List<IFtpEntry> Sel;
			if (GetSelection(Sel) > 0)
			{
				FtpCmd *Cmd = 0;

				for (IFtpEntry *e = Sel.First(); e; e = Sel.Next())
				{
					Cmd = NEW(FtpCmd(Mask));
					if (Cmd)
					{
						Cmd->Int = true;
						Cmd->Cmd = TASK_Transfer;
						Cmd->Entry = NEW(IFtpEntry(e));
						Cmd->View = this;
						(IsLocal() ? this : Opposite)->GetDir(Cmd->Arg2);
						PostCmd(Cmd);
					}
				}
				
				Cmd = NEW(FtpCmd);
				if (Cmd)
				{
					Cmd->Cmd = TASK_ListElements;
					Cmd->View = Opposite;
					Cmd->ForceRemote = NOT IsCmdLocal(TASK_ListElements);
					PostCmd(Cmd);
				}
			}
			break;
		}
	}
}

void GFileSystemView::OnClick(IFtpEntry *e, GMouse *m)
{
	if (e AND m->Right() AND m->Down())
	{
		GSubMenu *RClick = NEW(GSubMenu);
		if (RClick)
		{
			GetMenuCommands(e, RClick);

			if (GetMouse(*m, true))
			{
				int Cmd = RClick->Float(this, m->x, m->y, false);
				if (Cmd)
				{
					OnMenuCommand(e, Cmd);
				}
			}
		}
	}

	if (e AND m AND Worker)
	{
		FtpCmd *Cmd = NEW(FtpCmd);
		if (Cmd)
		{
			Cmd->Cmd = TASK_ProcessClick;
			Cmd->Entry = NEW(IFtpEntry(e));
			Cmd->Click = *m;
			Cmd->View = this;
			PostCmd(Cmd);
		}
	}
}

int GFileSystemView::OnNotify(GViewI *Ctrl, int Flags)
{
	if (Worker)
	{
		switch (Ctrl->GetId())
		{
			case IDC_DIR:
			{
				if (Edit)
				{
					LgiExecute(Edit->Name(), "", "");
				}
				break;
			}
			case IDC_MOUNT:
			{
				if (Mount AND Edit)
				{
					char *Drive = Mount->Name();
					if (Drive)
					{
						char Desk[256];
						if (stricmp(Drive, LgiLoadString(IDS_DESKTOP)) == 0)
						{
							LgiGetSystemPath(LSP_DESKTOP, Desk, sizeof(Desk));
							Drive = Desk;
						}

						FtpCmd *Cmd = NEW(FtpCmd(Drive));
						if (Cmd)
						{
							Cmd->Cmd = TASK_Cd;
							Cmd->View = this;
							PostCmd(Cmd);
						}
						Mount->Value(-1);
					}
				}
				break;
			}
			case IDM_LIST:
			{
				FtpCmd *Cmd = NEW(FtpCmd);
				if (Cmd)
				{
					Cmd->Cmd = TASK_ListElements;
					Cmd->View = this;
					PostCmd(Cmd);
				}
				break;
			}
			case GFSV_DIR:
			{
				if (Edit AND Flags == VK_RETURN)
				{
					FtpCmd *Cmd = NEW(FtpCmd(Edit->Name()));
					if (Cmd)
					{
						Cmd->Cmd = TASK_Cd;
						Cmd->View = this;
						PostCmd(Cmd);
					}
				}
				break;
			}
			case IDM_UPLOAD:
			{
				List<IFtpEntry> Sel;
				if (GetSelection(Sel) > 0)
				{
					for (IFtpEntry *e = Sel.First(); e; e = Sel.Next())
					{
						FtpCmd *Cmd = NEW(FtpCmd);
						if (Cmd)
						{
							Cmd->Cmd = TASK_Transfer;
							Cmd->Entry = NEW(IFtpEntry(e));
							Cmd->View = this;
							(IsLocal() ? this : Opposite)->GetDir(Cmd->Arg2);
							PostCmd(Cmd);
						}
					}

					FtpCmd *Cmd = NEW(FtpCmd);
					if (Cmd)
					{
						Cmd->Cmd = TASK_ListElements;
						Cmd->View = Opposite;
						Cmd->ForceRemote = NOT IsCmdLocal(TASK_ListElements);
						PostCmd(Cmd);
					}
				}
				break;
			}
			case IDM_UP:
			{
				FtpCmd *Cmd = NEW(FtpCmd);
				if (Cmd)
				{
					Cmd->Cmd = TASK_Cd;
					Cmd->View = this;
					PostCmd(Cmd);
				}
				break;
			}
			case IDM_CREATE_DIR:
			{
				GInput Inp(this, "", LgiLoadString(IDS_NEW_DIR_NAME), AppName);
				if (Inp.DoModal())
				{
					FtpCmd *Cmd = NEW(FtpCmd(Inp.Str));
					if (Cmd)
					{
						Cmd->Cmd = TASK_CreateDir;
						Cmd->View = this;
						PostCmd(Cmd);
					}
				}
				break;
			}
			case IDM_RENAME:
			{
				List<IFtpEntry> Sel;
				if (GetSelection(Sel) > 0 AND
					Sel.Length() == 1)
				{
					IFtpEntry *e = Sel.First();
					if (e)
					{
						GInput Inp(this, e->Name, LgiLoadString(IDS_NEW_FILE_NAME), AppName);
						if (Inp.DoModal())
						{
							FtpCmd *Cmd = NEW(FtpCmd);
							if (Cmd)
							{
								Cmd->Cmd = TASK_Rename;
								Cmd->View = this;
								strcpy(Cmd->Arg, e->Name); // from
								strcpy(Cmd->Arg2, Inp.Str); // to
								PostCmd(Cmd);
							}
						}
					}
				}
				break;
			}
			case IDM_DELETE:
			{
				List<IFtpEntry> Sel;
				if (GetSelection(Sel) > 0)
				{
					IFtpEntry *n;
					for (IFtpEntry *e = Sel.First(); e; e = n)
					{
						n = Sel.Next();

						FtpCmd *Cmd = NEW(FtpCmd);
						if (Cmd)
						{
							Cmd->Cmd = TASK_Delete;
							Cmd->View = this;
							Cmd->Entry = NEW(IFtpEntry(e));

							if (NOT n) // only on the last one should we list
							{
								FtpCmd *ListCmd = NEW(FtpCmd);
								if (ListCmd)
								{
									ListCmd->Cmd = TASK_ListElements;
									ListCmd->View = this;
									ListCmd->ForceRemote = NOT IsCmdLocal(TASK_ListElements);
									Cmd->OnSuccess.Insert(ListCmd);
								}
							}

							PostCmd(Cmd);
						}
					}
				}
				break;
			}
			case IDM_EXECUTE:
			{
				List<IFtpEntry> Sel;
				if (GetSelection(Sel) > 0)
				{
					for (IFtpEntry *e = Sel.First(); e; e = Sel.Next())
					{
						FtpCmd *Cmd = NEW(FtpCmd(e->Name));
						if (Cmd)
						{
							Cmd->Cmd = TASK_ExecuteFile;
							Cmd->View = this;
							PostCmd(Cmd);
						}
					}
				}
				break;
			}
			case IDM_VIEW_TEXT:
			{
				List<IFtpEntry> Sel;
				if (GetSelection(Sel) > 0)
				{
					for (IFtpEntry *e = Sel.First(); e; e = Sel.Next())
					{
						FtpCmd *Cmd = NEW(FtpCmd(e->Name));
						if (Cmd)
						{
							Cmd->Cmd = TASK_ViewTextFile;
							Cmd->View = this;
							PostCmd(Cmd);
						}
					}
				}
				break;
			}
			case IDM_SETPERMS:
			{
				List<IFtpEntry> Sel;
				if (GetSelection(Sel) > 0)
				{
					#ifdef WIN32
					bool Win32 = IsCmdLocal(IDM_SETPERMS);
					#else
					bool Win32 = false;
					#endif
					
					int Attributes[9];
					ZeroObj(Attributes);
					int Flags = Win32 ? 4 : 9;
					ushort *PlatformFlags = Win32 ? Win32Perms : UnixPerms;
					for (IFtpEntry *e = Sel.First(); e; e = Sel.Next())
					{
						for (int i=0; i<Flags; i++)
						{
							if (e->Attributes & PlatformFlags[i])
							{
								Attributes[i]++;
							}
						}
					}
					for (int i=0; i<Flags; i++)
					{
						if (Attributes[i] > 0)
						{
							if (Attributes[i] < Sel.Length())
							{
								Attributes[i] = 2;
							}
							else
							{
								Attributes[i] = 1;
							}
						}
					}

					IFtpEntry *First = Sel.First();
					PermsDlg Dlg(this, Win32, Sel.Length() > 1, Attributes, First->User, First->Group);
					if (Dlg.DoModal() == IDOK)
					{
						for (IFtpEntry *e = Sel.First(); e; e = Sel.Next())
						{
							FtpCmd *Cmd = NEW(FtpCmd);
							if (Cmd)
							{
								Cmd->Cmd = TASK_SetPerms;
								Cmd->View = this;
								Cmd->Entry = NEW(IFtpEntry(e));
								
								Cmd->Int = 0;
								for (int i=0; i<Flags; i++)
								{
									switch (Attributes[i])
									{
										case 0:
										{
											break;
										}
										case 2:
										{
											Cmd->Int |= e->Attributes & PlatformFlags[i];
											break;
										}
										default: // 1 or true
										{
											Cmd->Int |= PlatformFlags[i];
											break;
										}
									}
								}
								
								PostCmd(Cmd);
							}
						}

						FtpCmd *Cmd = NEW(FtpCmd);
						if (Cmd)
						{
							Cmd->Cmd = TASK_ListElements;
							Cmd->View = this;
							PostCmd(Cmd);
						}
					}
				}
				break;
			}
		}
	}

	return 0;
}

int GFileSystemView::GetSelection(List<IFtpEntry> &Sel)
{
	int Status = 0;
	if (Lst)
	{
		List<GListItem> Ls;
		if (Lst AND Lst->GetSelection(Ls))
		{
			for (DirListItem *i = dynamic_cast<DirListItem*>(Ls.First());
				i;
				i = dynamic_cast<DirListItem*>(Ls.Next()))
			{
				if (i->GetEntry())
				{
					Sel.Insert(i->GetEntry());
					Status++;
				}
			}
		}
	}
	return Status;
}

void GFileSystemView::GetDir(char *Path, char *Append)
{
	char *Dir = (Edit) ? Edit->Name() : 0;
	if (Path AND Dir)
	{
		strcpy(Path, Dir);
		if (Append)
		{
			if (Path[strlen(Path)-1] != DirChar())
			{
				char *e = Path+strlen(Path);
				e[0] = DirChar();
				e[1] = 0;
			}
			strcat(Path, Append);
		}
	}
}

void GFileSystemView::DoCmd(FtpCmd *Cmd)
{
	if (Cmd AND Cmd->View)
	{
		bool Status = false;

		switch (Cmd->Cmd)
		{
			case TASK_ListElements:
			{
				Status = Cmd->View->ListElements();
				break;
			}
			case TASK_Cd:
			{
				Status = Cmd->View->Cd(Cmd->Str);
				break;
			}
			case TASK_CreateDir:
			{
				Status = Cmd->View->CreateDir(Cmd->Str);
				break;
			}
			case TASK_Delete:
			{
				GVariant Confirm = 0;
				Wnd->GetOptions()->GetValue(OPT_ConfirmDeletes, Confirm);

				if (NOT Confirm.CastInt32() OR
					LgiMsg(Wnd, LgiLoadString(IDS_DELETE_FILES), AppName, MB_YESNO) == IDYES)
				{
					if (Cmd->Entry->IsDir())
					{
						Status = Cmd->View->DeleteDir(Cmd->Entry);
					}
					else
					{
						Status = Cmd->View->Delete(Cmd->Entry);
					}
				}
				break;
			}
			case TASK_ExecuteFile:
			{
				Status = Cmd->View->ExecuteFile(Cmd->Str);
				break;
			}
			case TASK_ViewTextFile:
			{
				Status = Cmd->View->ViewTextFile(Cmd->Str);
				break;
			}
			case TASK_ProcessClick:
			{
				Status = Cmd->View->ProcessClick(Cmd->Entry, &Cmd->Click);
				break;
			}
			case TASK_Login:
			{
				GToken T(Cmd->Str, ",");
				if (Wnd->GetFtp() AND
					T.Length() >= 4)
				{
					Wnd->EnableCommands(true);
					if (T[5])
					{
						Wnd->GetFtp()->IsForceActive(atoi(T[5]));
					}

					#define CONNECT_RETIES 6
					bool Open = false;
					for (int i=0; NOT Open AND i<CONNECT_RETIES AND Wnd->Connecting; i++)
					{
						if (i>0)
						{
							Wnd->LogPanel->Print(FTP_COL_INFO, LgiLoadString(IDS_RETRY), i, CONNECT_RETIES-1);
						}

						Wnd->CmdAbort.Enabled(true);
						Wnd->CmdConnect.Enabled(false);

						char *User = LgiDecodeUri(T[2]);
						char *Pass = LgiDecodeUri(T[3]);
						int Port = atoi(T[1]);

						FtpOpenStatus Status = Wnd->GetFtp()->Open
							(
								Cmd->Connection,
								T[0], // server
								Port, // port
								User,
								Pass
							);

						DeleteArray(User);
						DeleteArray(Pass);

						Open = Status == FO_Connected;

						if (Open)
						{
							Wnd->PostEvent(M_ON_CONNECT, (int)NewStr(T[0]), Port);
						}
						else if (Status != FO_ConnectFailed)
						{
							if (Cmd->Connection->IsOpen())
							{
								Cmd->Connection->Close();
							}
							break;
						}
					}

					Wnd->Connecting = false;

					if (Open)
					{
						char *RemoteDir = T[4];
						if (ValidStr(RemoteDir))
						{
							Wnd->GetFtp()->SetDir(RemoteDir);
						}

						Status = ListElements();
					}
				}
				break;
			}
			case TASK_Transfer:
			{
				Status = Cmd->View->TransferFile(Cmd);
				break;
			}
			case TASK_TransferPattern:
			{
				Status = Cmd->View->TransferFilePattern(Cmd);
				break;
			}
			case TASK_Rename:
			{
				Status = Cmd->View->RenameFile(Cmd->Arg, Cmd->Arg2);
				break;
			}
			case TASK_SetPerms:
			{
				Status = Cmd->View->SetPerms(Cmd->Entry, Cmd->Int);
				break;
			}
			case TASK_Noop:
			{
				Wnd->GetFtp()->Noop();
				Status = true;
				break;
			}
			case TASK_Exit:
			{
				Wnd->PostEvent(M_CLOSE);
				Status = true;
				break;
			}
		}

		if (Status)
		{
			FtpCmd *c;
			while (c = Cmd->OnSuccess.First())
			{
				Cmd->OnSuccess.Delete(c);
				PostCmd(c);
			}
		}
		else
		{
			FtpCmd *c;
			while (c = Cmd->OnFail.First())
			{
				Cmd->OnFail.Delete(c);
				PostCmd(c);
			}
		}

		if (Wnd)
		{
			Wnd->OnCmd(Cmd);
		}
	}
}

IFtpEntry *GFileSystemView::HasDir(char *Name)
{
	DirListItem *e = HasEntry(Name);
	if (e AND e->GetEntry()->IsDir())
	{
		return e->GetEntry();
	}
	return 0;
}

DirListItem *GFileSystemView::HasEntry(char *Name)
{
	if (Lst AND Name)
	{
		for (DirListItem *i = dynamic_cast<DirListItem*>(Lst->First());
			i; i = dynamic_cast<DirListItem*>(Lst->Next()))
		{
			if (i->GetEntry() AND
				i->GetEntry()->Name AND
				stricmp(i->GetEntry()->Name, Name) == 0)
			{
				return i;
			}
		}
	}

	return 0;
}

bool GFileSystemView::TransferFilePattern(FtpCmd *c)
{
	bool Status = false;

	if (c)
	{
		List<GListItem> All;
		Lst->GetAll(All);
		for (DirListItem *	i = dynamic_cast<DirListItem*>(All.First()); i;
							i = dynamic_cast<DirListItem*>(All.Next()))
		{
			if (NOT i->Entry->IsDir() AND
				MatchStr(c->Str, i->Entry->Name))
			{
				FtpCmd Cmd;
				Cmd.Int2 = c->Int2;
				Cmd.View = c->View;
				Cmd.Connection = c->Connection;
				Cmd.Entry = NEW(IFtpEntry(i->Entry));

				Status |= TransferFile(&Cmd);
			}
		}
	}

	return Status;
}

bool GFileSystemView::TransferDir(IFtpEntry *e, char *Mask)
{
	bool Status = false;
	if (Lst AND
		e AND
		e->IsDir())
	{
		char OpPath[300];
		Opposite->GetDir(OpPath, e->Name);		
	
		IFtpEntry *d = Opposite->HasDir(e->Name);
		if (d)
		{
			// change to
			Status = Opposite->Cd(d->Path);
		}
		else
		{
			// create
			if (Status = Opposite->CreateDir(e->Name))
			{
				Status = Opposite->Cd(OpPath);
				if (NOT Status)
				{
					LgiTrace("%s:%i - Cd(%s) failed.\n", __FILE__, __LINE__, OpPath);
				}
			}
			else
			{
				LgiTrace("%s:%i - CreateDir(%s) failed.\n", __FILE__, __LINE__, e->Name);
			}
		}

		if (Status AND
			(Status = Cd(e->Path)))
		{
			GToken T(Mask, " ");

			// For all items in directory
			List<IFtpEntry> Local;
			{
				// copy all the items in the current directory into a temp
				// list, so that as we traverse the tree we don't lose items
				// in the list.
				DirListItem *i;
				for (i = dynamic_cast<DirListItem*>(Lst->First()); i;
					 i = dynamic_cast<DirListItem*>(Lst->Next()))
				{
					IFtpEntry *e = i->GetEntry();
					if (e AND e->Name AND stricmp(e->Name, "..") != 0)
					{
						Local.Insert(NEW(IFtpEntry(i->GetEntry())));
					}
				}
			}

			{
				// iterate the saved dir listing and process each item in turn
				for (IFtpEntry *i = Local.First(); Status AND i; i = Local.First())
				{
					if (i->IsDir())
					{
						// Dir
						if (stricmp(i->Name, "..") != 0)
						{
							Status &= TransferDir(i, Mask);
						}
					}
					else
					{
						// File
						bool Transfer = Mask == 0;
						if (Mask)
						{
							for (int n=0; n<T.Length(); n++)
							{
								if (MatchStr(T[n], i->Name))
								{
									Transfer = true;
									break;
								}
							}
						}

						if (Transfer)
						{
							char Path[256];
							if (IsCmdLocal(TASK_ListElements))
							{
								// upload
								GetDir(Path, i->Name);
								Status &= Wnd->GetFtp()->UploadFile(Path, i->Name, Wnd->GetBinaryMode(i->Name));

								// destination is remote, we can't waste time/bandwidth
								// updating the UI as we go...
							}
							else
							{
								// download
								Opposite->GetDir(Path, i->Name);
								Status &= Wnd->GetFtp()->DownloadFile(Path, i, Wnd->GetBinaryMode(i->Name));

								// we can list the local contents as we go as the command
								// is local
								Opposite->ListElements();
							}
						}
					}

					Local.Delete(i);
					DeleteObj(i);
				}
			}

			if (Status)
			{
			Status &= Cd();
			Status &= Opposite->Cd();
		}
	}
	}
	return Status;
}

bool GFileSystemView::DeleteDir(IFtpEntry *e)
{
	bool Status = false;
	if (Lst AND
		e AND
		e->IsDir())
	{
		if (IsLocal())
		{
			Status = FileDev->RemoveDirectory(e->Path, true);
			DeleteObj(e);
		}
		else
		{
			// Save all current items...
			List<DirListItem> Init;
			DirListItem *i;
			for (i = dynamic_cast<DirListItem*>(Lst->First());
				i; i = dynamic_cast<DirListItem*>(Lst->Next()))
			{
				Init.Insert(i);
			}
			Lst->RemoveAll();

			// Change to dir...
			if (Status = Cd(e->Path))
			{
				// For all items in directory
				List<DirListItem> Local;
				for (i = dynamic_cast<DirListItem*>(Lst->First());
					i; i = dynamic_cast<DirListItem*>(Lst->Next()))
				{
					Local.Insert(i);
				}
				Lst->RemoveAll();

				for (i = Local.First();
					i; i = Local.First())
				{
					if (i->GetEntry())
					{
						char *n = i->GetEntry()->Name;
						if (stricmp(n, "..") != 0)
						{
							if (i->GetEntry()->IsDir())
							{
								// Dir
								Status &= DeleteDir(i->GetEntry());
							}
							else
							{
								// File
								Status &= Delete(i->GetEntry());
							}
						}
					}

					Local.Delete(i);
					DeleteObj(i);
				}

				Status &= Cd(0, false);
				Status &= Delete(e);
			}

			// ReLoad all current items.
			for (i=Init.First(); i; i=Init.Next())
			{
				if (i->GetEntry()->Path AND
					strcmp(i->GetEntry()->Path, e->Path) == 0)
				{
					DeleteObj(i);
				}
				else
				{
					Lst->Insert(i);
				}
			}
		}
	}
	return Status;
}

