#include "Lgi.h"
#include "Ftp.h"

ILog::ILog(GLog *log)
{
	Info = 0;
	Logger = log;
	Logging = true;
}

ILog::~ILog()
{
}

void ILog::WriteLog(char *s, int Len, COLOUR c)
{
	if (Logging AND
		Logger AND
		s AND
		Len > 0)
	{
		if (Len > 5 AND
			strnicmp(s, "PASS ", 5) == 0)
		{
			Logger->Write(c, "PASS *******", 12);
			return;
		}
		
		char *Eol;
		while ((Len > 0) AND (Eol = strnstr(s, "\r\n", Len)))
		{
			int l = (int)Eol-(int)s;
			Logger->Write(c, s, l);
			s += l + 2;
			Len -= l + 2;
		}

		if (Len > 0)
		{
			Logger->Write(c, s, Len);
		}
	}
}

void ILog::LogRead(char *Data, int Len)
{
	if (Info)
	{
		Info->Value(Info->Value() + Len);
	}

	if (Data)
	{
		bool Error = Data[0] == '5' AND
						IsDigit(Data[1]) AND
						IsDigit(Data[2]) AND
						Data[3] == ' ';

		WriteLog(Data, Len, Error ? FTP_COL_ERROR : FTP_COL_READ);
	}
}

void ILog::LogWrite(char *Data, int Len)
{
	if (Info)
	{
		Info->Value(Info->Value() + Len);
	}

	if (Data)
	{
		WriteLog(Data, Len, FTP_COL_WRITE);
	}
}

void ILog::LogError(int ErrorCode, char *ErrorDescription)
{
	if (Logger)
	{
		if (ErrorCode > 0)
		{
			Logger->Print(FTP_COL_ERROR, "%s(%i): %s", LgiLoadString(IDS_ERROR), ErrorCode, ErrorDescription);
		}
		else
		{
			Logger->Print(FTP_COL_ERROR, "%s: %s", LgiLoadString(IDS_ERROR), ErrorDescription);
		}
	}
}

void ILog::LogInformation(char *Str)
{
	if (Logger)
	{
		Logger->Print(FTP_COL_INFO, "%s: %s", LgiLoadString(IDS_INFORMATION), Str);
	}
}

int ILog::LogSetParameter(int Which, int What)
{
	switch (Which)
	{
		case GSocket_Log:
		{
			Logging = What != 0;
			break;
		}
		case GSocket_Progress:
		{
			Info = (Progress*) What;
			break;
		}
	}

	return 0;
}

///////////////////////////////////////////////////////////////
ILogConnection::ILogConnection(GLog *log, AppWnd *wnd) :
	ILog(log)
{
	Wnd = wnd;
}

void ILogConnection::OnDisconnect()
{
	if (Wnd)
	{
		Wnd->PostEvent(M_ON_DISCONNECT);
	}

	char s[256];
	sprintf(s, "[INET] %s", LgiLoadString(IDS_DISCONNECT));
	LogInformation(s);
}

void ILogConnection::OnRead(char *Data, int Len) { LogRead(Data, Len); }
void ILogConnection::OnWrite(char *Data, int Len) { LogWrite(Data, Len); }
void ILogConnection::OnError(int ErrorCode, char *ErrorDescription) { LogError(ErrorCode, ErrorDescription); }
void ILogConnection::OnInformation(char *Str) { LogInformation(Str); }
int ILogConnection::SetParameter(int Which, int What) { return LogSetParameter(Which, What); }

//////////////////////////////////////////////////////////////////////////////
ISocks5LogConnection::ISocks5LogConnection(	char *proxy,
											int port,
											char *username,
											char *password,
											GLog *log,
											AppWnd *wnd) :
	ILog(log),
	GSocks5Socket(proxy, port, username, password)
{
	Wnd = wnd;
}

void ISocks5LogConnection::OnDisconnect()
{
	if (Wnd)
	{
		Wnd->PostEvent(M_ON_DISCONNECT);
	}

	char s[256];
	sprintf(s, "[INET] %s", LgiLoadString(IDS_DISCONNECT));
	LogInformation(s);
}

void ISocks5LogConnection::OnRead(char *Data, int Len) { LogRead(Data, Len); }
void ISocks5LogConnection::OnWrite(char *Data, int Len) { LogWrite(Data, Len); }
void ISocks5LogConnection::OnError(int ErrorCode, char *ErrorDescription) { LogError(ErrorCode, ErrorDescription); }
void ISocks5LogConnection::OnInformation(char *Str) { LogInformation(Str); }
int ISocks5LogConnection::SetParameter(int Which, int What) { return LogSetParameter(Which, What); }

