#include "Lgi.h"
#include "Ftp.h"
#include "Schedule.h"
#include "GXmlTree.h"
#include "GList.h"

bool GetScheduleFile(char *File, int Len)
{
	if (LgiGetExePath(File, Len))
	{
		#ifdef WIN32
		if (stristr(File, "\\release") OR stristr(File, "\\debug"))
			LgiTrimDir(File);
		#endif
		return LgiMakePath(File, Len, File, "Schedule.xml");
	}

	return false;
}

class Event : public GListItem
{
public:
	GXmlTag *Tag;

	Event(GXmlTag *tag);
	~Event();
	char *GetText(int Col);
	bool DoUi(GView *p);
	void OnMouseClick(GMouse &m);
};

class EventDlg : public GDialog
{
	Event *e;

public:
	EventDlg(GView *p, Event *event)
	{
		SetParent(p);
		e = event;
		if (LoadFromResource(IDD_DOWNLOAD))
		{
			MoveToCenter();

			GViewI *Time = FindControl(IDC_TIME);
			if (Time)
			{
				GViewI *v = FindControl(IDC_SET_DATE);
				if (v) v->SetNotify(Time);
				v = FindControl(IDC_SET_TIME);
				if (v) v->SetNotify(Time);
			}

			SetCtrlName(IDC_URL, e->Tag->GetAttr(OPT_Url));
			SetCtrlName(IDC_TIME, e->Tag->GetAttr(OPT_Time));
			SetCtrlName(IDC_FOLDER, e->Tag->GetAttr(OPT_Folder));
		}
	}

	int OnNotify(GViewI *c, int f)
	{
		switch (c->GetId())
		{
			case IDC_SET_FOLDER:
			{
				GFileSelect s;
				s.Parent(this);
				s.Name(GetCtrlName(IDC_FOLDER));
				if (s.OpenFolder())
				{
					SetCtrlName(IDC_FOLDER, s.Name());
				}
				break;
			}
			case IDOK:
			{
				e->Tag->SetAttr(OPT_Url, GetCtrlName(IDC_URL));
				e->Tag->SetAttr(OPT_Time, GetCtrlName(IDC_TIME));
				e->Tag->SetAttr(OPT_Folder, GetCtrlName(IDC_FOLDER));
				
				// fall thru
			}
			case IDCANCEL:
			{
				EndModal(c->GetId() == IDOK);
				break;
			}
		}
		return 0;
	}
};

Event::Event(GXmlTag *tag)
{
	Tag = tag;
}

Event::~Event()
{
}

void Event::OnMouseClick(GMouse &m)
{
	if (m.Left() AND m.Double() AND m.Down())	
	{
		DoUi(GetList());
	}
}

char *Event::GetText(int Col)
{
	switch (Col)
	{
		case 0:
			return Tag->GetAttr(OPT_Url);
		case 1:
			return Tag->GetAttr(OPT_Folder);
	}

	return 0;
}

bool Event::DoUi(GView *p)
{
	EventDlg d(p, this);
	bool Status = d.DoModal();
	Update();
	return Status;
}

class ScheduleWndPrivate : public GLgiRes
{
public:
	char *File;
	GXmlTag *Root;
	GList *Lst;
	AppWnd *App;

	ScheduleWndPrivate()
	{
		App = 0;
		Root = NEW(GXmlTag("Schedule"));
		File = 0;
		Lst = 0;

		char e[256];
		if (GetScheduleFile(e, sizeof(e)))
		{
			File = NewStr(e);
			if (FileExists(e))
			{
				GFile f;
				if (f.Open(e, O_READ))
				{
					GXmlTree t;
					t.Read(Root, &f, 0);
				}
			}
		}
	}

	~ScheduleWndPrivate()
	{
		if (Root AND File)
		{
			GFile f;
			if (f.Open(File, O_WRITE))
			{
				f.SetSize(0);

				GXmlTree t;
				if (t.Write(Root, &f))
				{
					f.Close();

					App->OnScheduleChanged(File);
				}
			}
		}

		DeleteArray(File);
		DeleteObj(Root);
	}
};

ScheduleWnd::ScheduleWnd(AppWnd *app)
{
	d = NEW(ScheduleWndPrivate);
	d->App = app;

	char n[256];
	GRect p;
	if (d->LoadFromResource(IDD_SCHEDULE, Children, &p, n))
	{
		SetPos(p);
		MoveToCenter();
		Name(n);
		#ifdef WIN32
		CreateClass(Name());
		#endif

		if (Attach(0))
		{
			AttachChildren();
			if (d->Lst = dynamic_cast<GList*>(FindControl(IDC_DOWNLOADS)))
			{
				if (d->Root)
				{
					for (GXmlTag *t = d->Root->Children.First(); t; t = d->Root->Children.Next())
					{
						d->Lst->Insert(NEW(Event(t)));
					}
				}
			}

			Visible(true);
		}
	}
}

ScheduleWnd::~ScheduleWnd()
{
	DeleteObj(d);
}

void ScheduleWnd::OnPosChange()
{
}

int ScheduleWnd::OnNotify(GViewI *c, int f)
{
	switch (c->GetId())
	{
		case IDC_NEW:
		{
			if (d->Root AND d->Lst)
			{
				Event *e = NEW(Event(NEW(GXmlTag("Event"))));
				if (e)
				{
					if (e->DoUi(this))
					{
						d->Lst->Insert(e);
						d->Root->InsertTag(e->Tag);
					}
					else
					{
						DeleteObj(e->Tag);
						DeleteObj(e);
					}
				}
			}
			break;
		}
		case IDC_DELETE:
		{
			if (d->Lst)
			{
				List<GListItem> s;
				if (d->Lst->GetSelection(s))
				{
					for (Event *e = dynamic_cast<Event*>(s.First()); e;
						e = dynamic_cast<Event*>(s.Next()))
					{
						e->Tag->RemoveTag();
						DeleteObj(e->Tag);
					}

					s.DeleteObjects();
				}
			}
			break;
		}
	}

	return 0;
}

